IMPLEMENTATION MODULE M2Time ;


FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 GetScope, GetOffset, GetSize, GetUnitPrefix,
                 MaxDatum ;

FROM M2ALU IMPORT PopInt ;

FROM SymbolTable IMPORT NulSym,
                        ModeOfAddr,
                        GetMainModule,
                        IsConstString, GetString, GetStringLength,
                        IsProcedure, IsVar, IsModule, IsExported,
                        GetSymName, GetNth, GetMode, GetType,
                        PushValue ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT InternalError, NearToken ;
FROM M2Error IMPORT BeginError, EndError ;
FROM StdIO IMPORT Write, PushOutput, PopOutput ;
FROM StrIO IMPORT WriteString, WriteLn ;


(*
   Methods of timing calculation:

   (i)   I used the Intel microprocessor and peripheral handbook vol 1.
   (ii)  All timings taken from i386, protected virtual address mode
*)



(*
   BinInstReg - inst operating on a reg. (inst <-> Reg)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determine the operand register.

                Example:

                push   eax
*)

PROCEDURE BinInstReg (Inst: Instruction;
                      Unit: DatumPrefix;
                      Reg : Register) : CARDINAL ;
BEGIN
   CASE Inst OF

   push:  RETURN( 4 ) |
   pop :  RETURN( 6 ) |
   inc :  RETURN( 2 ) |
   dec :  RETURN( 2 ) |
   neg :  RETURN( 2 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END BinInstReg ;


(*
   InstStr - inst operating on a label (string). (inst  label)
             Inst   - determines the instruction.
             Str    - string label

             Example:

             jmp    L123
*)

PROCEDURE InstStr (Inst: Instruction;
                   Str : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   jmp :  RETURN( 7 ) |
   call:  RETURN( 17 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstStr ;


(*
   BinInstMD - inst operating on a memory location. (inst <-> Mem)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the operand memory location.
               Scope1 - determines the scope of the operand.
               Sym    - symbol requiring the memory reference.

               Example:

               push   (ebp+15)
*)

PROCEDURE BinInstMD (Inst: Instruction; Unit: DatumPrefix;
                     o1, go: INTEGER; Scope1: KindOfScope; Sym: CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   push:  RETURN( 9 ) |
   pop :  RETURN( 6 ) |
   inc :  RETURN( 9 ) |
   dec :  RETURN( 10 ) |
   neg :  RETURN( 10 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END BinInstMD ;


(*
   InstRegConstant - instruction constant and memory.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   12, Reg
                     add   12, Reg
*)

PROCEDURE InstRegConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Reg : Register;
                           Sym : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov :  RETURN( 2 ) |
   add :  RETURN( 2 ) |
   sub :  RETURN( 2 ) |
   mult:  RETURN( 13 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstRegConstant ;


(*
   InstRegReg - inst reg into reg. (Reg1 <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.

                Example:

                mov   ebp, eax
               ; mov   reg2, reg1
*)

PROCEDURE InstRegReg (Inst: Instruction; Unit: DatumPrefix;
                      Reg1, Reg2: Register) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov :  RETURN( 2 ) |
   add :  RETURN( 2 ) |
   sub :  RETURN( 2 ) |
   mult:  RETURN( 12 ) |
   idiv:  CASE Unit OF

          PByte :  RETURN( 19 ) |
          PWord :  RETURN( 27 ) |
          PDWord:  RETURN( 43 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   div :  CASE Unit OF

          PByte :  RETURN( 14 ) |
          PWord :  RETURN( 22 ) |
          PDWord:  RETURN( 38 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   sal :  RETURN( 3 ) |
   or  :  RETURN( 2 ) |
   and :  RETURN( 2 ) |
   test:  RETURN( 2 ) |
   xor :  RETURN( 2 )
   
   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstRegReg ;


(*
   InstRegMD - inst memory into reg. (Reg <- inst <- Mem1)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o2, go - determine the offset into element2.
               Reg    - determine the register.
               Scope2 - determines the scope of element2.
               Sym    - symbol requiring the memory reference.

               Example:

               ; inst  Mem{scope}+o1+go, reg
               mov   (ebp+15), eax
*)

PROCEDURE InstRegMD (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg   : Register;
                     o1, go: INTEGER ;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov :  RETURN( 4 ) |
   add :  RETURN( 6 ) |
   sub :  RETURN( 6 ) |
   mult:  RETURN( 17 ) |
   idiv:  CASE Unit OF     (* does the idiv exist?   *)

          PByte :  RETURN( 22 ) |
          PWord :  RETURN( 30 ) |
          PDWord:  RETURN( 46 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   div :  CASE Unit OF

          PByte :  RETURN( 17 ) |
          PWord :  RETURN( 25 ) |
          PDWord:  RETURN( 41 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   cmp :  RETURN( 5 ) |
   sal :  InternalError('sal does not exist for this mode (should have called InstMDReg)', __FILE__, __LINE__) |
   or  :  RETURN( 6 ) |
   and :  RETURN( 6 ) |
   test:  RETURN( 5 ) |
   xor :  RETURN( 6 ) |
   lea :  RETURN( 2 )
   
   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstRegMD ;


(*
   InstMDReg - inst reg into memory. (Mem1 <- inst <- Reg)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Reg    - determine the register.
               Scope1 - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   eax, (ebp+15)
*)

PROCEDURE InstMDReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     o1, go: INTEGER ;
                     Reg   : Register;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov :  RETURN( 4 ) |
   add :  RETURN( 7 ) |
   sub :  RETURN( 7 ) |
   idiv:  CASE Unit OF

          PByte :  RETURN( 22 ) |
          PWord :  RETURN( 30 ) |
          PDWord:  RETURN( 46 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   div :  CASE Unit OF

          PByte :  RETURN( 17 ) |
          PWord :  RETURN( 25 ) |
          PDWord:  RETURN( 41 )

          ELSE
             InternalError('incorrect datum prefix', __FILE__, __LINE__)
          END |
   cmp :  RETURN( 6 ) |
   sal :  RETURN( 7 ) |
   or  :  RETURN( 7 ) |
   and :  RETURN( 7 ) |
   test:  RETURN( 5 ) |
   xor :  RETURN( 7 )
   
   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstMDReg ;


(*
   InstMDConstant - instruction constant and memory.  (Mem <- inst <- Constant)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    o1, go - determine the offset into element1.
                    Scope  - determines the scope of element1.
                    Sym1   - element1 where the constant is to be placed.
                             Symbol requiring the memory reference.
                    Sym2   - constant element2.

                    Example:

                    mov   12, (ebp+15)
*)

PROCEDURE InstMDConstant (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          o1, go: INTEGER ;
                          Scope : KindOfScope;
                          Sym,
                          const : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov :  RETURN( 4 ) |
   add :  RETURN( 11 ) |
   sub :  RETURN( 11 ) |
   cmp :  RETURN( 7 ) |
   sal :  RETURN( 11 ) |
   or  :  RETURN( 11 ) |
   and :  RETURN( 11 ) |
   test:  RETURN( 7 ) |
   xor :  RETURN( 11 )
   
   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstMDConstant ;


(*
   InstMDOffset - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
                  Offset is the address of symbol Sym2.
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Sym1   - indicates the memory location which will
                           contain the address of Sym2.
                  Sym2   - symbol requiring the memory reference.

                  Example:

                  mov   $_varname, (ebp-15)
                  ; mov  $_varname, {unit prefix} {scope}
*)

PROCEDURE InstMDOffset (Inst: Instruction;
                        Unit: DatumPrefix;
                        Sym1, Sym2: CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov:  RETURN( 4 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstMDOffset ;


(*
   InstRegImmName -  instruction immediate name into register.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   $foo, Reg
*)

PROCEDURE InstRegImmName (Inst: Instruction;
                          Unit: DatumPrefix;
                          Reg : Register;
                          Sym : CARDINAL) : CARDINAL ;
BEGIN
   CASE Inst OF

   mov:  RETURN( 2 )

   ELSE
      InternalError('unknown assembly instruction in timing analysis', __FILE__, __LINE__)
   END
END InstRegImmName ;



(**************************************************************************
(*
   InstConstantReg - instruction register and constant.
      	       	     	   (cmp Constant, Reg)

                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   Reg, 12
                     add   Reg, 12
*)

PROCEDURE InstConstantReg (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL;
                           Reg : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  ') ;
   WriteRegister(Reg, Unit) ;
   WriteString(', $') ;
   WriteConstant(Sym) ;
   WriteLn
END InstConstantReg ;


(*
   InstMDConstant - instruction constant and memory.  (Mem <- inst <- Constant)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    o1, go - determine the offset into element1.
                    Scope  - determines the scope of element1.
                    Sym1   - element1 where the constant is to be placed.
                             Symbol requiring the memory reference.
                    Sym2   - constant element2.

                    Example:

                    mov   12, (ebp+15)
*)

PROCEDURE InstMDConstant (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          o1, go: INTEGER ;
                          Scope : KindOfScope;
                          Sym1,
                          Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym2) ;
   WriteString(' ,') ;
   WriteScope(Scope, o1, go, Sym1) ;
   WriteLn
END InstMDConstant ;


(*
   InstMRIConstant - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Sym2   - constant element2.

                     Example:

                     mov   12, (edi+15)
*)

PROCEDURE InstMRIConstant (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Sym2  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym2) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   WriteLn
END InstMRIConstant ;


(*
   InstMRICardinal - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Card   - cardinal constant.

                     Example:

                     mov   12H, (edi+15)
*)

PROCEDURE InstMRICardinal (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Card  : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteString('$0x') ; WriteHex(Card, 0) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   WriteLn
END InstMRICardinal ;


(*
   InstMRIReg - instruction constant and memory indexed by a register.
                (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                o1, go - determine the offset into element1.
                Reg1   - indexes memory.
                Reg2   - source of data.

                Example:

                mov   eax, (edi+15)
*)

PROCEDURE InstMRIReg (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      o1, go: INTEGER ;
                      Reg1  ,
                      Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', ') ;
   WritePositiveOffset(o1) ;
   WriteOffset(go) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   WriteLn
END InstMRIReg ;




(*
   InstRegMIO - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.
                Offset - offset to be added to the index register.

               Example:

               mov   (edi+4), eax
               ; mov   (reg2+offset), reg1
*)

PROCEDURE InstRegMIO (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      Reg1,
                      Reg2  : Register;
                      Offset: INTEGER) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('), ') ;
   WriteRegister(Reg1, Unit) ;
   WriteLn
END InstRegMIO ;


(*
   InstRegMI - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the destination register.
               Reg2   - determine the index register.

               Example:

               mov   (edi), eax
               ; mov   (reg2), reg1
*)

PROCEDURE InstRegMI (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' (') ;
   WriteRegister(Reg2, MaxDatum()) ;
   WriteString('), ') ;
   WriteRegister(Reg1, Unit) ;
   WriteLn
END InstRegMI ;


(*
   InstMIOReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the index register.
                Offset - determines the offset to be added to the index
                         register.
                Reg2   - determine the source register.

               Example:

               mov   eax, (edi)
               ; mov   reg2, (reg1)
*)

PROCEDURE InstMIOReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1  : Register;
                     Offset: INTEGER;
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', ') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   WriteLn
END InstMIOReg ;


(*
   InstMIReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the index register.
               Reg2   - determine the source register.

               Example:

               mov   eax, (edi)
               ; mov   reg2, (reg1)
*)

PROCEDURE InstMIReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   Write(' ') ;
   WriteRegister(Reg2, Unit) ;
   WriteString(', (') ;
   WriteRegister(Reg1, MaxDatum()) ;
   Write(')') ;
   WriteLn
END InstMIReg ;


(*
   InstRegInt - inst Reg with an Int. (Reg <- inst <- Int)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determines the destination register.
                Int    - the integer constant as the second operand.

                Example:

                mov   15, eax
               ; mov   Int, Reg
*)

PROCEDURE InstRegInt (Inst: Instruction; Unit: DatumPrefix;
                      Reg: Register; Int: INTEGER) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  $') ;
   WriteInt(Int, 0) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   WriteLn
END InstRegInt ;


(*
   InstRegHexCard - inst Reg with an cardinal. (Reg <- inst <- Card)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg    - determines the destination register.
                    Card   - the cardinal constant as the second operand.

                    Example:

                    and   $0FFFF0000H, eax
                   ; mov   Int, Reg
*)

PROCEDURE InstRegHexCard (Inst: Instruction; Unit: DatumPrefix;
                          Reg: Register; Card: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $0x') ;
   WriteHex(Card, 0) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   WriteLn
END InstRegHexCard ;


(*
   InstMDOffset - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
                  Offset is the address of symbol Sym2.
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Sym1   - indicates the memory location which will
                           contain the address of Sym2.
                  Sym2   - symbol requiring the memory reference.

                  Example:

                  mov   $_varname, (ebp-15)
                  ; mov  $_varname, {unit prefix} {scope}
*)

PROCEDURE InstMDOffset (Inst: Instruction; Unit: DatumPrefix;
                        Sym1, Sym2: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteAsmName(Sym2) ;
   WriteString(', ') ;
   WriteScope(GetScope(Sym1), GetOffset(Sym1), 0, Sym1) ;
   WriteLn
END InstMDOffset ;


(*
   InstMDInt - inst Mem with an Int. (Mem[Sym1] <- inst <- Int)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Int    - the integer constant as the second operand.
               Scope  - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   123, (ebp-15)
               ; mov   int, {unit prefix} {scope}
*)

PROCEDURE InstMDInt (Inst: Instruction; Unit: DatumPrefix;
                     o1, go: INTEGER; Int: INTEGER;
                     Scope: KindOfScope; Sym: CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteInt(Int, 0) ;
   WriteString(', ') ;
   WriteScope(Scope, o1, go, Sym) ;
   WriteLn
END InstMDInt ;


(*
   InstStr - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstStr (Inst: Instruction; String: ARRAY OF CHAR) ;
BEGIN
   WriteInstruction(Inst, MaxDatum()) ;
   WriteString('  ') ;
   WriteString(String) ;
   WriteLn
END InstStr ;


(*
   BinInst - write a sole inst.
             Inst   - determines the instruction.

             Example:

             ret
*)

PROCEDURE BinInst (Inst: Instruction) ;
BEGIN
   WriteInstruction(Inst, MaxDatum()) ;
   WriteLn
END BinInst ;


(*
   BinInstRegI - inst operating on a reg. ( inst <-> (Reg) )
                 Inst   - determines the instruction.
                 Unit   - determines the DatumPrefix of transfer.
                 Reg    - determine the operand register.

                 Example:

                 call   (eax)
*)

PROCEDURE BinInstRegI (Inst: Instruction;
                       Unit: DatumPrefix;
                       Reg : Register) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' (') ;
   WriteRegister(Reg, Unit) ;
   Write(')') ;
   WriteLn
END BinInstRegI ;


(*
   BinInstConstant - inst operating on a reg. (inst <-> Reg)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Sym    - determine the constant.

                     Example:

                     push   15
*)

PROCEDURE BinInstConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString(' $') ;
   WriteConstant(Sym) ;
   WriteLn
END BinInstConstant ;


(*
   BinInstHex - inst operating on a number. (inst <-> hex)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                c      - number.

                Example:

                push   0x1234
*)

PROCEDURE BinInstHex (Inst: Instruction;
                      Unit: DatumPrefix;
                      c   : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   IF (Unit=PByte) AND (Inst=push)
   THEN
      WriteError('cannot push a byte on the 8088, 8086, 80386 or 80486')
   END ;
   WriteString(' $0x') ;
   WriteHex(c, 0) ; WriteLn ;
END BinInstHex ;


(*
   WriteScope - writes the referencing information to a particular scope.
                Scope is either Global or Local.
                o1 is the offset of the element and go is the Delta
                offset. It is useful to contain both as separate
                entities as it makes human debugging easier!
                Sym - is the variable requiring memory access.
                      If global then the name Sym is written
                      otherwise the activation record offset in the form of
                      o1+go is written.
*)

PROCEDURE WriteScope (Scope: KindOfScope; o1, go: INTEGER; Sym: CARDINAL) ;
VAR
   Proc : CARDINAL ;
BEGIN
   IF Scope=Local
   THEN
      WritePositiveOffset(o1) ; WriteOffset(go) ;
      WriteString('(%ebp)')
   ELSE
      WriteAsmName(Sym) ; WriteOffset(go)
   END
END WriteScope ;


(*
   WritePositiveOffset - writes an integer offset. It only has a sign if
                         the offset is negative.
*)

PROCEDURE WritePositiveOffset (o: INTEGER) ;
BEGIN
   WriteInt(o, 0)
END WritePositiveOffset ;


(*
   WriteOffset - writes an integer offset with a sign beforehand.
*)

PROCEDURE WriteOffset (o: INTEGER) ;
BEGIN
   IF o>=0
   THEN
      Write('+')
   END ;
   WriteInt(o, 0)
END WriteOffset ;


(*
   WriteConstant - writes a constant from a symbol.
                   Place in SymbolTable - must write out the following consts:

                   (1)  Integer, Cardinal value constants.
                   (2)  Character constants.
                   (3)  String constants.
*)

PROCEDURE WriteConstant (Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
   Str : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF IsConstString(Sym)
   THEN
      Size := GetStringLength(Sym) ;
      GetKey(GetString(Sym), Str) ;
      IF Size=1
      THEN
         Write("'") ;
         IF Str[0]='\'   (* \ must be escaped by \ *)
         THEN
            Write('\')
         END ;
         Write(Str[0])
      ELSIF Size=0
      THEN
         WriteString(' 0x00') ;  (* Empty string *)
      ELSE
         WriteString(Str)
      END
   ELSE
      PushValue(Sym) ;
      WriteInt(PopInt(), 0)
   END
END WriteConstant ;


(*
   WriteInstruction - writes an instruction.
*)

PROCEDURE WriteInstruction (Inst: Instruction; Unit: DatumPrefix) ;
BEGIN
   WriteString('    ') ;
   CASE Inst OF

   mov    : WriteString('mov') ; WriteInstPrefix(Unit) |
   add    : WriteString('add') ; WriteInstPrefix(Unit) |
   sub    : WriteString('sub') ; WriteInstPrefix(Unit) |
   mult   : WriteString('imul') ; WriteInstPrefix(Unit) |
   idiv   : WriteString('idiv') ; WriteInstPrefix(Unit) |
   div    : WriteString('div') ; WriteInstPrefix(Unit) |
   cdq    : WriteString('cdq') ; WriteInstPrefix(Unit) |
   mod    : WriteString('mod') ; WriteInstPrefix(Unit) |
   neg    : WriteString('neg') ; WriteInstPrefix(Unit) |
   push   : WriteString('push') ; WriteInstPrefix(Unit) |
   pop    : WriteString('pop') ; WriteInstPrefix(Unit) |
   jmp    : WriteString('jmp') |
   je     : WriteString('je') |
   jne    : WriteString('jne') |
   jl     : WriteString('jl') |
   jg     : WriteString('jg') |
   jle    : WriteString('jle') |
   jge    : WriteString('jge') |
   jae    : WriteString('jae') |
   cmp    : WriteString('cmp') ; WriteInstPrefix(Unit) |
   repmovs: WriteString('rep') ; WriteLn ;
            WriteString('    movsb %esp:$0, %esp:$0') |
   not    : WriteString('not') ; WriteInstPrefix(Unit) |
   and    : WriteString('and') ; WriteInstPrefix(Unit) |
   or     : WriteString('or') ; WriteInstPrefix(Unit) |
   xor    : WriteString('xor') ; WriteInstPrefix(Unit) |
   test   : WriteString('test') ; WriteInstPrefix(Unit) |
   sal    : WriteString('sal') ; WriteInstPrefix(Unit) |
   ret    : WriteString('ret') |
   call   : WriteString('call') |
   leave  : WriteString('leave')

   ELSE
      WriteError('Instruction not known - expand case clause')
   END
END WriteInstruction ;


(*
   WriteInstPrefix - writes the instruction prefix, which depends
                     upon the Unit of data to be processed by this
                     instruction.
*)

PROCEDURE WriteInstPrefix (Unit: DatumPrefix) ;
BEGIN
   CASE Unit OF

   PByte : Write('b') |
   PWord : Write('w') |
   PDWord: Write('l')

   ELSE
      WriteError('Unrecognised prefix value in case clause')
   END
END WriteInstPrefix ;


(*
   WriteRegister - writes a register.
*)

PROCEDURE WriteRegister (Reg: Register; Unit: DatumPrefix) ;
BEGIN
   Write('%') ;
   CASE Unit OF

   PDWord: CASE Reg OF

           RegEAX: WriteString('eax') |
           RegEBX: WriteString('ebx') |
           RegECX: WriteString('ecx') |
           RegEDX: WriteString('edx') |
           RegEBP: WriteString('ebp') |
           RegESP: WriteString('esp') |
           RegESI: WriteString('esi') |
           RegEDI: WriteString('edi')

           ELSE
              WriteError('Register not known - expand case clause')
           END |

   PWord : CASE Reg OF

           RegEAX: WriteString('ax') |
           RegEBX: WriteString('bx') |
           RegECX: WriteString('cx') |
           RegEDX: WriteString('dx') |
           RegEBP: WriteString('bp') |
           RegESP: WriteString('sp') |
           RegESI: WriteString('si') |
           RegEDI: WriteString('di')

           ELSE
              WriteError('Register not known - expand case clause')
           END |

   PByte : CASE Reg OF

           RegEAX: WriteString('al') |
           RegEBX: WriteString('bl') |
           RegECX: WriteString('cl') |
           RegEDX: WriteString('dl') |
           RegEBP: WriteError('Cannot get byte access ebp') |
           RegESP: WriteError('Cannot get byte access esp') |
           RegESI: WriteError('Cannot get byte access esi') |
           RegEDI: WriteError('Cannot get byte access edi')

           ELSE
              WriteError('Register not known - expand case clause')
           END

   ELSE
      WriteError('DatumPrefix value not known')
   END
END WriteRegister ;


(*
   WriteSize - writes the Size in bytes, but in a human form.
*)

PROCEDURE WriteSize (Size: CARDINAL) ;
BEGIN
   WriteCard(Size, 0)
END WriteSize ;


(*
   InstRegImmName -  instruction immediate name into register.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   $foo, Reg
*)

PROCEDURE InstRegImmName (Inst: Instruction;
                          Unit: DatumPrefix;
                          Reg : Register;
                          Sym : CARDINAL) ;
BEGIN
   WriteInstruction(mov, Unit) ;
   WriteString('  $') ; (* $ means immediate in gas *)
   WriteAsmName(Sym) ;
   WriteString(', ') ;
   WriteRegister(Reg, Unit) ;
   WriteLn
END InstRegImmName ;


(*
   InstName - instruction to operate on a symbol name.
              Inst  - determines the instruction.
              Unit  - determines the DatumPrefix of manipulation
              Sym   - symbol whose name is used as the operand

              Example:

              call  $foo
*)

PROCEDURE InstName (Inst: Instruction;
                    Unit: DatumPrefix;
                    Sym : CARDINAL) ;
BEGIN
   WriteInstruction(Inst, Unit) ;
   WriteString('  ') ;
   WriteAsmName(Sym) ; WriteLn ;
END InstName ;


(*
   MovePackedStringTo - moves a packed string b into the memory location
                        indicated by DestSym.
                        GeneralOffset determines the number of bytes offset
                        from DestSym the packed string is to be placed.
                        Pack determines the number of packed characters
                        must be stored at location DestSym.
                        If the string, b, contains less characters than
                        Pack then the remaining characters must be filled
                        in as nul.
*)

PROCEDURE MovePackedStringTo (DestSym: CARDINAL; GeneralOffset: INTEGER;
                              b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   Offset := GetOffset(DestSym) ;
   IF GetMode(DestSym)=RightValue
   THEN
      WriteInstruction(mov, GetUnitPrefix(Pack)) ;
      WriteString(' $0x') ; WriteHex(Word, 0) ;
      WriteString(', ') ;
      WriteScope(GetScope(DestSym), Offset, GeneralOffset, DestSym) ;
      WriteString(' #  ') ; DisplayString(b, Pack-1) ;
      WriteLn
   ELSE
      Assert(GetMode(DestSym)=LeftValue) ;
      InstRegMD(mov, MaxDatum(), RegEDI, Offset, 0, GetScope(DestSym), DestSym) ;
      WriteInstruction(mov, GetUnitPrefix(Pack)) ;
      WriteString(' $0x') ; WriteHex(Word, 0) ;
      WriteString(', ') ;
      WritePositiveOffset(GeneralOffset) ;
      Write('(') ;
      WriteRegister(RegEDI, MaxDatum()) ;
      Write(')') ;
      WriteString(' #  ') ; DisplayString(b, Pack-1) ;
      WriteLn
   END
END MovePackedStringTo ;


(*
   PushPackedString - moves a packed string b onto the stack.
                      Pack determines the number of packed characters
                      must be stored on the stack.
                      If the string, b, contains less characters than
                      Pack then the remaining characters must be filled
                      in as nul.
*)

PROCEDURE PushPackedString (b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   WriteInstruction(push, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ;
   WriteHex(Word, 0) ;
   WriteString('  # ') ;
   DisplayString(b, Pack-1) ;
   WriteLn
END PushPackedString ;


(*
   PutStringMDReg - moves a packed string a into memory offset, Offset,
                    indexed by register, Reg.
                    Pack determines the number of packed characters
                    to be stored.
                    If the string, b, contains less characters than
                    Pack then the remaining characters must be filled
                    in as nul.
*)

PROCEDURE PutStringMDReg (Offset: INTEGER; Reg: Register;
                          a: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   WriteInstruction(mov, GetUnitPrefix(Pack)) ;
   WriteString(' $0x') ; WriteHex(Word, 0) ;
   WriteString(' ,') ;
   WritePositiveOffset(Offset) ;
   Write('(') ;
   WriteRegister(Reg, MaxDatum()) ;
   WriteString(')  #  ') ;
   DisplayString(a, Pack-1) ;
   WriteLn
END PutStringMDReg ;


(*
   DisplayString - displays a string, a, which contains, n, characters.
*)

PROCEDURE DisplayString (a: ARRAY OF CHAR; n: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   Write('"') ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString('"   ( ') ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         WriteString('<nul>')
      ELSE
         Write(a[i])
      END
   END ;
   WriteString(' )')
END DisplayString ;
*****************************************************************)

END M2Time.
