IMPLEMENTATION MODULE M2ALU ;


FROM ASCII IMPORT nul ;
FROM StrLib IMPORT StrLen ;
FROM NameKey IMPORT GetKey ;
FROM NumberIO IMPORT StrToInt, StrToCard, StrToHex, StrToOct, StrToBin, WriteCard, BinToStr, StrToBinInt ;
FROM FpuIO IMPORT StrToLongReal ;
FROM M2Lexical IMPORT WriteError, InternalError ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM M2Debug IMPORT Assert ;
FROM M2System IMPORT BitsetSize ;
FROM SYSTEM IMPORT BITSET, BYTE, ADR ;

CONST
   MaxStack      =  100 ;
   MaxBitsetSize =    4 ;  (* Max number of bytes for a bitset *)

TYPE
   TypeOfSign = (Negative, Positive) ;
   TypeOfValue= (Integer, Real) ;

   Mem  = RECORD
             Sign : TypeOfSign ;
             Type : TypeOfValue ;
             Value: LONGINT ;
             Float: LONGREAL ;
          END ;

   Cell = RECORD
             Content: Mem ;
             Solved : BOOLEAN ;
          END ;

VAR
   Stack    : ARRAY [0..MaxStack] OF Mem ;
   Cells    : ARRAY PtrToValue OF Cell ;
   CellsUsed: CARDINAL ;
   Ptr      : CARDINAL ;


(* %%%FORWARD%%%
PROCEDURE PushMem (v: Mem) ; FORWARD ;
PROCEDURE PopMem(VAR v: Mem) ; FORWARD ;
PROCEDURE IsReal (a: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE RealSub (Op1, Op2: Mem) ; FORWARD ;
PROCEDURE RealAdd (Op1, Op2: Mem) ; FORWARD ;
PROCEDURE RealMult (Op1, Op2: Mem) ; FORWARD ;
PROCEDURE RealDiv (Op1, Op2: Mem) ; FORWARD ;
PROCEDURE RealMod (Op1, Op2: Mem) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   InitValue - initializes a memory cell. v is set to the index
               of the initialized cell.
*)

PROCEDURE InitValue (VAR v: PtrToValue) ;
BEGIN
   INC(CellsUsed) ;
   IF CellsUsed>MaxCells
   THEN
      InternalError('increase MaxCells', __FILE__, __LINE__)
   ELSE
      WITH Cells[CellsUsed] DO
         Solved := FALSE
      END ;
      v := PtrToValue(CellsUsed)
   END
END InitValue ;


(*
   Pop - pops the top element from the stack and places its value
         into the memory cell indexed by v.
*)

PROCEDURE Pop (v: PtrToValue) ;
BEGIN
   WITH Cells[v] DO
      Solved := TRUE ;
      PopMem(Content)
   END
END Pop ;


(*
   Push - pushes the memory cell value indexed by v onto the stack.
*)

PROCEDURE Push (v: PtrToValue) ;
BEGIN
   WITH Cells[v] DO
      IF Solved
      THEN
         PushMem(Content)
      ELSE
         InternalError('M2ALU - value not declared', __FILE__, __LINE__)
      END
   END
END Push ;


(*
   PushCard - pushes a cardinal onto the stack.
*)
 
PROCEDURE PushCard (c: CARDINAL) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         Type := Integer ;
         Value := VAL(LONGINT, c) ;
         Sign := Positive
      END ;
      INC(Ptr)
   END
END PushCard ;

 
(*
   PopCard - attempts to return the top element from the stack as
             a cardinal.
             If the top element is negative then an error messsage
             is emitted.
*)

PROCEDURE PopCard () : CARDINAL ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      WITH Stack[Ptr] DO
         IF Sign=Negative
         THEN
            WriteError('Cannot convert a negative number into a CARDINAL')
         END ;
         IF Type=Real
         THEN
            WriteError('Cannot pop a REAL number into a CARDINAL')
         END ;
         RETURN( VAL(CARDINAL, Value) )
      END
   END
END PopCard ;
 

(*
   PushInt - pushes a cardinal onto the stack.
*)

PROCEDURE PushInt (i: INTEGER) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      IF i>=0
      THEN
         WITH Stack[Ptr] DO
            Sign := Positive ;
            Type := Integer ;
            Value := VAL(LONGINT, i)
         END
      ELSE
         WITH Stack[Ptr] DO
            Sign := Negative ;
            Type := Integer ;
            Value := -VAL(LONGINT, i)
         END
      END ;
      INC(Ptr)
   END
END PushInt ;

 
(*
   PopInt - attempts to return the top element from the stack as
            an integer.
*)
 
PROCEDURE PopInt () : INTEGER ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack Underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      WITH Stack[Ptr] DO
         IF Type=Real
         THEN
            WriteError('Cannot pop a REAL number into a INTEGER')
         END ;
         IF Sign=Positive
         THEN
            RETURN( VAL(INTEGER, Value) )
         ELSE
            RETURN( VAL(INTEGER, -Value) )
         END
      END
   END
END PopInt ;
 
 
(*
   PushChar - pushes a char onto the stack.
*)
 
PROCEDURE PushChar (c: CHAR) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         Value := VAL(LONGINT, ORD(c)) ;
         Type := Integer ;
         Sign := Positive
      END ;
      INC(Ptr)
   END
END PushChar ;
 
 
(*
   PopChar - attempts to return the top element from the stack as
             a char.
*)

PROCEDURE PopChar () : CHAR ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack Underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      WITH Stack[Ptr] DO
         IF Sign=Negative
         THEN
            WriteError('Cannot convert a negative number into a CHAR')
         END ;
         IF Type=Real
         THEN
            WriteError('Cannot pop a REAL number into a CHAR')
         END ;
         RETURN( VAL(CHAR, (Value MOD 0100H) ) )
      END
   END
END PopChar ;
 

(*
   PushReal - pushes a REAL, r, onto the stack.
*)

PROCEDURE PushReal (r: REAL) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         Float := VAL(LONGREAL, r) ;
         Type := Real ;
         IF r>=0.0
         THEN
            Sign := Positive
         ELSE
            Sign := Negative
         END
      END ;
      INC(Ptr)
   END
END PushReal ;


(*
   PushLongReal - pushes a LONGREAL, r, onto the stack.
*)

PROCEDURE PushLongReal (r: LONGREAL) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      WITH Stack[Ptr] DO
         Float := r ;
         Type := Real ;
         IF r>=0.0
         THEN
            Sign := Positive
         ELSE
            Sign := Negative
         END
      END ;
      INC(Ptr)
   END
END PushLongReal ;


(*
   PopReal - attempts to return the top element from the stack as
             a REAL.
*)

PROCEDURE PopReal () : REAL ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack Underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      WITH Stack[Ptr] DO
         IF Type=Real
         THEN
            RETURN( VAL(REAL, Float) )
         ELSE
            WriteError('Cannot pop a CARDINAL or INTEGER into a REAL number')
         END ;
      END
   END
END PopReal ;


(*
   PopLongReal - attempts to return the top element from the stack as
                 a LONGREAL.
*)

PROCEDURE PopLongReal () : LONGREAL ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack Underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      WITH Stack[Ptr] DO
         IF Type=Real
         THEN
            RETURN( Float )
         ELSE
            WriteError('Cannot pop a CARDINAL or INTEGER into a LONGREAL number')
         END ;
      END
   END
END PopLongReal ;


(*
   PushString - pushes the numerical human readable value of the string
                onto the stack.
*)
 
PROCEDURE PushString (s: CARDINAL) ;
CONST
   Max = 64 ;
VAR
   a      : ARRAY [0..Max] OF CHAR ;
   r      : LONGREAL ;
   i      : INTEGER ;
   v, high: CARDINAL ;
BEGIN
   GetKey(s, a) ;
   high := StrLen(a) ;
   IF a[high-1]='H'
   THEN
      a[high-1] := nul ;
      StrToHex(a, v) ;
      PushCard(v)
   ELSIF a[high-1]='A'
   THEN
      a[high-1] := nul ;
      StrToBin(a, v) ;
      PushCard(v)
   ELSIF a[high-1]='B'
   THEN
      a[high-1] := nul ;
      StrToOct(a, v) ;
      PushCard(v)
   ELSIF a[high-1]='C'
   THEN
      a[high-1] := nul ;
      StrToOct(a, v) ;
      PushCard(v)
   ELSIF IsReal(a)
   THEN
      StrToLongReal(a, r) ;
      PushLongReal(r)
   ELSE
      StrToInt(a, i) ;
      PushInt(i)
   END
END PushString ;


(*
   IsReal - returns TRUE if a is a REAL number.
*)

PROCEDURE IsReal (a: ARRAY OF CHAR) : BOOLEAN ;
VAR
   high,
   i   : CARDINAL ;
BEGIN
   high := StrLen(a) ;
   i := 0 ;
   WHILE i<high DO
      IF a[i]='.'
      THEN
         RETURN( TRUE )
      ELSE
         INC(i)
      END
   END ;
   RETURN( FALSE )
END IsReal ;


(*
   CoerseLongRealToCard - performs a coersion between a REAL to a CARDINAL
*)

PROCEDURE CoerseLongRealToCard ;
VAR
   r : REAL ;
   pa,
   pb: POINTER TO ARRAY [0..3] OF BYTE ;
   c : CARDINAL ;
BEGIN
   r := PopReal() ;
   pa := ADR(r) ;
   pb := ADR(c) ;

   (* this depends upon the endianness of the machine *)
   pb^[0] := pa^[0] ;
   pb^[1] := pa^[1] ;
   pb^[2] := pa^[2] ;
   pb^[3] := pa^[3] ;
   PushCard(c)
END CoerseLongRealToCard ;


(*
   OldCoerseLongRealToCard - performs a coersion between a REAL to a CARDINAL
*)

PROCEDURE OldCoerseLongRealToCard ;
VAR
   r : REAL ;
   pa,
   pb: POINTER TO ARRAY [0..3] OF BYTE ;
BEGIN
   WITH Stack[Ptr-1] DO
      IF Type=Real
      THEN
         r := Float ;
         pa := ADR(r) ;
         pb := ADR(Value) ;
         Value := 0 ;

         (* this depends upon the endianness of the machine *)
         pb^[0] := pa^[0] ;
         pb^[1] := pa^[1] ;
         pb^[2] := pa^[2] ;
         pb^[3] := pa^[3] ;

         Type := Integer ;
         Sign := Positive
      ELSE
         WriteError('expecting a REAL number')
      END
   END
END OldCoerseLongRealToCard ;


(*
   ConvertRealToInt - converts a REAL into an INTEGER
*)

PROCEDURE ConvertRealToInt ;
BEGIN
   WITH Stack[Ptr-1] DO
      IF Type=Real
      THEN
         IF Float<=0.0
         THEN
            Sign := Negative ;
            Value := VAL(LONGINT, TRUNC(-Float))
         ELSE
            Sign := Positive ;
            Value := VAL(LONGINT, TRUNC(Float))
         END ;
         Type := Integer
      ELSE
         WriteError('expecting a REAL number')
      END
   END
END ConvertRealToInt ;


(*
   IsSolved - returns true if the memory cell indicated by v
              has a set value.
*)
 
PROCEDURE IsSolved (v: PtrToValue) : BOOLEAN ;
BEGIN
   RETURN( Cells[v].Solved )
END IsSolved ;


(*
   SignReal - adds the correct sign to Mem, m.
*)

PROCEDURE SignReal (VAR m: Mem) ;
BEGIN
   WITH m DO
      IF Float>=0.0
      THEN
         Sign := Positive
      ELSE
         Sign := Negative
      END
   END
END SignReal ;


(*
   EitherReal - returns TRUE if either, Op1, or, Op2, are Real.
*)

PROCEDURE EitherReal (Op1, Op2: Mem) : BOOLEAN ;
BEGIN
   RETURN( (Op1.Type=Real) OR (Op2.Type=Real) )
END EitherReal ;


(*
   Add - adds the top two elements on the stack.
 
         The Stack:
 
         Entry             Exit
    
  Ptr ->
         +------------+
         | Op1        |                   <- Ptr
         |------------|    +------------+
         | Op2        |    | Op2 + Op1  |
         |------------|    |------------|
*)
 
PROCEDURE Add ;
VAR
   Temp,
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF EitherReal(Op1, Op2)
   THEN
      RealAdd(Op1, Op2)
   ELSIF (Op1.Sign=Negative) AND (Op2.Sign=Positive)
   THEN
      PushMem(Op2) ;
      Op1.Sign := Positive ;   (* flip sign *)
      PushMem(Op1) ;
      Sub
   ELSIF (Op1.Sign=Positive) AND (Op2.Sign=Negative)
   THEN
      PushMem(Op1) ;
      Op2.Sign := Positive ;   (* flip sign *)
      PushMem(Op2) ;
      Sub
   ELSE
      Temp.Sign := Op1.Sign ;
      Temp.Type := Op1.Type ;
      Temp.Value := Op1.Value + Op2.Value ;
      PushMem(Temp)
   END
END Add ;


(*
   RealAdd - adds two numbers. One of which is a Real.
*)

PROCEDURE RealAdd (Op1, Op2: Mem) ;
VAR
   Temp: Mem ;
BEGIN
   IF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      Temp.Float := Op1.Float + Op2.Float ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op1.Type=Real
   THEN
      IF Op2.Sign=Positive
      THEN
         Temp.Float := Op1.Float + VAL(LONGREAL, Op2.Value)
      ELSE
         Temp.Float := Op1.Float - VAL(LONGREAL, Op2.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op2.Type=Real
   THEN
      IF Op1.Sign=Positive
      THEN
         Temp.Float := Op2.Float + VAL(LONGREAL, Op1.Value)
      ELSE
         Temp.Float := Op2.Float - VAL(LONGREAL, Op1.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   END
END RealAdd ;


(*
   Sub - subtracts the top two elements on the stack.
 
         The Stack:
 
         Entry             Exit
 
  Ptr ->
         +------------+
         | Op1        |                   <- Ptr
         |------------|    +------------+
         | Op2        |    | Op2 - Op1  |
         |------------|    |------------|
*)
 
PROCEDURE Sub ;
VAR
   Temp,
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF EitherReal(Op1, Op2)
   THEN
      RealSub(Op1, Op2)
   ELSIF Op1.Sign=Negative
   THEN
      Op1.Sign := Positive ;
      PushMem(Op2) ;
      PushMem(Op1) ;
      Add
   ELSIF Op2.Sign=Negative
   THEN
      Temp.Sign := Negative ;
      Temp.Type := Op1.Type ;
      Temp.Value := Op1.Value + Op2.Value ;
      PushMem(Temp)
   ELSE
      (* Both positive *)
      IF Op2.Value>=Op1.Value
      THEN
         Temp.Sign := Positive ;
         Temp.Type := Op1.Type ;
         Temp.Value := Op2.Value - Op1.Value ;
         PushMem(Temp)
      ELSE
         Temp.Sign := Negative ;
         Temp.Type := Op1.Type ;
         Temp.Value := Op1.Value - Op2.Value ;
         PushMem(Temp)
      END
   END
END Sub ;


(*
   RealSub - subtracts two numbers. One of which is a Real.
*)

PROCEDURE RealSub (Op1, Op2: Mem) ;
VAR
   Temp: Mem ;
BEGIN
   IF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      Temp.Float := Op2.Float - Op1.Float ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op1.Type=Real
   THEN
      IF Op2.Sign=Positive
      THEN
         Temp.Float := VAL(LONGREAL, Op2.Value) - Op1.Float
      ELSE
         Temp.Float := - VAL(LONGREAL, Op2.Value) - Op1.Float
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op2.Type=Real
   THEN
      IF Op1.Sign=Positive
      THEN
         Temp.Float := Op2.Float - VAL(LONGREAL, Op1.Value)
      ELSE
         Temp.Float := Op2.Float + VAL(LONGREAL, Op1.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   END
END RealSub ;


(*
   Mult - multiplies the top two elements on the stack.
 
          The Stack:
 
          Entry             Exit
 
   Ptr ->
          +------------+
          | Op1        |                   <- Ptr
          |------------|    +------------+
          | Op2        |    | Op2 * Op1  |
          |------------|    |------------|
*)

PROCEDURE Mult ;
VAR
   Temp,
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF EitherReal(Op1, Op2)
   THEN
      RealMult(Op1, Op2)
   ELSIF Op1.Sign#Op2.Sign
   THEN
      Temp.Sign := Negative ;
      Temp.Value := Op1.Value * Op2.Value ;
      Temp.Type := Op1.Type ;
      PushMem(Temp)
   ELSE
      Temp.Sign := Positive ;
      Temp.Value := Op1.Value * Op2.Value ;
      Temp.Type := Op1.Type ;
      PushMem(Temp)
   END
END Mult ;


(*
   RealMult - multiplies two numbers. One of which is a Real.
*)

PROCEDURE RealMult (Op1, Op2: Mem) ;
VAR
   Temp: Mem ;
BEGIN
   IF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      Temp.Float := Op1.Float * Op2.Float ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op1.Type=Real
   THEN
      IF Op2.Sign=Positive
      THEN
         Temp.Float := Op1.Float * VAL(LONGREAL, Op2.Value)
      ELSE
         Temp.Float := -Op1.Float * VAL(LONGREAL, Op2.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op2.Type=Real
   THEN
      IF Op1.Sign=Positive
      THEN
         Temp.Float := Op2.Float * VAL(LONGREAL, Op1.Value)
      ELSE
         Temp.Float := -Op2.Float * VAL(LONGREAL, Op1.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   END
END RealMult ;


(*
   Div - divides the top two elements on the stack.

         The Stack:

         Entry             Exit

  Ptr ->
         +------------+
         | Op1        |                     <- Ptr
         |------------|    +--------------+
         | Op2        |    | Op2 DIV Op1  |
         |------------|    |--------------|
*)

PROCEDURE Div ;
VAR
   Temp,
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF EitherReal(Op1, Op2)
   THEN
      RealDiv(Op1, Op2)
   ELSIF Op1.Sign#Op2.Sign
   THEN
      Temp.Sign := Negative ;
      Temp.Value := Op2.Value DIV Op1.Value ;
      Temp.Type := Integer ;
      PushMem(Temp)
   ELSE
      Temp.Sign := Positive ;
      Temp.Value := Op2.Value DIV Op1.Value ;
      Temp.Type := Integer ;
      PushMem(Temp)
   END
END Div ;


(*
   RealDiv - divides two numbers. One of which is a Real.
*)

PROCEDURE RealDiv (Op1, Op2: Mem) ;
VAR
   Temp: Mem ;
BEGIN
   IF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      Temp.Float := Op2.Float / Op1.Float ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op1.Type=Real
   THEN
      IF Op2.Sign=Positive
      THEN
         Temp.Float := VAL(LONGREAL, Op2.Value) / Op1.Float
      ELSE
         Temp.Float := -VAL(LONGREAL, Op2.Value) / Op1.Float
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op2.Type=Real
   THEN
      IF Op1.Sign=Positive
      THEN
         Temp.Float := Op2.Float / VAL(LONGREAL, Op1.Value)
      ELSE
         Temp.Float := -Op2.Float / VAL(LONGREAL, Op1.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   END
END RealDiv ;


(*
   Mod - modulus of the top two elements on the stack.

         The Stack:

         Entry             Exit

  Ptr ->
         +------------+
         | Op1        |                     <- Ptr
         |------------|    +--------------+
         | Op2        |    | Op2 MOD Op1  |
         |------------|    |--------------|
*)

PROCEDURE Mod ;
VAR
   Temp,
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF EitherReal(Op1, Op2)
   THEN
      WriteError('cannot yet perform MOD on REALs') ;
      RealMod(Op1, Op2)
   ELSIF Op1.Sign#Op2.Sign
   THEN
      Temp.Sign := Negative ;
      Temp.Value := Op2.Value MOD Op1.Value ;
      Temp.Type := Integer ;
      PushMem(Temp)
   ELSE
      Temp.Sign := Positive ;
      Temp.Value := Op2.Value MOD Op1.Value ;
      Temp.Type := Integer ;
      PushMem(Temp)
   END
END Mod ;


(*
   RealMod - calculates the modulus between two numbers. One of which is a Real.
*)

PROCEDURE RealMod (Op1, Op2: Mem) ;
VAR
   Temp: Mem ;
BEGIN
(*
   IF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      Temp.Float := Op2.Float MOD Op1.Float ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op1.Type=Real
   THEN
      IF Op2.Sign=Positive
      THEN
         Temp.Float := VAL(LONGREAL, Op2.Value) MOD Op1.Float
      ELSE
         Temp.Float := -VAL(LONGREAL, Op2.Value) MOD Op1.Float
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   ELSIF Op2.Type=Real
   THEN
      IF Op1.Sign=Positive
      THEN
         Temp.Float := Op2.Float MOD VAL(LONGREAL, Op1.Value)
      ELSE
         Temp.Float := Op2.Float MOD -VAL(LONGREAL, Op1.Value)
      END ;
      Temp.Type := Real ;
      SignReal(Temp) ;
      PushMem(Temp)
   END
*)
END RealMod ;


(*
   Equ - returns true if the top two elements on the stack
         are identical.

         The Stack:

         Entry             Exit

  Ptr ->
         +------------+
         | Op1        |  
         |------------|
         | Op2        |  
         |------------|    Empty
 
         RETURN( Op2 = Op1 )
*)
 
PROCEDURE Equ () : BOOLEAN ;
VAR
   Op1, Op2: Mem ;
BEGIN
   PopMem(Op1) ;
   PopMem(Op2) ;
   IF (Op1.Type=Integer) AND (Op2.Type=Integer)
   THEN
      RETURN( (Op1.Value=Op2.Value) AND (Op1.Sign=Op2.Sign) )
   ELSIF (Op1.Type=Real) AND (Op2.Type=Real)
   THEN
      RETURN( Op1.Float=Op2.Float )
   ELSE
      IF Op1.Type=Integer
      THEN
         IF Op1.Sign=Positive
         THEN
            RETURN( FLOAT(Op1.Value) = Op2.Float )
         ELSE
            RETURN( -FLOAT(Op1.Value) = Op2.Float )
         END
      ELSE
         IF Op2.Sign=Positive
         THEN
            RETURN( FLOAT(Op2.Value) = Op1.Float )
         ELSE
            RETURN( -FLOAT(Op2.Value) = Op1.Float )
         END
      END
   END
END Equ ;


(*
   NotEqu - returns true if the top two elements on the stack
            are not identical.
 
            The Stack:
 
            Entry             Exit
 
     Ptr ->
            +------------+
            | Op1        |
            |------------|
            | Op2        |
            |------------|    Empty
 
            RETURN( Op2 # Op1 )
*)

PROCEDURE NotEqu () : BOOLEAN ;
BEGIN
   RETURN( NOT Equ() )
END NotEqu ;


(*
   Less - returns true if Op2 > Op1
          are not identical.
 
          The Stack:
 
          Entry             Exit
 
   Ptr ->
          +------------+
          | Op1        | 
          |------------|
          | Op2        | 
          |------------|    Empty
 
          RETURN( Op2 > Op1 )
*)
 
PROCEDURE Less () : BOOLEAN ;
VAR
   Result: Mem ;
BEGIN
   Sub ;
   PopMem(Result) ;
   RETURN( Result.Sign=Negative )
END Less ;


(*
   Gre - returns true if Op2 > Op1
         are not identical.

         The Stack:

         Entry             Exit

  Ptr ->
         +------------+
         | Op1        |  
         |------------|
         | Op2        |  
         |------------|    Empty

         RETURN( Op2 > Op1 )
*)

PROCEDURE Gre () : BOOLEAN ;
VAR
   Result: Mem ;
BEGIN
   Sub ;
   PopMem(Result) ;
   RETURN(
          (Result.Sign=Positive) AND (((Result.Type=Integer) AND (Result.Value>0)) OR
                                      ((Result.Type=Real) AND (Result.Float>0.0)))
          )
END Gre ;


(*
   LessEqu - returns true if Op2<Op1
             are not identical.

            The Stack:

            Entry             Exit

     Ptr ->
            +------------+
            | Op1        |
            |------------|
            | Op2        |
            |------------|    Empty

            RETURN( Op2 < Op1 )
*)

PROCEDURE LessEqu () : BOOLEAN ;
VAR
   Result: Mem ;
BEGIN
   Sub ;
   PopMem(Result) ;
   RETURN(
          (Result.Sign=Negative) OR (((Result.Type=Integer) AND (Result.Value=0)) OR
                                     ((Result.Type=Real) AND (Result.Float=0.0)))
         )
END LessEqu ;


(*
   GreEqu - returns true if Op2 >= Op1
            are not identical.
 
            The Stack:
 
            Entry             Exit
 
     Ptr ->
            +------------+
            | Op1        |
            |------------|
            | Op2        |
            |------------|    Empty
 
            RETURN( Op2 >= Op1 )
*)

PROCEDURE GreEqu () : BOOLEAN ;
VAR
   Result: Mem ;
BEGIN
   Sub ;
   PopMem(Result) ;
   RETURN( Result.Sign=Positive )
END GreEqu ;


(*
   Bit -  performs 1 shifted right by the value of the number on the stack.

                 The Stack:

                 Entry             Exit

          Ptr ->                                 <- Ptr

                 +------------+    +------------+
                 | Op1        |    | 1 << Op1   |
                 |------------|    |------------|

*)

PROCEDURE Bit ;
VAR
   Op1: INTEGER ;
   c  : CARDINAL ;
BEGIN
   Op1 := PopInt() ;
   c := 1 ;
   IF Op1<BitsetSize*8
   THEN
      WHILE Op1>0 DO
         c := c*2 ;
         DEC(Op1)
      END
   ELSE
      WriteError('Bit value exceeds wordlength')
   END ;
   PushCard(c)
END Bit ;


(*
   BitRange -  evaluates the mask defined by Op1..Op2

                    The Stack:

                    Entry             Exit

               Ptr ->
                      +------------+
                      | Op1        |                   <- Ptr
                      |------------|    +------------+
                      | Op2        |    | Op1..Op2   |
                      |------------|    |------------|
*)

PROCEDURE BitRange ;
VAR
   i,
   Op1,
   Op2: CARDINAL ;
   c  : CARDINAL ;
BEGIN
   Op1 := PopCard() ;
   Op2 := PopCard() ;
   IF Op1>Op2
   THEN
      c := Op2 ;
      Op2 := Op1 ;
      Op1 := c
   END ;
   IF Op2<BitsetSize*8
   THEN
      (* in correct bitset range *)
      c := 0 ;
      FOR i := Op1 TO Op2 DO
         PushCard(i) ;
         Bit ;
         INC(c, PopCard())
      END ;
      PushCard(c)
   ELSE
      WriteError('bitrange exceeds wordlength')
   END
END BitRange ;


(*
   SetIn - returns true if the Op2 IN Op1

           The Stack:

           Entry             Exit

    Ptr ->
           +------------+
           | Op1        |
           |------------|
           | Op2        |
           |------------|    Empty

           RETURN( Op2 IN Op1 )
*)

PROCEDURE SetIn () : BOOLEAN ;
VAR
   Op1,
   Op2 : INTEGER ;
   res,
   i   : CARDINAL ;
   a, b: ARRAY [0..MaxBitsetSize*8] OF CHAR ;
BEGIN
   Assert(BitsetSize<=MaxBitsetSize) ;
   Op1 := PopInt() ;
   Op2 := PopInt() ;
   BinToStr(CARDINAL(Op1), 32, b) ;
   BinToStr(CARDINAL(Op2), 32, a) ;
   FOR i := 0 TO BitsetSize*8-1 DO
      IF (a[i]='1') AND (b[i]#'1')
      THEN
         RETURN( FALSE )
      END
   END ;
   RETURN( TRUE )
END SetIn ;


(*
   SetOr -  performs an inclusive OR of the top two elements on the stack.

            The Stack:

            Entry             Exit

     Ptr ->
            +------------+
            | Op1        |                   <- Ptr
            |------------|    +------------+
            | Op2        |    | Op2 + Op1  |
            |------------|    |------------|

*)

PROCEDURE SetOr ;
VAR
   Op1,
   Op2: INTEGER ;
   res: INTEGER ;
   i  : CARDINAL ;
   a, b, c: ARRAY [0..MaxBitsetSize*8] OF CHAR ;
BEGIN
   Assert(BitsetSize<=MaxBitsetSize) ;
   Op1 := PopInt() ;
   Op2 := PopInt() ;
   BinToStr(CARDINAL(Op1), 32, a) ;
   BinToStr(CARDINAL(Op2), 32, b) ;
   FOR i := 0 TO BitsetSize*8-1 DO
      IF (a[i]='1') OR (b[i]='1')
      THEN
         c[i] := '1'
      ELSE
         c[i] := '0'
      END
   END ;
   c[32] := nul ;
   StrToBinInt(c, res) ;
   PushInt(res)
END SetOr ;


(*
   SetAnd - performs a set AND the top two elements on the stack.

            The Stack:

            Entry             Exit

     Ptr ->
            +------------+
            | Op1        |                   <- Ptr
            |------------|    +------------+
            | Op2        |    | Op2 * Op1  |
            |------------|    |------------|
*)

PROCEDURE SetAnd ;
VAR
   Op1,
   Op2: INTEGER ;
   res: INTEGER ;
   i  : CARDINAL ;
   a, b, c: ARRAY [0..MaxBitsetSize*8] OF CHAR ;
BEGIN
   Assert(BitsetSize<=MaxBitsetSize) ;
   Op1 := PopInt() ;
   Op2 := PopInt() ;
   BinToStr(CARDINAL(Op1), 32, a) ;
   BinToStr(CARDINAL(Op2), 32, b) ;
   FOR i := 0 TO BitsetSize*8-1 DO
      IF (a[i]='1') AND (b[i]='1')
      THEN
         c[i] := '1'
      ELSE
         c[i] := '0'
      END
   END ;
   c[32] := nul ;
   StrToBinInt(c, res) ;
   PushInt(res)
END SetAnd ;


(*
   SetDifference - performs a set difference of the top two elements on the stack.

                   The Stack:

                   Entry             Exit

            Ptr ->
                   +------------+
                   | Op1        |                   <- Ptr
                   |------------|    +------------+
                   | Op2        |    | Op2 - Op1  |
                   |------------|    |------------|
*)

PROCEDURE SetDifference ;
VAR
   Op1,
   Op2: INTEGER ;
   res: INTEGER ;
   i  : CARDINAL ;
   a, b, c: ARRAY [0..MaxBitsetSize*8] OF CHAR ;
BEGIN
   Assert(BitsetSize<=MaxBitsetSize) ;
   Op1 := PopInt() ;
   Op2 := PopInt() ;
   BinToStr(CARDINAL(Op1), 32, a) ;
   BinToStr(CARDINAL(Op2), 32, b) ;
   FOR i := 0 TO BitsetSize*8-1 DO
      IF (a[i]='0') AND (b[i]='1')
      THEN
         c[i] := '1'
      ELSE
         c[i] := '0'
      END
   END ;
   c[32] := nul ;
   StrToBinInt(c, res) ;
   PushInt(res)
END SetDifference ;


(*
   SetSymmetricDifference - performs a set difference of the top two elements on the stack.

                            The Stack:

                            Entry             Exit

                     Ptr ->
                            +------------+
                            | Op1        |                   <- Ptr
                            |------------|    +------------+
                            | Op2        |    | Op2 - Op1  |
                            |------------|    |------------|
*)

PROCEDURE SetSymmetricDifference ;
VAR
   Op1,
   Op2: INTEGER ;
   res: INTEGER ;
   i  : CARDINAL ;
   a, b, c: ARRAY [0..MaxBitsetSize*8] OF CHAR ;
BEGIN
   Assert(BitsetSize<=MaxBitsetSize) ;
   Op1 := PopInt() ;
   Op2 := PopInt() ;
   BinToStr(CARDINAL(Op1), 32, a) ;
   BinToStr(CARDINAL(Op2), 32, b) ;
   FOR i := 0 TO BitsetSize*8-1 DO
      IF a[i]#b[i]
      THEN
         c[i] := '1'
      ELSE
         c[i] := '0'
      END
   END ;
   c[32] := nul ;
   StrToBinInt(c, res) ;
   PushInt(res)
END SetSymmetricDifference ;


(*
   PushMem - pushes the contents of v onto the stack.
*)

PROCEDURE PushMem (v: Mem) ;
BEGIN
   IF Ptr=MaxStack
   THEN
      InternalError('MaxStack exceeded', __FILE__, __LINE__)
   ELSE
      Stack[Ptr] := v ;
      INC(Ptr)
   END
END PushMem ;


(*
   PopMem - pops the top element from the stack into v.
*)

PROCEDURE PopMem(VAR v: Mem) ;
BEGIN
   IF Ptr=0
   THEN
      InternalError('Stack Underflow', __FILE__, __LINE__)
   ELSE
      DEC(Ptr) ;
      v := Stack[Ptr]
   END
END PopMem ;


(*
   Init - initialises the stack and the number of memory cells used
          to empty.
*)

PROCEDURE Init ;
BEGIN
   CellsUsed := 0 ;
   Ptr       := 0
END Init ;


BEGIN
   Init
END M2ALU.
