IMPLEMENTATION MODULE M2Alias ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT InternalError ;
FROM Storage IMPORT ALLOCATE ;
FROM StdIO IMPORT Write ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM SymbolTable IMPORT NulSym, GetSymName ;
FROM NameKey IMPORT WriteKey ;


CONST
   MaxSyms  = 100 ;

TYPE
   AliasList = POINTER TO AList ;

   (*
      AList has two main components: List and Syms.

            List is contains the alias through its pairing of indices in
                 the To and From fields.

            Syms contains the actual symbols that the List indices reference
                 together with their mode.
   *)

   SymIndex  = [0..MaxSyms] ;

   Alias     = RECORD
                  From, To: SymIndex ;    (* index into syms, showing the alias        *)
               END ;

   AList     = RECORD
                  List    : ARRAY [1..MaxAlias] OF Alias ;
                  AliasPtr: CARDINAL ;    (* points to the top of the List array       *)
                  Syms    : ARRAY [1..MaxSyms] OF CARDINAL ;  (* SymbolTable Ids       *)
                  SymPtr  : SymIndex ;    (* points to the top of the Syms array       *)
                  Next    : AliasList ;   (* used to store old lists on the free queue *)
               END ;

VAR
   FreeList: AliasList ;


(* %%%FORWARD%%%
PROCEDURE PutAlias (a: AliasList; f, t: SymIndex) ; FORWARD ;
PROCEDURE GetSym (a: AliasList; Sym: CARDINAL) : SymIndex ; FORWARD ;
PROCEDURE CollectAliases (a: AliasList; VAR l: ARRAY OF AliasType; Sym: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE AddTo (VAR a: ARRAY OF SymIndex; VAR High: CARDINAL; s: SymIndex) ; FORWARD ;
PROCEDURE DumpAliases (a: AliasList) ; FORWARD ;
PROCEDURE CollectAlias (a: AliasList;
                        VAR g: ARRAY OF SymIndex; VAR High: CARDINAL;
                        s: SymIndex) ; FORWARD ;
PROCEDURE CombineAlias (a: AliasList; f, t: SymIndex) ; FORWARD ;
   %%%FORWARD%%% *)


(*
   DumpAliases - 
*)

PROCEDURE DumpAliases (a: AliasList) ;
VAR
   al        : ARRAY [0..MaxAlias] OF CARDINAL ;
   b         : ARRAY [0..MaxAlias] OF CARDINAL ;
   i, j, n, m: CARDINAL ;
BEGIN
   WriteLn ;
   WriteString('Alias dump') ; WriteLn ;
   n := CollectAllAliases(a, al) ;
   i := 0 ;
   WHILE i<n DO
      m := CollectAliases(a, b, al[i]) ;
      j := 0 ;
      WHILE j<m DO
         WriteKey(GetSymName(al[j])) ; WriteString('  ') ;
         INC(j)
      END ;
      WriteLn ;
      INC(i)
   END
END DumpAliases ;


(*
   Init - initializes module data structures.
*)

PROCEDURE Init ;
BEGIN
   FreeList := NIL ;
END Init ;


(*
   New - creates a new AliasList entity either from the heap or from the, FreeList.
*)

PROCEDURE New () : AliasList ;
VAR
   a: AliasList ;
BEGIN
   IF FreeList=NIL
   THEN
      NEW(a)
   ELSE
      a := FreeList ;
      FreeList := FreeList^.Next
   END ;
   Assert(a#NIL) ;
   RETURN( a )
END New ;


(*
   InitAlias - creates a new empty AliasList and returns it.
*)

PROCEDURE InitAlias () : AliasList ;
VAR
   a: AliasList ;
BEGIN
   a := New() ;
   WITH a^ DO
      AliasPtr := 0 ;
      SymPtr := 0
   END ;
   RETURN( a )
END InitAlias ;


(*
   KillAlias - destroys an AliasList.
*)

PROCEDURE KillAlias (a: AliasList) : AliasList ;
BEGIN
   a^.Next := FreeList ;
   FreeList := a ;
   RETURN( NIL )
END KillAlias ;


(*
   DuplicateAlias - duplicates an AliasList and returns the duplicate.
*)

PROCEDURE DuplicateAlias (a: AliasList) : AliasList ;
VAR
   b: AliasList ;
   i: CARDINAL ;
   j: CARDINAL ;
BEGIN
   b := New() ;
   (* it may well be faster simply to perform b^ := a^ ?? *)
   i := a^.AliasPtr ;
   b^.AliasPtr := i ;
   WHILE i>0 DO
      b^.List[i] := a^.List[i] ;
      DEC(i)
   END ;
   j := a^.SymPtr ;
   b^.SymPtr := j ;
   WHILE j>0 DO
      b^.Syms[j] := a^.Syms[j] ;
      DEC(j)
   END ;
   RETURN( b )
END DuplicateAlias ;


(*
   AliasSymbol - alias a symbol [FromSymbol, FromMode] -> [ToSymbol, ToMode]
*)

PROCEDURE AliasSymbol (a: AliasList;
                       FromSymbol: CARDINAL; ToSymbol: CARDINAL) ;
VAR
   t, f: SymIndex ;
BEGIN
   IF FromSymbol#ToSymbol
   THEN
      f := GetSym(a, FromSymbol) ;
      t := GetSym(a, ToSymbol) ;
      PutAlias(a, f, t)
(*    ; DumpAliases(a) *)
   END
END AliasSymbol ;


(*
   PutAlias - places an alias between symbol id, f, to symbol id, t.
              f -> t
*)

PROCEDURE PutAlias (a: AliasList; f, t: SymIndex) ;
VAR
   o, i: CARDINAL ;
BEGIN
   IF f#t
   THEN
      i := 1 ;
      o := 0 ;
      WITH a^ DO
         WHILE i<=AliasPtr DO
            IF (List[i].From=f) AND (List[i].To=t)
            THEN
               (* found alias - we dont want another alias so we *)
               RETURN
            ELSE
               IF (o=0) AND (List[i].From=NulSym) AND (List[i].To=NulSym)
               THEN
                  o := i
               END ;
               INC(i)
            END
         END ;
         (* alias not found - we will create a new one *)
         IF o=0
         THEN
            INC(AliasPtr) ;
            IF AliasPtr>MaxAlias
            THEN
               InternalError('limit reached in M2Alias - increase this constant', __FILE__, __LINE__)
            ELSE
               WITH List[AliasPtr] DO
                  From := f ;
                  To := t
               END ;
               CombineAlias(a, f, t)
            END
         ELSE
            (* reuse an old alias *)
            WITH List[o] DO
               From := f ;
               To := t
            END ;
            CombineAlias(a, f, t)
         END
      END
   END
END PutAlias ;


(*
   CombineAlias - combines all aliases between, f, and, t.
*)

PROCEDURE CombineAlias (a: AliasList; f, t: SymIndex) ;
VAR
   gf, gt   : ARRAY [0..MaxAlias] OF SymIndex ;
   i, ht, hf: CARDINAL ;
BEGIN
   ht := 0 ;
   CollectAlias(a, gt, ht, t) ;
   hf := 0 ;
   CollectAlias(a, gf, hf, f) ;
   i := 0 ;
   WHILE i<ht DO
      PutAlias(a, gt[i], f) ;
      INC(i)
   END ;
   i := 0 ;
   WHILE i<hf DO
      PutAlias(a, gf[i], t) ;
      INC(i)
   END
END CombineAlias ;


(*
   GetSym - returns the SymIndex to a symbol, [Sym:m].
*)

PROCEDURE GetSym (a: AliasList; Sym: CARDINAL) : SymIndex ;
VAR
   i: SymIndex ;
BEGIN
   WITH a^ DO
      i := 1 ;
      WHILE i<=SymPtr DO
         IF Syms[i]=Sym
         THEN
            (* found symbol - it was placed here before *)
            RETURN( i )
         ELSE
            INC(i)
         END
      END ;
      (* have not found symbol, we will add it to our list *)
      INC(SymPtr) ;
      IF SymPtr>MaxSyms
      THEN
         InternalError('MaxSyms - limit reached - increase this constant', __FILE__, __LINE__)
      ELSE
         Syms[SymPtr] := Sym
      END ;
      RETURN( SymPtr )
   END
END GetSym ;


(*
   RemoveAliasTo - removes all alias' to a symbol in the AliasList.
*)

PROCEDURE RemoveAliasTo (a: AliasList; Sym: CARDINAL) ;
VAR
   i: CARDINAL ;
   s: SymIndex ;
BEGIN
   s := GetSym(a, Sym) ;
   WITH a^ DO
      i := 1 ;
      WHILE i<=AliasPtr DO
         IF (List[i].To=s) OR (List[i].From=s)
         THEN
            List[i].To := NulSym ;
            List[i].From := NulSym
         END ;
         INC(i)
      END
   END
END RemoveAliasTo ;


(*
   CollectAliases - collect all aliases to Sym and Mode into the list, l.
*)

PROCEDURE CollectAliases (a: AliasList; VAR l: ARRAY OF CARDINAL;
                          Sym: CARDINAL) : CARDINAL ;
VAR
   s: SymIndex ;
   b: ARRAY [0..MaxAlias] OF SymIndex ;
   i, h: CARDINAL ;
BEGIN
   s := GetSym(a, Sym) ;
   h := 0 ;
   CollectAlias(a, b, h, s) ;
   i := 0 ;
   WHILE i<h DO
      l[i] := a^.Syms[b[i]] ;
      CollectAlias(a, b, h, GetSym(a, l[i])) ;
      INC(i)
   END ;
   RETURN( h )
END CollectAliases ;


(*
   CollectAllAliases - collect all aliases in, a.
*)

PROCEDURE CollectAllAliases (a: AliasList; VAR l: ARRAY OF CARDINAL) : CARDINAL ;
VAR
   i, h: CARDINAL ;
BEGIN
   WITH a^ DO
      h := CARDINAL(SymPtr) ;
      i := 0 ;
      WHILE i<h DO
         l[i] := Syms[i+1] ;
         INC(i)
      END
   END ;
   RETURN( h )
END CollectAllAliases ;


(*
   CollectAlias - collects all aliases from, a, into array, g, for a symbol, s.
*)

PROCEDURE CollectAlias (a: AliasList;
                        VAR g: ARRAY OF SymIndex; VAR High: CARDINAL;
                        s: SymIndex) ;
VAR
   i: CARDINAL ;
BEGIN
   AddTo(g, High, s) ;  (* s can always be an alias to itself *)
   WITH a^ DO
      i := 1 ;
      WHILE i<=AliasPtr DO
         IF List[i].To=s
         THEN
            AddTo(g, High, List[i].From)
         ELSIF List[i].From=s
         THEN
            AddTo(g, High, List[i].To)
         END ;
         INC(i)
      END
   END
END CollectAlias ;


(*
   AddTo - add symbol index, s, to the array, a, incrementing the pointer, High.
*)

PROCEDURE AddTo (VAR a: ARRAY OF SymIndex; VAR High: CARDINAL; s: SymIndex) ;
VAR
   i: CARDINAL ;
BEGIN
   (* just check to make sure, s, is not already in the array *)
   i := 0 ;
   WHILE i<High DO
      IF a[i]=s
      THEN
         (* yes it is here, we will return as we have nothing else to do *)
         RETURN
      ELSE
         INC(i)
      END
   END ;
   (* no, s, is not currently inside, a *)
   IF High=HIGH(a)
   THEN
      InternalError('increase array size', __FILE__, __LINE__)
   ELSE
      a[High] := s ;
      INC(High)
   END
END AddTo ;


(*
   IsAliased - returns TRUE whether there exists an alias in, a, which matches
               [FromSymbol, FromMode] -> [ToSymbol, ToMode]
*)

PROCEDURE IsAliased (a: AliasList;
                     FromSymbol: CARDINAL; ToSymbol: CARDINAL) : BOOLEAN ;
VAR
   f, t: SymIndex ;
   i   : CARDINAL ;
BEGIN
   f := GetSym(a, FromSymbol) ;
   t := GetSym(a, ToSymbol) ;
   WITH a^ DO
      i := 1 ;
      WHILE i<=AliasPtr DO
         IF ((List[i].From=f) AND (List[i].To=t)) OR
            ((List[i].From=t) AND (List[i].To=f))
         THEN
            RETURN( TRUE )
         END ;
         INC(i)
      END
   END ;
   RETURN( FALSE )
END IsAliased ;


END M2Alias.
