IMPLEMENTATION MODULE M2Comp ;


FROM M2Options IMPORT QuietString, QuietLn, QuietName, Statistics ;

FROM M2Pass IMPORT SetPassToPass1, SetPassToPass2, SetPassToPass3,
                   SetPassToNoPass ;

FROM M2Search IMPORT FindSourceFile ;
FROM M2Code IMPORT Code ;
FROM M2Lexical IMPORT OpenSource, CloseSource, WriteError, WriteErrorFormat1, NearToken ;
FROM M2FileName IMPORT CalculateFileName ;

IMPORT P1Compile ;
IMPORT P2Compile ;
IMPORT P3Compile ;
IMPORT PHCompile ;

FROM M2Batch IMPORT GetSource, GetModuleNo ;

FROM SymbolTable IMPORT GetSymName, IsDefImp, NulSym,
                        IsHiddenTypeDeclared, GetFirstUsed,
                        FinalSymbol ;

FROM NameKey IMPORT GetKey ;
FROM NumberIO IMPORT CardToStr ;
FROM StrLib IMPORT StrConCat ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM StdIO IMPORT Write ;


CONST
   MaxStringLen = 79 ;


VAR
   ModuleType : (None, Definition, Implementation, Program) ;


(* %%%FORWARD%%%
PROCEDURE DoPass1 (Main: CARDINAL) ; FORWARD ;
PROCEDURE DoPass2 (Main: CARDINAL) ; FORWARD ;
PROCEDURE DoPass3 (Main: CARDINAL) ; FORWARD ;
   %%%FORWARD%%% *)

(*
   CompilingDefinitionModule - returns true if the current module being
                               compiled is a definition module.
*)

PROCEDURE CompilingDefinitionModule() : BOOLEAN ;
BEGIN
   RETURN( ModuleType=Definition )
END CompilingDefinitionModule ;


(*
   CompilingImplementationModule - returns true if the current module being
                                   compiled is an implementation module.
*)

PROCEDURE CompilingImplementationModule() : BOOLEAN ;
BEGIN
   RETURN( ModuleType=Implementation )
END CompilingImplementationModule ;


(*
   CompilingProgramModule - returns true if the current module being
                            compiled is a program module.
*)

PROCEDURE CompilingProgramModule() : BOOLEAN ;
BEGIN
   RETURN( ModuleType=Program )
END CompilingProgramModule ;


(*
   EmitSymbolStatistics - generates a simple report on the total
                          number of symbols currently used during compilation.
*)

PROCEDURE EmitSymbolStatistics (a: ARRAY OF CHAR) ;
VAR
   string: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   IF Statistics
   THEN
      QuietString(a) ;
      CardToStr(FinalSymbol(), 0, string) ;
      QuietString(' (') ;
      QuietString(string) ;
      QuietString(')') ;
      QuietLn
   END
END EmitSymbolStatistics ;


(*
   Compile - compiles the required modules by using a 3 pass compilation
             technique. Code generation is achieved by translating quadruples
             into assembly language.
*)

PROCEDURE Compile (Sym: CARDINAL) ;
BEGIN
   QuietString('Pass 1') ; QuietLn ;
   DoPass1(Sym) ;
   QuietString('Pass 2') ; QuietLn ;
   DoPass2(Sym) ;
   QuietString('Pass 3') ; QuietLn ;
   DoPass3(Sym) ;
   QuietString('Pass 4') ; QuietLn ;
   Code ;
   EmitSymbolStatistics('Total number of symbols used       ')
END Compile ;


(*
   DoPass1 - parses the sources of all modules necessary to compile
             the required module, Main.
*)

PROCEDURE DoPass1 (Main: CARDINAL) ;
VAR
   Sym : CARDINAL ;
   n, i: CARDINAL ;
   Name,
   FileName: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   SetPassToPass1 ;
   i := 1 ;
   Sym := GetModuleNo(i) ;
   WHILE Sym#NulSym DO
      n := GetSymName(Sym) ;
      GetKey(n, Name) ;
      (*
         Attempt to open <Name>.def, cannot test IsDefImp because it may not
         be set yet.
      *)
      CalculateFileName(Name, 'def', FileName) ;
      IF OpenSource(FileName)
      THEN
         QuietString('   Module ') ; QuietName(Name) ; QuietString(' : ') ;
      	 IF FindSourceFile(FileName, FileName)
      	 THEN
            QuietString(FileName) ; QuietLn
      	 ELSE
      	    WriteError('inconsistancy between opening file and finding file')
      	 END ;
         ModuleType := Definition ;
         IF NOT P1Compile.CompilationUnit()
         THEN
            WriteError('Compilation Failed')
         END ;
         CloseSource
      ELSIF Sym#Main
      THEN
         NearToken('', GetFirstUsed(Sym)) ;
         WriteErrorFormat1('cannot find definition module %s.def', n)
      ELSE
         ModuleType := Program
      END ;
      IF (Main=Sym) OR (IsDefImp(Sym) AND IsHiddenTypeDeclared(Sym))
      THEN
      	 CalculateFileName(Name, 'mod', FileName) ;
         IF OpenSource(FileName)
         THEN
            QuietString('   Module ') ; QuietName(Name) ; QuietString(' : ') ;
            IF FindSourceFile(FileName, FileName)
            THEN
               QuietString(FileName) ; QuietLn
            ELSE
               WriteError('inconsistancy between opening file and finding file')
            END ;
            IF ModuleType=Definition
            THEN
               ModuleType := Implementation
            ELSE
               ModuleType := Program
            END ;
            IF NOT P1Compile.CompilationUnit()
            THEN
               WriteError('compilation failed')
            END ;
            CloseSource
         ELSE
            NearToken('', GetFirstUsed(Sym)) ;
            WriteErrorFormat1('module %s.mod cannot be found', n)
         END
      END ;
      INC(i) ;
      Sym := GetModuleNo(i)
   END ;
   SetPassToNoPass
END DoPass1 ;


(*
   DoPass2 - parses the sources of all modules necessary to compile
             the required module, Main.
*)

PROCEDURE DoPass2 (Main: CARDINAL) ;
VAR
   Sym : CARDINAL ;
   i,
   n   : CARDINAL ;
   Name,
   FileName: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   SetPassToPass2 ;
   i := 1 ;
   Sym := GetModuleNo(i) ;
   WHILE Sym#NulSym DO
      n := GetSymName(Sym) ;
      GetKey(n, Name) ;
      IF IsDefImp(Sym)
      THEN
      	 CalculateFileName(Name, 'def', FileName) ;
         IF OpenSource(FileName)
         THEN
            ModuleType := Definition ;
            IF NOT P2Compile.CompilationUnit()
            THEN
               WriteError('compilation Failed')
            END ;
            CloseSource
         ELSE
            NearToken('', GetFirstUsed(Sym)) ;
            WriteErrorFormat1('module %s.def cannot be found', n)
         END
      ELSE
         ModuleType := Program
      END ;
      IF (Main=Sym) OR (IsDefImp(Sym) AND IsHiddenTypeDeclared(Sym))
      THEN
      	 CalculateFileName(Name, 'mod', FileName) ;
         IF OpenSource(FileName)
         THEN
            IF ModuleType=Definition
            THEN
               ModuleType := Implementation
            ELSE
               ModuleType := Program
            END ;
            IF NOT P2Compile.CompilationUnit()
            THEN
               WriteError('Compilation Failed')
            END ;
            CloseSource
         ELSE
            NearToken('', GetFirstUsed(Sym)) ;
            WriteErrorFormat1('module %s.mod cannot be found', n)
         END
      END ;
      INC(i) ;
      Sym := GetModuleNo(i)
   END ;
   SetPassToNoPass
END DoPass2 ;


(*
   DoPass3 - parses the sources of all modules necessary to compile
             the required module, Main.
*)

PROCEDURE DoPass3 (Main: CARDINAL) ;
VAR
   Sym : CARDINAL ;
   i,
   n   : CARDINAL ;
   Name,
   FileName: ARRAY [0..MaxStringLen] OF CHAR ;
BEGIN
   SetPassToPass3 ;
   i := 1 ;
   Sym := GetModuleNo(i) ;
   WHILE Sym#NulSym DO
      n := GetSymName(Sym) ;
      GetKey(n, Name) ;
      IF IsDefImp(Sym)
      THEN
      	 CalculateFileName(Name, 'def', FileName) ;
         IF OpenSource(FileName)
         THEN
            ModuleType := Definition ;
            IF NOT P3Compile.CompilationUnit()
            THEN
               WriteError('compilation failed')
            END ;
            CloseSource
         ELSE
            NearToken('', GetFirstUsed(Sym)) ;
            WriteErrorFormat1('module %s.def cannot be found', n)
         END
      ELSE
         ModuleType := Program
      END ;
      IF (Main=Sym) OR (IsDefImp(Sym) AND IsHiddenTypeDeclared(Sym))
      THEN
      	 CalculateFileName(Name, 'mod', FileName) ;
         IF OpenSource(FileName)
         THEN
            IF ModuleType=Definition
            THEN
               ModuleType := Implementation
            ELSE
               ModuleType := Program
            END ;
            IF Main=Sym
            THEN
               IF NOT P3Compile.CompilationUnit()
               THEN
                  WriteError('compilation failed')
               END
            ELSE
               (*
                  not the main module .mod therefore must be implementing
                  a hidden type - we dont want to generate any
                  StatementSequence quadrupes but we do want to build TYPEs
                  and ConstExpressions.
               *)
               IF NOT PHCompile.CompilationUnit()
               THEN
                  WriteError('compilation failed')
               END
            END ;
            CloseSource
         ELSE
            NearToken('', GetFirstUsed(Sym)) ;
            WriteErrorFormat1('module %s.mod cannot be found', n)
         END
      END ;
      INC(i) ;
      Sym := GetModuleNo(i)
   END ;
   SetPassToNoPass
END DoPass3 ;


BEGIN
   ModuleType := None
END M2Comp.
