IMPLEMENTATION MODULE M2Pipe586 ;


(*
   This module provides the same interface as M2AS, but
   internally it performs pipeline scheduling and peephole
   optimization and also it sums the pentium cycles.
   The amount of cycles required by an i586 is then used
   by the is32 code generator when we emit workload models
   for the ismene simulator.
*)

FROM SYSTEM IMPORT BITSET ;
FROM NumberIO IMPORT WriteCard, WriteInt, WriteHex, WriteBin ;
FROM M2AssemIO IMPORT OpenAssemblyFile, CloseAssemblyFile ;
FROM SymbolTable IMPORT NulSym, ModeOfAddr, IsConst, GetScopeAuthor, IsProcedure, GetMode, GetSymName, IsVarAParam ;
FROM M2Debug IMPORT Assert ;
FROM NameKey IMPORT WriteKey ;

FROM i486 IMPORT DWORDSIZE, WORDSIZE, BYTESIZE,
                 GetScope, GetOffset, GetSize, GetUnitPrefix,
                 GetUnitOfTransfer, GetSizeOfPrefix,
                 MaxDatum,
                 LowReg, HighReg, IsRegister,
                 EAX, EBX, ECX, EDX,
                 ESI, EDI, ESP, EBP ;

FROM M2Options IMPORT Optimize586, Optimize686, OptimizeCache, Profiling,
                      APUCodeGenerator, OptimizeLength, OptimizeWorst, Pedantic,
                      IgnorePipeLine ;

FROM StrIO IMPORT WriteString, WriteLn ;
FROM M2Lexical IMPORT InternalError, FormatWarningMessage2 ;
FROM M2System IMPORT IsPseudoSystemFunction ;
FROM M2Time IMPORT Pairings ;
FROM M2GenAPU IMPORT Time ;
FROM StdIO IMPORT PushOutput, PopOutput ;
FROM Environment IMPORT GetEnvironment ;
FROM NameKey IMPORT MakeKey ;
FROM StdIO IMPORT Write ;
FROM M2Quads IMPORT IsBackReference, IsBackReferenceConditional, QuadToTokenNo, GetLastQuadNo ;

FROM M2GDB IMPORT FileInfo, TypeInfo,
                  VariableInfo, BeginInfo, EndInfo, MainInfo ;

FROM Lists IMPORT List, InitList, KillList,
                  GetIndexOfList, IsItemInList,
                  NoOfItemsInList, IncludeItemIntoList,
                  RemoveItemFromList, DuplicateList, GetItemFromList ;

IMPORT IO ;
IMPORT M2AS ;
IMPORT M2Time ;

CONST
   MaxInstruction          =      500 ;
   Infinity                = 10000000 ;   (* sufficient for a basic block cycles *)
   WarningTheshold         =    10000 ;   (* warn if execution time exceeds this *)
   AGIWithConstantOperands =    FALSE ;   (* I don't think this is true? (check) *)
   NopIndex                =        0 ;   (* instruction 0 is always a nop       *)
   CacheBucket             =       32 ;   (* difference in address which causes  *)
                                          (* a conflict and the size of cache    *)
                                          (* entity.                             *)
   CacheMissPenilty        =        3 ;   (* min penility 3 cycles               *)
   CacheWriteNoReadPenilty =        3 ;   (* min penility for write cycles       *)
                                          (* this is an average as it depends on *)
                                          (* the next read/write to memory.      *)
   VerboseDebugging        =     TRUE ;
   NoOfLinesFor586         =        2 ;   (* number of cache lines for various   *)
   NoOfLinesFor686         =        4 ;   (* members of the pentium family       *)

   NoOfBytesFor586         =     8192 ;   (* data cache size for 586             *)
   NoOfBytesFor686         =   2*8192 ;   (* data cache size for 686             *)

   MaxCacheLines           = NoOfBytesFor686 DIV CacheBucket ;   (* use maximum  *)
   MaxSameLines            = NoOfLinesFor686 ;            (* again maximum       *)

(* *********************************************************************** *)
(*                       C a c h e    I n t e r f a c e                    *)
(* *********************************************************************** *)

TYPE
   CacheState    = (empty, filled) ;

   DataBank      = RECORD
                      AddressLo: INTEGER ;           (* relative to symbol       *)
                      AddressHi: INTEGER ;           (* relative to symbol       *)
                      Father   : CARDINAL ;
                      ReadBytes: BITSET ;
                   END ;

   CacheLine     = RECORD
                      State   : CacheState ;
                      SameLine: ARRAY [0..MaxSameLines] OF DataBank ;
                      SetValue: CARDINAL ;      (* cache line setvalue           *)
                      MRULine : CARDINAL ;      (* the most recently used line.  *)
                   END ;

   CacheArray    = ARRAY [0..MaxCacheLines-1] OF CacheLine ;

   CacheInfo = RECORD
                  Lines        : CacheArray ;
                  NoOfLinesUsed: CARDINAL ;     (* how many lines are used?      *)
                  StackPtr     : INTEGER ;      (* model of esp value, so handle *)
                                                (* stack cache                   *)
               END ;



   InstrIndex = [0..MaxInstruction] ;
   PipeIndex  = [0..MaxInstruction] ;
   TreeIndex  = [0..MaxInstruction] ;

   MetaInstruction = (instregconstant, instconstantreg, instregreg,
                      instmdconstant, instmdint, instname, instmdoffset,
                      instregoffset, instkey, instregimmname,
                      instregint, instregmd, instmdreg,
                      instmriconstant, instmrisymconstant, instmricardinal, instmrireg,
                      instregmio, instregmisym, instregmi, instmioreg, instmisymreg,
                      instmireg,
                      bininst, bininstunit, bininstreg, bininstmd, bininsthex,
                      bininstregi, bininstconstant) ;

   ClassInstKey      = RECORD
                          inst   : Instruction ;
                          string : CARDINAL ;
                       END ;

   ClassInstRegOffset= RECORD
                          inst   : Instruction ;
                          unit   : DatumPrefix ;
                          destreg: Register ;
                          srcsym : CARDINAL ;
                       END ;

   ClassInstMDOffset = RECORD
                          inst   : Instruction ;
                          unit   : DatumPrefix ;
                          srcsym : CARDINAL ;
                          destsym: CARDINAL ;
                       END ;

   ClassInstName   = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        sym    : CARDINAL ;
                     END ;

   ClassBinInst    = RECORD
                        inst   : Instruction ;
                     END ;

   ClassBinInstUnit= RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        length : CARDINAL ;
                     END ;

   ClassBinInstReg = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        reg    : Register ;
                     END ;

   ClassBinInstConstant = RECORD
                             inst   : Instruction ;
                             unit   : DatumPrefix ;
                             srcsym : CARDINAL ;
                          END ;

   ClassBinInstMD  = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        o1, go : INTEGER ;
                        scope  : KindOfScope ;
                        sym    : CARDINAL ;
                     END ;

   ClassBinInstHex     = RECORD
                            inst   : Instruction ;
                            unit   : DatumPrefix ;
                            card   : CARDINAL ;
                         END ;

   ClassBinInstI       = RECORD
                            inst   : Instruction ;
                            unit   : DatumPrefix ;
                            reg    : Register ;
                         END ;

   ClassInstMDConstant = RECORD
                            inst   : Instruction ;
                            unit   : DatumPrefix ;
                            srcsym : CARDINAL ;
                            o1, go : INTEGER ;
                            scope  : KindOfScope ;
                            destsym: CARDINAL ;
                         END ;

   ClassInstMDInt      = RECORD
                            inst   : Instruction ;
                            unit   : DatumPrefix ;
                            srcint : INTEGER ;
                            o1, go : INTEGER ;
                            scope  : KindOfScope ;
                            destsym: CARDINAL ;
                         END ;

   ClassInstRegMD  = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        srcsym : CARDINAL ;
                        o1, go : INTEGER ;
                        scope  : KindOfScope ;
                        destreg: Register ;
                     END ;

   ClassInstMDReg  = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        o1, go : INTEGER ;
                        srcreg : Register ;
                        scope  : KindOfScope ;
                        destsym: CARDINAL ;
                     END ;

   ClassInstMRISymConstant = RECORD
                                inst   : Instruction ;
                                unit   : DatumPrefix ;
                                destsym: CARDINAL ;
                                o1, go : INTEGER ;
                                destreg: Register ;
                                srcsym : CARDINAL ;
                             END ;

   ClassInstMRIConstant = RECORD
                             inst   : Instruction ;
                             unit   : DatumPrefix ;
                             o1, go : INTEGER ;
                             destreg: Register ;
                             srcsym : CARDINAL ;
                          END ;

   ClassInstMRISym = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        o1, go : INTEGER ;
                        destreg: Register ;
                        srcreg : Register ;
                        srcsym : CARDINAL ;
                     END ;

   ClassInstMRICardinal = RECORD
                             inst   : Instruction ;
                             unit   : DatumPrefix ;
                             o1, go : INTEGER ;
                             destreg: Register ;
                             srccard: CARDINAL ;
                          END ;

   ClassInstMRIReg = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        o1, go : INTEGER ;
                        destreg: Register ;
                        srcreg : Register ;
                     END ;

   ClassInstRegMIO = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        o1, go : INTEGER ;
                        srcreg : Register ;
                     END ;

   ClassInstMIOReg = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        o1, go : INTEGER ;
                        srcreg : Register ;
                     END ;

   ClassInstMISymReg = RECORD
                          inst   : Instruction ;
                          unit   : DatumPrefix ;
                          destreg: Register ;
                          destsym: CARDINAL ;
                          o1, go : INTEGER ;
                          srcreg : Register ;
                       END ;

   ClassInstRegMISym = RECORD
                          inst   : Instruction ;
                          unit   : DatumPrefix ;
                          destreg: Register ;
                          srcreg : Register ;
                          o1, go : INTEGER ;
                          srcsym : CARDINAL ;
                       END ;

   ClassInstRegMI = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        srcreg : Register ;
                     END ;

   ClassInstMIReg = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        srcreg : Register ;
                     END ;

   ClassInstRegInt = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        srcint : INTEGER ;
                     END ;

   ClassInstRegImmName = RECORD
                            inst   : Instruction ;
                            unit   : DatumPrefix ;
                            destreg: Register ;
                            srcsym : CARDINAL ;
                         END ;

   ClassInstRegReg = RECORD
                        inst   : Instruction ;
                        unit   : DatumPrefix ;
                        destreg: Register ;
                        srcreg : Register ;
                     END ;

   ClassInstRegConstant = RECORD
                             inst   : Instruction ;
                             unit   : DatumPrefix ;
                             destreg: Register ;
                             srcsym : CARDINAL ;
                          END ;

   ClassInstConstantReg = RECORD
                             inst   : Instruction ;
                             unit   : DatumPrefix ;
                             srcreg : Register ;
                             destsym: CARDINAL ;
                          END ;

   CostMatrix    = RECORD
                      Cost,
                      Cache,
                      AGI,
                      Conflict,
                      Operand : CARDINAL ;
                   END ;

   InstDefn = RECORD
                 Time   : CostMatrix ;
                 Branch1,
                 Branch2,
                 Parent : InstrIndex ;
                 CASE InstructionType: MetaInstruction OF

                 bininst           :  Bininst           : ClassBinInst |
                 bininstunit       :  Bininstunit       : ClassBinInstUnit |
                 bininstreg        :  Bininstreg        : ClassBinInstReg |
                 bininstmd         :  Bininstmd         : ClassBinInstMD |
                 bininsthex        :  Bininsthex        : ClassBinInstHex |
                 bininstregi       :  Bininstregi       : ClassBinInstI |
                 bininstconstant   :  Bininstconstant   : ClassBinInstConstant |
                 instname          :  Instname          : ClassInstName |
                 instregimmname    :  Instregimmname    : ClassInstRegImmName |
                 instmdoffset      :  Instmdoffset      : ClassInstMDOffset |
                 instregoffset     :  Instregoffset     : ClassInstRegOffset |
                 instregconstant   :  Instregconstant   : ClassInstRegConstant |
                 instconstantreg   :  Instconstantreg   : ClassInstConstantReg |
                 instregint        :  Instregint        : ClassInstRegInt |
                 instregreg        :  Instregreg        : ClassInstRegReg |
                 instregmd         :  Instregmd         : ClassInstRegMD |
                 instregmi         :  Instregmi         : ClassInstRegMI |
                 instregmio        :  Instregmio        : ClassInstRegMIO |
                 instregmisym      :  Instregmisym      : ClassInstRegMISym |
                 instmireg         :  Instmireg         : ClassInstMIReg |
                 instmrireg        :  Instmrireg        : ClassInstMRIReg |
                 instmricardinal   :  Instmricardinal   : ClassInstMRICardinal |
                 instmriconstant   :  Instmriconstant   : ClassInstMRIConstant |
                 instmrisymconstant:  Instmrisymconstant: ClassInstMRISymConstant |
                 instmioreg        :  Instmioreg        : ClassInstMIOReg |
                 instmisymreg      :  Instmisymreg      : ClassInstMISymReg |
                 instregmi         :  Insrregmi         : ClassInstRegMI |
                 instmdreg         :  Instmdreg         : ClassInstMDReg |
                 instmdconstant    :  Instmdconstant    : ClassInstMDConstant |
                 instmdint         :  Instmdint         : ClassInstMDInt |
                 instkey           :  Instkey           : ClassInstKey

                 END
              END ;

   ChoiceIndex   = [0..20] ;

   PipeInfo      = RECORD
                      U,
                      V   : InstrIndex ;
                      Time: CostMatrix ;
                   END ;

   PipeNext      = (v, u) ;

   PipeLineState = RECORD
                      Cost       : CARDINAL ;
                      Pipe       : ARRAY PipeIndex OF PipeInfo ;
                      Head       : PipeIndex ;
                      Next       : PipeNext ;
                      Cache      : CacheInfo ;
                   END ;

   PipeType      = (Unknown, Upipe, Vpipe) ;

VAR
   DebugOn              : BOOLEAN ;
   Instr                : ARRAY InstrIndex OF InstDefn ;
   NoOfInstructions     : InstrIndex ;
   Trees                : ARRAY TreeIndex OF InstrIndex ;
   NoOfTrees            : TreeIndex ;
   CurrentPipeLine      : PipeLineState ;
   BestPipeLine         : PipeLineState ;
   AccumulatedCost      : CARDINAL ;
   OldCache             : CacheInfo ;
   ActualCacheLines     : CARDINAL ;   (* total cache lines for this processor *)
   ActualSameCacheLines : CARDINAL ;   (* no of line with same set values      *)
   LeftOverInstruction  : InstrIndex ;
   MustFlushInstruction : BOOLEAN ;
   LastInstructionPipe  : PipeType ;   (* this variable is to be deleted gaius *)
   LastInstructionMustBe: InstrIndex ; (* the last instruction must be         *)
   NewCacheLineCost     : CacheState ; (* how do we cost new cache lines?      *)


(* %%%FORWARD%%%
PROCEDURE DisplayCache ; FORWARD ;
PROCEDURE ResetCurrentCache (new: CacheState) ; FORWARD ;
PROCEDURE IsCacheLineConflict (a, sizea, b, sizeb: INTEGER) : BOOLEAN ; FORWARD ;
PROCEDURE Min (a, b: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE AddChoices (item: InstrIndex; ChoiceList: List) ; FORWARD ;
PROCEDURE DisplayInstruction (i: InstrIndex) ; FORWARD ;
PROCEDURE DumpChoiceList (ChoiceList: List) ; FORWARD ;
PROCEDURE DumpBestPipeLine ; FORWARD ;
PROCEDURE IsBetter (a, b: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE ReworkForward ; FORWARD ;
PROCEDURE DumpCurrentPipeLine ; FORWARD ;
PROCEDURE WriteMatrix (m: CostMatrix) ; FORWARD ;
PROCEDURE AddDependant (n: InstrIndex) ; FORWARD ;
PROCEDURE FlushInstruction (FixLastInstruction: BOOLEAN) ; FORWARD ;
PROCEDURE AddToCurrentPipeLine (i: InstrIndex) ; FORWARD ;
PROCEDURE FlushPipeline (FixLastInstruction: BOOLEAN) ; FORWARD ;
PROCEDURE GetInstruction (i: InstrIndex) : Instruction ; FORWARD ;
PROCEDURE GetInstructionType (i: InstrIndex) : MetaInstruction ; FORWARD ;
PROCEDURE GetComplexSrc (i: InstrIndex;
                         VAR Father: CARDINAL ;
                         VAR Unit  : CARDINAL ;
                         VAR Off   : INTEGER) ; FORWARD ;
PROCEDURE GetComplexDest (i: InstrIndex;
                          VAR Father: CARDINAL ;
                          VAR Unit  : CARDINAL ;
                          VAR Off   : INTEGER) ; FORWARD ;
PROCEDURE IsDestSrc (from, to: InstrIndex) : BOOLEAN ; FORWARD ;
PROCEDURE IsDestDest (from, to: InstrIndex) : BOOLEAN ; FORWARD ;
PROCEDURE IsStackBasedInst (i: InstrIndex) : BOOLEAN ; FORWARD ;
PROCEDURE IsStackBasedPair (i, j: InstrIndex) : BOOLEAN ; FORWARD ;
PROCEDURE IsSubSet (i1, i2, j1, j2: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE Dummy ; FORWARD ;
PROCEDURE CacheCost (h: InstrIndex) : CARDINAL ; FORWARD ;
PROCEDURE AddToCache (sym: CARDINAL; size: CARDINAL; offset: INTEGER; IsWrite: BOOLEAN) : CARDINAL ; FORWARD ;
   %%%FORWARD%%% *)


(*
   CopyPipeLine - copies pipeline, from, to, to.
*)

PROCEDURE CopyPipeLine (VAR from: PipeLineState; VAR to: PipeLineState) ;
VAR
   i, j: CARDINAL ;
BEGIN
   i := from.Head ;
   j := 0 ;
   WHILE j<=i DO
      to.Pipe[j] := from.Pipe[j] ;
      INC(j)
   END ;
   to.Cost := from.Cost ;
   to.Head := from.Head ;
   to.Next := from.Next ;
   IF OptimizeCache
   THEN
      i := from.Cache.NoOfLinesUsed ;
      j := 0 ;
      WHILE j<i DO
         to.Cache.Lines[j] := from.Cache.Lines[j] ;
         INC(j)
      END ;
      to.Cache.NoOfLinesUsed := from.Cache.NoOfLinesUsed ;
      to.Cache.StackPtr      := from.Cache.StackPtr
   END
END CopyPipeLine ;


(*
   Evaluate - tries each choice in turn to find out the optimal
              instruction scheduling order.
*)

PROCEDURE Evaluate (ChoiceList: List; VAR BestCost: CARDINAL) ;
VAR
   OldPipeLine: PipeLineState ;
   item       : InstrIndex ;
   i, c       : CARDINAL ;
   NewList    : List ;
BEGIN
   c := NoOfItemsInList(ChoiceList) ;
   IF c=0
   THEN
      ReworkForward ;
      
      IF IsBetter(CurrentPipeLine.Cost, BestCost) OR
         ((CurrentPipeLine.Cost=BestCost) AND
          IsBetter(BestPipeLine.Head, CurrentPipeLine.Head))
      THEN
         IF DebugOn
         THEN
            DumpBestPipeLine
         END ;
         BestCost     := CurrentPipeLine.Cost ;
         BestPipeLine := CurrentPipeLine ;
         IF DebugOn
         THEN
            ; WriteString('# beaten by') ; WriteLn ;
            ; DumpBestPipeLine ;
         END
      END
   ELSE
      i := 1 ;
      WHILE i<=c DO
         item := GetItemFromList(ChoiceList, i) ;
         IF item=0
         THEN
            InternalError('wrong node value', __FILE__, __LINE__)
         END ;
         IF DebugOn
         THEN
            WriteString('# -------------------------------------------') ; WriteLn ;
            WriteString('# we have ') ; WriteCard(c, 0) ; WriteString(' choices') ; WriteLn ;
            DumpChoiceList(ChoiceList) ;
            WriteString('consider adding: ') ; DisplayInstruction(item) ;
            DumpCurrentPipeLine
         END ;
         NewList := DuplicateList(ChoiceList) ;
         RemoveItemFromList(NewList, CARDINAL(item)) ;
         AddChoices(item, NewList) ;
         CopyPipeLine(CurrentPipeLine, OldPipeLine) ;
         AddToCurrentPipeLine(item) ;
         IF OptimizeWorst
         THEN
            Evaluate(NewList, BestCost)
         ELSE
            IF IsBetter(BestCost, CurrentPipeLine.Cost)
            THEN
               IF DebugOn
               THEN
                  WriteString('# no point in evaluating further (cost)') ; WriteLn
               END ;
               (* do not evaluate furthur *)
            ELSIF (CurrentPipeLine.Cost=BestCost) AND
                  IsBetter(BestPipeLine.Head, CurrentPipeLine.Head)
            THEN
               IF DebugOn
               THEN
                  WriteString('# no point in evaluating further  (instr length)') ; WriteLn
               END ;
               (* always choose better instruction length given same cost *)
            ELSE
               Evaluate(NewList, BestCost)
            END
         END ;
         CopyPipeLine(OldPipeLine, CurrentPipeLine) ;
         KillList(NewList) ;
         INC(i)
      END
   END
END Evaluate ;


PROCEDURE TouchCache (index: CARDINAL; line: CARDINAL;
                      offset: INTEGER; size: CARDINAL; IsWrite: BOOLEAN) : CARDINAL ;
VAR
   t   : DataBank ;
   i, j: CARDINAL ;
   cost: CARDINAL ;
   old : BITSET ;
BEGIN
   WITH CurrentPipeLine.Cache.Lines[index] DO
      IF MRULine#line
      THEN
         t := SameLine[line] ;   (* copy of this line *)
         i := line ;
         j := (line+1) MOD ActualSameCacheLines ;
         WHILE i#MRULine DO
            SameLine[i] := SameLine[j] ;
            i := (i+1) MOD ActualSameCacheLines ;
            j := (j+1) MOD ActualSameCacheLines
         END ;
         SameLine[MRULine] := t
      END ;
      WITH SameLine[MRULine] DO
         IF IsWrite
         THEN
            old := ReadBytes ;
            FOR i := offset MOD CacheBucket TO Min(offset MOD CacheBucket + size, CacheBucket-1) DO
               INCL(ReadBytes, i)
            END ;
            IF old={}
            THEN
               RETURN( CacheWriteNoReadPenilty )
            END
         ELSE
            IF (ReadBytes={}) AND (State=filled)
            THEN
               cost := CacheMissPenilty
            ELSE
               cost := 0
            END ;
            FOR i := offset MOD CacheBucket TO Min(offset MOD CacheBucket + size, CacheBucket-1) DO
               INCL(ReadBytes, i)
            END ;
            RETURN( cost )
         END
      END
   END ;
   RETURN( 0 )
END TouchCache ;

(*
   Max - returns the highest of two values.
*)

PROCEDURE Max (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a>b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Max ;


(*
   Min - returns the lowest of two values.
*)

PROCEDURE Min (a, b: CARDINAL) : CARDINAL ;
BEGIN
   IF a<b
   THEN
      RETURN( a )
   ELSE
      RETURN( b )
   END
END Min ;


(*
   GetMDReg - returns the register associated with the base index of symbol, Sym.
*)

PROCEDURE GetMDReg (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF GetScope(Sym)=Local
   THEN
      RETURN( EBP )
   ELSE
      RETURN( NulSym )
   END
END GetMDReg ;


(*
   GetBaseReg - returns the symbol table level symbol which
                represents the destination for this instruction.
                NOTE a register might be returned.
*)

PROCEDURE GetBaseReg (i: InstrIndex) : CARDINAL ;
BEGIN
   IF (GetInstruction(i)=pop) OR (GetInstruction(i)=push) OR
      (GetInstruction(i)=call) OR (GetInstruction(i)=ret) OR
      (GetInstruction(i)=leave)
   THEN
      RETURN( ESP )
   END ;
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  RETURN( NulSym ) |
      bininstunit       :  RETURN( NulSym ) |
      bininstreg        :  RETURN( NulSym ) |
      bininsthex        :  RETURN( NulSym ) |
      bininstregi       :  RETURN( NulSym ) |
      bininstconstant   :  RETURN( NulSym ) |
      instname          :  RETURN( NulSym ) |
      instregimmname    :  RETURN( NulSym ) |
      instregconstant   :  RETURN( NulSym ) |
      instconstantreg   :  RETURN( NulSym ) |
      instmriconstant   :  RETURN( HighReg(Instmriconstant.destreg) ) |
      instmrisymconstant:  RETURN( HighReg(Instmrisymconstant.destreg) ) |
      instmricardinal   :  RETURN( HighReg(Instmricardinal.destreg) ) |
      instmrireg        :  RETURN( HighReg(Instmrireg.destreg) ) |
      instregmio        :  RETURN( HighReg(Instregmio.srcreg) ) |
      instregmisym      :  RETURN( HighReg(Instregmisym.srcreg) ) |
      instregmi         :  RETURN( HighReg(Instregmi.srcreg) ) |
      instmioreg        :  RETURN( HighReg(Instmioreg.destreg) ) |
      instmisymreg      :  RETURN( HighReg(Instmisymreg.destreg) ) |  (* ** *)
      instregreg        :  RETURN( NulSym ) |
      instregint        :  RETURN( NulSym ) |
      instregmd         :  RETURN( GetMDReg(Instregmd.srcsym) ) |
      instmdoffset      :  RETURN( GetMDReg(Instmdoffset.destsym) ) |
      instmdconstant    :  RETURN( GetMDReg(Instmdconstant.destsym ) ) |
      instmdint         :  RETURN( GetMDReg(Instmdint.destsym) ) |
      instregoffset     :  RETURN( NulSym ) |
      instkey           :  RETURN( NulSym ) |
      instmdreg         :  RETURN( GetMDReg(Instmdreg.destsym) ) |
      bininstmd         :  RETURN( GetMDReg(Bininstmd.sym) )

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetBaseReg ;


(*
   GetDest - returns the symbol table level symbol which
             represents the destination for this instruction.
             NOTE a register might be returned.
*)

PROCEDURE GetDest (i: InstrIndex) : CARDINAL ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  RETURN( NulSym ) |
      bininstunit       :  RETURN( NulSym ) |
      bininstreg        :  RETURN( HighReg(Bininstreg.reg) ) |
      bininsthex        :  RETURN( NulSym ) |
      bininstregi       :  RETURN( NulSym ) |
      bininstconstant   :  RETURN( NulSym ) |
      instname          :  RETURN( Instname.sym ) |
      instregimmname    :  RETURN( HighReg(Instregimmname.destreg) ) |
      instregconstant   :  RETURN( HighReg(Instregconstant.destreg) ) |
      instconstantreg   :  RETURN( Instconstantreg.destsym ) |
(*      instmrisym        :  RETURN( HighReg(Instmrisym.destreg) ) | *)
      instmriconstant   :  RETURN( HighReg(Instmriconstant.destreg) ) |
      instmrisymconstant:  RETURN( Instmrisymconstant.destsym ) |  (* destreg as well.. *)
      instmricardinal   :  RETURN( HighReg(Instmricardinal.destreg) ) |
      instmrireg        :  RETURN( HighReg(Instmrireg.destreg) ) |
      instregmio        :  RETURN( HighReg(Instregmio.destreg) ) |
      instregmisym      :  RETURN( HighReg(Instregmisym.destreg) ) |
      instregmi         :  RETURN( HighReg(Instregmi.destreg) ) |
      instmioreg        :  RETURN( HighReg(Instmioreg.destreg) ) |
      instmisymreg      :  RETURN( Instmisymreg.destsym ) |  (* destreg as well.. *)
      instregreg        :  RETURN( HighReg(Instregreg.destreg) ) |
      instregint        :  RETURN( HighReg(Instregint.destreg) ) |
      instregmd         :  RETURN( HighReg(Instregmd.destreg) ) |
      instmdoffset      :  RETURN( Instmdoffset.destsym ) |
      instmdconstant    :  RETURN( Instmdconstant.destsym ) |
      instmdint         :  RETURN( Instmdint.destsym ) |
      instregoffset     :  RETURN( HighReg(Instregoffset.destreg) ) |
      instkey           :  RETURN( NulSym ) |
      instmdreg         :  RETURN( NulSym ) |
      bininstmd         :  RETURN( NulSym )

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetDest ;


(*
   GetSrc - returns the symbol table level symbol which
            represents the destination for this instruction.
            NOTE a register might be returned.
*)

PROCEDURE GetSrc (i: InstrIndex) : CARDINAL ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  RETURN( NulSym ) |
      bininstunit       :  RETURN( NulSym ) |
      bininstreg        :  RETURN( HighReg(Bininstreg.reg) ) |
      bininsthex        :  RETURN( NulSym ) |
      bininstregi       :  RETURN( HighReg(Bininstregi.reg) ) |
      bininstconstant   :  RETURN( Bininstconstant.srcsym ) |
      instname          :  RETURN( Instname.sym ) |
      instregconstant   :  RETURN( Instregconstant.srcsym ) |
      instregimmname    :  RETURN( Instregimmname.srcsym ) |
      instconstantreg   :  RETURN( HighReg(Instconstantreg.srcreg) ) |
      instregreg        :  RETURN( HighReg(Instregreg.srcreg) ) |
      instregint        :  RETURN( NulSym ) |
      instmdreg         :  RETURN( HighReg(Instmdreg.srcreg) ) |
(*      instmrisym        :  RETURN( Instmrisym.srcsym ) |   (* srcreg as well.. *) *)
      instmriconstant   :  RETURN( Instmriconstant.srcsym ) |
      instmrisymconstant:  RETURN( Instmrisymconstant.srcsym ) |
      instmricardinal   :  RETURN( NulSym ) |   (* srccard is a constant *)
      instmrireg        :  RETURN( HighReg(Instmrireg.srcreg) ) |
      instregmio        :  RETURN( HighReg(Instregmio.srcreg) ) |
      instregmisym      :  RETURN( Instregmisym.srcsym ) |   (* and srcreg as well.. *)
      instregmi         :  RETURN( HighReg(Instregmi.srcreg) ) |
      instmioreg        :  RETURN( HighReg(Instmioreg.srcreg) ) |
      instmisymreg      :  RETURN( HighReg(Instmisymreg.srcreg) ) |
      instmdoffset      :  RETURN( Instmdoffset.srcsym ) |
      instmdint         :  RETURN( NulSym ) |
      instregoffset     :  RETURN( NulSym ) |
      instmdconstant    :  RETURN( Instmdconstant.srcsym ) |
      instkey           :  RETURN( NulSym ) |
      instregmd         :  RETURN( NulSym ) |
      bininstmd         :  RETURN( NulSym )

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetSrc ;


(*
   GetInstruction - returns the i486 representation of the instruction.
*)

PROCEDURE GetInstruction (i: InstrIndex) : Instruction ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  RETURN( Bininst.inst ) |
      bininstunit       :  RETURN( Bininstunit.inst ) |
      bininstreg        :  RETURN( Bininstreg.inst ) |
      bininstmd         :  RETURN( Bininstmd.inst ) |
      bininsthex        :  RETURN( Bininsthex.inst ) |
      bininstregi       :  RETURN( Bininstregi.inst ) |
      bininstconstant   :  RETURN( Bininstconstant.inst ) |
      instname          :  RETURN( Instname.inst ) |
      instregconstant   :  RETURN( Instregconstant.inst ) |
      instconstantreg   :  RETURN( Instconstantreg.inst ) |
      instregreg        :  RETURN( Instregreg.inst ) |
      instregint        :  RETURN( Instregint.inst ) |
      instregimmname    :  RETURN( Instregimmname.inst ) |
      instregmd         :  RETURN( Instregmd.inst ) |
      instmdreg         :  RETURN( Instmdreg.inst ) |
      instmdoffset      :  RETURN( Instmdoffset.inst ) |
      instmriconstant   :  RETURN( Instmriconstant.inst ) |
      instmrisymconstant:  RETURN( Instmrisymconstant.inst ) |
      instmricardinal   :  RETURN( Instmricardinal.inst ) |
      instmrireg        :  RETURN( Instmrireg.inst ) |
      instregmio        :  RETURN( Instregmio.inst ) |
      instregmisym      :  RETURN( Instregmisym.inst ) |
      instregmi         :  RETURN( Instregmi.inst ) |
      instmioreg        :  RETURN( Instmioreg.inst ) |
      instmisymreg      :  RETURN( Instmisymreg.inst ) |
      instmireg         :  RETURN( Instmireg.inst ) |
      instregoffset     :  RETURN( Instregoffset.inst ) |
      instmdconstant    :  RETURN( Instmdconstant.inst ) |
      instmdint         :  RETURN( Instmdint.inst ) |
      instkey           :  RETURN( Instkey.inst )

      ELSE
         InternalError('unexpected instruction', __FILE__, __LINE__)
      END
   END
END GetInstruction ;


(*
   GetUnit - returns the unit field for any instruction type.
*)

PROCEDURE GetUnit (i: InstrIndex) : CARDINAL ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  RETURN( 0 ) |
      bininstunit       :  RETURN( GetSizeOfPrefix(Bininstunit.unit) ) |
      bininstreg        :  RETURN( GetSizeOfPrefix(Bininstreg.unit) ) |
      bininsthex        :  RETURN( GetSizeOfPrefix(Bininsthex.unit) ) |
      bininstregi       :  RETURN( GetSizeOfPrefix(Bininstregi.unit) ) |
      bininstconstant   :  RETURN( GetSizeOfPrefix(Bininstconstant.unit) ) |
      instname          :  RETURN( GetSizeOfPrefix(Instname.unit) ) |
      instregimmname    :  RETURN( GetSizeOfPrefix(Instregimmname.unit) ) |
      instregconstant   :  RETURN( GetSizeOfPrefix(Instregconstant.unit) ) |
      instconstantreg   :  RETURN( GetSizeOfPrefix(Instconstantreg.unit) ) |
      instregreg        :  RETURN( GetSizeOfPrefix(Instregreg.unit) ) |
      instregint        :  RETURN( GetSizeOfPrefix(Instregint.unit) ) |
      instregmd         :  RETURN( GetSizeOfPrefix(Instregmd.unit) ) |
      instmdoffset      :  RETURN( GetSizeOfPrefix(Instmdoffset.unit) ) |
      instmdconstant    :  RETURN( GetSizeOfPrefix(Instmdconstant.unit) ) |
      instmdint         :  RETURN( GetSizeOfPrefix(Instmdint.unit) ) |
      instmriconstant   :  RETURN( GetSizeOfPrefix(Instmriconstant.unit) ) |
      instmrisymconstant:  RETURN( GetSizeOfPrefix(Instmrisymconstant.unit) ) |
      instmricardinal   :  RETURN( GetSizeOfPrefix(Instmricardinal.unit) ) |
      instmrireg        :  RETURN( GetSizeOfPrefix(Instmrireg.unit) ) |
      instregmio        :  RETURN( GetSizeOfPrefix(Instregmio.unit) ) |
      instregmisym      :  RETURN( GetSizeOfPrefix(Instregmisym.unit) ) |
      instregmi         :  RETURN( GetSizeOfPrefix(Instregmi.unit) ) |
      instmioreg        :  RETURN( GetSizeOfPrefix(Instmioreg.unit) ) |
      instmisymreg      :  RETURN( GetSizeOfPrefix(Instmisymreg.unit) ) |
      instregoffset     :  RETURN( GetSizeOfPrefix(Instregoffset.unit) ) |
      instkey           :  RETURN( 0 ) |
      instmdreg         :  RETURN( GetSizeOfPrefix(Instmdreg.unit) ) |
      bininstmd         :  RETURN( GetSizeOfPrefix(Bininstmd.unit) )

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetUnit ;


(*
   IsDestAlsoSrcInstruction - returns TRUE if the instruction, inst,
                              uses the destination operand also as a source
                              operand.
*)

PROCEDURE IsDestAlsoSrcInstruction (inst: Instruction) : BOOLEAN ;
BEGIN
   RETURN(
          (inst=add) OR
          (inst=sub) OR
          (inst=inc) OR
          (inst=dec) OR
          (inst=sar) OR
          (inst=sal) OR
          (inst=shl) OR
          (inst=shr) OR
          (inst=not) OR
          (inst=and) OR
          (inst=or ) OR
          (inst=neg) OR
          (inst=xor)
         )
END IsDestAlsoSrcInstruction ;


(*
   IsDestAlsoSrc - returns TRUE if instruction, i, uses
                   the destination operand also as a source
                   operand.
*)

PROCEDURE IsDestAlsoSrc (i: InstrIndex) : BOOLEAN ;
BEGIN
   RETURN( IsDestAlsoSrcInstruction(GetInstruction(i)) )
END IsDestAlsoSrc ;


(*
   DisplayInstruction - displays the instruction, i.
*)

PROCEDURE DisplayInstruction (i: InstrIndex) ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  WITH Bininst DO
                              M2AS.BinInst(inst)
                           END |
      bininstunit       :  WITH Bininstunit DO
                              M2AS.BinInstUnit(inst, unit, length)
                           END |
      bininstreg        :  WITH Bininstreg DO
                              M2AS.BinInstReg(inst, unit, reg)
                           END |
      bininstmd         :  WITH Bininstmd DO
                              M2AS.BinInstMD(inst, unit, o1, go, scope, sym)
                           END |
      bininsthex        :  WITH Bininsthex DO
                              M2AS.BinInstHex(inst, unit, card)
                           END |
      bininstregi       :  WITH Bininstregi DO
                              M2AS.BinInstRegI(inst, unit, reg)
                           END |
      bininstconstant   :  WITH Bininstconstant DO
                              M2AS.BinInstConstant(inst, unit, srcsym)
                           END |
      instname          :  WITH Instname DO
                              M2AS.InstName(inst, unit, sym)
                           END |
      instregimmname    :  WITH Instregimmname DO
                              M2AS.InstRegImmName(inst, unit, destreg, srcsym)
                           END |
      instregconstant   :  WITH Instregconstant DO
                              M2AS.InstRegConstant(inst, unit, destreg, srcsym)
                           END |
      instconstantreg   :  WITH Instconstantreg DO
                              M2AS.InstConstantReg(inst, unit, destsym, srcreg)
                           END |
      instregreg        :  WITH Instregreg DO
                              M2AS.InstRegReg(inst, unit, destreg, srcreg)
                           END |
      instregint        :  WITH Instregint DO
                              M2AS.InstRegInt(inst, unit, destreg, srcint)
                           END |
      instregmd         :  WITH Instregmd DO
                              M2AS.InstRegMD(inst, unit, destreg,
                                             o1, go, scope, srcsym)
                           END |
      instmdreg         :  WITH Instmdreg DO
                              M2AS.InstMDReg(inst, unit, o1, go,
                                             srcreg, scope, destsym)
                           END |
      instmdconstant    :  WITH Instmdconstant DO
                              M2AS.InstMDConstant(inst, unit, o1, go,
                                                  scope, destsym, srcsym)
                           END |
      instmdint         :  WITH Instmdint DO
                              M2AS.InstMDInt(inst, unit, o1, go,
                                             srcint, scope, destsym)
                           END |
      instmrireg        :  WITH Instmrireg DO
                              M2AS.InstMRIReg(inst, unit, o1, go,
                                              destreg, srcreg)
                           END | 
      instmrisymconstant:  WITH Instmrisymconstant DO
                              M2AS.InstMRISymConstant(inst, unit, destsym,
                                                      o1+go, destreg, srcsym)
                           END |
      instmriconstant   :  WITH Instmriconstant DO
                              M2AS.InstMRIConstant(inst, unit, o1, go,
                                                   destreg, srcsym)
                           END |
      instmricardinal   :  WITH Instmricardinal DO
                              M2AS.InstMRICardinal(inst, unit, o1, go,
                                                   destreg, srccard)
                           END |
      instregmio        :  WITH Instregmio DO
                              M2AS.InstRegMIO(inst, unit,
                                              destreg, srcreg, go+o1)
                           END |
      instregmisym      :  WITH Instregmisym DO
                              M2AS.InstRegMISym(inst, unit,
                                                destreg, srcreg, go+o1, srcsym)
                           END |
      instregmi         :  WITH Instregmi DO
                              M2AS.InstRegMI(inst, unit, destreg, srcreg)
                           END |
      instmioreg        :  WITH Instmioreg DO
                              M2AS.InstMIOReg(inst, unit, destreg, go+o1, srcreg)
                           END |
      instmisymreg      :  WITH Instmisymreg DO
                              M2AS.InstMISymReg(inst, unit,
                                                srcreg, destreg, go+o1, destsym)
                           END |
      instmdoffset      :  WITH Instmdoffset DO
                              M2AS.InstMDOffset(inst, unit, destsym, srcsym)
                           END |
      instregoffset     :  WITH Instregoffset DO
                              M2AS.InstRegOffset(inst, unit, destreg, srcsym)
                           END |
      instkey           :  WITH Instkey DO
                              M2AS.InstKey(inst, string)
                           END

      ELSE
         InternalError('unknown instruction', __FILE__, __LINE__)
      END
   END
END DisplayInstruction ;


(*
   WriteNSpaces - writes, n, spaces.
*)

PROCEDURE WriteNSpaces (n: CARDINAL) ;
BEGIN
   WHILE n>0 DO
      Write(' ') ;
      DEC(n)
   END
END WriteNSpaces ;


(*
   DisplayList - displays the list of instructions from, i.
*)

PROCEDURE DisplayList (i: InstrIndex; indent: CARDINAL) ;
BEGIN
   IF i#0
   THEN
      IF Instr[i].Branch1#0
      THEN
         WriteNSpaces(indent) ;
         WriteString('# start of branch1') ; WriteLn ;
         DisplayList(Instr[i].Branch1, indent+8) ;
         WriteNSpaces(indent) ;
         WriteString('# end of branch') ; WriteLn
      END ;
      IF Instr[i].Branch2#0
      THEN
         WriteNSpaces(indent) ;
         WriteString('# start of branch2') ; WriteLn ;
         DisplayList(Instr[i].Branch2, indent+8) ;
         WriteNSpaces(indent) ;
         WriteString('# end of branch') ; WriteLn
      END ;
      DisplayList(Instr[i].Parent, indent+8) ;
      WriteNSpaces(indent) ;
      WriteString('# ') ; DisplayInstruction(i)
   END
END DisplayList ;


(*
   DisplayPipeTrees - displays all the pipeline trees.
*)

PROCEDURE DisplayPipeTrees ;
VAR
   t: InstrIndex ;
BEGIN
   t := 1 ;
   WHILE t<=NoOfTrees DO
      WriteString('# tree') ; WriteCard(CARDINAL(t), 0) ; WriteLn ;
      DisplayList(Trees[t], 0) ;
      INC(t)
   END
END DisplayPipeTrees ;


(*
   InitChoices - creates an initial list of instruction choices
                 from the tree list. It also resets all node
                 costs to Infinity.
*)

PROCEDURE InitChoices () : List ;
VAR
   t: InstrIndex ;
   l: List ;
BEGIN
   InitList(l) ;
   t := NoOfTrees ;
   WHILE t>0 DO
      IF Trees[t]#0
      THEN
         IncludeItemIntoList(l, CARDINAL(Trees[t]))
      END ;
      DEC(t)
   END ;
   RETURN( l )
END InitChoices ;


(*
   GetPairings - returns the pairing of an instruction, i.
*)

PROCEDURE GetPairings (i: InstrIndex) : Pairings ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  WITH Bininst DO
                              RETURN( M2Time.PairBinInst(inst) )
                           END |
      bininstunit       :  WITH Bininstunit DO
                              RETURN( M2Time.PairBinInstUnit(inst, unit) )
                           END |
      bininstreg        :  WITH Bininstreg DO
                              RETURN( M2Time.PairBinInstReg(inst, unit, reg) )
                           END |
      bininstmd         :  WITH Bininstmd DO
                              RETURN( M2Time.PairBinInstMD(inst, unit, o1, go, scope, sym) )
                           END |
      bininsthex        :  WITH Bininsthex DO
                              RETURN( M2Time.PairBinInstHex(inst, unit, card) )
                           END |
      bininstregi       :  WITH Bininstregi DO
                              RETURN( M2Time.PairBinInstRegI(inst, unit, reg) )
                           END |
      bininstconstant   :  WITH Bininstconstant DO
                              RETURN( M2Time.PairBinInstConstant(inst, unit, srcsym) )
                           END |
      instname          :  WITH Instname DO
                              RETURN( M2Time.PairInstName(inst, unit, sym) )
                           END |
      instregimmname    :  WITH Instregimmname DO
                              RETURN( M2Time.PairInstRegImmName(inst, unit, destreg, srcsym) )
                           END |
      instregconstant   :  WITH Instregconstant DO
                              RETURN( M2Time.PairInstRegConstant(inst, unit,
                                                                 destreg, srcsym) )
                           END |
      instconstantreg   :  WITH Instconstantreg DO
                              RETURN( M2Time.PairInstConstantReg(inst, unit,
                                                              srcreg, destsym) )
                           END |
      instregreg        :  WITH Instregreg DO
                              RETURN( M2Time.PairInstRegReg(inst, unit,
                                                            destreg, srcreg) )
                           END |
      instregint        :  WITH Instregint DO
                              RETURN( M2Time.PairInstRegInt(inst, unit,
                                                            destreg, srcint) )
                           END |
      instregmd         :  WITH Instregmd DO
                              RETURN( M2Time.PairInstRegMD(inst, unit,
                                                           destreg, o1, go,
                                                           scope, srcsym) )
                           END |
      instmdreg         :  WITH Instmdreg DO
                              RETURN( M2Time.PairInstMDReg(inst, unit,
                                                           o1, go, srcreg,
                                                           scope, destsym) )
                           END |
      instmdconstant    :  WITH Instmdconstant DO
                              RETURN( M2Time.PairInstMDConstant(inst, unit,
                                                                o1, go, scope,
                                                                destsym, srcsym) )
                           END |
      instmdint         :  WITH Instmdint DO
                              RETURN( M2Time.PairInstMDInt(inst, unit,
                                                           o1, go, srcint,
                                                           scope, destsym) )
                           END |
      instmdoffset      :  WITH Instmdoffset DO
                              RETURN( M2Time.PairInstMDOffset(inst, unit, destsym, srcsym) )
                           END |
      instmrireg        :  WITH Instmrireg DO
                              RETURN( M2Time.PairInstMRIReg(inst, unit, o1, go,
                                                            destreg, srcreg) )
                           END | 
      instmricardinal   :  WITH Instmricardinal DO
                              RETURN( M2Time.PairInstMRICardinal(inst, unit, o1, go,
                                                                 destreg, srccard) )
                           END |
      instmriconstant   :  WITH Instmriconstant DO
                              RETURN( M2Time.PairInstMRIConstant(inst, unit, o1, go,
                                                                 destreg, srcsym) )
                           END |
      instmrisymconstant:  WITH Instmrisymconstant DO
                              RETURN( M2Time.PairInstMRISymConstant(inst, unit, destsym,
                                                                    o1+go, destreg, srcsym) )
                           END |
      instregmio        :  WITH Instregmio DO
                              RETURN( M2Time.PairInstRegMIO(inst, unit,
                                                        destreg, srcreg, go+o1) )
                           END |
      instregmisym      :  WITH Instregmisym DO
                              RETURN( M2Time.PairInstRegMISym(inst, unit,
                                                          destreg, srcreg, go+o1, srcsym) )
                           END |
      instregmi         :  WITH Instregmi DO
                              RETURN( M2Time.PairInstRegMI(inst, unit, destreg, srcreg) )
                           END |
      instmioreg        :  WITH Instmioreg DO
                              RETURN( M2Time.PairInstMIOReg(inst, unit, destreg, go+o1, srcreg) )
                           END |
      instmisymreg      :  WITH Instmisymreg DO
                              RETURN( M2Time.PairInstMISymReg(inst, unit,
                                                              srcreg, destreg, go+o1, destsym) )
                           END |
      instregoffset     :  WITH Instregoffset DO
                              RETURN( M2Time.PairInstRegOffset(inst, unit, destreg, srcsym) )
                           END |
      instkey           :  WITH Instkey DO
                              RETURN( M2Time.PairInstKey(inst, string) )
                           END

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetPairings ;


(*
   GetCost - returns the cost of an instruction, i.
*)

PROCEDURE GetCost (i: InstrIndex) : CARDINAL ;
BEGIN
   WITH Instr[i] DO
      CASE InstructionType OF

      bininst           :  WITH Bininst DO
                              RETURN( M2Time.BinInst(inst) )
                           END |
      bininstunit       :  WITH Bininstunit DO
                              RETURN( M2Time.BinInstUnit(inst, unit, length) )
                           END |
      bininstreg        :  WITH Bininstreg DO
                              RETURN( M2Time.BinInstReg(inst, unit, reg) )
                           END |
      bininstmd         :  WITH Bininstmd DO
                              RETURN( M2Time.BinInstMD(inst, unit, o1, go, scope, sym) )
                           END |
      bininsthex        :  WITH Bininsthex DO
                              RETURN( M2Time.BinInstHex(inst, unit, card) )
                           END |
      bininstregi       :  WITH Bininstregi DO
                              RETURN( M2Time.BinInstRegI(inst, unit, reg) )
                           END |
      bininstconstant   :  WITH Bininstconstant DO
                              RETURN( M2Time.BinInstConstant(inst, unit, srcsym) )
                           END |
      instname          :  WITH Instname DO
                              RETURN( M2Time.InstName(inst, unit, sym) )
                           END |
      instregimmname    :  WITH Instregimmname DO
                              RETURN( M2Time.InstRegImmName(inst, unit, destreg, srcsym) )
                           END |
      instregconstant   :  WITH Instregconstant DO
                              RETURN( M2Time.InstRegConstant(inst, unit,
                                                             destreg, srcsym) )
                           END |
      instconstantreg   :  WITH Instconstantreg DO
                              RETURN( M2Time.InstConstantReg(inst, unit,
                                                             srcreg, destsym) )
                           END |
      instregreg        :  WITH Instregreg DO
                              RETURN( M2Time.InstRegReg(inst, unit,
                                                        destreg, srcreg) )
                           END |
      instregint        :  WITH Instregint DO
                              RETURN( M2Time.InstRegInt(inst, unit,
                                                        destreg, srcint) )
                           END |
      instregmd         :  WITH Instregmd DO
                              RETURN( M2Time.InstRegMD(inst, unit,
                                                       destreg, o1, go,
                                                       scope, srcsym) )
                           END |
      instmdreg         :  WITH Instmdreg DO
                              RETURN( M2Time.InstMDReg(inst, unit,
                                                       o1, go, srcreg,
                                                       scope, destsym) )
                           END |
      instmdconstant    :  WITH Instmdconstant DO
                              RETURN( M2Time.InstMDConstant(inst, unit,
                                                            o1, go, scope,
                                                            destsym, srcsym) )
                           END |
      instmrireg        :  WITH Instmrireg DO
                              RETURN( M2Time.InstMRIReg(inst, unit, o1, go,
                                                        destreg, srcreg) )
                           END | 
      instmricardinal   :  WITH Instmricardinal DO
                              RETURN( M2Time.InstMRICardinal(inst, unit, o1, go,
                                                             destreg, srccard) )
                           END |
      instmriconstant   :  WITH Instmriconstant DO
                              RETURN( M2Time.InstMRIConstant(inst, unit, o1, go,
                                                             destreg, srcsym) )
                           END |
      instmrisymconstant:  WITH Instmrisymconstant DO
                              RETURN( M2Time.InstMRIConstant(inst, unit, destsym,
                                                             o1+go, destreg, srcsym) )
                           END |
      instregmio        :  WITH Instregmio DO
                              RETURN( M2Time.InstRegMIO(inst, unit,
                                                        destreg, srcreg, go+o1) )
                           END |
      instregmisym      :  WITH Instregmisym DO
                              RETURN( M2Time.InstRegMISym(inst, unit,
                                                          destreg, srcreg, go+o1, srcsym) )
                           END |
      instregmi         :  WITH Instregmi DO
                              RETURN( M2Time.InstRegMI(inst, unit, destreg, srcreg) )
                           END |
      instmioreg        :  WITH Instmioreg DO
                              RETURN( M2Time.InstMIOReg(inst, unit, destreg, go+o1, srcreg) )
                           END |
      instmisymreg      :  WITH Instmisymreg DO
                              RETURN( M2Time.InstMISymReg(inst, unit,
                                                          srcreg, destreg, go+o1, destsym) )
                           END |
      instmdint         :  WITH Instmdint DO
                              RETURN( M2Time.InstMDInt(inst, unit,
                                                       o1, go, srcint,
                                                       scope, destsym) )
                           END |
      instmdoffset      :  WITH Instmdoffset DO
                              RETURN( M2Time.InstMDOffset(inst, unit, destsym, srcsym) )
                           END |
      instregoffset     :  WITH Instregoffset DO
                              RETURN( M2Time.InstRegOffset(inst, unit, destreg, srcsym) )
                           END |
      instkey           :  WITH Instkey DO
                              RETURN( M2Time.InstKey(inst, string) )
                           END

      ELSE
         InternalError('unknown instruction type', __FILE__, __LINE__)
      END
   END
END GetCost ;


(*
   OperandPair - returns TRUE if the operands to instructions, i,
                 and, j, pair.
*)

PROCEDURE OperandPair (i, j: InstrIndex) : BOOLEAN ;
BEGIN
(*
   WriteString('# ') ; DisplayInstruction(i) ;
   WriteString(' ; ') ; DisplayInstruction(j) ;
   WriteString(' operands do') ;
   IF (IsDestSrc(j, i) OR IsDestDest(i, j)) OR
      (IsStackBasedInst(i) AND IsStackBasedInst(j))
   THEN
      WriteString(' NOT')
   END ;
   WriteString(' pair') ; WriteLn ;
*)
   IF IsDestSrc(j, i) OR IsDestDest(i, j)
   THEN
      RETURN( FALSE )
   ELSIF IsStackBasedInst(i) AND IsStackBasedInst(j) AND
         (NOT IsStackBasedPair(i, j))
   THEN
      RETURN( FALSE )
   ELSE
      RETURN( TRUE )
   END
END OperandPair ;


(*
   InstructionUsesConstantOperand - returns TRUE if instruction uses
                                    a constant operand.
*)

PROCEDURE InstructionUsesConstantOperand (t: MetaInstruction) : BOOLEAN ;
BEGIN
   RETURN(
          (t=instregconstant) OR (t=instconstantreg) OR (t=instmdconstant) OR
          (t=instmdint) OR (t=instmdoffset) OR (t=instregoffset) OR (t=instregimmname) OR
          (t=instregint) OR (t=instmriconstant) OR (t=instmrisymconstant) OR (t=instmricardinal)
         )
END InstructionUsesConstantOperand ;


(*
   AGI - returns TRUE if instruction, i, and instruction, j, do
         cause an AGI. pi and pj are the previous instructions.
*)

PROCEDURE AGI (i, j, pi, pj: InstrIndex) : BOOLEAN ;
VAR
   iT, jT  : MetaInstruction ;
   BaseRegI,
   BaseRegJ,
   RegI,
   RegJ    : CARDINAL ;
   AgiFound: BOOLEAN ;
BEGIN
   iT := GetInstructionType(i) ;
   jT := GetInstructionType(j) ;
   IF AGIWithConstantOperands
   THEN
      (*  ** do AGIs really occur when we have two constant operands? **  *)
      AgiFound := InstructionUsesConstantOperand(iT) AND InstructionUsesConstantOperand(jT) ;
      IF DebugOn
      THEN
         IF AgiFound
         THEN
            M2AS.TurnLineFeedOff ;
            WriteString(' # AGI (constants) found  ' ) ; DisplayInstruction(i) ; DisplayInstruction(j) ;
            WriteString(' yes') ; WriteLn ;
            M2AS.TurnLineFeedOn
         END
      END ;
      IF AgiFound
      THEN
         RETURN( TRUE )
      END
   END ;
   RegI     := GetDest(pi) ;
   RegJ     := GetDest(pj) ;
   BaseRegI := GetBaseReg(i) ;
   BaseRegJ := GetBaseReg(j) ;
   AgiFound := IsRegister(RegI) AND ((RegI=BaseRegI) OR (RegI=BaseRegJ)) ;
   IF DebugOn AND AgiFound
   THEN
      M2AS.TurnLineFeedOff ;
      WriteString(' # AGI (registers) found between  ') ;
      DisplayInstruction(i) ; DisplayInstruction(pi) ; DisplayInstruction(pj) ;
      WriteLn ;
      M2AS.TurnLineFeedOn
   END ;
   IF AgiFound
   THEN
      RETURN( TRUE )
   ELSE
      AgiFound := IsRegister(RegJ) AND ((RegJ=BaseRegI) OR (RegJ=BaseRegJ)) ;
      IF AgiFound AND DebugOn
      THEN
         M2AS.TurnLineFeedOff ;
         WriteString(' # AGI (registers) found between  ') ;
         DisplayInstruction(j) ; DisplayInstruction(pi) ; DisplayInstruction(pj) ;
         WriteLn ;
         M2AS.TurnLineFeedOn
      END
   END ;
   RETURN( AgiFound )
END AGI ;


(*
   ABS - returns the absolute number of i.
*)

PROCEDURE ABS (i: INTEGER) : INTEGER ;
BEGIN
   IF i<0
   THEN
      RETURN( -i )
   ELSE
      RETURN( i )
   END
END ABS ;


(*
   NoCacheConflict - returns TRUE if there will be no cache line conflict
                     between the operands, i, and, j.
                     This occurs if the addresses of, i, and, j, differ
                     by a multiple of 32.
*)

PROCEDURE NoCacheConflict (i, j: InstrIndex) : BOOLEAN ;
VAR
   iFather, jFather: CARDINAL ;
   iUnit  , jUnit  : CARDINAL ;
   iOff   , jOff   : INTEGER ;
BEGIN
   IF IsRegister(GetSrc(i)) OR IsRegister(GetDest(j))
   THEN
      (* no cache conflict if one of the potential conflict operands is a register *)
      RETURN( TRUE )
   ELSE
      GetComplexSrc(i, iFather, iUnit, iOff) ;
      GetComplexDest(j, jFather, jUnit, jOff) ;
      RETURN(
             (iFather#jFather) OR
             (NOT IsCacheLineConflict(iOff, INTEGER(iUnit), jOff, INTEGER(jUnit)))
            )
   END
END NoCacheConflict ;


(*
   OperandIsRegisterOnly - returns TRUE if the operand is only a register.
                           (I've also included constants here as well,
                            not sure whether this is correct?  -- FIXME --)
*)

PROCEDURE OperandIsRegisterOnly (i: InstrIndex) : BOOLEAN ;
BEGIN
   CASE GetInstructionType(i) OF

   instregconstant,
   instconstantreg,
   instregreg,
   instregint,
   bininst,
   bininstreg,
   bininsthex     :  RETURN( TRUE )

   ELSE
      RETURN( FALSE )
   END
END OperandIsRegisterOnly ;


(*
   OperandReadModify - returns TRUE if instruction, i, causes a read from
                       memory with computation. Non pairable instructions return FALSE.
*)

PROCEDURE OperandReadModify (i: InstrIndex) : BOOLEAN ;
BEGIN
   CASE GetInstruction(i) OF

   add, sub, and, or,
   sal, xor, sar, shl, shr: CASE GetInstructionType(i) OF

                            instregmd:  RETURN( TRUE )

                            ELSE
                               RETURN( FALSE )
                            END
   ELSE
      RETURN( FALSE )
   END
END OperandReadModify ;


(*
   OperandReadModifyWrite - returns TRUE if instruction, i, causes a read from
                            memory with computation and a write.
                            Non pairable instructions return FALSE.
*)

PROCEDURE OperandReadModifyWrite (i: InstrIndex) : BOOLEAN ;
BEGIN
   CASE GetInstruction(i) OF

   not, neg, inc, dec: CASE GetInstructionType(i) OF

                       bininstmd:  RETURN( TRUE )

                       ELSE
                          RETURN( FALSE )
                       END
   ELSE
      RETURN( FALSE )
   END
END OperandReadModifyWrite ;


(*
   SetsCondition - returns TRUE if, inst, sets the conditional flags.
*)

PROCEDURE SetsCondition (inst: Instruction) : BOOLEAN ;
BEGIN
   RETURN( (inst=cmp) OR (inst=test) )
END SetsCondition ;


(*
   IsConditional - returns TRUE if instruction, inst, is a conditional instruction.
*)

PROCEDURE IsConditional (inst: Instruction) : BOOLEAN ;
BEGIN
   RETURN(
          (inst=je) OR (inst=jne) OR (inst=jg) OR (inst=jl) OR
          (inst=jge) OR (inst=jle) OR (inst=jae)
         )
END IsConditional ;


(*
   IsSimpleInst - returns TRUE if the instruction, j, will pair without
                  penilty.
*)

PROCEDURE IsSimpleInst (j: InstrIndex) : BOOLEAN ;
VAR
   jInst: Instruction ;
   jType: MetaInstruction ;
BEGIN
   jInst := GetInstruction(j) ;
   jType := GetInstructionType(j) ;
   RETURN(
          (jInst=cmp) OR (jInst=mov) OR (jType=instname) OR (jInst=lea) OR
          (jInst=call) OR (jInst=jmp) OR IsConditional(jInst)
         )
END IsSimpleInst ;


(*
   SetCost - sets the cost matrix for an instruction.
*)

PROCEDURE SetCost (i: PipeIndex; ConflictCost, AGICost, OperandCost, CacheMiss, Total: CARDINAL) ;
BEGIN
   WITH CurrentPipeLine.Pipe[i].Time DO
      Cost     := Total ;
      Conflict := ConflictCost ;
      AGI      := AGICost ;
      Operand  := OperandCost ;
      Cache    := CacheMiss
   END
END SetCost ;


(*
   SecondInstructionAdditionalCycles - returns the number of additional instruction
                                       cycles needed for this pipeline.
*)

PROCEDURE SecondInstructionAdditionalCycles (CurrentCycles: CARDINAL;
                                             j            : InstrIndex) : CARDINAL ;
BEGIN
   IF OperandIsRegisterOnly(j) OR IsSimpleInst(j)
   THEN
      RETURN( 0+CurrentCycles )          (* 0 additional cycles *)
   ELSIF OperandReadModify(j)
   THEN
      RETURN( 1+CurrentCycles )
   ELSIF OperandReadModifyWrite(j)
   THEN
      RETURN( 2+CurrentCycles )
   ELSE
      InternalError('unknown operand of instruction', __FILE__, __LINE__)
   END
END SecondInstructionAdditionalCycles ;


(*
   ImperfectPairing - returns the additional cost of imperfect pairing if it
                      should occur.

                      From the pentium optimization document

   10.4 If a read/modify/write instruction is paired with a read/modify or
        read/modify/write instruction, then they will pair imperfectly.

   The number of clock cycles used is given in the following table:

                         |             Second instruction
                         | MOV or             read/       read/modify/
   First  instruction    | register only      modify      write
   ----------------------|----------------------------------------------
   MOV or register only  |      1               2              3
   read/modify           |      2               2              3
   read/modify/write     |      3               4              5
   ----------------------|-----------------------------------------------

   Example:
   ADD [mem1], EAX / ADD EBX, [mem2]  ; 4 clock cycles
   ADD EBX, [mem2] / ADD [mem1], EAX  ; 3 clock cycles
*)

PROCEDURE ImperfectPairing (i, j: InstrIndex) : CARDINAL ;
BEGIN
   IF OperandIsRegisterOnly(i) OR IsSimpleInst(i)
   THEN
      RETURN( SecondInstructionAdditionalCycles(0, j) )
   ELSIF OperandReadModify(i)
   THEN
      RETURN( SecondInstructionAdditionalCycles(1, j) )
   ELSIF OperandReadModifyWrite(i)
   THEN
      RETURN( SecondInstructionAdditionalCycles(2, j) )
   ELSE
      InternalError('unknown first instruction category (add instruction to the cases above)', __FILE__, __LINE__)
   END
END ImperfectPairing ;


(*
   SpecialCaseAGI - checks for instructions which will generate AGIs
*)

PROCEDURE SpecialCaseAGI (i: InstrIndex; VAR AGICost: CARDINAL) ;
BEGIN
   IF GetInstruction(i)=leave
   THEN
      INC(AGICost)
   END
END SpecialCaseAGI ;


(*
   CostOfUV - returns the total cost of pairing instructions, IU, IV.
              PIV is the previous V piped instruction.
*)

PROCEDURE CostOfUV (IU, IV, PIU, PIV: InstrIndex; h, head: PipeIndex;
                    VAR ConflictCost, OperandCost,
                        CacheMiss, PairCost, AGICost: CARDINAL) : CARDINAL ;
VAR
   PairAbilityU,
   PairAbilityV: Pairings ;
   CostI,
   CostU,
   CostV       : CARDINAL ;
BEGIN
   PairAbilityU := GetPairings(IU) ;
   PairAbilityV := GetPairings(IV) ;
   IF AGI(IU, IV, PIU, PIV)
   THEN
      INC(AGICost, 1)
   END ;
   IF ((GetInstruction(IU)=call) OR IsConditional(GetInstruction(IU))) AND ((GetInstruction(IV)#nop) OR (h>1))
   THEN
      RETURN( Infinity )     (* we do not allow a call, cmp or test before another instruction *)
   ELSIF ((GetInstruction(IV)=call) OR IsConditional(GetInstruction(IV))) AND (h>1)
   THEN
      (* same again but now we check that the V pipe is not being used for a call with instructions
       following *)
      RETURN( Infinity )
   ELSIF IsConditional(GetInstruction(IV)) AND (h=1) AND
         (NOT SetsCondition(GetInstruction(IU))) AND (head>1) AND (IU#NopIndex)
   THEN
      (* we must test ; jcc  and not put an instruction inbetween.. *)
      RETURN( Infinity )
   ELSIF IsConditional(GetInstruction(IU)) AND (h=1) AND (NOT SetsCondition(GetInstruction(PIV)))
   THEN
      (*
         we must test ; jcc  and not put an instruction inbetween.
         This time U contains jcc and hopefully PIV the test.
      *)
      RETURN( Infinity )
   ELSIF (((PairAbilityU=uv) OR (PairAbilityU=pu)) AND
          ((PairAbilityV=pv) OR (PairAbilityV=uv))) OR
         IsStackBasedPair(IU, IV)
   THEN
      CostU := GetCost(IU) ;
      CostV := GetCost(IV) ;
      PairCost := Max(CostU, CostV) ;
      IF NOT OperandPair(IU, IV)
      THEN
         INC(OperandCost)
      END ;
      IF NOT NoCacheConflict(IU, IV)
      THEN
         INC(ConflictCost)
      END ;
      RETURN( Min(OperandCost + AGICost + ImperfectPairing(IU, IV), 5) + ConflictCost + PairCost )
   ELSIF PairAbilityU=np
   THEN
      Assert(IV=NopIndex) ;   (* v slot should not be used if np *)
      SpecialCaseAGI(IU, AGICost) ;
      RETURN( GetCost(IU) + AGICost )
   ELSIF PairAbilityV=np
   THEN
      Assert(IU=NopIndex) ;   (* u slot should not be used if np *)
      SpecialCaseAGI(IV, AGICost) ;
      RETURN( GetCost(IV) + AGICost )
   ELSE
      PairCost := GetCost(IU) + GetCost(IV) ;
      Assert(OptimizeWorst) ;   (* should never reach here if performing real optimization *)
      RETURN( PairCost )   (* this is not quite accurate as it does not know about IV pairing with the next IU instruction *)
   END
END CostOfUV ;


(*
   DetermineCost - returns the total cost of the pipeline, at point, h
*)

PROCEDURE DetermineCost (h: PipeIndex) : CARDINAL ;
VAR
   PIU,
   PIV,
   IU,
   IV          : InstrIndex ;
   ConflictCost,
   OperandCost,
   CacheMiss,
   PairCost,
   AGICost,
   Total       : CARDINAL ;
BEGIN
   ConflictCost := 0 ;
   AGICost      := 0 ;
   OperandCost  := 0 ;
   CacheMiss    := 0 ;
   PairCost     := 0 ;

   WITH CurrentPipeLine DO
      IV := Pipe[h].V ;
      IU := Pipe[h].U ;
      IF h<Head
      THEN
         PIV := Pipe[h+1].V ;
         PIU := Pipe[h+1].U ;
      ELSE
         PIV := NopIndex ;
         PIU := NopIndex
      END ;

      Total := CostOfUV(IU, IV, PIU, PIV, h, Head, ConflictCost, OperandCost, CacheMiss, PairCost, AGICost)
   END ;

   CacheMiss := CacheCost(h) ;
   INC(Total, CacheMiss) ;
   IF (AGICost>0) AND (CacheMiss>0)
   THEN
      (* agi and cache miss can occur in parallel *)
      DEC(Total, AGICost)
   END ;
   SetCost(h, ConflictCost, AGICost, OperandCost, CacheMiss, Total) ;
   IF DebugOn AND FALSE
   THEN
      M2AS.TurnLineFeedOff ;
      WITH CurrentPipeLine DO
         WriteString(' # determine cost ' ) ; DisplayInstruction(IU) ; DisplayInstruction(IV)
      END ;
      WriteMatrix(CurrentPipeLine.Pipe[h].Time) ; WriteLn ;
      M2AS.TurnLineFeedOn
   END ;
   RETURN( Total )
END DetermineCost ;


(*
   RecalculateCost - returns the total cost of the current pipe.
*)

PROCEDURE RecalculateCost () : CARDINAL ;
VAR
   i   : PipeIndex ;
   cost: CARDINAL ;
BEGIN
   WITH CurrentPipeLine DO
      Cache := OldCache ;
      i := Head ;
      cost := 0 ;
      WHILE i>0 DO
         INC(cost, DetermineCost(i)) ;
         DEC(i)
      END
   END ;
   RETURN( cost )
END RecalculateCost ;


(*
   CostPushCache - returns the cost of touching the stack for an instruction.
*)

PROCEDURE CostPushCache (i: InstrIndex) : CARDINAL ;
VAR
   unit: CARDINAL ;
BEGIN
   IF GetInstruction(i)=push
   THEN
      unit := GetUnit(i) ;
      WITH CurrentPipeLine.Cache DO
         DEC(StackPtr, unit) ;
         RETURN( AddToCache(ESP, unit, StackPtr, FALSE) )   (* we assume that we have read from stack sometime *)
      END
   ELSE
      RETURN( 0 )
   END
END CostPushCache ;


(*
   CostPopCache - returns the cost of touching the stack for an instruction.
*)

PROCEDURE CostPopCache (i: InstrIndex) : CARDINAL ;
VAR
   unit: CARDINAL ;
BEGIN
   IF GetInstruction(i)=pop
   THEN
      unit := GetUnit(i) ;
      WITH CurrentPipeLine.Cache DO
         INC(StackPtr, unit) ;
         RETURN( AddToCache(ESP, unit, StackPtr+unit, FALSE) )
      END
   ELSE
      RETURN( 0 )
   END
END CostPopCache ;


(*
   IsIndirectDest - returns TRUE if instruction, i, has an indirect
                    destination address.
*)

PROCEDURE IsIndirectDest (i: InstrIndex) : BOOLEAN ;
VAR
   t  : MetaInstruction ;
   reg: CARDINAL ;
BEGIN
   t := GetInstructionType(i) ;
   IF (t=instmdconstant) OR (t=instmdint) OR (t=instmdoffset) OR (t=instmdreg) OR
      (t=instmriconstant) OR (t=instmrisymconstant) OR (t=instmricardinal) OR (t=instmrireg) OR
      (t=instmioreg) OR (t=instmisymreg) OR (t=instmireg) OR (t=bininstregi)
   THEN
      reg := GetDest(i) ;
      IF (reg#EBP) AND (reg#NulSym)
      THEN
         IF DebugOn
         THEN
            WriteString('# found indirect dest  ') ; DisplayInstruction(i)
         END
      END ;
      RETURN( (reg#EBP) AND (reg#NulSym) )
   ELSE
      RETURN( FALSE )
   END
END IsIndirectDest ;


(*
   IsIndirectSrc - returns TRUE if instruction, i, has an indirect
                   source address.
*)

PROCEDURE IsIndirectSrc (i: InstrIndex) : BOOLEAN ;
VAR
   t  : MetaInstruction ;
   reg: CARDINAL ;
BEGIN
   t := GetInstructionType(i) ;
   IF (t=instregmio) OR (t=instregmisym) OR (t=instregmi) OR (t=bininstregi)
   THEN
      reg := GetSrc(i) ;
      IF (reg#EBP) AND (reg#NulSym)
      THEN
         IF DebugOn
         THEN
            WriteString('# found indirect src  ') ; DisplayInstruction(i)
         END
      END ;
      RETURN( (reg#EBP) AND (reg#NulSym) )
   ELSE
      RETURN( FALSE )
   END
END IsIndirectSrc ;


(*
   CacheCostInstr - returns the cache cost of instruction, i.
                    NOTE that this has the side effect of adding
                    the operand to the cache line model.
*)

PROCEDURE CacheCostInstr (i: InstrIndex) : CARDINAL ;
VAR
   sFather,
   sUnit  : CARDINAL ;
   sOff   : INTEGER ;
   cost   : CARDINAL ;
   iInst  : Instruction ;
BEGIN
   IF (i=NopIndex) OR (GetInstruction(i)=lea)
   THEN
      RETURN( 0 )
   ELSE
      (* cost := CostPopCache(i) ; *)
(*
      IF GetInstruction(i)=call
      THEN
         ResetCurrentCache(filled)
      END ;
*)
      IF GetInstruction(i)=repmovs
      THEN
         ResetCurrentCache(filled)
      END ;

      cost := 0 ;
      IF cost>100
      THEN
         InternalError('why so large?', __FILE__, __LINE__)
      END ;

      (* first the read *)
      IF NOT IsRegister(GetSrc(i))
      THEN
         GetComplexSrc(i, sFather, sUnit, sOff) ;
         IF (sFather#NulSym) AND ((NOT IsProcedure(sFather)) OR (sOff>0))
         THEN
            cost := AddToCache(sFather, sUnit, sOff, FALSE)
            ; IF cost>100
            THEN
               InternalError('why so large?', __FILE__, __LINE__)
            END ;
         END ;
(*
         IF (cost<CacheMissPenilty) AND IsIndirectSrc(i) AND (NOT IsProcedure(sFather))
         THEN
            cost := CacheMissPenilty ;
            IF DebugOn
            THEN
               WriteLn ;
               WriteString('# ') ; DisplayInstruction(i) ; WriteString('  <-------- causes read cache problems') ;
               WriteLn ;
               WriteString('# addr = ') ; WriteInt(sOff, 0) ; WriteString('..') ; WriteInt(sOff+sUnit, 0) ; WriteLn ;
               WriteLn
            END
         END
*)
      END ;

      (* and now the write *)
      IF NOT IsRegister(GetDest(i))
      THEN
         GetComplexDest(i, sFather, sUnit, sOff) ;
         IF sFather#NulSym
         THEN
            IF IsProcedure(sFather) AND (sOff<0)
            THEN
               (* always assume that paramaters have been read/pushed to stack, thus suffer no write miss *)
               cost := AddToCache(sFather, sUnit, sOff, FALSE)
            ELSE
               IF DebugOn
               THEN
                  DisplayCache ;
                  IF (i=6) OR (i=5)
                  THEN
                     DumpCurrentPipeLine ;   (* note the cost maybe incorrect *)
                     Dummy
                  END
               END ;
               cost := AddToCache(sFather, sUnit, sOff, TRUE)
               ; IF (cost>=3) AND DebugOn
               THEN
                  WriteString('# ') ; DisplayInstruction(i) ; WriteString('  <-------- causes write cache problems') ;
                  WriteLn ;
                  WriteString('# addr = ') ; WriteInt(sOff, 0) ; WriteString('..') ; WriteInt(sOff+sUnit, 0) ; WriteLn ;
                  WriteLn
               END
            END ;
            ; IF cost>100
            THEN
               InternalError('why so large?', __FILE__, __LINE__)
            END
         END
      END ;
      IF IsIndirectDest(i) AND (cost<CacheWriteNoReadPenilty)
      THEN
         cost := CacheWriteNoReadPenilty
      END ;

      (* INC(cost, CostPushCache(i)) ; *)
      IF cost>100
      THEN
         InternalError('why so large?', __FILE__, __LINE__)
      END ;
      RETURN( cost )
   END
END CacheCostInstr ;


(*
   CacheCost - returns the cost in cycles for a cache miss for the two instructions
               at position, Head, in the current pipe line.
*)

PROCEDURE CacheCost (h: PipeIndex) : CARDINAL ;
BEGIN
   WITH CurrentPipeLine.Pipe[h] DO
      RETURN( CacheCostInstr(U) + CacheCostInstr(V) )
   END
END CacheCost ;


(*
   CanMakeSpecialPair - checks for special normally np instructions which
                        when combined with another instruction will pair.
                        -- FIXME -- this is true for the Pentium NOT sure
                        about the Pentium Pro
*)

PROCEDURE CanMakeSpecialPair (h: CARDINAL) : CARDINAL ;
VAR
   PairV: Pairings ;
   CurV,
   LastU,
   LastV: InstrIndex ;
BEGIN
   IF Optimize686
   THEN
      (* -- FIXME --   does the 686 have special pair instructions ?? *)
   END ;
   IF h>0
   THEN
      WITH CurrentPipeLine DO
         LastV := Pipe[h-1].V ;
         LastU := Pipe[h-1].U ;
         CurV  := Pipe[h].V ;
         PairV := GetPairings(LastV) ;
         IF (PairV=np) AND (LastU=NopIndex) AND IsStackBasedPair(LastV, CurV)
         THEN
            IF DebugOn
            THEN
               WriteString('# special pair found: ') ;
               DisplayInstruction(Pipe[h-1].V) ;
               WriteString(' ; ') ; DisplayInstruction(Pipe[h].V) ; WriteLn
            END ;

            (* ahh, they will pair so lets move CurV to the U slot of the LastV *)
            Assert(Pipe[h].U=NopIndex) ;    (* Last U must be nop, right..      *)
            Pipe[h-1].U := CurV ;
            Pipe[h].V   := NopIndex ;
            DEC(h) ;
            Head := h
         END
      END
   END ;
   RETURN( h )
END CanMakeSpecialPair ;


(*
   AddToVPipeWorst - adds instruction, i, to the V pipe and
                     calculate the cost of the instruction.
                     We add to the V pipe before we add to the U pipe
                     and thus assume the U pipe to be empty.

                     We purposly make the worst decisions...
*)

PROCEDURE AddToVPipeWorst (i: InstrIndex) ;
VAR
   cache: CARDINAL ;
   PairV: Pairings ;
   old  : PipeLineState ;
BEGIN
   WITH CurrentPipeLine DO
      INC(Head) ;
      Pipe[Head].U := NopIndex ;   (* blank out U slot as it is uninitialized *)
      Pipe[Head].V := i ;          (* instruction will go in V pipe *)
      PairV := GetPairings(i) ;

      IF (PairV=uv) OR (PairV=pv)
      THEN
         Next := u
      ELSIF PairV=np
      THEN
         Head := CanMakeSpecialPair(Head) ;
         (*
            either we cannot use the U slot, or we have paired special instructions,
            in both cases we choose the v pipe for the next instruction 
         *)
         Next := v
      ELSE
         Assert(PairV=pu) ;
         (* goodie lets create a pipeline stall.., (pu in v slot) *)
         Next := u
      END ;
      Cost := RecalculateCost() ;
   END
END AddToVPipeWorst ;


(*
   AddToUPipeWorst - adds instruction, i, to the U pipe and
                     calculate the cost of the instruction.
*)

PROCEDURE AddToUPipeWorst (i: InstrIndex) ;
VAR
   PairU: Pairings ;
BEGIN
   WITH CurrentPipeLine DO
      (* instruction will go in U pipe *)
      PairU := GetPairings(i) ;
      Next := v ;
      Assert(Pipe[Head].U=NopIndex) ;
      IF (PairU=np) OR (PairU=pv)
      THEN
         (* must place it in V *)
         AddToCurrentPipeLine(i)
      ELSE
         Pipe[Head].U := i ;
         Cost := RecalculateCost()
      END
   END
END AddToUPipeWorst ;


(*
   AddToVPipe - adds instruction, i, to the V pipe and
                calculate the cost of the instruction.
                We add to the V pipe before we add to the U pipe
                and thus assume the U pipe to be empty.
*)

PROCEDURE AddToVPipe (i: InstrIndex) ;
VAR
   cache: CARDINAL ;
   PairV: Pairings ;
   old  : PipeLineState ;
BEGIN
   WITH CurrentPipeLine DO
      INC(Head) ;
      Pipe[Head].U := NopIndex ;   (* blank out U slot as it is uninitialized *)
      Pipe[Head].V := i ;          (* instruction will go in V pipe *)
      PairV := GetPairings(i) ;

      IF ((PairV=uv) OR (PairV=pv)) AND OperandPair(Pipe[Head].U, Pipe[Head].V)
      THEN
         Cost := RecalculateCost() ;
         Next := u
      ELSIF PairV=np
      THEN
         Head := CanMakeSpecialPair(Head) ;
         Cost := RecalculateCost() ;
         (*
            either we cannot use the U slot, or we have paired special instructions,
            in both cases we choose the v pipe for the next instruction 
         *)
         Next := v
      ELSE
         Assert(PairV=pu) ;
         Pipe[Head].V := NopIndex ;    (* nop *)
         Next := u ;
         AddToCurrentPipeLine(i)                   (* adds, i, to U pipe *)
      END
   END
END AddToVPipe ;


(*
   AddToUPipe - adds instruction, i, to the U pipe and
                calculate the cost of the instruction.
*)

PROCEDURE AddToUPipe (i: InstrIndex) ;
VAR
   cache: CARDINAL ;
   PairU: Pairings ;
BEGIN
   WITH CurrentPipeLine DO
      (* instruction will go in U pipe *)
      PairU := GetPairings(i) ;
      Next := v ;   (* whatever the pairability, we will always try adding the next instruction to V *)
      IF ((PairU=pu) OR (PairU=uv)) AND ((Pipe[Head].V=NopIndex) OR OperandPair(i, Pipe[Head].V))
      THEN
         Assert(Pipe[Head].U=NopIndex) ;
         Pipe[Head].U := i ;
         Cost := RecalculateCost() ;
      ELSE
         (* cannot pipeline, i, in U pipe, it already contains a nop so we just try adding it to V *)
         AddToCurrentPipeLine(i)
      END
   END
END AddToUPipe ;


(*
   AddToCurrentPipeLine - adds instruction, i, to the current pipeline
                          and determine the cost.
*)

PROCEDURE AddToCurrentPipeLine (i: InstrIndex) ;
BEGIN
   WITH CurrentPipeLine DO
      IF Next=v
      THEN
         IF OptimizeWorst
         THEN
            AddToVPipeWorst(i)
         ELSE
            AddToVPipe(i)
         END
      ELSE
         IF OptimizeWorst
         THEN
            AddToUPipeWorst(i)
         ELSE
            (* instruction will go in U pipe *)
            AddToUPipe(i)
         END
      END
   END
END AddToCurrentPipeLine ;


(*
   VerifyCost - verifies the cost of the current pipeline.
*)

PROCEDURE VerifyCost ;
VAR
   i   : CARDINAL ;
   cost: CARDINAL ;
   old : PipeLineState ;
BEGIN
   old := CurrentPipeLine ;
   cost := RecalculateCost() ;
   IF old.Cost#cost
   THEN
      WriteString('# recalculated as ') ; WriteCard(cost, 0) ; WriteLn ;
      WriteString('# was') ; WriteLn ;
      DumpCurrentPipeLine ;
      CurrentPipeLine := old ;
      DumpCurrentPipeLine ;
      InternalError('inconsistant cost in pipeline', __FILE__, __LINE__)
   END ;
   old := CurrentPipeLine ;
   cost := RecalculateCost() ;
   IF old.Cost#cost
   THEN
      WriteString('# recalculated as ') ; WriteCard(cost, 0) ; WriteLn ;
      DumpCurrentPipeLine ;
      WriteString('# was') ; WriteLn ;
      CurrentPipeLine := old ;
      DumpCurrentPipeLine ;
      InternalError('inconsistant cost in pipeline', __FILE__, __LINE__)
   END ;
END VerifyCost ;


(*
   ReworkForward - examines a pipeline and reworks the pairings forward.
                   This only is used if the top instruction does not
                   form a pair.
*)

PROCEDURE ReworkForward ;
VAR
   PairU,
   PairV: Pairings ;
   i    : CARDINAL ;
BEGIN
   WITH CurrentPipeLine DO
      (*
         might have filled the V slot without U, better check for
         possible pairings between top V and top-1 U
      *)
      i := Head ;
      WHILE i>0 DO
         IF Pipe[i].U=NopIndex
         THEN
            PairU := GetPairings(Pipe[i].V) ;
            PairV := GetPairings(Pipe[i-1].U) ;  (* yes really switch U, V as no top U.. *)
            IF (* (((PairU=pu) OR (PairU=uv)) AND ((PairV=uv) OR (PairV=pv))) OR *)
               IsStackBasedPair(Pipe[i].V, Pipe[i-1].U)
            THEN
               IF DebugOn
               THEN
                  WriteString('# found forward pairing: ') ;
                  DisplayInstruction(Pipe[i].V) ; DisplayInstruction(Pipe[i-1].U) ; WriteLn
               END ;
               Pipe[i].U   := Pipe[i].V ;
               Pipe[i].V   := Pipe[i-1].U ;
               Pipe[i-1].U := NopIndex
            END
         ELSIF Pipe[i].V=NopIndex
         THEN
            PairU := GetPairings(Pipe[i].U) ;
            PairV := GetPairings(Pipe[i-1].U) ;  (* yes really switch U, V as no top V.. *)
            IF (* (((PairU=pu) OR (PairU=uv)) AND ((PairV=uv) OR (PairV=pv))) OR *)
               IsStackBasedPair(Pipe[i].U, Pipe[i-1].U)
            THEN
               IF DebugOn
               THEN
                  WriteString('# found forward pairing: ') ;
                  DisplayInstruction(Pipe[i].U) ; DisplayInstruction(Pipe[i-1].U) ; WriteLn
               END ;
               Pipe[i].V   := Pipe[i-1].U ;
               Pipe[i-1].U := NopIndex
            END
         END ;
         DEC(i)
      END
   END ;
   CurrentPipeLine.Cost := RecalculateCost()
END ReworkForward ;


(*
   AddChoices - adds future choices dependant upon, item.
*)

PROCEDURE AddChoices (item: InstrIndex; ChoiceList: List) ;
BEGIN
   IF Instr[item].Parent#0
   THEN
      IncludeItemIntoList(ChoiceList, Instr[item].Parent)
   END ;
   IF Instr[item].Branch1#0
   THEN
      IncludeItemIntoList(ChoiceList, Instr[item].Branch1)
   END ;
   IF Instr[item].Branch2#0
   THEN
      IncludeItemIntoList(ChoiceList, Instr[item].Branch2)
   END
END AddChoices ;


(*
   IsBetter - returns TRUE if, a, is better than, b.
*)

PROCEDURE IsBetter (a, b: CARDINAL) : BOOLEAN ;
BEGIN
   IF OptimizeWorst AND (a<Infinity) AND (b<Infinity)
   THEN
      RETURN( a>b )
   ELSE
      RETURN( a<b )
   END
END IsBetter ;


(*
   DumpChoiceList - display the choice of instructions available.
*)

PROCEDURE DumpChoiceList (ChoiceList: List) ;
VAR
   c, i: CARDINAL ;
BEGIN
   WriteString('# choices are: ') ;
   c := NoOfItemsInList(ChoiceList) ;
   i := 1 ;
   WHILE i<=c DO
      DisplayInstruction(GetItemFromList(ChoiceList, i)) ;
      INC(i)
   END ;
   WriteLn
END DumpChoiceList ;


(* Evaluate and CopyCache here *)

(*
   ResetCurrentCache - to a particular state.
*)

PROCEDURE ResetCurrentCache (new: CacheState) ;
VAR
   i, j: CARDINAL ;
BEGIN
   NewCacheLineCost := new ;
   WITH CurrentPipeLine.Cache DO
      StackPtr := 0 ;
      NoOfLinesUsed := 0
   END
(*
      FOR i := 0 TO MaxCacheLines-1 DO
         WITH Lines[i] DO
            FOR j := 0 TO MaxSameLines DO
               WITH SameLine[j] DO
                  AddressLo := 0 ;
                  AddressHi := 0 ;
                  State     := new
               END
            END ;
            MRULine := 1
         END
      END
   END
*)
END ResetCurrentCache ;


(*
   FindLastInstruction - finds the requested last instruction and places
                         it into the correct pipeline slot.
*)

PROCEDURE FindLastInstruction (ChoiceList: List) ;
BEGIN
   IF LastInstructionMustBe#NopIndex
   THEN
      IF IsItemInList(ChoiceList, CARDINAL(LastInstructionMustBe))
      THEN
         (* first time InitializeCurrentPipe has been called *)
         RemoveItemFromList(ChoiceList, CARDINAL(LastInstructionMustBe))
      END ;
      AddToCurrentPipeLine(LastInstructionMustBe) ;
      AddChoices(LastInstructionMustBe, ChoiceList)
   END
END FindLastInstruction ;


(*
   InitializeCurrentPipeLine - initialize the CurrentPipeLine variable,
                               UpipeLast, determines whether the U pipe
                               is to be filled last.
*)

PROCEDURE InitializeCurrentPipeLine (p: PipeNext; ChoiceList: List) ;
VAR
   i: CARDINAL ;
BEGIN
   WITH CurrentPipeLine DO
      Cache := OldCache ;
      Cost := 0 ;
      IF p=u
      THEN
         Head := 1
      ELSE
         Head := 0
      END ;
      Next := p ;
      Pipe[1].U := NopIndex ;
      Pipe[1].V := NopIndex ;
      FindLastInstruction(ChoiceList)
   END
END InitializeCurrentPipeLine ;


(*
   IsUPipeNops - returns TRUE if the U pipe only contains NOPs.
*)

PROCEDURE IsUPipeNops () : BOOLEAN ;
VAR
   i: InstrIndex ;
BEGIN
   WITH BestPipeLine DO
      i := Head ;
      WHILE i>0 DO
         IF GetInstruction(Pipe[i].U)#nop
         THEN
            RETURN( FALSE )
         END ;
         DEC(i)
      END
   END ;
   RETURN( TRUE )
END IsUPipeNops ;


(*
   DisplayLastInstructionPipe - displays the value of the global variable,
                                LastInstructionPipe.
*)
(*
PROCEDURE DisplayLastInstructionPipe ;
BEGIN
   CASE LastInstructionPipe OF

   Upipe  :  Comment('last instruction was in U pipe') |
   Vpipe  :  Comment('last instruction was in V pipe') |
   Unknown:  Comment('unknown last instruction pipe')

   ELSE 
      InternalError('unknown value of LastInstructionPipe', __FILE__, __LINE__)
   END
END DisplayLastInstructionPipe ;
*)


(*
   DumpBestPipeLine - debugging routine, display raw pipeline information.
*)

PROCEDURE DumpBestPipeLine ;
VAR
   i        : PipeIndex ;
   LocalCost: CARDINAL ;
BEGIN
   (* we always emit instructions in U,V pairs *)
   WriteString('# start of RAW pipeline') ; WriteLn ;
   M2AS.TurnLineFeedOff ;
   LocalCost := 0 ;
   WITH BestPipeLine DO
      WriteString('# U Pipe          ;  V Pipe') ; WriteLn ;
      i := Head ;
      WHILE i>0 DO
         WITH Pipe[i] DO
            IF (U=NopIndex) AND (V=NopIndex)
            THEN
               (* ignore if the last pair is two nops - ReworkForward might cause this *)
               IF i#1
               THEN
                  InternalError('why are u and v both nops?', __FILE__, __LINE__)
               END
            ELSE
               WriteCard(CARDINAL(i), 0) ; WriteString('  ') ;
               DisplayInstruction(U) ; WriteString('  ;  ') ;
               DisplayInstruction(V) ;
               WriteString('  #  ') ; WriteMatrix(Time) ;
               WriteLn ;
               INC(LocalCost, Time.Cost)
            END
         END ;
         DEC(i)
      END ;
   END ;
   WriteString('# end of pipeline: this pipeline will take ') ; WriteCard(LocalCost, 0) ;
   WriteString(' cycles to execute') ; WriteLn ;
   M2AS.TurnLineFeedOn
END DumpBestPipeLine ;


(*
   DumpCurrentPipeLine - debugging routine, display raw pipeline information.
*)

PROCEDURE DumpCurrentPipeLine ;
VAR
   old: PipeLineState ;
BEGIN
   (* a horrible, but safe hack.. *)
   old := BestPipeLine ;
   BestPipeLine := CurrentPipeLine ;
   DumpBestPipeLine ;
   BestPipeLine := old
END DumpCurrentPipeLine ;


(*
   WriteMatrix - writes out the matix of instruction costs.
*)

PROCEDURE WriteMatrix (m: CostMatrix) ;
BEGIN
   WriteString('  ') ;
   WITH m DO
      WriteString(' (') ;
      WriteCard(Cost, 0) ;
      WriteString(', ') ; WriteCard(Cache, 0) ;
      WriteString(', ') ; WriteCard(AGI, 0) ;
      WriteString(', ') ; WriteCard(Conflict, 0) ;
      WriteString(', ') ; WriteCard(Operand, 0) ;
      Write(')')
   END
END WriteMatrix ;


(*
   DisplayBestPipeLine - displays the best scheduled instructions.
*)

PROCEDURE DisplayBestPipeLine ;
VAR
   h, i       : PipeIndex ;
   IgnoreUPipe: BOOLEAN ;
   pu, pv     : Pairings ;
BEGIN
   CurrentPipeLine := BestPipeLine ;
   IF DebugOn
   THEN
      DumpBestPipeLine
   END ;
   IF BestPipeLine.Head=0
   THEN
      InternalError('pipe line failed to find a legal ordering?', __FILE__, __LINE__)
   END ;
   M2AS.TurnLineFeedOff ;
   WITH BestPipeLine DO
      IF (NOT APUCodeGenerator) AND (Optimize586 OR Optimize686)
      THEN
         WriteString('# this pipeline will take ') ; WriteCard(Cost, 0) ;
         WriteString(' cycles to execute') ;
         WriteLn ;
         WriteString('# U Pipe          ;  V Pipe               (cost, cache, agi, conf, opr)') ; WriteLn
      END ;
      IF Pedantic AND (Cost>=WarningTheshold)
      THEN
         PushOutput(IO.Write) ;
         FormatWarningMessage2('block of code will take a long time to execute (%d execution cycles)', Cost, 0,
                               QuadToTokenNo(GetLastQuadNo())) ;
         PopOutput
      END ;
      i := Head ;
      h := i ;
      WHILE i>0 DO
         WITH Pipe[i] DO
            IF NOT APUCodeGenerator
            THEN
               pu := GetPairings(U) ;
               pv := GetPairings(V) ;
               IF pu=np
               THEN
                  IF (U#NopIndex) OR Optimize586 OR Optimize686
                  THEN
                     DisplayInstruction(U)
                  END ;
                  IF V=NopIndex
                  THEN
                     IF Optimize586 OR Optimize686
                     THEN
                        WriteString(' # non pairable')
                     END
                  ELSE
                     (* this is not an error, some instructions can form special pairs ie push ; push *)
                     WriteString('  ;  ') ;
                     DisplayInstruction(V) ;
                     WriteString(' # special pair')
                  END ;
                  LeftOverInstruction := NopIndex
               ELSIF pv=np
               THEN
                  IF U=NopIndex
                  THEN
                     DisplayInstruction(V) ;
                     IF Optimize586 OR Optimize686
                     THEN
                        WriteString(' # non pairable')
                     END
                  ELSE
                     (* this is not an error, some instructions can form special pairs ie push ; push *)
                     DisplayInstruction(U) ;
                     IF (V#NopIndex) OR Optimize586 OR Optimize686
                     THEN
                        WriteString('  ;  ') ;
                        DisplayInstruction(V) ;
                        WriteString(' # special pair')
                     END
                  END ;
                  LeftOverInstruction := NopIndex
               ELSE
                  IF (i=1) AND (V=NopIndex) AND (NOT MustFlushInstruction)
                  THEN
                     (* last possible pipeline pair, are there a pair or a left over instruction? *)
                     LeftOverInstruction := U
                  ELSE
                     LeftOverInstruction := NopIndex ;
                     IF (U=NopIndex) AND (V=NopIndex)
                     THEN
                        IF i#1
                        THEN
                           InternalError('why are both U and V nops?', __FILE__, __LINE__)
                        END
                     ELSE
                        IF (U#NopIndex) OR Optimize586 OR Optimize686
                        THEN
                           DisplayInstruction(U) ;
                           IF (V#NopIndex) OR Optimize586 OR Optimize686
                           THEN
                              WriteString('  ;  ') ;
                              DisplayInstruction(V)
                           END
                        ELSE
                           DisplayInstruction(V)
                        END
                     END
                  END
               END
            END ;
            IF NOT APUCodeGenerator
            THEN
               IF Optimize586 OR Optimize686
               THEN
                  WriteString(' # ' ) ; WriteMatrix(Time)
               END ;
               WriteLn
            END ;
            DEC(i)
         END
      END
   END ;
   IF (NOT APUCodeGenerator) AND (Optimize586 OR Optimize686)
   THEN
      WriteString('# end of pipeline') ; WriteLn
   END ;
   M2AS.TurnLineFeedOn
END DisplayBestPipeLine ;


(*
   SchedulePipeLine - schedules the instructions in an reasonably
                      efficient manner.
*)

PROCEDURE SchedulePipeLine ;
VAR
   ChoiceList : List ;
   BestCost   : CARDINAL ;
BEGIN
   OldCache          := CurrentPipeLine.Cache ;
   ChoiceList        := InitChoices() ;
   BestCost          := Infinity ;
   BestPipeLine.Head := 0 ;
   InitializeCurrentPipeLine(v, ChoiceList) ;
   Evaluate(ChoiceList, BestCost) ;
   InitializeCurrentPipeLine(u, ChoiceList) ;
   Evaluate(ChoiceList, BestCost) ;
   DisplayBestPipeLine ;
   CurrentPipeLine := BestPipeLine
END SchedulePipeLine ;


(*
   FlushPipe - flushes the outstanding instructions held in the
               pipeline data structures.
*)

PROCEDURE FlushPipe (cycles: CARDINAL) ;
BEGIN
   INC(AccumulatedCost, cycles) ;
   FlushInstruction(FALSE)
END FlushPipe ;


(*
   DetermineCacheLines - 
*)

PROCEDURE DetermineCacheLines ;
BEGIN
   IF Optimize686
   THEN
      ActualSameCacheLines := NoOfLinesFor686 ;
      ActualCacheLines     := NoOfBytesFor686 DIV CacheBucket
   ELSE
      ActualSameCacheLines := NoOfLinesFor586 ;
      ActualCacheLines     := NoOfBytesFor586 DIV CacheBucket
   END
END DetermineCacheLines ;


(*
   FlushPipeline - flushes the outstanding instructions help in the
                   pipeline data structures.
*)

PROCEDURE FlushPipeline (FixLastInstruction: BOOLEAN) ;
BEGIN
   IF FixLastInstruction
   THEN
      LastInstructionMustBe := NoOfInstructions
   ELSE
      LastInstructionMustBe := NopIndex
   END ;
   IF NoOfTrees>0
   THEN
      IF DebugOn
      THEN
         PushOutput(IO.Write) ; (* debugging purposes *)
         DisplayPipeTrees ;
      END ;
      DetermineCacheLines ;
      SchedulePipeLine ;
      IF DebugOn
      THEN
         PopOutput
      END
   END ;
   IF Profiling AND APUCodeGenerator
   THEN
      Time(BestPipeLine.Cost+AccumulatedCost) ;
      BestPipeLine.Cost := 0
   END ;
   AccumulatedCost  := 0 ;
   NoOfTrees        := 0 ;
   NoOfInstructions := 0 ;
   IF LeftOverInstruction#NopIndex
   THEN
      Assert(NOT MustFlushInstruction) ;
      Instr[1] := Instr[LeftOverInstruction] ;
      LeftOverInstruction := 1 ;
      NoOfInstructions    := 1 ;
      AddDependant(LeftOverInstruction) ;
      LeftOverInstruction := 0
   END
END FlushPipeline ;


(*
   FlushInstruction - flushes the remaining solo instruction, if it exists.
                      It also fixes the last instruction if necessary.
*)

PROCEDURE FlushInstruction (FixLastInstruction: BOOLEAN) ;
BEGIN
   IF NOT IgnorePipeLine
   THEN
      MustFlushInstruction := TRUE ;
      FlushPipeline(FixLastInstruction) ;
      MustFlushInstruction := FALSE
   END
END FlushInstruction ;


(*
   GetInstructionType - 
*)

PROCEDURE GetInstructionType (i: InstrIndex) : MetaInstruction ;
BEGIN
   RETURN( Instr[i].InstructionType )
END GetInstructionType ;


(*
   NewInstruction - returns an index to a new instruction.
*)

PROCEDURE NewInstruction () : InstrIndex ;
BEGIN
   IF (NoOfInstructions=MaxInstruction) OR (NoOfInstructions>=OptimizeLength)
   THEN
      FlushPipeline(FALSE)
   END ;
   INC(NoOfInstructions) ;
   RETURN( NoOfInstructions )
END NewInstruction ;


(*
   MakeLink - assigns the parent field of the instruction of,
              from, to, to.
*)

PROCEDURE MakeLink (from, to: InstrIndex) ;
BEGIN
   WITH Instr[from] DO
      Parent  := to ;
      Branch1 := 0 ;
      Branch2 := 0
   END
END MakeLink ;


(*
   MakeBranch - makes a branch between, from, to, to.
*)

PROCEDURE MakeBranch (from, to: InstrIndex) ;
BEGIN
   WITH Instr[from] DO
      IF Branch1=0
      THEN
         Branch1 := to
      ELSIF Branch2=0
      THEN
         Branch2 := to
      ELSE
         InternalError('both branches have been used', __FILE__, __LINE__)
      END
   END
END MakeBranch ;


(*
   IsIntersection - returns TRUE if the range a..b intersects with c..d.
*)

PROCEDURE IsIntersection (a, b, c, d: INTEGER) : BOOLEAN ;
BEGIN
   (* easier to prove not outside limits *)
   RETURN( NOT ((a>=d) OR (b<=c)) )
END IsIntersection ;


(*
   IsSubSet - returns true if the range i1..i2 is a subset of j1..j2.
*)

PROCEDURE IsSubSet (i1, i2, j1, j2: INTEGER) : BOOLEAN ;
BEGIN
   RETURN( (j1<=i1) AND (j2>=i2) )
END IsSubSet ;


(*
   GetComplexSrc - fills in the various parameters which reflect
                   the source operand for instruction, i.
*)

PROCEDURE GetComplexSrc (i: InstrIndex;
                         VAR Father: CARDINAL ;
                         VAR Unit  : CARDINAL ;
                         VAR Off   : INTEGER) ;
BEGIN
   IF GetInstruction(i)=nop
   THEN
      Father := NulSym ;
      Unit   := 0 ;
      Off    := 0
   ELSIF GetInstructionType(i)=instregmd
   THEN
      WITH Instr[i].Instregmd DO
         Father := GetScopeAuthor(srcsym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSIF GetInstructionType(i)=bininstmd
   THEN
      WITH Instr[i].Bininstmd DO
         Father := GetScopeAuthor(sym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSE
      IF (GetInstructionType(i)=instregint) OR (GetInstructionType(i)=instmdint) OR
         (GetSrc(i)=NulSym) OR IsConst(GetSrc(i)) OR IsProcedure(GetSrc(i))
      THEN
         Unit   := 0 ;
         Off    := 0 ;
         Father := NulSym
      ELSE
         Father := GetScopeAuthor(GetSrc(i)) ;
         Unit   := GetSize(GetSrc(i)) ;
         Off    := GetOffset(GetSrc(i))
      END
   END
END GetComplexSrc ;


(*
   GetComplexDest - fills in the various parameters which reflect
                    the dest operand for instruction, i.
*)

PROCEDURE GetComplexDest (i: InstrIndex;
                          VAR Father: CARDINAL ;
                          VAR Unit  : CARDINAL ;
                          VAR Off   : INTEGER) ;
BEGIN
   IF GetInstruction(i)=nop
   THEN
      Father := NulSym ;
      Unit   := 0 ;
      Off    := 0
   ELSIF GetInstruction(i)=leave
   THEN
      (* implicit destination register ebp *)
      Father := GetScopeAuthor(EBP) ;
      Unit   := GetSize(EBP) ;
      Off    := GetOffset(EBP)
   ELSIF GetInstructionType(i)=instmdreg
   THEN
      WITH Instr[i].Instmdreg DO
         Father := GetScopeAuthor(destsym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSIF GetInstructionType(i)=instmdint
   THEN
      WITH Instr[i].Instmdint DO
         Father := GetScopeAuthor(destsym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSIF GetInstructionType(i)=bininstmd
   THEN
      WITH Instr[i].Bininstmd DO
         Father := GetScopeAuthor(sym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSIF GetInstructionType(i)=instmrisymconstant
   THEN
      WITH Instr[i].Instmrisymconstant DO
         Father := GetScopeAuthor(destsym) ;
         Unit   := GetSizeOfPrefix(unit) ;
         Off    := o1+go
      END
   ELSE
      IF (GetDest(i)=NulSym) OR IsConst(GetDest(i)) OR IsProcedure(GetDest(i))
      THEN
         Unit   := 0 ;
         Off    := 0 ;
         Father := NulSym
      ELSE
         Unit   := GetSize(GetDest(i)) ;
         Off    := GetOffset(GetDest(i)) ;
         Father := GetScopeAuthor(GetDest(i))
      END
   END
END GetComplexDest ;


(*
   IsDestSrc - returns TRUE if the nodes, from, to,
               form a dependancy (from.src -> to.dest)
*)

PROCEDURE IsDestSrc (from, to: InstrIndex) : BOOLEAN ;
VAR
   tFather,
   fFather,
   tUnit,
   fUnit  : CARDINAL ;
   tOff,
   fOff   : INTEGER ;
BEGIN
   IF (from=0) OR (to=0)
   THEN
      RETURN( FALSE )
   ELSE
      GetComplexSrc(from, fFather, fUnit, fOff) ;
      GetComplexDest(to, tFather, tUnit, tOff) ;
      IF (fFather=tFather) AND (fFather#NulSym) AND
         IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
      THEN
         RETURN( TRUE )
      ELSIF IsDestAlsoSrc(from)
      THEN
         (* add and sub use right most operand as both dest and src *)
         GetComplexDest(from, fFather, fUnit, fOff) ;
         GetComplexDest(to, tFather, tUnit, tOff) ;
         RETURN(
                (fFather=tFather) AND (fFather#NulSym) AND
                IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
               )
      ELSE
         RETURN( FALSE )
      END
   END
END IsDestSrc ;


(*
   IsDestDest - returns TRUE if the nodes, from, to,
                form a dependancy (from.dest -> to.dest)
*)

PROCEDURE IsDestDest (from, to: InstrIndex) : BOOLEAN ;
VAR
   tFather,
   fFather,
   tUnit,
   fUnit  : CARDINAL ;
   tOff,
   fOff   : INTEGER ;
BEGIN
   IF (from=0) OR (to=0)
   THEN
      RETURN( FALSE )
   ELSE
      GetComplexDest(from, fFather, fUnit, fOff) ;
      GetComplexDest(to, tFather, tUnit, tOff) ;
      IF (fFather=tFather) AND (fFather#NulSym) AND
         IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
      THEN
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   END
END IsDestDest ;


(*
   IsInstrDependant - returns TRUE if the nodes, from, to,
                      form a dependancy.
*)

PROCEDURE IsInstrDependant (from, to: InstrIndex) : BOOLEAN ;
VAR
   tFather,
   fFather,
   tUnit,
   fUnit  : CARDINAL ;
   tOff,
   fOff   : INTEGER ;
BEGIN
   IF (from#0) AND (to#0)
   THEN
      GetComplexSrc(from, fFather, fUnit, fOff) ;
      GetComplexSrc(to, tFather, tUnit, tOff) ;
      IF (fFather=tFather) AND (fFather#NulSym) AND
         IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
      THEN
         RETURN( TRUE )
      ELSE
         GetComplexSrc(from, fFather, fUnit, fOff) ;
         GetComplexDest(to, tFather, tUnit, tOff) ;
         IF (fFather=tFather) AND (fFather#NulSym) AND
            IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
         THEN
            RETURN( TRUE )
         ELSE
            GetComplexDest(from, fFather, fUnit, fOff) ;
            GetComplexSrc(to, tFather, tUnit, tOff) ;
            IF (fFather=tFather) AND (fFather#NulSym) AND
               IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
            THEN
               RETURN( TRUE )
            ELSE
               GetComplexDest(from, fFather, fUnit, fOff) ;
               GetComplexDest(to, tFather, tUnit, tOff) ;
               IF (fFather=tFather) AND (fFather#NulSym) AND
                  IsIntersection(tOff, tOff+tUnit, fOff, fOff+fUnit)
               THEN
                  RETURN( TRUE )
               ELSE
                  (* now check for base register dependence *)
                  tFather := GetDest(to) ;
                  fFather := GetBaseReg(from) ;
                  IF (tFather#NulSym) AND IsRegister(tFather) AND (tFather=fFather)
                  THEN
                     RETURN( TRUE )
                  END ;
                  tFather := GetBaseReg(to) ;
                  fFather := GetDest(from) ;
                  IF (tFather#NulSym) AND IsRegister(tFather) AND (tFather=fFather)
                  THEN
                     RETURN( TRUE )
                  END
               END
            END
         END
      END
   END ;
   RETURN( FALSE )
END IsInstrDependant ;


(*
   IsSrcReg - returns TRUE if the source of, i, is register, reg.
*)

PROCEDURE IsSrcReg (i: InstrIndex; reg: CARDINAL) : BOOLEAN ;
VAR
   rFather, rUnit,
   iFather, iUnit: CARDINAL ;
   rOff, iOff    : INTEGER ;
BEGIN
   GetComplexSrc(i, iFather, iUnit, iOff) ;
   rFather := GetScopeAuthor(reg) ;
   rOff    := GetOffset(reg) ;
   rUnit   := GetSize(reg) ;
   RETURN(
          (rFather=iFather) AND (iFather#NulSym) AND
          IsIntersection(iOff, iOff+iUnit, rOff, rOff+rUnit)
         )
END IsSrcReg ;


(*
   IsDestReg - returns TRUE if the destination of, i, is register, reg.
*)

PROCEDURE IsDestReg (i: InstrIndex; reg: CARDINAL) : BOOLEAN ;
VAR
   rFather, rUnit,
   iFather, iUnit: CARDINAL ;
   rOff, iOff    : INTEGER ;
BEGIN
   GetComplexDest(i, iFather, iUnit, iOff) ;
   rFather := GetScopeAuthor(reg) ;
   rOff    := GetOffset(reg) ;
   rUnit   := GetSize(reg) ;
   RETURN(
          (rFather=iFather) AND (iFather#NulSym) AND
          IsIntersection(iOff, iOff+iUnit, rOff, rOff+rUnit)
         )
END IsDestReg ;


(*
   IsStackBasedInst - returns TRUE if instruction, i, accesses
                      or manipulates the stack pointer.
*)

PROCEDURE IsStackBasedInst (i: InstrIndex) : BOOLEAN ;
BEGIN
   RETURN(
          (GetInstruction(i)=pop) OR (GetInstruction(i)=push) OR
          (GetInstruction(i)=call) OR (GetInstruction(i)=ret) OR
          (GetInstruction(i)=leave) OR
          IsDestReg(i, ESP) OR IsSrcReg(i, ESP)
         )
END IsStackBasedInst ;


(*
   IsSimplePush - returns TRUE if instruction, i, is a push constant
                  or push register.
*)

PROCEDURE IsSimplePush (i: InstrIndex) : BOOLEAN ;
BEGIN
   RETURN(
          (GetInstruction(i)=push) AND
          ((GetInstructionType(i)=bininstreg) OR
           (GetInstructionType(i)=bininsthex))
         )
END IsSimplePush ;


(*
   IsStackBasedPair - returns TRUE if the stack based instructions, i,
                      and, j, pair. This handles the special case
                      pairing of:

                      push reg; call
                      pop reg1; pop reg2
                      push #1 ; call
*)

PROCEDURE IsStackBasedPair (i, j: InstrIndex) : BOOLEAN ;
BEGIN
   IF IsSimplePush(i) AND IsSimplePush(j)
   THEN
      RETURN( TRUE )   (* really, even through both modify esp *)
   ELSIF IsSimplePush(i) AND (GetInstruction(j)=call)
   THEN
      RETURN( TRUE )
   ELSIF (GetInstruction(i)=pop) AND (GetInstruction(j)=pop) AND
         IsRegister(GetSrc(i)) AND IsRegister(GetSrc(j))
   THEN
      RETURN( TRUE )
   ELSE
      RETURN( FALSE )
   END
END IsStackBasedPair ;


(*
   IsStackBased - returns TRUE if instruction, i, is stack based
                  and also instruction list, j, is stack based.
                  We must keep pushes and pops and other stack
                  based instructions mov %eax, %esp etc in the
                  same pipeline for consistancy.
*)

PROCEDURE IsStackBased (i, j: InstrIndex) : BOOLEAN ;
BEGIN
   IF IsStackBasedInst(i)
   THEN
      WHILE j#0 DO
         IF IsStackBasedInst(j)
         THEN
            RETURN( TRUE )
         ELSIF (Instr[j].Branch1#0) AND IsStackBased(i, Instr[j].Branch1)
         THEN
            RETURN( TRUE )
         ELSIF (Instr[j].Branch2#0) AND IsStackBased(i, Instr[j].Branch2)
         THEN
            RETURN( TRUE )
         ELSE
            j := Instr[j].Parent
         END
      END
   END ;
   RETURN( FALSE )
END IsStackBased ;


PROCEDURE Dummy ; BEGIN END Dummy ;


(*
   IsMRIDependant - returns TRUE if instruction (i is dependant upon j):

                    i, is an instmri.. instruction and instruction,
                    j, uses a register that, i, uses.
*)

PROCEDURE IsMRIDependant (i, j: InstrIndex) : BOOLEAN ;
BEGIN
   (*
      note we have already checked the symbol and memory in InstrDependant,
      so all we do here is check whether the register associated with instmrisymconst
      matches the source register of, j.
   *)
   (* dest dependants first *)

   CASE GetInstructionType(i) OF

   instmrisymconstant: WITH Instr[i].Instmrisymconstant DO
                          RETURN( IsSrcReg(j, HighReg(destreg)) )
                       END |
   instmisymreg      : WITH Instr[i].Instmisymreg DO
                          RETURN( IsSrcReg(j, HighReg(destreg)) )
                       END

   ELSE
      (* keeps p2c happy.. *)
   END ;
   (* now srcdependants *)
   CASE GetInstructionType(j) OF

(*
   instmrisym  : WITH Instr[j].Instmrisym DO
                    RETURN( IsDestReg(i, HighReg(srcreg)) )
                 END |
*)
   instregmisym: WITH Instr[j].Instregmisym DO
                    RETURN( IsDestReg(i, HighReg(srcreg)) )
                 END

   ELSE
      RETURN( FALSE )
   END
END IsMRIDependant ;


(*
   IsDivModDependant - returns TRUE if instruction:

                       i, is a div/mod/idiv and instruction j uses
                       registers %eax or %edx as its source.

                       or

                       j, is a div/mod/idiv and instruction i uses
                       registers %eax or %edx as its dest.
*)

PROCEDURE IsDivModDependant (i, j: InstrIndex) : BOOLEAN ;
VAR
   instJ,
   instI: Instruction ;
BEGIN
   instI := GetInstruction(i) ;
   instJ := GetInstruction(j) ;
   IF (instI=div) OR (instI=idiv) OR (instI=mod) OR (instI=mult)
   THEN
      RETURN(
              IsSrcReg(j, EAX) OR IsSrcReg(j, EDX) OR
              IsDestReg(j, EAX) OR IsDestReg(j, EDX) OR
              (instJ=cdq)
            )
   ELSIF (instJ=div) OR (instJ=idiv) OR (instJ=mod) OR (instJ=mult)
   THEN
      RETURN(
              IsSrcReg(i, EAX)  OR IsSrcReg(i, EDX) OR
              IsDestReg(i, EAX) OR IsDestReg(i, EDX) OR
              (instI=cdq)
            )
   ELSE
      RETURN( FALSE )
   END
END IsDivModDependant ;


(*
   IsRepmovsDependant - returns TRUE if instruction:

                        i, is a repmovs and instruction j uses
                        registers %ecx or %edi or %esi

                        or

                        j, is a repmovs and instruction i uses
                        registers %ecx or %edi or %esi
*)

PROCEDURE IsRepmovsDependant (i, j: InstrIndex) : BOOLEAN ;
VAR
   instJ,
   instI: Instruction ;
BEGIN
   instI := GetInstruction(i) ;
   instJ := GetInstruction(j) ;
   IF instI=repmovs
   THEN
      RETURN(
              IsSrcReg(j, ECX) OR IsSrcReg(j, EDI) OR IsSrcReg(j, ESI) OR
              IsDestReg(j, ECX) OR IsDestReg(j, EDI) OR IsDestReg(j, ESI)
            )
   ELSIF instJ=repmovs
   THEN
      RETURN(
              IsSrcReg(i, ECX) OR IsSrcReg(i, EDI) OR IsSrcReg(i, ESI) OR
              IsDestReg(i, ECX) OR IsDestReg(i, EDI) OR IsDestReg(i, ESI)
            )
   ELSE
      RETURN( FALSE )
   END
END IsRepmovsDependant ;


(*
   IsCdqDependant - returns TRUE if instruction:

                    i, is a cdq and instruction j uses
                    registers %edx or %eax

                    or

                    j, is a cdq and instruction i uses
                    registers %edx or %eax
*)

PROCEDURE IsCdqDependant (i, j: InstrIndex) : BOOLEAN ;
VAR
   instJ,
   instI: Instruction ;
BEGIN
   instI := GetInstruction(i) ;
   instJ := GetInstruction(j) ;
   IF instI=cdq
   THEN
      RETURN(
              IsSrcReg(j, EAX) OR IsSrcReg(j, EDX) OR
              IsDestReg(j, EAX) OR IsDestReg(j, EDX) OR
              (instJ=div) OR (instJ=idiv) OR (instJ=mod) OR
              (instJ=mult)
            )
   ELSIF instJ=cdq
   THEN
      RETURN(
              IsSrcReg(i, EAX) OR IsSrcReg(i, EDX) OR
              IsDestReg(i, EAX) OR IsDestReg(i, EDX) OR
              (instI=div) OR (instI=idiv) OR (instI=mod) OR
              (instI=mult)
            )
   ELSE
      RETURN( FALSE )
   END
END IsCdqDependant ;


(*
   IsDependant - returns TRUE if instruction, i, is dependant
                 on the list ending with, j.
*)

PROCEDURE IsDependant (i, j: InstrIndex) : BOOLEAN ;
BEGIN
   WHILE j#0 DO
      IF IsInstrDependant(i, j)
      THEN
         IF DebugOn
         THEN
            WriteString('# dependancy found between ') ;
            DisplayInstruction(i) ;
            WriteString('# and                      ') ;
            DisplayInstruction(j) ;
         END ;
         RETURN( TRUE )
      ELSIF IsMRIDependant(i, j)
      THEN
         RETURN( TRUE )
      ELSIF IsDivModDependant(i, j)
      THEN
         RETURN( TRUE )
      ELSIF IsRepmovsDependant(i, j)
      THEN
         RETURN( TRUE )
      ELSIF IsCdqDependant(i, j)
      THEN
         RETURN( TRUE )
      ELSIF IsConditional(GetInstruction(i)) AND SetsCondition(GetInstruction(j))
      THEN
         RETURN( TRUE )
      ELSIF ((GetInstruction(i)=cmp) OR (GetInstruction(i)=test)) AND IsDestDest(j, i)
      THEN
         RETURN( TRUE )    (* the compare instruction destination is really a source *)
      ELSIF IsStackBased(i, j)
      THEN
         IF DebugOn
         THEN
            WriteString('# found stack based instruction') ; WriteLn
         END ;
         RETURN( TRUE )
      ELSE
         IF (Instr[j].Branch1#0) AND IsDependant(i, Instr[j].Branch1)
         THEN
            RETURN( TRUE )
         ELSIF (Instr[j].Branch2#0) AND IsDependant(i, Instr[j].Branch2)
         THEN
            RETURN( TRUE )
         ELSE
            j := Instr[j].Parent
         END
      END
   END ;
   RETURN( FALSE )
END IsDependant ;


(*
   VerifyNotInTreeList - checks to make sure that instruction, i, is not
                         in the tree list, t.
*)

PROCEDURE VerifyNotInTreeList (t, i: InstrIndex) ;
BEGIN
   IF i#NopIndex
   THEN
      WHILE t#0 DO
         IF t=i
         THEN
            WriteString('duplication problem with instruction # ') ; WriteCard(CARDINAL(i), 0) ; WriteLn ;
            DisplayPipeTrees ;
            InternalError('instruction already placed into tree list', __FILE__, __LINE__)
         END ;
         IF Instr[t].Branch1#0
         THEN
            VerifyNotInTreeList(Instr[t].Branch1, i)
         END ;
         IF Instr[t].Branch2#0
         THEN
            VerifyNotInTreeList(Instr[t].Branch2, i)
         END ;
         t := Instr[t].Parent
      END
   END
END VerifyNotInTreeList ;


(*
   VerifyNotInTree - verifies that instruction, i, is not in any of the trees.
*)

PROCEDURE VerifyNotInTree (i: InstrIndex) ;
VAR
   t: InstrIndex ;
BEGIN
   t := NoOfTrees ;
   WHILE t>0 DO
      VerifyNotInTreeList(Trees[t], i) ;
      DEC(t)
   END
END VerifyNotInTree ;


(*
   AddDependant - adds a new instruction into the forest
*)

PROCEDURE AddDependant (n: InstrIndex) ;
VAR
   Conflict  : BOOLEAN ;    (* does inst, n, depend on 2 lists *)
   i, j, k, l: InstrIndex ;
BEGIN
   VerifyNotInTree(n) ;
   i := NoOfTrees ;
   j := 0 ;
   k := 0 ;
   l := 0 ;
   Conflict := FALSE ;
   WHILE i>0 DO
      IF IsDependant(n, Trees[i])
      THEN
         IF j=0
         THEN
            j := i
         ELSIF k=0
         THEN
            k := i
         ELSE
            l := i
         END
      END ;
      DEC(i)
   END ;
   IF j#0
   THEN
      (* found a tree which instruction, n, is dependant. *)
      IF k=0
      THEN
         MakeLink(n, Trees[j]) ;
         Trees[j] := n ;    (* remember the tree points to the last inst *)
         (* all done *)
         RETURN
      ELSE
         (* instruction, n, is dependant upon two trees *)
         MakeLink(n, Trees[j]) ;
         Trees[j] := n ;
         MakeBranch(n, Trees[k]) ;
         Trees[k] := 0 ;
         IF DebugOn
         THEN
            WriteString('# made branch') ;
            DisplayList(Trees[j], 0) ;
         END ;
         IF l#0
         THEN
            MakeBranch(n, Trees[l]) ;
            Trees[l] := 0 ;
            IF DebugOn
            THEN
               WriteString('# made branch') ;
               DisplayList(Trees[j], 0)
            END
         END ;
         RETURN
      END
   END ;
   (*
      we could not add, n, to an old tree, we need to create
      a new tree.
   *)
   IF NoOfTrees=MaxInstruction
   THEN
      InternalError('increase MaxInstruction', __FILE__, __LINE__)
   ELSE
      INC(NoOfTrees) ;
      Trees[NoOfTrees] := n ;
      MakeLink(n, 0) ;
      IF DebugOn
      THEN
         WriteString('# new tree created') ;
         DisplayList(n, 0)
      END
   END
END AddDependant ;


(*
   CheckProfiling - checks to see whether profiling is turned on
                    and if so it emits a call to _mcount.
*)

PROCEDURE CheckProfiling ;
BEGIN
   IF NOT APUCodeGenerator
   THEN
      IF Profiling
      THEN
         FlushInstruction(FALSE) ;
         M2AS.CheckProfiling
      END
   END
END CheckProfiling ;


(*
   BuildInstRegConstant - places an inst reg, const
                          instruction into the DAG ready for
                          pipelining.
*)

PROCEDURE BuildInstRegConstant (Inst: Instruction;
                                Unit: DatumPrefix;
                                Reg : Register;
                                Sym : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregconstant ;
      WITH Instregconstant DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg ;
         srcsym  := Sym
      END
   END ;
   AddDependant(i)
END BuildInstRegConstant ;


(*
   InstRegConstant - instruction constant and memory.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   12, Reg
                     add   12, Reg
*)

PROCEDURE InstRegConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Reg : Register;
                           Sym : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegConstant(Inst, Unit, Reg, Sym)
   ELSE
      BuildInstRegConstant(Inst, Unit, Reg, Sym)
   END
END InstRegConstant ;


(*
   BuildInstRegConstant - places an inst reg, const
                          instruction into the DAG ready for
                          pipelining.
*)

PROCEDURE BuildInstConstantReg (Inst: Instruction;
                                Unit: DatumPrefix;
                                Sym : CARDINAL;
                                Reg : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instconstantreg ;
      WITH Instconstantreg DO
         inst    := Inst ;
         unit    := Unit ;
         srcreg  := Reg ;
         destsym := Sym
      END
   END ;
   AddDependant(i)
END BuildInstConstantReg ;


(*
   InstConstantReg - instruction register and constant.
                     (cmp Constant, Reg)

                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

*)

PROCEDURE InstConstantReg (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL;
                           Reg : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstConstantReg(Inst, Unit, Sym, Reg)
   ELSE
      BuildInstConstantReg(Inst, Unit, Sym, Reg)
   END
END InstConstantReg ;


(*
   BuildInstMDConstant - places an inst constant, md
                         instruction into the DAG ready for
                         pipelining.
*)

PROCEDURE BuildInstMDConstant (Inst  : Instruction;
                               Unit  : DatumPrefix;
                               O1, GO: INTEGER ;
                               Scope : KindOfScope;
                               Sym1  : CARDINAL;
                               Sym2  : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmdconstant ;
      WITH Instmdconstant DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         scope   := Scope ;
         destsym := Sym1 ;
         srcsym  := Sym2
      END
   END ;
   AddDependant(i)
END BuildInstMDConstant ;


(*
   InstMDConstant - instruction constant and memory.  (Mem <- inst <- Constant)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    o1, go - determine the offset into element1.
                    Scope  - determines the scope of element1.
                    Sym1   - element1 where the constant is to be placed.
                             Symbol requiring the memory reference.
                    Sym2   - constant element2.

                    Example:

                    mov   12, (ebp+15)
*)

PROCEDURE InstMDConstant (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          o1, go: INTEGER ;
                          Scope : KindOfScope;
                          Sym1,
                          Sym2  : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMDConstant(Inst, Unit, o1, go, Scope, Sym1, Sym2)
   ELSE
      BuildInstMDConstant(Inst, Unit, o1, go, Scope, Sym1, Sym2)
   END
END InstMDConstant ;


(*
   BuildInstMRIConstant - places an inst constant, mri
                          instruction into the DAG ready for
                          pipelining.
*)

PROCEDURE BuildInstMRIConstant (Inst  : Instruction;
                                Unit  : DatumPrefix;
                                O1, GO: INTEGER ;
                                Reg1  : Register ;
                                Sym2  : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmriconstant ;
      WITH Instmriconstant DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         destreg := Reg1 ;
         srcsym  := Sym2
      END
   END ;
   AddDependant(i)
END BuildInstMRIConstant ;


(*
   InstMRIConstant - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Sym2   - constant element2.

                     Example:

                     mov   12, (edi+15)
*)

PROCEDURE InstMRIConstant (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Sym2  : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMRIConstant(Inst, Unit, o1, go, Reg1, Sym2)
   ELSE
      BuildInstMRIConstant(Inst, Unit, o1, go, Reg1, Sym2) ;
      FlushInstruction(TRUE)
   END
END InstMRIConstant ;


(*
   BuildInstMRISymConstant - places an inst constant, mri
                             instruction into the DAG ready for
                             pipelining.
*)

PROCEDURE BuildInstMRISymConstant (Inst  : Instruction;
                                   Unit  : DatumPrefix;
                                   Sym1  : CARDINAL;
                                   GO    : INTEGER;
                                   Reg1  : Register;
                                   Sym2  : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmrisymconstant ;
      WITH Instmrisymconstant DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := 0 ;   (* really as we dont pass O1 into this procedure *)
         go      := GO ;
         destsym := Sym1 ;
         destreg := Reg1 ;
         srcsym  := Sym2
      END
   END ;
   AddDependant(i)
END BuildInstMRISymConstant ;


(*
   InstMRISymConstant - instruction constant and memory indexed by a register.
                        (Mem[Reg1] <- inst <- Constant)
                        Inst   - determines the instruction.
                        Unit   - determines the DatumPrefix of transfer.
                        Sym1   - global variable.
                        go     - determine the offset into element1.
                        Sym2   - constant element2.

                        Example:

                        mov   12, myglobal+3(edi)
*)

PROCEDURE InstMRISymConstant (Inst  : Instruction;
                              Unit  : DatumPrefix;
                              Sym1  : CARDINAL;
                              go    : INTEGER ;
                              Reg1  : Register;
                              Sym2  : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMRISymConstant(Inst, Unit, Sym1, go, Reg1, Sym2)
   ELSE
      BuildInstMRISymConstant(Inst, Unit, Sym1, go, Reg1, Sym2) ;
      FlushInstruction(TRUE)
   END
END InstMRISymConstant ;


(*
   BuildInstMRICardinal - places an inst cardinal, mri
                          instruction into the DAG ready for
                          pipelining.
*)

PROCEDURE BuildInstMRICardinal (Inst  : Instruction;
                                Unit  : DatumPrefix;
                                O1, GO: INTEGER ;
                                Reg1  : Register ;
                                Card  : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmricardinal ;
      WITH Instmricardinal DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         destreg := Reg1 ;
         srccard := Card
      END
   END ;
   AddDependant(i)
END BuildInstMRICardinal ;


(*
   InstMRICardinal - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Constant)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Card   - cardinal constant.

                     Example:

                     mov   12H, (edi+15)
*)

PROCEDURE InstMRICardinal (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           o1, go: INTEGER ;
                           Reg1  : Register;
                           Card  : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMRICardinal(Inst, Unit, o1, go, Reg1, Card)
   ELSE
      BuildInstMRICardinal(Inst, Unit, o1, go, Reg1, Card) ;
      FlushInstruction(TRUE)
   END
END InstMRICardinal ;


(*
   BuildInstMRIReg - instruction constant and memory indexed by a register.
                     (Mem[Reg1] <- inst <- Reg2)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     o1, go - determine the offset into element1.
                     Reg1   - indexes memory.
                     Reg2   - source of data.

                     Example:

                     mov   eax, (edi+15)
*)

PROCEDURE BuildInstMRIReg (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           O1, GO: INTEGER ;
                           Reg1  ,
                           Reg2  : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmrireg ;
      WITH Instmrireg DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         destreg := Reg1 ;
         srcreg  := Reg2
      END
   END ;
   AddDependant(i)
END BuildInstMRIReg ;


(*
   InstMRIReg - instruction constant and memory indexed by a register.
                (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                o1, go - determine the offset into element1.
                Reg1   - indexes memory.
                Reg2   - source of data.

                Example:

                mov   eax, (edi+15)
*)

PROCEDURE InstMRIReg (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      o1, go: INTEGER ;
                      Reg1  ,
                      Reg2  : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMRIReg(Inst, Unit, o1, go, Reg1, Reg2)
   ELSE
      BuildInstMRIReg(Inst, Unit, o1, go, Reg1, Reg2) ;
      FlushInstruction(TRUE)
   END
END InstMRIReg ;


(*
   BuildInstMDReg - places an inst md, reg
                    instruction into the DAG ready for
                    pipelining.
*)

PROCEDURE BuildInstMDReg (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          O1, GO: INTEGER ;
                          Reg   : Register;
                          Scope : KindOfScope;
                          Sym   : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmdreg ;
      WITH Instmdreg DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         srcreg  := Reg ;
         scope   := Scope ;
         destsym := Sym
      END
   END ;
   AddDependant(i)
END BuildInstMDReg ;


(*
   InstMDReg - inst reg into memory. (Mem1 <- inst <- Reg)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into the element.
               Reg    - determine the register.
               Scope  - determines the scope of the element.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   eax, (ebp+15)
*)

PROCEDURE InstMDReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     O1, GO: INTEGER ;
                     Reg   : Register;
                     Scope : KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMDReg(Inst, Unit, O1, GO, Reg, Scope, Sym)
   ELSE
      BuildInstMDReg(Inst, Unit, O1, GO, Reg, Scope, Sym)
   END
END InstMDReg ;


(*
   BuildInstRegMD - places an inst reg, md
                    instruction into the DAG ready for
                    pipelining.
*)

PROCEDURE BuildInstRegMD (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          Reg   : Register;
                          O1, GO: INTEGER ;
                          Scope1: KindOfScope;
                          Sym   : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregmd ;
      WITH Instregmd DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg ;
         o1      := O1 ;
         go      := GO ;
         scope   := Scope1 ;
         srcsym  := Sym
      END
   END ;
   AddDependant(i)
END BuildInstRegMD ;


(*
   InstRegMD - inst memory into reg. (Reg <- inst <- Mem1)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o2, go - determine the offset into element2.
               Reg    - determine the register.
               Scope2 - determines the scope of element2.
               Sym    - symbol requiring the memory reference.

               Example:

               ; inst  Mem{scope}+o1+go, reg
               mov   (ebp+15), eax
*)

PROCEDURE InstRegMD (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg   : Register;
                     O1, GO: INTEGER ;
                     Scope1: KindOfScope;
                     Sym   : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegMD(Inst, Unit, Reg, O1, GO, Scope1, Sym)
   ELSE
      BuildInstRegMD(Inst, Unit, Reg, O1, GO, Scope1, Sym)
   END
END InstRegMD ;


(*
   BuildInstRegMIO - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Reg1   - determine the destination register.
                     Reg2   - determine the index register.
                     Offset - offset to be added to the index register.

                     Example:

                     mov   (edi+4), eax
                     ; mov   (reg2+offset), reg1
*)

PROCEDURE BuildInstRegMIO (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           Reg1,
                           Reg2  : Register;
                           Offset: INTEGER) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregmio ;
      WITH Instregmio DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2 ;
         go      := Offset ;
         o1      := 0
      END
   END ;
   AddDependant(i)
END BuildInstRegMIO ;


(*
   InstRegMIO - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.
                Offset - offset to be added to the index register.

                Example:

                mov   (edi+4), eax
                ; mov   (reg2+offset), reg1
*)

PROCEDURE InstRegMIO (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      Reg1,
                      Reg2  : Register;
                      Offset: INTEGER) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegMIO(Inst, Unit, Reg1, Reg2, Offset)
   ELSE
      BuildInstRegMIO(Inst, Unit, Reg1, Reg2, Offset) ;
      FlushInstruction(TRUE)
   END
END InstRegMIO ;


(*
   BuildInstRegMISym - inst memory into reg. (Reg1 <- inst <- global[Reg2])
                       Inst   - determines the instruction.
                       Unit   - determines the DatumPrefix of transfer.
                       Reg1   - determine the destination register.
                       Reg2   - determine the index register.
                       Offset - offset to be added to the index register.
                       Sym    - the global variable.

                       Example:

                       mov   myglobal+4(edi), eax
                     ; mov   Sym+offset(reg2), reg1
*)

PROCEDURE BuildInstRegMISym (Inst  : Instruction;
                             Unit  : DatumPrefix;
                             Reg1,
                             Reg2  : Register;
                             Offset: INTEGER;
                             Sym   : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregmisym ;
      WITH Instregmisym DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2 ;
         go      := Offset ;
         srcsym  := Sym ;
         o1      := 0
      END
   END ;
   AddDependant(i)
END BuildInstRegMISym ;


(*
   InstRegMISym - inst memory into reg. (Reg1 <- inst <- global[Reg2])
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   myglobal+4(edi), eax
                ; mov   Sym+offset(reg2), reg1
*)

PROCEDURE InstRegMISym (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegMISym(Inst, Unit, Reg1, Reg2, Offset, Sym)
   ELSE
      BuildInstRegMISym(Inst, Unit, Reg1, Reg2, Offset, Sym) ;
      FlushInstruction(TRUE)
   END
END InstRegMISym ;


(*
   BuildInstRegMI - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg1   - determine the destination register.
                    Reg2   - determine the index register.

                    Example:

                    mov   (edi), eax
                    ; mov   (reg2), reg1
*)

PROCEDURE BuildInstRegMI (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          Reg1,
                          Reg2  : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregmi ;
      WITH Instregmi DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2
      END
   END ;
   AddDependant(i)
END BuildInstRegMI ;


(*
   InstRegMI - inst memory into reg. (Reg1 <- inst <- Mem[Reg2])
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the destination register.
               Reg2   - determine the index register.

               Example:

               mov   (edi), eax
               ; mov   (reg2), reg1
*)

PROCEDURE InstRegMI (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegMI(Inst, Unit, Reg1, Reg2)
   ELSE
      BuildInstRegMI(Inst, Unit, Reg1, Reg2) ;
      FlushInstruction(TRUE)
   END
END InstRegMI ;


(*
   BuildInstMIOReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Reg1   - determine the index register.
                     Offset - determines the offset to be added to the index
                              register.
                     Reg2   - determine the source register.

                     Example:

                     mov   eax, (edi)
                     ; mov   reg2, (reg1)
*)

PROCEDURE BuildInstMIOReg (Inst  : Instruction;
                           Unit  : DatumPrefix;
                           Reg1  : Register;
                           Offset: INTEGER;
                           Reg2  : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmioreg ;
      WITH Instmioreg DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2 ;
         go      := Offset ;
         o1      := 0
      END
   END ;
   AddDependant(i)
END BuildInstMIOReg ;


(*
   InstMIOReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the index register.
                Offset - determines the offset to be added to the index
                         register.
                Reg2   - determine the source register.

                Example:

                mov   eax, (edi)
                ; mov   reg2, (reg1)
*)

PROCEDURE InstMIOReg (Inst  : Instruction;
                      Unit  : DatumPrefix;
                      Reg1  : Register;
                      Offset: INTEGER;
                      Reg2  : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMIOReg(Inst, Unit, Reg1, Offset, Reg2)
   ELSE
      BuildInstMIOReg(Inst, Unit, Reg1, Offset, Reg2) ;
      FlushInstruction(TRUE)
   END
END InstMIOReg ;


(*
   BuildInstMISymReg - inst reg into memory. (global[Reg2] <- inst <- Reg1)
                       Inst   - determines the instruction.
                       Unit   - determines the DatumPrefix of transfer.
                       Reg1   - determine the destination register.
                       Reg2   - determine the index register.
                       Offset - offset to be added to the index register.
                       Sym    - the global variable.

                       Example:

                       mov   %eax, myglobal+4(edi),
                     ; mov   reg1, Sym+offset(reg2)
*)

PROCEDURE BuildInstMISymReg (Inst  : Instruction;
                             Unit  : DatumPrefix;
                             Reg1,
                             Reg2  : Register;
                             Offset: INTEGER;
                             Sym   : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmisymreg ;
      WITH Instmisymreg DO
         inst    := Inst ;
         unit    := Unit ;
         srcreg  := Reg1 ;
         destreg := Reg2 ;
         go      := Offset ;
         o1      := 0 ;
         destsym := Sym
      END
   END ;
   AddDependant(i)
END BuildInstMISymReg ;


(*
   InstMISymReg - inst reg into memory. (global[Reg2] <- inst <- Reg1)
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Reg1   - determine the destination register.
                  Reg2   - determine the index register.
                  Offset - offset to be added to the index register.
                  Sym    - the global variable.

                  Example:

                  mov   %eax, myglobal+4(edi),
                ; mov   reg1, Sym+offset(reg2)
*)

PROCEDURE InstMISymReg (Inst  : Instruction;
                        Unit  : DatumPrefix;
                        Reg1,
                        Reg2  : Register;
                        Offset: INTEGER;
                        Sym   : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMISymReg(Inst, Unit, Reg1, Reg2, Offset, Sym)
   ELSE
      BuildInstMISymReg(Inst, Unit, Reg1, Reg2, Offset, Sym) ;
      FlushInstruction(TRUE)
   END
END InstMISymReg ;


(*
   BuildInstMIReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg1   - determine the index register.
                    Reg2   - determine the source register.

                    Example:

                    mov   eax, (edi)
                    ; mov   reg2, (reg1)
*)

PROCEDURE BuildInstMIReg (Inst  : Instruction;
                          Unit  : DatumPrefix;
                          Reg1,
                          Reg2  : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmireg ;
      WITH Instmireg DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2
      END
   END ;
   AddDependant(i)
END BuildInstMIReg ;


(*
   InstMIReg - inst memory into reg. (Mem[Reg1] <- inst <- Reg2)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               Reg1   - determine the index register.
               Reg2   - determine the source register.

               Example:

               mov   eax, (edi)
               ; mov   reg2, (reg1)
*)

PROCEDURE InstMIReg (Inst  : Instruction;
                     Unit  : DatumPrefix;
                     Reg1,
                     Reg2  : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMIReg(Inst, Unit, Reg1, Reg2)
   ELSE
      BuildInstMIReg(Inst, Unit, Reg1, Reg2) ;
      FlushInstruction(TRUE)
   END
END InstMIReg ;


(*
   BuildInstRegReg - places an inst reg, reg
                     instruction into the DAG ready for
                     pipelining.
*)

PROCEDURE BuildInstRegReg (Inst: Instruction; Unit: DatumPrefix;
                           Reg1, Reg2: Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregreg ;
      WITH Instregreg DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg1 ;
         srcreg  := Reg2
      END
   END ;
   AddDependant(i)
END BuildInstRegReg ;


(*
   InstRegReg - inst reg into reg. (Reg1 <- inst <- Reg2)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg1   - determine the destination register.
                Reg2   - determine the index register.

                Example:

                mov   ebp, eax
               ; mov   reg2, reg1
*)

PROCEDURE InstRegReg (Inst: Instruction; Unit: DatumPrefix;
                      Reg1, Reg2: Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegReg(Inst, Unit, Reg1, Reg2)
   ELSE
      BuildInstRegReg(Inst, Unit, Reg1, Reg2)
   END
END InstRegReg ;


(*
   BuildInstRegInt - places an inst reg, int
                     instruction into the DAG ready for
                     pipelining.
*)

PROCEDURE BuildInstRegInt (Inst: Instruction; Unit: DatumPrefix;
                           Reg: Register; Int: INTEGER) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregint ;
      WITH Instregint DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg ;
         srcint  := Int
      END
   END ;
   AddDependant(i)
END BuildInstRegInt ;


(*
   InstRegInt - inst Reg with an Int. (Reg <- inst <- Int)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determines the destination register.
                Int    - the integer constant as the second operand.

                Example:

                mov   15, eax
               ; mov   Int, Reg
*)

PROCEDURE InstRegInt (Inst: Instruction; Unit: DatumPrefix;
                      Reg: Register; Int: INTEGER) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegInt(Inst, Unit, Reg, Int)
   ELSE
      BuildInstRegInt(Inst, Unit, Reg, Int)
   END
END InstRegInt ;


(*
   InstRegHexCard - inst Reg with an cardinal. (Reg <- inst <- Card)
                    Inst   - determines the instruction.
                    Unit   - determines the DatumPrefix of transfer.
                    Reg    - determines the destination register.
                    Card   - the cardinal constant as the second operand.

                    Example:

                    and   $0FFFF0000H, eax
                   ; mov   Int, Reg
*)

PROCEDURE InstRegHexCard (Inst: Instruction; Unit: DatumPrefix;
                          Reg: Register; Card: CARDINAL) ;
BEGIN
   (* uses above procedure *)
   InstRegInt(Inst, Unit, Reg, INTEGER(Card))
END InstRegHexCard ;


(*
   BuildInstMDOffset - places an inst $sym1, sym2 into the instruction
                       pipeline.
*)

PROCEDURE BuildInstMDOffset (Inst: Instruction; Unit: DatumPrefix;
                             Sym1, Sym2: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmdoffset ;
      WITH Instmdoffset DO
         inst    := Inst ;
         unit    := Unit ;
         destsym := Sym1 ;
         srcsym  := Sym2
      END
   END ;
   AddDependant(i)
END BuildInstMDOffset ;


(*
   InstMDOffset - inst Mem with an offset Int. (Mem[Sym1] <- inst <- OFFSET Sym2)
                  Offset is the address of symbol Sym2.
                  Inst   - determines the instruction.
                  Unit   - determines the DatumPrefix of transfer.
                  Sym1   - indicates the memory location which will
                           contain the address of Sym2.
                  Sym2   - symbol requiring the memory reference.

                  Example:

                  mov   $_varname, (ebp-15)
                  ; mov  $_varname, {unit prefix} {scope}
*)

PROCEDURE InstMDOffset (Inst: Instruction; Unit: DatumPrefix;
                        Sym1, Sym2: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMDOffset(Inst, Unit, Sym1, Sym2)
   ELSE
      BuildInstMDOffset(Inst, Unit, Sym1, Sym2)
   END
END InstMDOffset ;


(*
   BuildInstRegOffset - places an inst $sym, reg into the instruction
                        pipeline.
*)

PROCEDURE BuildInstRegOffset (Inst: Instruction; Unit: DatumPrefix;
                              Reg: Register; Sym: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregoffset ;
      WITH Instregoffset DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg ;
         srcsym  := Sym
      END
   END ;
   AddDependant(i)
END BuildInstRegOffset ;


(*
   InstRegOffset - inst reg with an offset Int. (Reg <- inst <- OFFSET Sym)
                   Offset is the address of symbol Sym.
                   Inst   - determines the instruction.
                   Unit   - determines the DatumPrefix of transfer.
                   Reg    - register which contain the address of Sym2.
                   Sym    - symbol requiring the memory reference.

                   Example:

                   mov   $_varname, $eax
                  ; mov  $_varname, {unit prefix} {reg}
*)

PROCEDURE InstRegOffset (Inst: Instruction; Unit: DatumPrefix;
                         Reg: Register; Sym: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegOffset(Inst, Unit, Reg, Sym)
   ELSE
      BuildInstRegOffset(Inst, Unit, Reg, Sym)
   END
END InstRegOffset ;


(*
   BuildInstMDInt - adds an instruction, inst  int, sym+o1+go to the pipeline.
*)

PROCEDURE BuildInstMDInt (Inst: Instruction; Unit: DatumPrefix;
                          O1, GO: INTEGER; Int: INTEGER;
                          Scope: KindOfScope; Sym: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instmdint ;
      WITH Instmdint DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         scope   := Scope ;
         destsym := Sym ;
         srcint  := Int
      END
   END ;
   AddDependant(i)
END BuildInstMDInt ;


(*
   InstMDInt - inst Mem with an Int. (Mem[Sym1] <- inst <- Int)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the offset into element1.
               Int    - the integer constant as the second operand.
               Scope  - determines the scope of element1.
               Sym    - symbol requiring the memory reference.

               Example:

               mov   123, (ebp-15)
               ; mov   int, {unit prefix} {scope}
*)

PROCEDURE InstMDInt (Inst: Instruction; Unit: DatumPrefix;
                     o1, go: INTEGER; Int: INTEGER;
                     Scope: KindOfScope; Sym: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstMDInt(Inst, Unit, o1, go, Int, Scope, Sym)
   ELSE
      BuildInstMDInt(Inst, Unit, o1, go, Int, Scope, Sym)
   END
END InstMDInt ;


(*
   InstStr - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstStr (Inst: Instruction; String: ARRAY OF CHAR) ;
BEGIN
   InstKey(Inst, MakeKey(String)) ;
   IF (Inst=jmp) AND (NOT APUCodeGenerator)
   THEN
      (* unconditional jump, align code after this, Intel Pentium Optimization (3.4.1) *)
      M2AS.CodeAlignment
   END
END InstStr ;


(*
   BuildInstKey - builds instruction of the type: inst string
                  into the pipeline data structure.
*)

PROCEDURE BuildInstKey (Inst: Instruction; String: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instkey ;
      WITH Instkey DO
         inst    := Inst ;
         string  := String
      END
   END ;
   AddDependant(i)
END BuildInstKey ;


(*
   InstKey - inst followed by a string:
             inst String
             Inst   - determines the instruction.
             String - string that follows the Inst

             Example:

             Inst  String
             ; call  String
*)

PROCEDURE InstKey (Inst: Instruction; string: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstKey(Inst, string)
   ELSE
      BuildInstKey(Inst, string) ;
      FlushInstruction(TRUE)
   END
END InstKey ;


(*
   BuildBinInst - add a single instruction to the pipeline.
*)

PROCEDURE BuildBinInst (Inst: Instruction) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininst ;
      WITH Bininst DO
         inst := Inst
      END
   END ;
   AddDependant(i)
END BuildBinInst ;


(*
   BinInst - write a sole inst.
             Inst   - determines the instruction.

             Example:

             ret
*)

PROCEDURE BinInst (Inst: Instruction) ;
BEGIN
   IF Inst=leave
   THEN
      (* as leave ; ret     causes an AGIs we substitute it with pop ebp *)
      BinInstReg(pop, PDWord, RegEBP)       (* simple peep hole optimization *)
   ELSE
      IF IgnorePipeLine
      THEN
         M2AS.BinInst(Inst)
      ELSE
         BuildBinInst(Inst)
      END
   END
END BinInst ;


(*
   BuildBinInstUnit - builds a single instruction which has a unit
                      specifier.
*)

PROCEDURE BuildBinInstUnit (Inst: Instruction; Unit: DatumPrefix; Length: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininstunit ;
      WITH Bininstunit DO
         inst    := Inst ;
         unit    := Unit ;
         length  := Length
      END
   END ;
   AddDependant(i)
END BuildBinInstUnit ;


(*
   BinInstUnit - write a sole inst.
                 Inst   - determines the instruction.

                 Example:

                 repsl edi, esi
*)

PROCEDURE BinInstUnit (Inst: Instruction; Unit: DatumPrefix; Length: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstUnit(Inst, Unit, Length)
   ELSE
      BuildBinInstUnit(Inst, Unit, Length)
   END
END BinInstUnit ;


(*
   BuildBinInstReg - places a binary inst reg
                     instruction into the DAG ready for
                     pipelining.
*)

PROCEDURE BuildBinInstReg (Inst: Instruction; Unit: DatumPrefix;
                           Reg: Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininstreg ;
      WITH Bininstreg DO
         inst    := Inst ;
         unit    := Unit ;
         reg     := Reg
      END
   END ;
   AddDependant(i)
END BuildBinInstReg ;


(*
   BinInstReg - inst operating on a reg. (inst <-> Reg)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                Reg    - determine the operand register.

                Example:

                push   eax
*)

PROCEDURE BinInstReg (Inst: Instruction;
                      Unit: DatumPrefix;
                      Reg : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstReg(Inst, Unit, Reg)
   ELSE
      BuildBinInstReg(Inst, Unit, Reg) ;
      IF Inst=call
      THEN
         FlushInstruction(TRUE)
      END
   END
END BinInstReg ;


(*
   BuildBinInstRegI - inst operating on a reg. ( inst <-> (Reg) )
                      Inst   - determines the instruction.
                      Unit   - determines the DatumPrefix of transfer.
                      Reg    - determine the operand register.

                      Example:

                      call   (eax)
*)

PROCEDURE BuildBinInstRegI (Inst: Instruction;
                            Unit: DatumPrefix;
                            Reg : Register) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininstregi ;
      WITH Bininstregi DO
         inst    := Inst ;
         unit    := Unit ;
         reg     := Reg ;
      END
   END ;
   AddDependant(i)
END BuildBinInstRegI ;


(*
   BinInstRegI - inst operating on a reg. ( inst <-> (Reg) )
                 Inst   - determines the instruction.
                 Unit   - determines the DatumPrefix of transfer.
                 Reg    - determine the operand register.

                 Example:

                 call   (eax)
*)

PROCEDURE BinInstRegI (Inst: Instruction;
                       Unit: DatumPrefix;
                       Reg : Register) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstRegI(Inst, Unit, Reg)
   ELSE
      BuildBinInstRegI(Inst, Unit, Reg) ;
      FlushInstruction(TRUE)
   END
END BinInstRegI ;


(*
   BuildBinInstConstant - inst operating on a reg. (inst <-> Reg)
                          Inst   - determines the instruction.
                          Unit   - determines the DatumPrefix of transfer.
                          Sym    - determine the constant.

                          Example:

                          push   15
*)

PROCEDURE BuildBinInstConstant (Inst: Instruction;
                                Unit: DatumPrefix;
                                Sym : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininstconstant ;
      WITH Bininstconstant DO
         inst    := Inst ;
         unit    := Unit ;
         srcsym  := Sym
      END
   END ;
   AddDependant(i)
END BuildBinInstConstant ;


(*
   BinInstConstant - inst operating on a reg. (inst <-> Reg)
                     Inst   - determines the instruction.
                     Unit   - determines the DatumPrefix of transfer.
                     Sym    - determine the constant.

                     Example:

                     push   15
*)

PROCEDURE BinInstConstant (Inst: Instruction;
                           Unit: DatumPrefix;
                           Sym : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstConstant(Inst, Unit, Sym)
   ELSE
      BuildBinInstConstant(Inst, Unit, Sym)
   END
END BinInstConstant ;


(*
   BuildBinInstMD - builds instructions of the form, inst (ebp+15)
*)

PROCEDURE BuildBinInstMD (Inst: Instruction; Unit: DatumPrefix;
                          O1, GO: INTEGER; Scope: KindOfScope; Sym: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininstmd ;
      WITH Bininstmd DO
         inst    := Inst ;
         unit    := Unit ;
         o1      := O1 ;
         go      := GO ;
         scope   := Scope ;
         sym     := Sym
      END
   END ;
   AddDependant(i)
END BuildBinInstMD ;


(*
   BinInstMD - inst operating on a memory location. (inst <-> Mem)
               Inst   - determines the instruction.
               Unit   - determines the DatumPrefix of transfer.
               o1, go - determine the operand memory location.
               Scope1 - determines the scope of the operand.
               Sym    - symbol requiring the memory reference.

               Example:

               push   (ebp+15)
*)

PROCEDURE BinInstMD (Inst: Instruction; Unit: DatumPrefix;
                     o1, go  : INTEGER; Scope1: KindOfScope; Sym: CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstMD(Inst, Unit, o1, go, Scope1, Sym)
   ELSE
      BuildBinInstMD(Inst, Unit, o1, go, Scope1, Sym)
   END
END BinInstMD ;


(*
   BuildBinInstHex - 
*)

PROCEDURE BuildBinInstHex (Inst: Instruction;
                           Unit: DatumPrefix;
                           c   : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := bininsthex ;
      WITH Bininsthex DO
         inst    := Inst ;
         unit    := Unit ;
         card    := c
      END
   END ;
   AddDependant(i)
END BuildBinInstHex ;


(*
   BinInstHex - inst operating on a number. (inst <-> hex)
                Inst   - determines the instruction.
                Unit   - determines the DatumPrefix of transfer.
                c      - number.

                Example:

                push   0x1234
*)

PROCEDURE BinInstHex (Inst: Instruction;
                      Unit: DatumPrefix;
                      c   : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.BinInstHex(Inst, Unit, c)
   ELSE
      BuildBinInstHex(Inst, Unit, c)
   END
END BinInstHex ;


(*
   WriteConstant - writes a constant from a symbol.
                   writes out the following consts:

                   (1)  Integer, Cardinal value constants.
                   (2)  Character constants.
                   (3)  String constants.
*)

PROCEDURE WriteConstant (Sym: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.WriteConstant(Sym)
   END
END WriteConstant ;


(*
   Comment - places string, a, into the assembly file.
*)

PROCEDURE Comment (a: ARRAY OF CHAR) ;
BEGIN
   IF NOT APUCodeGenerator
   THEN
      (* not interested in i586 comments if generating apu code *)
      M2AS.Comment(a)
   END
END Comment ;


(*
   DeclareLocalVariables - lists the Local variables for Procedure Sym
                           together with their offset.
*)

PROCEDURE DeclareLocalVariables (Sym: CARDINAL) ;
BEGIN
   IF NOT APUCodeGenerator
   THEN
      M2AS.DeclareLocalVariables(Sym)
   END
END DeclareLocalVariables ;


(*
   GenerateQuadLabel - writes a label refering to quad, QuadNo.
*)

PROCEDURE GenerateQuadLabel (QuadNo: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   IF IsBackReference(QuadNo)
   THEN
      IF NOT APUCodeGenerator
      THEN
         IF NOT IsBackReferenceConditional(QuadNo)
         THEN
            (* align unconditional branch target loop starts *)
            M2AS.CodeAlignment
         END ;
         M2AS.GenerateQuadLabel(QuadNo)
      END ;

      IF DebugOn
      THEN
         PushOutput(IO.Write)    (* debugging purposes *)
      END ;
      Comment('loop found, setting cache to empty') ;
      IF DebugOn
      THEN
         PopOutput
      END ;
      ResetCurrentCache(empty)   (* likely to be a loop, in which case allow initial phase no cache misses *)
   ELSE
      IF NOT APUCodeGenerator
      THEN
         M2AS.GenerateQuadLabel(QuadNo)
      END
   END
END GenerateQuadLabel ;


(*
   GenerateEntryProcedureLabel - generates a procedure label for procedure, Sym.
                                 The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateEntryProcedureLabel (CurrentQuad: CARDINAL; Sym: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(empty) ;  (* was filled (gaius) *)
   IF NOT APUCodeGenerator
   THEN
      M2AS.GenerateEntryProcedureLabel(CurrentQuad, Sym)
   END
END GenerateEntryProcedureLabel ;


(*
   GenerateExitProcedureLabel - generates a procedure label for procedure, Sym.
                                The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateExitProcedureLabel (Sym: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(filled) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.GenerateExitProcedureLabel(Sym)
   END
END GenerateExitProcedureLabel ;


(*
   OpenAssemblyFileAndInitialize - displays the initialization code.
*)

PROCEDURE OpenAssemblyFileAndInitialize ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(filled) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.OpenAssemblyFileAndInitialize
   END ;
   IgnorePipeLine := NOT (Optimize586 OR Optimize686)
END OpenAssemblyFileAndInitialize ;


(*
   CloseAssemblyFileAndTerminate - displays termination code
*)

PROCEDURE CloseAssemblyFileAndTerminate ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(filled) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.CloseAssemblyFileAndTerminate
   END
END CloseAssemblyFileAndTerminate ;


(*
   WritePrologue - writes the prologue code before the
                   initialization code of a module, ModSym.
*)

PROCEDURE WritePrologue (ModSym: CARDINAL;
                         CurrentQuad: CARDINAL; Line: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(filled) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.WritePrologue(ModSym, CurrentQuad, Line)
   END ;
   LastInstructionPipe := Upipe
END WritePrologue ;


(*
   WriteEpilogue - writes the termination code after the initialization
                   of a module.
*)

PROCEDURE WriteEpilogue (ModSym: CARDINAL) ;
BEGIN
   FlushInstruction(FALSE) ;
   ResetCurrentCache(filled) ;
   IF NOT APUCodeGenerator
   THEN
      M2AS.WriteEpilogue(ModSym)
   END
END WriteEpilogue ;


(*
   BuildInstRegImmName - 
*)

PROCEDURE BuildInstRegImmName (Inst: Instruction;
                               Unit: DatumPrefix;
                               Reg : Register;
                               Sym : CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instregimmname ;
      WITH Instregimmname DO
         inst    := Inst ;
         unit    := Unit ;
         destreg := Reg ;
         srcsym  := Sym
      END
   END ;
   AddDependant(i)
END BuildInstRegImmName ;


(*
   InstRegImmName -  instruction immediate name into register.  (Reg <- inst <- Constant)
                     Inst  - determines the instruction.
                     Unit  - determines the DatumPrefix of transfer.
                     Reg   - destination register
                     Sym   - constant element.

                     Example:

                     mov   $foo, Reg
*)

PROCEDURE InstRegImmName (Inst: Instruction;
                          Unit: DatumPrefix;
                          Reg : Register;
                          Sym : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstRegImmName(Inst, Unit, Reg, Sym)
   ELSE
      BuildInstRegImmName(Inst, Unit, Reg, Sym)
   END
END InstRegImmName ;


(*
   BuildInstName - 
*)

PROCEDURE BuildInstName (Inst: Instruction;
                         Unit: DatumPrefix;
                         Sym: CARDINAL) ;
VAR
   i: InstrIndex ;
BEGIN
   i := NewInstruction() ;
   WITH Instr[i] DO
      InstructionType := instname ;
      WITH Instname DO
         inst := Inst ;
         unit := Unit ;
         sym  := Sym
      END
   END ;
   AddDependant(i)
END BuildInstName ;


(*
   InstName - instruction to operate on a symbol name.
              Inst  - determines the instruction.
              Unit  - determines the DatumPrefix of manipulation
              Sym   - symbol whose name is used as the operand

              Example:

              call  $foo
*)

PROCEDURE InstName (Inst: Instruction;
                    Unit: DatumPrefix;
                    Sym : CARDINAL) ;
BEGIN
   IF IgnorePipeLine
   THEN
      M2AS.InstName(Inst, Unit, Sym)
   ELSE
      BuildInstName(Inst, Unit, Sym) ;
      IF Inst=call
      THEN
         FlushInstruction(TRUE)
      END
   END
END InstName ;


(*
   MovePackedStringTo - moves a packed string b into the memory location
                        indicated by DestSym.
                        GeneralOffset determines the number of bytes offset
                        from DestSym the packed string is to be placed.
                        Pack determines the number of packed characters
                        must be stored at location DestSym.
                        If the string, b, contains less characters than
                        Pack then the remaining characters must be filled
                        in as nul.
*)

PROCEDURE MovePackedStringTo (DestSym: CARDINAL; GeneralOffset: INTEGER;
                              b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   Offset := GetOffset(DestSym) ;
   IF GetMode(DestSym)=RightValue
   THEN
      InstMDInt(mov, GetUnitPrefix(Pack), Offset, GeneralOffset, INTEGER(Word), GetScope(DestSym), DestSym)
   ELSE
      Assert(GetMode(DestSym)=LeftValue) ;
      InstRegMD(mov, MaxDatum(), RegEDI, Offset, 0, GetScope(DestSym), DestSym) ;
      InstMRICardinal(mov, GetUnitPrefix(Pack), Offset, GeneralOffset, RegEDI, Word)
   END
END MovePackedStringTo ;


(*
   MovePackedStringToIReg - moves a packed string b into the memory location
                            indicated by (Reg)
                            GeneralOffset determines the number of bytes offset
                            from IReg the packed string is to be placed.
                            Pack determines the number of packed characters
                            must be stored at location (Reg+GeneralOffset).
                            If the string, b, contains less characters than
                            Pack then the remaining characters must be filled
                            in as nul.
*)

PROCEDURE MovePackedStringToIReg (Reg: Register; GeneralOffset: INTEGER;
                                  b: ARRAY OF CHAR; Pack: CARDINAL) ;
VAR
   Offset: INTEGER ;
   i,
   Word  : CARDINAL ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(b)
      THEN
         INC(Word, ORD(b[i]))
      END
   END ;
   InstMRICardinal(mov, GetUnitPrefix(Pack), GeneralOffset, 0, Reg, Word)
END MovePackedStringToIReg ;


(* *********************************************************************** *)
(*                        C a c h e     M o d e l                          *)
(* *********************************************************************** *)


(*
   TouchCache - touches the cache, updates the LRU data structure of the cache line.
*)



(*
   IsCacheLineConflict - returns TRUE if a..a+sizea conflicts with b..b+sizeb.
*)

PROCEDURE IsCacheLineConflict (a, sizea, b, sizeb: INTEGER) : BOOLEAN ;
VAR
   i, j: INTEGER ;
BEGIN
   FOR i := a TO a+sizea-1 DO
      FOR j := b TO b+sizeb-1 DO
         IF ABS(i-j) DIV CacheBucket=0
         THEN
            RETURN( TRUE )
         END
      END
   END ;
   RETURN( FALSE )
END IsCacheLineConflict ;


(*
   DisplayCache - displays the cache
*)

PROCEDURE DisplayCache ;
VAR
   line,
   index: CARDINAL ;
BEGIN
   WriteString('# cache model') ; WriteLn ;
   index := 0 ;
   WHILE index<CurrentPipeLine.Cache.NoOfLinesUsed DO
      line := 0 ;
      WHILE line<ActualSameCacheLines DO
         WriteString('# cache line ') ; WriteCard(CurrentPipeLine.Cache.Lines[index].SetValue, 0) ; Write(' ') ;
         WriteString(' [') ; WriteCard(line, 0) ; WriteString(']  ') ;
         WITH CurrentPipeLine.Cache.Lines[index].SameLine[line] DO
            IF Father#NulSym
            THEN
               WriteString(' father = ') ; WriteKey(GetSymName(Father)) ;
               WriteString(' addr = ') ; WriteInt(AddressLo, 0) ; WriteString('..') ; WriteInt(AddressHi, 0) ;
               WriteString(' MRU = ') ; WriteCard(CurrentPipeLine.Cache.Lines[index].MRULine, 1) ;
               WriteString(' read ') ; WriteBin(ReadBytes, 32)
            END
         END ;
         INC(line) ;
         WriteString(' SP = ') ; WriteHex(CARDINAL(CurrentPipeLine.Cache.StackPtr), 8) ;
         WriteLn
      END ;
      INC(index)
   END
END DisplayCache ;


(*
   SetCache - 
*)

PROCEDURE SetCache (index: CARDINAL; line: CARDINAL;
                    setvalue: CARDINAL;
                    sym: CARDINAL; size: CARDINAL; offset: INTEGER) ;
VAR
   addr,
   cost: CARDINAL ;
BEGIN
   CurrentPipeLine.Cache.Lines[index].SetValue := setvalue ;
   WITH CurrentPipeLine.Cache.Lines[index].SameLine[line] DO
      AddressLo := (offset DIV 32) * 32 ;
      AddressHi := AddressLo + 31 ;
      addr      := AddressHi ;
      Father    := sym ;
      IF DebugOn
      THEN
         DisplayCache
      END
      (*
         actually we do not consider data larger than 32 bytes (cache data bucket) since:
         - an array access will still be considered as sizeof(array) (wrong..)
         - also these cases are, hopefully, rare.

         the purist would fix the array case above and make sure multiple cache adds work. (fixme)
      *)
   END
END SetCache ;


(*
   CacheLineExists - returns TRUE if a cache line exists with, setvalue.
*)

PROCEDURE CacheLineExists (setvalue: CARDINAL; VAR index: CARDINAL) : BOOLEAN ;
BEGIN
   index := 0 ;
   WHILE index<CurrentPipeLine.Cache.NoOfLinesUsed DO
      WITH CurrentPipeLine.Cache.Lines[index] DO
         IF setvalue=SetValue
         THEN
            RETURN( TRUE )
         ELSE
            INC(index)
         END
      END
   END ;
   RETURN( FALSE )
END CacheLineExists ;


(*
   CacheLineMatch - returns TRUE if cache line, index, contains address: offset..offset+size.
*)

PROCEDURE CacheLineMatch (setvalue: CARDINAL;
                          offset: INTEGER; size: CARDINAL; sym: CARDINAL;
                          VAR index: CARDINAL; VAR line: CARDINAL) : BOOLEAN ;
BEGIN
   index := 0 ;
   IF CacheLineExists(setvalue, index)
   THEN
      (* yes cache line exists, but it maybe for a different address *)
      line := 0 ;
      WHILE line<ActualSameCacheLines DO
         WITH CurrentPipeLine.Cache.Lines[index].SameLine[line] DO
            IF (sym=Father) AND IsSubSet(offset, offset+size-1, AddressLo, AddressHi)
            THEN
               RETURN( TRUE )
            ELSE
               INC(line)
            END
         END
      END
   END ;
   RETURN( FALSE )
END CacheLineMatch ;


(*
   CalculateSetValue - works out the setvalue given an address, offset.
*)

PROCEDURE CalculateSetValue (offset: INTEGER) : CARDINAL ;
BEGIN
   IF offset>=0
   THEN
      RETURN( (offset DIV 32) MOD 4096 )
   ELSE
      offset := 4096-(ABS(offset) MOD 2048) ;
      RETURN( (offset DIV 32) MOD 4096 )
   END
END CalculateSetValue ;


(*
   CheckCacheFilled - checks to see whether the cache bucket should be set to filled.
*)

PROCEDURE CheckCacheFilled (index: CARDINAL) ;
BEGIN
   WITH CurrentPipeLine.Cache.Lines[index] DO
      IF MRULine=0
      THEN
         State := filled
      END
   END
END CheckCacheFilled ;


(*
   AddCache - adds a new address, sym:offset..offset+size-1 to the cache model.
*)

PROCEDURE AddCache (setvalue: CARDINAL; offset: INTEGER; size: CARDINAL; sym: CARDINAL; IsWrite: BOOLEAN) : CARDINAL ;
VAR
   line,
   index: CARDINAL ;
BEGIN
   IF CacheLineExists(setvalue, index)
   THEN
      line := 0 ;
      WITH CurrentPipeLine.Cache.Lines[index] DO
         Assert(setvalue=SetValue) ;
         MRULine := (MRULine+ActualSameCacheLines-1) MOD ActualSameCacheLines ;
         (* add new cache lines in reverse *)
         SetCache(index, MRULine, setvalue, sym, size, offset) ;
         (* no reads logged yet *)
         CurrentPipeLine.Cache.Lines[index].SameLine[MRULine].ReadBytes := {} ;
         CheckCacheFilled(index) ;
         RETURN( TouchCache(index, MRULine, offset, size, IsWrite) )
      END
   ELSE
      index := CurrentPipeLine.Cache.NoOfLinesUsed ;
      INC(CurrentPipeLine.Cache.NoOfLinesUsed) ;
      IF CurrentPipeLine.Cache.NoOfLinesUsed=ActualCacheLines
      THEN
         InternalError('cache line overflow', __FILE__, __LINE__)
      END ;
      CurrentPipeLine.Cache.Lines[index].MRULine := 0 ;
      (* and blank out lines *)
      FOR line := 0 TO ActualSameCacheLines DO
         CurrentPipeLine.Cache.Lines[index].SameLine[line].Father := NulSym
      END ;
      CurrentPipeLine.Cache.Lines[index].State := NewCacheLineCost ;
      WITH CurrentPipeLine.Cache.Lines[index] DO
         SetCache(index, MRULine, setvalue, sym, size, offset) ;
         (* no reads logged yet *)
         CurrentPipeLine.Cache.Lines[index].SameLine[MRULine].ReadBytes := {} ;
         RETURN( TouchCache(index, MRULine, offset, size, IsWrite) )
      END
   END
END AddCache ;

      
(*
   AddToCache - returns the number of cycles consumed by a cache miss.
*)

PROCEDURE AddToCache (sym: CARDINAL; size: CARDINAL; offset: INTEGER; IsWrite: BOOLEAN) : CARDINAL ;
VAR
   setvalue: CARDINAL ;
   split   : INTEGER ;
   line,
   index   : CARDINAL ;
BEGIN
   IF OptimizeCache
   THEN
      setvalue := CalculateSetValue(offset) ;
      IF CacheLineMatch(setvalue, offset, size, sym, index, line)
      THEN
         (* exists for the same address *)
         RETURN( TouchCache(index, line, offset, size, IsWrite) )
      ELSE
         RETURN( AddCache(setvalue, offset, size, sym, IsWrite) )
      END
   ELSE
      RETURN( 0 )
   END
END AddToCache ;


(* *********************************************************************** *)


(*
   Init - initialize this modules global variables.
*)

PROCEDURE Init ;
VAR
   a: ARRAY [0..100] OF CHAR ;
BEGIN
   WITH Instr[NopIndex] DO
      Branch1         := 0 ;
      Branch2         := 0 ;
      Parent          := 0 ;
      InstructionType := bininst ;
      Bininst.inst    := nop ;
   END ;
   LeftOverInstruction  := NopIndex ;
   MustFlushInstruction := FALSE ;
   NoOfTrees            := 0 ;
   NoOfInstructions     := 0 ;
   AccumulatedCost      := 0 ;
   LastInstructionPipe  := Unknown ;
   DebugOn              := GetEnvironment('DEBUGPIPE', a) ;
   ResetCurrentCache(filled)
END Init ;


BEGIN
   Init
END M2Pipe586.
(*
 * Local variables:
 * compile-command: "m2f -quiet -g -M \"../libs 586 .\" M2Pipe586"
 * End:
 *)
