IMPLEMENTATION MODULE SymbolKey ;


FROM Storage IMPORT ALLOCATE, DEALLOCATE ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard ;
FROM NameKey IMPORT WriteKey ;
FROM Debug IMPORT Halt ;


TYPE
(*
   SymbolTree = POINTER TO Node ;
   Node       = RECORD
                   KeyName : CARDINAL ;   (* The sorted entity *)
                   KeySym  : CARDINAL ;   (* The value entity  *)
                   Left    : SymbolTree ;
                   Right   : SymbolTree ;
                END ;
*)

   Relation= (less, equal, greater) ;


(* %%%FORWARD%%%
PROCEDURE Kill (t: SymbolTree) ; FORWARD ;
PROCEDURE KillTree (VAR t: SymbolTree) ; FORWARD ;
PROCEDURE PrintItem (t: SymbolTree) ; FORWARD ;
PROCEDURE Rel (a, b: CARDINAL) : Relation ; FORWARD ;
PROCEDURE Search (t: SymbolTree; n: CARDINAL;
                  VAR id, father: SymbolTree; VAR r: Relation) ; FORWARD ;
PROCEDURE SearchAndDo (t: SymbolTree; P: PerformOperation) ; FORWARD ;
PROCEDURE SearchForAny (t: SymbolTree; P: IsSymbol) : BOOLEAN ; FORWARD ;
PROCEDURE TraverseTree (t: SymbolTree) ; FORWARD ;

PROCEDURE InitTree (VAR t: SymbolTree) ; FORWARD ;
PROCEDURE KillTree (VAR t: SymbolTree) ; FORWARD ;
PROCEDURE Kill (t: SymbolTree) ; FORWARD ;
PROCEDURE KillTree (VAR t: SymbolTree) ; FORWARD ;
PROCEDURE GetSymKey (t: SymbolTree; NameKey: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE PutSymKey (t: SymbolTree; NameKey: CARDINAL; SymKey: CARDINAL) ; FORWARD ;
PROCEDURE DelSymKey (t: SymbolTree; NameKey: CARDINAL) ; FORWARD ;
PROCEDURE IsEmptyTree (t: SymbolTree) : BOOLEAN ; FORWARD ;
PROCEDURE Tabulate (t: SymbolTree) ; FORWARD ;
PROCEDURE PrintItem (t: SymbolTree) ; FORWARD ;
PROCEDURE DoesTreeContainAny (t: SymbolTree; P: IsSymbol) : BOOLEAN ; FORWARD ;
   %%%FORWARD%%% *)


PROCEDURE InitTree (VAR t: SymbolTree) ;
BEGIN
   NEW(t) ;
   WITH t^ DO
      Left := NIL ;
      Right := NIL
   END
END InitTree ;


PROCEDURE KillTree (VAR t: SymbolTree) ;
BEGIN
   IF t#NIL
   THEN
      Kill(t) ;  (* Would like to place Kill in here but the compiler *)
                 (* gives a type incompatible error... so i've split  *)
                 (* the procedure into two. - Problem i think with    *)
                 (* VAR t at the top?                                 *)
      t := NIL
   END
END KillTree ;


PROCEDURE Kill (t: SymbolTree) ;
BEGIN
   IF t#NIL
   THEN
      Kill(t^.Left) ;
      Kill(t^.Right) ;
      DISPOSE(t)
   END
END Kill ;

(*
   Would really like KillTree to look like this...

PROCEDURE KillTree (VAR t: SymbolTree) ;
BEGIN
   IF t#NIL
   THEN
      KillTree(t^.Left) ;
      KillTree(t^.Right) ;
      DISPOSE(t) ;
      t := NIL
   END
END KillTree ;

  But the compiler gives type incompatibilities under the parameters
  for:
       KillTree(t^.Left)
       KillTree(t^.Right)

*)


PROCEDURE GetSymKey (t: SymbolTree; NameKey: CARDINAL) : CARDINAL ;
VAR
   r     : Relation ;
   f, c  : SymbolTree ;
   SymKey: CARDINAL ;
BEGIN
   Search( t, NameKey, c, f, r ) ;
   IF c=NIL
   THEN
      SymKey := NulKey
   ELSE
      SymKey := c^.KeySym
   END ;
   RETURN( SymKey )
END GetSymKey ;


PROCEDURE PutSymKey (t: SymbolTree; NameKey: CARDINAL; SymKey: CARDINAL) ;
VAR
   r     : Relation ;
   f, c  : SymbolTree ;
BEGIN
   Search( t, NameKey, c, f, r ) ;
   IF r=equal
   THEN
      Halt('symbol already stored', __LINE__, __FILE__)
   ELSE
      IF c=NIL
      THEN
         IF r=less
         THEN
            NEW( f^.Left ) ;
            c := f^.Left
         ELSIF r=greater
         THEN
            NEW( f^.Right ) ;
            c := f^.Right
         END ;
         WITH c^ DO
            Right := NIL ;
            Left := NIL ;
            KeySym := SymKey ;
            KeyName := NameKey
         END
      END
   END
END PutSymKey ;


(*
   NB in order for this to work we must ensure that the InitTree sets
      both Left and Right to NIL.
*)

PROCEDURE DelSymKey (t: SymbolTree; NameKey: CARDINAL) ;
VAR
   i, c, f: SymbolTree ;
   r      : Relation ;
BEGIN
   Search(t, NameKey, c, f, r) ;  (* Find father and child of the node *)
   IF r=equal
   THEN
      (* Have found the node to be deleted *)
      IF f^.Right=c
      THEN
         (* Node is c and this is greater than the father, f.  *)
         (* Greater being on the right.                        *)
         (* Connect c^.Left onto the f^.Right. Connect         *)
         (* c^.Right onto the end of the right most branch of  *)
         (* c^.Left.                                           *)
         IF c^.Left#NIL
         THEN
            (* Scan for Right most node of c^.Left *)
            i := c^.Left ;
            WHILE i^.Right#NIL DO
               i := i^.Right
            END ;
            i^.Right := c^.Right ;
            f^.Right := c^.Left
         ELSE
            (* No c^.Left node therefore link over c (as in a  *)
            (* single linked list) to c^.Right.                *)
            f^.Right := c^.Right
         END ;
         DISPOSE(c)
      ELSE
         (* Assert that f^.Left=c will always be true! *)
         (* Perform exactly the mirror immage of the above code *)

         (* Connect c^.Right onto the f^.Left. Connect         *)
         (* c^.Left onto the end of the Left most branch of    *)
         (* c^.Right.                                          *)
         IF c^.Right#NIL
         THEN
            (* Scan for Left most node of c^.Right *)
            i := c^.Right ;
            WHILE i^.Left#NIL DO
               i := i^.Left
            END ;
            i^.Left := c^.Left ;
            f^.Left := c^.Right
         ELSE
            (* No c^.Right node therefore link over c (as in a *)
            (* single linked list) to c^.Left.                 *)
            f^.Left := c^.Left
         END ;
         DISPOSE(c)
      END
   ELSE
      Halt('Trying to delete a symbol that is not in the tree - the compiler algorithm never expects this to occur',
            __LINE__, __FILE__)
   END
END DelSymKey ;


(*
   Search table, t,with name equal to n. id is set to the node if
   an entry is found, father is set to the node above id.
*)

PROCEDURE Search (t: SymbolTree; n: CARDINAL;
                  VAR id, father: SymbolTree; VAR r: Relation) ;
BEGIN
   father := t ;
   IF t=NIL
   THEN
      Halt('parameter t should never be NIL', __LINE__, __FILE__)
   END ;
   id := t^.Left ;
   IF id=NIL
   THEN
      r := less
   ELSE
      REPEAT
         r := Rel( n, id^.KeyName ) ;
         IF r=less
         THEN
            father := id ;
            id := id^.Left
         ELSIF r=greater
         THEN
            father := id ;
            id := id^.Right
         END
      UNTIL (id=NIL) OR (r=equal)
   END
END Search ;


(*
   Compare a and b returning the comparison
*)

PROCEDURE Rel (a, b: CARDINAL) : Relation ;
VAR
   r: Relation ;
BEGIN
   IF a<b
   THEN
      r := less
   ELSIF a>b
   THEN
      r := greater
   ELSE
      r := equal
   END ;
   RETURN( r )
END Rel ;


(*
   IsEmptyTree - returns true if SymbolTree, t, is empty.
*)

PROCEDURE IsEmptyTree (t: SymbolTree) : BOOLEAN ;
BEGIN
   RETURN( t^.Left=NIL )
END IsEmptyTree ;


(*
   Tabulate - displays in alphabetical order the symbols contained
              in tree, t.
*)

PROCEDURE Tabulate (t: SymbolTree) ;
BEGIN
   WriteString('Symbol   Name') ; WriteLn ;
   WriteLn ;
   TraverseTree(t^.Left)
END Tabulate ;


PROCEDURE TraverseTree (t: SymbolTree) ;
BEGIN
   IF t#NIL
   THEN
      TraverseTree(t^.Left) ;
      PrintItem(t) ;
      TraverseTree(t^.Right)
   END
END TraverseTree ;


PROCEDURE PrintItem (t: SymbolTree) ;
BEGIN
   WriteCard(t^.KeySym, 6) ;
   WriteString('   ') ;
   WriteKey(t^.KeyName) ;
   WriteLn
END PrintItem ;


(*
   DoesTreeContainAny - returns true if SymbolTree, t, contains any
                        symbols which in turn return true when procedure,
                        P, is called with a symbol as its parameter.
                        The SymbolTree root is empty apart from the field,
                        Left, hence we need two procedures.
*)

PROCEDURE DoesTreeContainAny (t: SymbolTree; P: IsSymbol) : BOOLEAN ;
BEGIN
   RETURN( SearchForAny(t^.Left, P) )
END DoesTreeContainAny ;


(*
   SearchForAny - performs the search required for DoesTreeContainAny.
                  The root node always contains a nul data value,
                  therefore we must skip over it.
*)

PROCEDURE SearchForAny (t: SymbolTree; P: IsSymbol) : BOOLEAN ;
BEGIN
   IF t=NIL
   THEN
      RETURN( FALSE )
   ELSE
      RETURN( P(t^.KeySym) OR SearchForAny(t^.Left, P) OR
                              SearchForAny(t^.Right, P)
            )
   END
END SearchForAny ;


(*
   ForeachNodeDo - for each node in SymbolTree, t, a procedure, P,
                   is called with the node symbol as its parameter.
                   The tree root node only contains a legal Left pointer,
                   therefore we need two procedures to examine this tree.
*)

PROCEDURE ForeachNodeDo (t: SymbolTree; P: PerformOperation) ;
BEGIN
   SearchAndDo(t^.Left, P)
END ForeachNodeDo ;


(*
   SearchAndDo - searches all the nodes in SymbolTree, t, and
                 calls procedure, P, with a node as its parameter.
*)

PROCEDURE SearchAndDo (t: SymbolTree; P: PerformOperation) ;
BEGIN
   IF t#NIL
   THEN
      WITH t^ DO
         SearchAndDo(Right, P) ;
         P(KeySym) ;
         SearchAndDo(Left, P)
      END
   END
END SearchAndDo ;


END SymbolKey.
