IMPLEMENTATION MODULE M2BasicBlock ;


FROM Storage IMPORT ALLOCATE, DEALLOCATE ;

FROM M2Quads IMPORT IsReferenced, IsConditional, IsUnConditional, IsCall,
                    IsReturn, IsModFile, IsKillLocalVar,
                    IsPseudoQuad,
                    GetNextQuad, GetQuad, QuadOperator,
                    SubQuad ;

FROM M2Pass IMPORT SetPassToCodeGeneration ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard ;
FROM M2Pipe586 IMPORT FlushPipe ;

IMPORT M2GenAPU ;
IMPORT M2GenDyn486 ;

CONST
   FunctionEndTime   = 8 ;
   FunctionBeginTime = 5 ;     (* i586 cycles *)

TYPE
   PtrToBasicBlock = POINTER TO BasicBlock ;
   BasicBlock      = RECORD
                        StartQuad  : CARDINAL ;  (* First Quad in Basic Block *)
                        EndQuad    : CARDINAL ;  (* End Quad in Basic Block   *)
                        CPUCycles  : CARDINAL ;  (* Number of CPU cycles the  *)
                                                 (* Basic Block requires for  *)
                                                 (* execution.                *)
                                                 (* Next Basic Block in list  *)
                        Right      : PtrToBasicBlock ;
                                                 (* Last Basic Block in list  *)
                        Left       : PtrToBasicBlock ;
                     END ;

VAR
   HeadOfBB    : PtrToBasicBlock ;   (* First in the list of Basic Blocks *)
   FreeBB      : PtrToBasicBlock ;   (* Free list of Basic Blocks         *)


(* %%%FORWARD%%%
PROCEDURE Add (VAR Head: PtrToBasicBlock;
               b : PtrToBasicBlock) ; FORWARD ;
PROCEDURE CodeBasicBlocks (Head: CARDINAL) ; FORWARD ;
PROCEDURE CodeBasicBlocks (Head: CARDINAL) ; FORWARD ;
PROCEDURE CodeMainBB (Head: CARDINAL) ; FORWARD ;
PROCEDURE CodeProceduresBB (Head: CARDINAL) ; FORWARD ;
PROCEDURE ConvertQuads2BasicBlock (VAR Head: CARDINAL) ; FORWARD ;
PROCEDURE DisplayBasicBlocks ; FORWARD ;
PROCEDURE DisplayBlock (b: PtrToBasicBlock) ; FORWARD ;
PROCEDURE EndBB (b: PtrToBasicBlock; Quad: CARDINAL) ; FORWARD ;
PROCEDURE InitializeBasicBlock ; FORWARD ;
PROCEDURE InsideProcedure (VAR CurrentlyInside: BOOLEAN;
                           b: PtrToBasicBlock) : BOOLEAN ; FORWARD ;
PROCEDURE StartBB (b: PtrToBasicBlock; Quad: CARDINAL) ; FORWARD ;
PROCEDURE Sub (VAR Head: PtrToBasicBlock;
               b: PtrToBasicBlock) ; FORWARD ;
   %%%FORWARD%%% *)

(*
   BasicBlock - Converts a list of quadruples, Head, to a list of Basic
                Blocks, each Basic Block is then passed on to the
                specific target code generation phase.

                Analageous with a protocol:

                +------------------------------------------------+
                |     Generate Quadruples from Source            |
                +------------------------------------------------+
                     |
                     v
                +------------------------------------------------+
                | Convert List of Quadruples to Basic Blocks     |
                +------------------------------------------------+
                     |            |
                     v            v
                +-----------+
                |  Code     |    ....
                +-----------+
*)

PROCEDURE GenBasicBlocks (VAR Head: CARDINAL) ;
BEGIN
   InitializeBasicBlock ;
   ConvertQuads2BasicBlock(Head)
   (* DisplayBasicBlocks *)
END GenBasicBlocks ;


(*
   DestroyBasicBlocks - destroys the list of Basic Blocks - so that
                        the quadruples can be furthur optimized.
*)

PROCEDURE DestroyBasicBlocks (VAR Head: CARDINAL) ;
VAR
   b, c: PtrToBasicBlock ;
BEGIN
   b := HeadOfBB ;
   IF b#NIL
   THEN
      REPEAT
         c := b^.Right ;
         DISPOSE(b) ;
         b := c
      UNTIL b=HeadOfBB
   END ;
   HeadOfBB := NIL
END DestroyBasicBlocks ;


(*
   GenBasicBlockCode - converts the Quadruples within Basic Blocks into
                       Code.
*)

PROCEDURE GenBasicBlockCode (Head: CARDINAL) ;
BEGIN
   SetPassToCodeGeneration ;
   CodeBasicBlocks(Head)
END GenBasicBlockCode ;


(*
   QuadMinusOne - returns the End-1 quadruple.
*)

PROCEDURE QuadMinusOne (Start, End: CARDINAL) : CARDINAL ;
VAR
   Last: CARDINAL ;
BEGIN
   Last := Start ;
   WHILE Start#End DO
      Last := Start ;
      Start := GetNextQuad(Start)
   END ;
   RETURN( Last )
END QuadMinusOne ;


(*
   TwoGenCode - invokes two code generators: the first
                to generate semantics of code; the second to
                generate execution times.
*)

PROCEDURE TwoGenCode (Head, Start, End: CARDINAL) ;
VAR
   sop, eop        : QuadOperator ;
   sop1, sop2, sop3,
   eop1, eop2, eop3: CARDINAL ;
   i               : CARDINAL ;
BEGIN
   IF Start<=End
   THEN
      GetQuad(Start, sop, sop1, sop2, sop3) ;
      GetQuad(End, eop, eop1, eop2, eop3) ;
      M2GenAPU.SetLocalFlags(Start) ;
      (* general rule is to generate timing before the equivalent smp code *)

      IF (sop=NewLocalVarOp) OR (sop=StartOp)
      THEN
         M2GenDyn486.GenCode(Head, Start, End) ;
         IF (eop=KillLocalVarOp) OR (eop=EndOp) OR (eop=ReturnOp)
         THEN
            M2GenAPU.GenCode(Head, Start, Start) ;
            FlushPipe(0) ;
            M2GenAPU.FlushTime ;
            M2GenAPU.GenCode(Head, GetNextQuad(Start), End) 
         ELSE
            M2GenAPU.GenCode(Head, Start, End) ;
            FlushPipe(0) ;
            M2GenAPU.FlushTime ;
         END
      ELSIF IsReferenced(Head, Start)
      THEN
         (* must emit the first apu quadruple before the timing calculation *)
         FlushPipe(0) ;
         M2GenAPU.FlushTime ;   (* flush old timing values *)
         M2GenDyn486.GenCode(Head, Start, End) ;
         IF (eop=KillLocalVarOp) OR (eop=EndOp) OR (eop=ReturnOp)
         THEN
            M2GenAPU.GenCode(Head, Start, Start) ;
            FlushPipe(0) ;
            M2GenAPU.FlushTime ;
            M2GenAPU.GenCode(Head, GetNextQuad(Start), End)
         ELSE
            M2GenAPU.GenCode(Head, Start, End) ;
            FlushPipe(0) ;
            M2GenAPU.FlushTime
         END
      ELSE
         M2GenDyn486.GenCode(Head, Start, End) ;
         FlushPipe(0) ;
         M2GenAPU.FlushTime ;
         M2GenAPU.GenCode(Head, Start, End)
      END
   END
END TwoGenCode ;


(*
   ConvertQuads2BasicBlock - converts a list of quadruples to a list of
                             Basic Blocks.
                             A Basic Block is defined as a list of quadruples
                             which has only has one entry and exit point.
*)

PROCEDURE ConvertQuads2BasicBlock (VAR Head: CARDINAL) ;
VAR
   LastQuadMod,
   LastQuadConditional,
   LastQuadCall,
   LastQuadReturn     : BOOLEAN ;
   Quad               : CARDINAL ;
   b                  : PtrToBasicBlock ;
   CurrentBB          : PtrToBasicBlock ;
   LastBB             : PtrToBasicBlock ;
BEGIN
   (*
      Algorithm to perform Basic Block:

      For every quadruple establish a set of leaders.
      A Leader is defined as a quadruple which is
      either:

      (i)   The first quadruple.
      (ii)  Any quadruple which is the target of a jump or unconditional jump.
      (iii) Any statement which follows a conditional jump

      For each leader construct a basic block.
      A Basic Block starts with a leader quadruple and ends with either:

      (i)  Another Leader
      (ii) An unconditional Jump.

      Any quadruples that do not fall into a Basic Block can be thrown away
      since they will never be executed.
   *)
   LastBB := NIL ;
   Quad := Head ;
   LastQuadConditional := TRUE ;  (* Force Rule (i) *)
   LastQuadCall := FALSE ;
   LastQuadReturn := FALSE ;
   LastQuadMod := FALSE ;
   (* Scan all quadruples *)
   WHILE Quad#0 DO
      IF LastQuadConditional OR LastQuadCall OR LastQuadReturn OR
         LastQuadMod OR IsReferenced(Head, Quad)
      THEN
         (* Rule (ii) *)
         NEW(CurrentBB) ;           (* Get a new Basic Block *)
         StartBB(CurrentBB, Quad) ; (* At least one quad in this Basic Block  *)
         EndBB(CurrentBB, Quad)
      ELSIF CurrentBB#NIL
      THEN
         (* We have a Basic Block - therefore add quad to this Block  *)
         EndBB(CurrentBB, Quad)
      ELSIF IsPseudoQuad(Quad)
      THEN
         (* Add Quad to the Last BB since Pseudo Quads - compiler directives *)
         (* must not be thrown away.                                         *)
         EndBB(LastBB, Quad)
      ELSIF IsReturn(Quad) OR IsKillLocalVar(Quad)
      THEN
         (* we must leave the ReturnOp alone as it indictes end of procedure *)
         EndBB(LastBB, Quad)
      ELSE
         (* Chuck this Quad since it will never be reached by the processor *)
         SubQuad(Head, Quad)
      END ;
      LastQuadConditional := IsConditional(Quad) ;
      LastQuadCall := IsCall(Quad) ;
      LastQuadReturn := IsReturn(Quad) ;
      LastQuadMod := IsModFile(Quad) ;
      IF IsUnConditional(Quad)
      THEN
         LastBB := CurrentBB ;
         CurrentBB := NIL
      END ;
      Quad := GetNextQuad(Quad)
   END
END ConvertQuads2BasicBlock ;


(*
   ForeachBasicBlockDo - for each basic block call procedure, p.
*)

PROCEDURE ForeachBasicBlockDo (p: BasicBlockProc) ;
VAR
   b: PtrToBasicBlock ;
BEGIN
   IF HeadOfBB#NIL
   THEN
      b := HeadOfBB ;
      REPEAT
         WITH b^ DO
            p(StartQuad, EndQuad)
         END ;
         b := b^.Right
      UNTIL HeadOfBB=b
   END
END ForeachBasicBlockDo ;


(*
   CodeBasicBlocks - codes all basic blocks individually.
*)

PROCEDURE CodeBasicBlocks (Head: CARDINAL) ;
BEGIN
   M2GenAPU.InitializationCode ;
   (* Code the procedures *)
   CodeProceduresBB(Head) ;
   (* Code Main *)
   CodeMainBB(Head) ;
   M2GenAPU.TerminationCode
END CodeBasicBlocks ;


(*
   CodeMainBB - Codes the Basic Blocks which are not inside procedures.
*)

PROCEDURE CodeMainBB (Head: CARDINAL) ;
VAR
   b              : PtrToBasicBlock ;
   CurrentlyInside: BOOLEAN ;
   Delta          : BOOLEAN ;
BEGIN
   CurrentlyInside := FALSE ;
   IF HeadOfBB#NIL
   THEN
      b := HeadOfBB ;
      REPEAT
         Delta := InsideProcedure(CurrentlyInside, b) ;
         IF NOT (CurrentlyInside OR Delta)
         THEN
            (*
               The Level has not changed since last time and still not
               inside a procedure
            *)
            WITH b^ DO
               TwoGenCode(Head, StartQuad, EndQuad)
            END
         END ;
         b := b^.Right
      UNTIL HeadOfBB=b
   END
END CodeMainBB ;


(*
   CodeProceduresBB - Code all basic blocks which implement the procedures.
*)

PROCEDURE CodeProceduresBB (Head: CARDINAL) ;
VAR
   b              : PtrToBasicBlock ;
   CurrentlyInside: BOOLEAN ;
   Delta          : BOOLEAN ;
BEGIN
   CurrentlyInside := FALSE ;
   IF HeadOfBB#NIL
   THEN
      b := HeadOfBB ;
      REPEAT
         Delta := InsideProcedure(CurrentlyInside, b) ;
         IF CurrentlyInside OR Delta
         THEN
            WITH b^ DO
               TwoGenCode(Head, StartQuad, EndQuad)
            END
         END ;
         b := b^.Right
      UNTIL HeadOfBB=b
   END
END CodeProceduresBB ;


(*
   InsideProcedure - Determines whether a Basic Block is inside a procedure.
                     A Delta boolean is returned to determine whether
                     the basic block has entered or exited from a procedure.
*)

PROCEDURE InsideProcedure (VAR CurrentlyInside: BOOLEAN;
                           b: PtrToBasicBlock) : BOOLEAN ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   q, f    : CARDINAL ;
   Delta   : BOOLEAN ;
BEGIN
   Delta := FALSE ;
   WITH b^ DO
      q := StartQuad ;
      f := GetNextQuad(EndQuad) ;
      WHILE q#f DO
         GetQuad(q, Operator, Operand1, Operand2, Operand3) ;
         IF Operator=NewLocalVarOp
         THEN
            (* The begining of a procedure *)
            CurrentlyInside := TRUE ;
            Delta := TRUE
         ELSIF Operator=ReturnOp
         THEN
            (* The end of a procedure *)
            CurrentlyInside := FALSE ;
            Delta := TRUE
         END ;
         q := GetNextQuad(q)
      END
   END ;
   RETURN( Delta )
END InsideProcedure ;


(*
   StartBB - Initially fills a Basic Block, b, with a start quad Quad.
             The Basic Block is then added to the end of Basic Block list.
*)

PROCEDURE StartBB (b: PtrToBasicBlock; Quad: CARDINAL) ;
BEGIN
   WITH b^ DO
      StartQuad := Quad ;
      EndQuad := Quad ;
      CPUCycles := 0
   END ;
   Add(HeadOfBB, b)   (* Add b to the end of the Basic Block list *)
END StartBB ;


(*
   EndBB - Fills a Basic Block, b, with an end quad Quad.
*)

PROCEDURE EndBB (b: PtrToBasicBlock; Quad: CARDINAL) ;
BEGIN
   WITH b^ DO
      EndQuad := Quad ;
   END
END EndBB ;


(*
   Add adds a specified element to the end of a queue.
*)
 
PROCEDURE Add (VAR Head: PtrToBasicBlock;
               b : PtrToBasicBlock) ;
BEGIN
   IF Head=NIL
   THEN
      Head := b ;
      b^.Left := b ;
      b^.Right := b
   ELSE
      b^.Right := Head ;
      b^.Left := Head^.Left ;
      Head^.Left^.Right := b ;
      Head^.Left := b
   END
END Add ;

 
(*
   Sub deletes an element from the specified queue.
*)
 
PROCEDURE Sub (VAR Head: PtrToBasicBlock;
               b: PtrToBasicBlock) ;
BEGIN
   IF (b^.Right=Head) AND (b=Head)
   THEN
      Head := NIL
   ELSE
      IF Head=b
      THEN
         Head := Head^.Right
      END ;
      b^.Left^.Right := b^.Right ;
      b^.Right^.Left := b^.Left
   END
END Sub ;


PROCEDURE InitializeBasicBlock ;
BEGIN
   HeadOfBB := NIL
END InitializeBasicBlock ;


(*
   DisplayBasicBlocks - displays the basic block data structure.
*)

PROCEDURE DisplayBasicBlocks ;
VAR
   b: PtrToBasicBlock ;
BEGIN
   b := HeadOfBB ;
   IF b#NIL
   THEN
      REPEAT
         DisplayBlock(b) ;
         b := b^.Right
      UNTIL b=HeadOfBB
   END
END DisplayBasicBlocks ;


PROCEDURE DisplayBlock (b: PtrToBasicBlock) ;
BEGIN
   WITH b^ DO
      WriteString('Start ') ; WriteCard(StartQuad, 6) ;
      WriteString(' End   ') ; WriteCard(EndQuad, 6) ;
      WriteString(' Cycles') ; WriteCard(CPUCycles, 6) ; WriteLn
   END
END DisplayBlock ;


END M2BasicBlock.
