IMPLEMENTATION MODULE M2GenAPU ;


FROM libc IMPORT exit ;
FROM ASCII IMPORT nul ;
FROM M2Configure IMPORT ReturnArgumentOnStack ;
FROM M2FileName IMPORT CalculateFileName ;
FROM M2AssemIO IMPORT OpenAssemblyFile, CloseAssemblyFile ;
FROM NameKey IMPORT WriteKey, GetKey, MakeKey, NulName ;
FROM M2AsmUtil IMPORT WriteAsmName, GetAsmName, GetModuleInitName ;
FROM Environment IMPORT GetEnvironment ;
FROM StdIO IMPORT PushOutput, PopOutput, Write ;

FROM M2Quads IMPORT QuadOperator, GetQuad, IsReferenced, GetNextQuad,
                    IsProfileOn, IsOptimizeOn, IsCodeOn,
                    QuadToLineNo, DisplayQuadList ;

FROM SymbolTable IMPORT PushOffset, PushSize, PushValue, PopValue,
                        PushSumOfLocalVarSize,
                        PushSumOfParamSize, PushVarSize,
                        ForeachExportedDo, ForeachImportedDo,
                        ForeachInnerModuleDo,
                        GetCurrentModule,  GetMainModule, GetScopeAuthor,
                        GetVarFather, GetSymName, ModeOfAddr, GetMode,
                        IsConstString, GetString, GetStringLength,
                        IsConst, IsProcedure, IsProcType, IsProcedureNested,
                        IsVar, IsVarParam,
                        IsUnbounded, IsArray, IsTemporary,
                        IsProcedureVariable, IsExported, IsModule,
                        GetType, GetNth, GetNthParam,
                        GetVarReadQuads,
                        GetSubrange, NoOfElements,
                        NulSym ;

FROM M2Debug IMPORT Assert ;
FROM M2Lexical IMPORT WriteError, InternalError ;

FROM M2ALU IMPORT PopCard, PushCard, PopInt, PushInt,
      	       	  SetOr, SetAnd, SetIn, Bit,
                  SetDifference, SetSymmetricDifference,
                  Add, Sub, Mult, Div, Mod,
                  Equ, NotEqu, Less, Gre, LessEqu, GreEqu ;

FROM M2System IMPORT Address, AddressSize, WordSize, ByteSize, IsPseudoSystemFunction ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM NumberIO IMPORT WriteCard, CardToStr, IntToStr, WriteInt ;
FROM M2Base IMPORT Cardinal, ActivationPointer ;
FROM M2Options IMPORT Optimizing, APUUseMemCopy, Alignment ;
FROM M2Inline IMPORT WriteInlineBuffer ;
FROM StrLib IMPORT StrEqual, StrConCat, StrCopy, StrLen ;

FROM M2GDB IMPORT LineInfo, ProcedureInfo, TypeInfo, MainInfo,
                  VariableInfo, BeginInfo, EndInfo, FileInfo ;

FROM M2Time IMPORT BinInstUnit ;

IMPORT i486 ;
IMPORT IO ;


CONST
   MaxString  =    255 ;
   DWORDSIZE  =      4 ;
   WORDSIZE   =      2 ;
   BYTESIZE   =      1 ;

TYPE
   Instructions = (add, sub, mult, div, mod, load, store, pr, equ, notequ,
                   less, gre, lessequ, greequ, memcopy, nop,
                   or, and, xor, not, sal, sar) ;
   Modes        = (direct, immediate, internal, indirect, indexed) ;
   Sizes        = (byte, word, dword) ;
   Registers    = (ra, rb, rc, rd, pc, sp, no, im, em, st) ;

VAR
   TotalDataSize    : CARDINAL ;
   CurrentQuad      : CARDINAL ;
   AbsoluteHead     : CARDINAL ;
   LocalCoding,                    (* Local compiler flags set by quadruples *)
   LocalOptimizing,                (* that are encountered by generating     *)
   LocalProfiling   : BOOLEAN ;    (* code.                                  *)
   UnboundedOnStack : BOOLEAN ;    (* Determines whether Unbounded Array     *)
                                   (* Data has been placed on the stack,     *)
                                   (* if true then the SP must be reduced    *)
                                   (* after the procedure call - to destroy  *)
                                   (* this data. Old SP is kept in RegC.     *)
   Zero, One          : CARDINAL ; (* names for 0, 1                         *)
   CompilingMainModule: BOOLEAN ;  (* are we compiling the main module?      *)
   LastOperator       : QuadOperator ; (* The last operator processed.       *)
   OutstandingDuration: CARDINAL ; (* No of cycles for current basic block   *)
   Debugging          : BOOLEAN ;  (* Are we debugging the apu code gen?     *)
   CurrentLineNo      : CARDINAL ; (* last line number seen                  *)
   CurrentProcedure,
   PreviousScope      : CARDINAL ; (* the current procedure being compiled   *)


(*
   Rules for Quadruples and Modes
   ==============================

   Symbols have four modes

   Mode = (None, Value, Offset, Immediate)

   A symbol can only change its mode from:
   None -> [Value, Offset, Immediate]
     any other alteration should be regarded as an error.


   Description of Mode
   ===================

   Sym<O> => Offset    - the value of symbol is contained at location Offset.
   Sym<V> => Value     - the symbol value is referenced indirectly via the
                         memory location offset.
   Sym<I> => Immediate - symbol is a constant and the value is in Symbol Table.


   Rules
   =====

   All program declared variables are given the mode, Offset.
   All constants have mode, Immediate.

   Operators
   =========

------------------------------------------------------------------------------
   Array Operators
------------------------------------------------------------------------------
   Sym<I>   Base   a            Delivers a constant result if a is a
                                Global variable. If a is a local variable
                                then the Frame pointer needs to be added.
                                Base yields the effective location in memory
                                of, a, array [0,0, .. ,0] address.
   Sym<I>   ElementSize 1       Always delivers a constant. The number
                                indicates which specified element is chosen.
                                ElementSize is the TypeSize for that element.
   Unbounded  Op1 Op3           Initializes the Op1 StartAddress of the array
                                Op3. Op3 can be a normal array or unbounded array.
                                This operator pushes the value of the array
                                onto the stack and sets Op1 to the
                                start of this array copy.
                                Op1 (is the Unbounded.ArrayAddress) :=
                                ADR(CopyOfOp3)
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   Addr Operator  - contains the address of a variable - may need to add
------------------------------------------------------------------------------
   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
------------------------------------------------------------------------------
   Xindr Operator  ( *a = b)
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   IndrX Operator  (a = *b)   where <X> means any value
------------------------------------------------------------------------------
   Sym1<X>   IndrX  Sym2<I>     meaning     Mem[Sym1<I>] := Mem[constant]
   Sym1<X>   IndrX  Sym2<I>     meaning     Mem[Sym1<I>] := Mem[constant]

   Sym1<X>   IndrX  Sym2<X>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<X>   IndrX  Sym2<X>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   + - / * Operators
------------------------------------------------------------------------------
   Sym1<I>   +      Sym2<I>  Sym3<I>  meaning Sym1<I> := Sym2<I> + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<I>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<O>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<O>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<V>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
------------------------------------------------------------------------------
   Base Operator
------------------------------------------------------------------------------
   Sym1<O>   Base  Sym2   Sym3<O>     meaning     Mem[Sym1<I>] := Sym3<I>
   Sym1<V>   Base  Sym2   Sym3<O>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Sym3<I>
   Sym1<O>   Base  Sym2   Sym3<V>     meaning     Mem[Sym1<I>] := Mem[Sym3<I>]
   Sym1<V>   Base  Sym2   Sym3<V>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Mem[Sym3<I>]
                   Sym2 is the array type
------------------------------------------------------------------------------

Notes

1)   All references to memory Mem[Sym1<I>] may need to add the base
     pointer of the activation record if the symbol is a procedure variable.

2)   Addr provides the address of variables and therefore can be used to
     perform the pseudo system call function ADR( variable ), provided
     one obeys rule (1).
*)


(*
   Problems/improvements:

   Access registers from SYSTEM.regname
   SYSTEM pseudo procedures not implemented ie
   SIZE, ADR, TSIZE, VAL, CARDINAL, INTEGER etc
*)

(*
   Register Usage:

   RegRA points to the base of the local Variables on the stack.
   RegRB is used to dereference pointers, arrays and indirect procedures.

   Parameter passing:

   The Call Frame is built,
   All parameters are pushed onto the stack,
   And the procedure is called.

   Build Frame
   Put P0 into Param0 ; return function result
   Put P1 into Param1 ;
   Put P2 into Param2 ;
   Put P3 into Param3 ;
   Put Pn into Param n ;
   Save Volitiles
   Call Procedure
   Restore Volitiles
   Destroy Frame ; Frame -SIZE(P0)  if Procedure is a function
   ; result is now on the stack
*)


(*
   GenCode - Codes the list of quadruples described by Head.
             Codes quadruples Start .. End in APU assembly language.
*)

PROCEDURE GenCode (Head, Start, End: CARDINAL) ;
VAR
   LastQuad: CARDINAL ;
BEGIN
   SetLocalFlags(Start) ;
   AbsoluteHead := Head ;
   CurrentQuad := Start ;
   LastOperator := AddOp ;   (* we must disable the line no opt *)
   REPEAT
      LastQuad := CurrentQuad ;
      CodeStatement(Start, End)
   UNTIL LastQuad=End ;
   Flush(Start, End)
END GenCode ;


(*
   Init ;
   InitializationCode ;
   ...
   End
*)

   
(*
   CodeMain - Codes all the modules quadruples which are not in a procedure.
*)

PROCEDURE CodeMain (Start, End: CARDINAL) ;
VAR
   Level   : CARDINAL ;
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   Level := 0 ;
   WHILE CurrentQuad#0 DO
      GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
      IF Operator=NewLocalVarOp
      THEN
         INC(Level)
      END ;
      IF Level=0
      THEN
         CodeStatement(Start, End)
      ELSE
         CurrentQuad := GetNextQuad(CurrentQuad)
      END ;
      IF Operator=KillLocalVarOp
      THEN
         DEC(Level)
      END
   END
END CodeMain ;


(*
   CodeProcedures - Codes quadruples which are in a procedure.
*)

PROCEDURE CodeProcedures (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Level   : CARDINAL ;
BEGIN
   Level := 0 ;
   WHILE CurrentQuad#0 DO
      GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
      IF Operator=NewLocalVarOp
      THEN
         INC(Level)
      END ;
      IF Level#0
      THEN
         CodeStatement(Start, End)
      ELSE
         CurrentQuad := GetNextQuad(CurrentQuad)
      END ;
      IF Operator=KillLocalVarOp
      THEN
         DEC(Level)
      END
   END
END CodeProcedures ;


(*
   SetLocalFlags - sets the local compiler flags relative to the
                   quadruple QuadNo.
*)

PROCEDURE SetLocalFlags (QuadNo: CARDINAL) ;
BEGIN
   LocalCoding     := IsCodeOn(QuadNo) ;
   LocalProfiling  := IsProfileOn(QuadNo) ;
   LocalOptimizing := IsOptimizeOn(QuadNo)
END SetLocalFlags ;


(*
   CodeStatement - A Multi-Way decision call depending on the current
                   quadruple statement.
*)

PROCEDURE CodeStatement (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsReferenced(AbsoluteHead, CurrentQuad) AND (Operator#NewLocalVarOp)
   THEN
      GenerateQuadLabel(Start, End)
   END ;
   CASE Operator OF

   LogicalOrOp        : CodeLogicalOr |
   LogicalAndOp       : CodeLogicalAnd |
   LogicalXorOp       : CodeSymmetricDifference |
   AddOp              : CodeAdd |
   SubOp              : CodeSub |
   MultOp             : CodeMult |
   DivOp              : CodeDiv |
   ModOp              : CodeMod |
   NegateOp           : CodeNegate |
   BitOp              : CodeBit |
   ConvertOp          : CodeConvert |
   CoerceOp           : CodeCoerce |
   NewLocalVarOp      : CodeNewLocalVar(Start, End) |
   KillLocalVarOp     : CodeKillLocalVar(Start, End) |
   CallOp             : CodeCall(Start, End) |
   ParamOp            : CodeParam |
   AddrOp             : CodeAddr |
   BecomesOp          : CodeBecomes |
   XIndrOp            : CodeXIndr |
   IndrXOp            : CodeIndrX |
   GotoOp             : CodeGoto(Start, End) |
   EndOp              : CodeEnd |
   IfInOp             : CodeIfIn(Start, End) |
   IfNotInOp          : CodeIfNotIn(Start, End) |
   IfEquOp            : CodeIfEqu(Start, End) |
   IfNotEquOp         : CodeIfNotEqu(Start, End) |
   IfGreEquOp         : CodeIfGreEqu(Start, End) |
   IfLessEquOp        : CodeIfLessEqu(Start, End) |
   IfLessOp           : CodeIfLess(Start, End) |
   IfGreOp            : CodeIfGre(Start, End) |
   BaseOp             : CodeBase |
   ElementSizeOp      : CodeElementSize |
   OffsetOp           : CodeOffset |
   FunctValueOp       : CodeFunctValue |
   ReturnOp           : CodeReturn(Start, End) |
   ReturnValueOp      : CodeReturnValue |
   UnboundedOp        : CodeUnbounded(Start, End) |
   StartDefFileOp     : CodeStartDefFile |
   StartModFileOp     : CodeStartModFile |
   EndFileOp          : CodeEndFile |
   StartOp            : CodeStart |
   EndOp              : CodeEnd |
   DummyOp            : WriteString('something has gone very wrong at quad') ; WriteCard(CurrentQuad, 4) ; WriteLn ;
                        DisplayQuadList(1) ; exit(1) |

   (* And the Compiler Flag Pseudo Quadruples *)
   LineNumberOp       : CodeLineNumber(Start, End) |
   InlineOp           : CodeInline(Start, End) |
   CodeOnOp           : CodeCodeOn |
   CodeOffOp          : CodeCodeOff |
   ProfileOnOp        : CodeProfileOn |
   ProfileOffOp       : CodeProfileOff |
   OptimizeOnOp       : CodeOptimizeOn |
   OptimizeOffOp      : CodeOptimizeOff

   ELSE
      WriteString('Quad') ; WriteCard(CurrentQuad, 6) ; WriteLn ;
      WriteError('Not implemented yet')
   END ;
   LastOperator := Operator ;
   CurrentQuad := GetNextQuad(CurrentQuad)
END CodeStatement ;


(*
   CodeLineNumber - LineNumberOp is a quadruple which has the following format:

                    LineNumberOp   NulSym  NulSym  LineNo

                    The line number, LineNo, is ommited if we are producing
                    runtime debugging information.
*)

PROCEDURE CodeLineNumber (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF LastOperator#LineNumberOp
   THEN
      IF CurrentInst>0
      THEN
         Instruction[CurrentInst].LineNo := Operand3 ;
      END ;
      CurrentLineNo := Operand3
   END
END CodeLineNumber ;


(*
   CodeStartModFile - StartModFileOp is a quadruple which has the following
                      format:

                      StartModFileOp  _  _  ModuleSym

                      Its function is to reset the source file to another
                      file, hence all line numbers emmitted with the
                      generated code will be relative to this source file.
*)

PROCEDURE CodeStartModFile ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   CompilingMainModule := GetMainModule()=Operand3
END CodeStartModFile ;


(*
   CodeStartDefFile - StartDefFileOp is a quadruple which has the following
                      format:

                      StartDefFileOp  _  _  ModuleSym

                      Its function is to reset the source file to another
                      file, hence all line numbers emmitted with the
                      generated code will be relative to this source file.
*)

PROCEDURE CodeStartDefFile ;
BEGIN
   CompilingMainModule := FALSE
END CodeStartDefFile ;


(*
   CodeEndFile - FileOp is a quadruple which has the following format:

                 EndFileOp  _  _  ModuleSym

                 Its function is to reset the source file to another
                 file, hence all line numbers emmitted with the
                 generated code will be relative to this source file.
*)

PROCEDURE CodeEndFile ;
BEGIN
   (* CompilingMainModule := FALSE *)
END CodeEndFile ;


(*
   CallInnerInit - produce a call to inner module initialization routine.
*)

PROCEDURE CallInnerInit (Sym: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   WriteString('  load.d.d  $') ; WriteKey(GetModuleInitName(Sym)) ;
   WriteLn ;
   WriteString('  call') ; WriteLn
END CallInnerInit ;


(*
   CodeStart - emits starting code before the main BEGIN END of the
               current module.
*)

PROCEDURE CodeStart ;
VAR
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, op, op1, op2, op3) ;
   IF CompilingMainModule
   THEN
      WritePrologue(op3, CurrentQuad, op1) ;
      ForeachInnerModuleDo(op3, CallInnerInit)
   END
END CodeStart ;


(*
   CodeEnd - emits terminating code after the main BEGIN END of the
             current module.
*)

PROCEDURE CodeEnd ;
VAR
   op           : QuadOperator ;
   op1, op2, op3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, op, op1, op2, op3) ;
   FlushTime ;
   WriteEpilogue(op3)
END CodeEnd ;


(*
   CodeInline - InlineOp is a quadruple which has the following format:

                InlineOp   NulSym  NulSym  Buf

                The Inline Buffer, Buf, is written to standard output.
*)

PROCEDURE CodeInline (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   (*
      Must Flush the out standing instructions in the code buffer
      otherwise the Inline code will be emmitted in the incorrect
      order.
   *)
   Flush(Start, End) ;
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   WriteInlineBuffer(Operand3)
END CodeInline ;


(*
   CodeCodeOn - CodeOnOp is a quadruple which switches the compiler
                Code flag to On.
*)

PROCEDURE CodeCodeOn ;
BEGIN
   LocalCoding := TRUE
END CodeCodeOn ;


(*
   CodeCodeOff - CodeOffOp is a quadruple which switches the compiler
                 Code flag to Off.
*)

PROCEDURE CodeCodeOff ;
BEGIN
   LocalCoding := FALSE
END CodeCodeOff ;


(*
   CodeProfileOn - ProfileOnOp is a quadruple which switches the compiler
                   Profile flag to On.
*)

PROCEDURE CodeProfileOn ;
BEGIN
   LocalProfiling := TRUE
END CodeProfileOn ;


(*
   CodeProfileOff - CodeProfileOp is a quadruple which switches the compiler
                    Profile flag to Off.
*)

PROCEDURE CodeProfileOff ;
BEGIN
   LocalProfiling := FALSE
END CodeProfileOff ;


(*
   CodeOptimizeOn - OptimizeOnOp is a quadruple which switches the compiler
                    Optimize flag to On.
*)

PROCEDURE CodeOptimizeOn ;
BEGIN
   LocalOptimizing := TRUE
END CodeOptimizeOn ;


(*
   CodeOptimizeOff - CodeOptimizeOp is a quadruple which switches the compiler
                     Optimize flag to Off.
*)

PROCEDURE CodeOptimizeOff ;
BEGIN
   LocalOptimizing := FALSE
END CodeOptimizeOff ;


(*
   CodeBit - Briefly:
             Operand1 :=  (1 << Operand3)
             Checks for Constant folding - otherwise apu code is generated.
*)

PROCEDURE CodeBit ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   i       : INTEGER ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      Bit ;
      PopValue(Operand1)
   ELSIF IsConst(Operand3)
   THEN
      IF GetSize(GetType(Operand1)) > WordSize
      THEN
         WriteError('Cannot find the bit shift of a variable larger than register size (4 bytes)')
      END ;
      PushValue(Operand3) ;
      Bit ;
      i := PopInt() ;
      PushImmSize(i, DWORDSIZE) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   ELSE
      PushImmSize(1, DWORDSIZE) ;
      PushAPUTo(Operand3, DWORDSIZE) ;
      EmitZero(sal) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeBit ;


(*
   CodeLogicalOr - Briefly:
                            Operand1 := Operand2 LOR Operand3
                   Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeLogicalOr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetOr ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(or) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeLogicalOr ;


(*
   CodeLogicalAnd - Briefly:
                             Operand1 := Operand2 LAND Operand3
                    Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeLogicalAnd ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetAnd ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(and) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeLogicalAnd ;


(*
   CodeSymmetricDifference - Briefly:
                                      Operand1 := Operand2 / Operand3
                             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeSymmetricDifference ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      SetSymmetricDifference ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(xor) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeSymmetricDifference ;


(*
   CodeAdd - Briefly:
                      Operand1 := Operand2 + Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeAdd ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Add ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(add) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeAdd ;


(*
   CodeSub - Briefly:
                      Operand1 := Operand2 - Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeSub ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand2) ;
      PushValue(Operand3) ;
      Sub ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand2, DWORDSIZE) ;
      PushAPUTo(Operand3, DWORDSIZE) ;
      EmitZero(sub) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeSub ;


(*
   CodeNegate - Briefly:
                         Operand1 := - Operand3
                Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeNegate ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushCard(0) ;
      PushValue(Operand3) ;
      Sub ;
      PopValue(Operand1)
   ELSE
      PushImmSize(0, DWORDSIZE) ;
      PushAPUTo(Operand3, DWORDSIZE) ;
      EmitZero(sub) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeNegate ;


(*
   CodeMult - Briefly:
                       Operand1 := Operand2 * Operand3
              Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeMult ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Mult ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(mult) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeMult ;


(*
   CodeDiv - Briefly:
                      Operand1 := Operand2 / Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeDiv ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Div ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(div) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeDiv ;


(*
   CodeMod - Briefly:
                      Operand1 := Operand2 MOD Operand3
             Checks for Constant folding - otherwise APU code is generated.
*)

PROCEDURE CodeMod ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushValue(Operand3) ;
      PushValue(Operand2) ;
      Mod ;
      PopValue(Operand1)
   ELSE
      PushAPUTo(Operand3, DWORDSIZE) ;
      PushAPUTo(Operand2, DWORDSIZE) ;
      EmitZero(mod) ;
      PopAPUFrom(Operand1, DWORDSIZE)
   END
END CodeMod ;


(*
   CodeConvert - Converts Operand3 to type Operand2 placing the result into
                 Operand1.
                 Convert will, if need be, alter the machine representation
                 of Operand3 to comply with TYPE Operand2.
                 However with the simple APU all base types are the same size,
                 this simple code generator does not implement changing
                 more complex types.
                 This must be added later.
*)

PROCEDURE CodeConvert ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   FromSize: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   FromSize := PushAPU(Operand3) ;
   IF (FromSize#BYTESIZE) AND (FromSize#WORDSIZE) AND (FromSize#DWORDSIZE)
   THEN
      WriteError('implementation restriction - can only convert between 1, 2 and 4 byte entities')
   END ;
   PopAPUFrom(Operand1, FromSize)
END CodeConvert ;


(*
   CodeCoerce - Coerce Operand3 to type Operand2 placing the result into
                Operand1.
                Coerce will NOT alter the machine representation
                of Operand3 to comply with TYPE Operand2.
                Therefore it INSISTS that under all circumstances that the
                type sizes of Operand1 and Operand3 are the same.
                CONVERT will perform machine manipulation to change variable
                types, coerce does no such thing.
*)

PROCEDURE CodeCoerce ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand3)
   THEN
      PushSize(GetType(Operand1)) ;
      Size := PopCard() ;
      PushValue(Operand3) ;
      Emit(load, immediate, no, dword, CardToSize(Size), CreateNumber(PopInt()), NulSym) ;
      IF Size#PopOperandMode(Operand1)
      THEN
         InternalError('sizes should be the same', __FILE__, __LINE__)
      END
   ELSE
      IF PushOperandMode(Operand3)#PopOperandMode(Operand1)
      THEN
         WriteError('TYPE Coersion can only be achieved with types of the same size')
      END
   END
END CodeCoerce ;


(*
   Notes on procedure call using Unbounded arrays.

   (i)   The Unbounded data is pushed onto the stack. Unbounded Op1
         is then initialised to contain the address of this data.
         Op1 is effectively (Unbounded.ArrayAddress).
   (ii)  The old sp ie sp' is copied into RegC.
   (iii) RegC and all other registers are preserved when calling the
         required procedure.
   (iv)  After the procedure returns the Unbounded data must be destroyed
         ie SP := RegC
   (v)   This should be performed in different places depending whether
         the procedure has a return argument (function).
         (a)  Functions perform this after FunctValue quadruple.
         (b)  Procedures perform this after the Call quadruple.

   register usage:

   ra = frame pointer
   rb = contains previous static scope if nested procedures are used
        (only used when entering a procedure) after entry it is general
        purpose
   rc = used for dynamic arrays on the stack see above
   rd = return value of a function and general purpose during procedures.
*)

(*
   The Call Mechanism:


   Stack of the APU - grows downwards!


   Before Call
   ===========
   .             .
   .             .
   .             .
   |-------------|
   | Return Arg  | = (Param 0)        (or in RegRD if NOT ReturnArgumentOnStack)
   |-------------|
   | Parameter n |
   |-------------|
   | Parameter 1 |
   |-------------|
   | pc''        |
   |-------------|
   | ra''        |   <- ra'
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | rc''        |
   |-------------|
   .             .
   .             .
   .             .


   Stacking Parameters                    res := proc(a, b, c)
   ===================

   .             .
   .             .
   .             .
   |             |
   |-------------|
   | Return Arg  |
   |-------------|
   | Parameter n |
   |-------------|
   | Parameter 1 |
   |-------------|
   | pc''        |
   |-------------|
   | ra''        |   <- ra'
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | rc''        |
   |-------------|
   | Unbounded   |   <- sp'     <- rc
   :             :
   | Array space |
   |-------------|
   | res         |
   |-------------|
   | c           |
   |-------------|
   | b           |
   |-------------|
   | a           |
   |-------------|
   |             |   <- sp
   .             .
   .             .
   .             .


   NewLocalVar  (adds the following)
   ===========


   .             .
   .             .
   .             .
   |-------------|
   | pc'         |
   |-------------|
   | ra'         |   <- ra
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | rc'         |
   |-------------|
   .             .
   .             .
   .             .



   KillLocalVar
   ============

   .             .
   .             .
   .             .
   |             |
   |-------------|
   | Return Arg  |
   |-------------|
   | Parameter n |
   |-------------|
   | Parameter 1 |
   |-------------|
   | pc''        |
   |-------------|
   | ra''        |   <- ra'
   |-------------|
   |             |
   | Local       |
   |             |
   | Variables   |
   |             |
   |-------------|
   | rc''        |
   |-------------|
   | Unbounded   |   <- sp'     <- rc
   :             :
   | Array space |
   |-------------|
   | res         |
   |-------------|
                     <- sp
*)



(*
   GenerateEntryProcedureLabel - generates a procedure label for procedure, Sym.
                                 The procedure is defined at CurrentQuad.
*)

PROCEDURE GenerateEntryProcedureLabel (Sym: CARDINAL) ;
BEGIN
   IF IsExported(GetMainModule(), Sym)
   THEN
      WriteString('   .global  ') ;    (* Exported procedures defined as global *)
      WriteAsmName(Sym) ;
      WriteLn ;
   END ;
   WriteAsmName(Sym) ; Write(':') ; WriteLn
END GenerateEntryProcedureLabel ;


(*
   CodeNewLocalVar - Builds a new frame on the stack to contain the procedure
                     local variables.
*)

PROCEDURE CodeNewLocalVar (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Name    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, PreviousScope, CurrentProcedure) ;
   Flush(Start, End) ;
   ProcedureInfo(CurrentProcedure, Operand1) ;   (* Generate GDB information *)
   GenerateEntryProcedureLabel(CurrentProcedure) ;
   LineInfo(Operand1) ;         (* Generate GDB information *)

   PushReg(ra) ;                (* save old frame pointer *)
   PushReg(sp) ;                (* make new frame pointer *)
   PopReg(ra) ;

   PushSumOfLocalVarSize(CurrentProcedure) ;

   IF IsProcedureNested(CurrentProcedure)
   THEN
      PushReg(rb)               (* last static scope (outer procedure) *)
   END ;

   DecreaseStack(PopCard()) ;   (* make room for locals   *)

   (* Push Register *)
   PushReg(rc) ;

   (* Generate debugging trace infomation *)
   Flush(Start, End) ;
   BeginInfo

END CodeNewLocalVar ;


(*
   CodeKillLocalVar - Resets RA to the old value before the CodeNewLocalVar
                      and kills off local variables.
*)

PROCEDURE CodeKillLocalVar (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;

   (* Generate debugging trace infomation *)

   Flush(Start, End) ;
   EndInfo ;

   (* Pop Registers *)
   PopReg(rc) ;                 (* restore registers         *)

   PushSumOfLocalVarSize(Operand3) ;
   IncreaseStack(PopCard()) ;   (* remove local variables    *)

   (* Kill activation record - (call frame) *)
   IF IsProcedureNested(CurrentProcedure)
   THEN
      PopReg(rb)
   END ;

   PopReg(ra) ;                 (* restore old frame pointer *)
   CurrentProcedure := NulSym ;
   PreviousScope    := NulSym
END CodeKillLocalVar ;


(*
   DecreaseStack - decreases the runtime APU stack by c.
*)

PROCEDURE DecreaseStack (c: CARDINAL) ;
BEGIN
   IF c>0
   THEN
      PushReg(sp) ;
      PushImmSize(INTEGER(c), DWORDSIZE) ;
      EmitZero(sub) ;
      PopReg(sp)
   END
END DecreaseStack ;


(*
   IncreaseStack - increases the runtime APU stack by c.
*)

PROCEDURE IncreaseStack (c: CARDINAL) ;
BEGIN
   IF c>0
   THEN
      PushReg(sp) ;
      PushImmSize(INTEGER(c), DWORDSIZE) ;
      EmitZero(add) ;
      PopReg(sp)
   END
END IncreaseStack ;


(*
   Consider when time allows ...
   Perhaps implement functions, procedures and parameterless
   variables less procedures differently..

   Re code this later
*)


(*
   CodeCall - determines whether the procedure call is a direct call
              or an indirect procedure call.
*)

PROCEDURE CodeCall (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Operator : CallOp
      Operand3 : Procedure
   *)
   IF IsProcedure(Operand3)
   THEN
      CodeDirectCall(Start, End)
   ELSIF IsProcType(GetType(Operand3))
   THEN
      CodeIndirectCall(Start, End)
   ELSE
      InternalError('expecting Procedure or ProcType', __FILE__, __LINE__)
   END
END CodeCall ;


(*
   CheckNestedProcedure - checks whether we are about to call a nested procedure.
                          If the procedure is nested we must set up the activation
                          record pointer:

                            o    if we are the statically defined parent scope then we
                                 pass our frame pointer in rb
                            o    otherwise we find out the previous statically defined parent
                                 and pass this in rb.
*)

PROCEDURE CheckNestedProcedure (Sym: CARDINAL) ;
VAR
   BackScope: CARDINAL ;
   s        : CARDINAL ;
BEGIN
   IF IsProcedureNested(Sym)
   THEN
      (*
         following Advanced Compiler Design and Implementation by Steven S.Muchnick p114

         (i)   if the procedure being called is nested directly within its caller, its
               static link points to its callers frame.
         (ii)  if the procedure is at the same level of nesting as its caller, then its
               static link is the same as the callers static link
         (iii) if the procedure being called is n levels higher than the caller in the
               nesting structure, then its static link can be determined by following, n,
               static links back from the callers static link and copying the static link
               found there.
      *)
      IF GetScopeAuthor(Sym)=CurrentProcedure
      THEN
         (* found (i) *)
         PushReg(ra) ;
         PopReg(rb)
      ELSIF GetScopeAuthor(Sym)=PreviousScope
      THEN
         (* found (ii) *)
         s := PushAPU(ActivationPointer) ;
         PopReg(rb)
      ELSE
         (* must be (iii) *)
         s := PushAPU(ActivationPointer) ;
         PopReg(rb) ;
         BackScope := PreviousScope ;
         WHILE BackScope#GetScopeAuthor(Sym) DO
            PushOffset(ActivationPointer) ;
            PushUsingDword(ActivationPointer, PopInt(), WordSize, indexed, rb) ;
            PopReg(rb) ;
            BackScope := GetScopeAuthor(BackScope)
         END
      END
   END
END CheckNestedProcedure ;


(*
   CodeDirectCall - saves all volitiles and jumps to a subroutine.
*)

PROCEDURE CodeDirectCall (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Name    : CARDINAL ;
   NameStr : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   FlushTime ;

   (*
      At a Later Stage:

      If we are calling a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         push current interrupt mask
         set current interrupt mask to that of the called procedure MODULE
      end
   *)

   CheckNestedProcedure(Operand3) ;

   (*
      Generate a procedure label.
   *)
   PushImmLabel(GetAsmName(Operand3)) ;
   Flush(Start, End) ;
   WriteString('  call') ; WriteLn ;

   (*
      At a later stage we could add MODULE priority here:

      If we have called a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         pop current interrupt mask
      end
   *)

   (*
      Now restore sp to above param 0.
      The Stack Pointer will be just above the procedure Parameters
   *)

   PushSumOfParamSize(Operand3) ;
   IF GetType(Operand3)=NulSym   (* Is it a function *)
   THEN
      (* Procedure *)
      IncreaseStack(PopCard()) ;
      RemoveAnyUnboundedData
   ELSE
      (* Function *)
      IF ReturnArgumentOnStack
      THEN
         PushSize(GetType(Operand3)) ;
         Sub
         (* Stack Should be above Param 0 *)
      END ;
      IncreaseStack(PopCard())
   END

END CodeDirectCall ;


(*
   CodeIndirectCall - saves all volitiles and jumps to a subroutine.
*)

PROCEDURE CodeIndirectCall (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Proc    : CARDINAL ;
   s       : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   FlushTime ;

   (*
      At a Later Stage:

      If we are calling a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         push current interrupt mask
         set current interrupt mask to that of the called procedure MODULE
      end
   *)

   s := PushOperandMode(Operand3) ;           (* Address of procedure     *)
   Flush(Start, End) ;
   WriteString('  call') ; WriteLn ;

   (*
      At a later stage we could add MODULE priority here:

      If we have called a procedure which is in
      a different interrupt priority MODULE to the current MODULE
      then
         pop current interrupt mask
      end
   *)

   (*
      Now restore sp to above param 0.
      The Stack Pointer will be just above the procedure Parameters
   *)

   Proc := GetType(Operand3) ;
   Assert(IsProcType(Proc)) ;
   PushSumOfParamSize(Proc) ;
   IF GetType(Proc)=NulSym   (* Is it a function *)
   THEN
      (* Procedure *)
      IncreaseStack(PopCard()) ;
      RemoveAnyUnboundedData
   ELSE
      (* Function *)
      IF ReturnArgumentOnStack
      THEN
         PushSize(GetType(Operand3)) ;
         Sub
         (* Stack Should be above Param 0 *)
      END ;
      IncreaseStack(PopCard())
   END

END CodeIndirectCall ;


(*
   CodeParam - pushes the parameter onto the stack, if the parameter,
               number is greater than 0, otherwise the parameter
               is effectively space for a return value.
*)

PROCEDURE CodeParam ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF Operand1=0
   THEN
      IF ReturnArgumentOnStack
      THEN
         (*
            Parameter 0 is the return value for a function. (Operand3)
            - must leave space for the value on the stack.
         *)
         PushSize(GetType(Operand3)) ;
         DecreaseStack(PopCard())
      END
   ELSE
      IF IsVarParam(Operand2, Operand1)
      THEN
         IF IsConst(Operand3)
         THEN
            WriteError('cannot pass a constant as a VAR parameter')
         ELSIF IsProcedure(Operand3)
         THEN
            WriteError('cannot pass a procedure as a VAR parameter')
         ELSE
            Size := PushAPU(Operand3)
         END
      ELSE
         IF IsProcedure(Operand3) OR IsConst(Operand3)
         THEN
            IF Alignment
            THEN
               PushAPUTo(Operand3, DWORDSIZE)
            ELSE
               PushAPUTo(Operand3, GetSize(GetType(GetNthParam(Operand2, Operand1))))
            END
         ELSE
            (*
               basically, if alignment is set then we must ensure operands with size < DWORDSIZE
               are pushed as a DWORD, all other sizes will be rounded up inside M2EvalSym.
            *)
            IF Alignment AND (GetMode(Operand3)=RightValue)
            THEN
               PushSize(Operand3) ;
               IF PopCard()<DWORDSIZE
               THEN
                  PushAPUTo(Operand3, DWORDSIZE)
               ELSE
                  Size := PushAPU(Operand3)
               END
            ELSE
               Size := PushAPU(Operand3)
            END
         END
      END
   END
END CodeParam ;


(*
   CodeFunctValue - retrieves the function return value from the stack.
                    The result of the function should be on top of the
                    APU stack.
*)

PROCEDURE CodeFunctValue ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Operator : FunctValueOp
      Operand1 : The Returned Variable
      Operand2 : The Function Returning this Variable
   *)

   (*
      move return value into Symbol Operand1 ignoring its mode (always assumed to be offset, see M2Quads.mod).
      note that the variable Operand1 might actually have a different mode (LeftValue) but we will
      treat it as RightValue since this is what M2Quads.mod expects. M2SubExp might reorder temporary
      usage, but it will still expect NO indirection to occur.
   *)

   IF ReturnArgumentOnStack
   THEN
      Size := PopOperandMode(Operand1)
   ELSE
      PushVarSize(Operand1) ;
      Size := PopCard() ;
      CASE Size OF

      BYTESIZE :  Emit(load, internal, rd, dword, byte, Zero, NulSym) |
      WORDSIZE :  Emit(load, internal, rd, dword, word, Zero, NulSym) |
      DWORDSIZE:  Emit(load, internal, rd, dword, dword, Zero, NulSym)

      ELSE
         WriteError('implementation restriction - can only implement return values of 1, 2, 4 byte sizes')
      END ;
      Assert(PopAPU(Operand1)=Size)   (* FunctValue ignores the mode of the variable (RightValue=LeftValue) *)
   END ;
   RemoveAnyUnboundedData
END CodeFunctValue ;


(*
   CodeReturn - performs a return from subroutine in APU assembly
                language.
*)

PROCEDURE CodeReturn (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   Ret(Start, End)                       (* pc := pc' *)
END CodeReturn ;


(*
   CodeReturnValue - places the operand into the return value space
                     allocated by the function call.
*)

PROCEDURE CodeReturnValue ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size, i : CARDINAL ;
   Num     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Operator : ReturnValueOp
      Operand1 : The Variable Begin Returned from the procedure
      Operand3 : The procedure
   *)
   Size := PushOperandMode(Operand1) ;
   (* Now check that the return value is the same size as the variable *)
   IF GetType(Operand3)=NulSym
   THEN
      WriteError('procedure does not have a return value (not a function)')
   END ;
   PushSize(GetType(Operand3)) ;
   IF PopCard()#Size
   THEN
      WriteError('Attempting to return a value of a different size to the function type')
   END ;
   IF ReturnArgumentOnStack
   THEN
      PushSize(GetType(Operand3)) ;
      PushSumOfParamSize(Operand3) ;
      Add ;
      PushCard(AddressSize) ;   (* ra' *)
      Add ;
      PushCard(AddressSize) ;   (* pc' *)
      Add ;
      PopUsing(NulSym, PopInt(), Size, indexed, ra)
   ELSE
      IF Size>AddressSize
      THEN
         WriteError('cannot return a value whose type size does not fit a register')
      END ;
      (* essentially perform a PopReg(rd) but taking the pushed size into account *)
      Emit(store, internal, rd, CardToSize(Size), dword, Zero, NulSym)    (* return value in rd *)
   END
END CodeReturnValue ;


(*
   RepsTime - returns the number of cycles for a single reps instruction.
*)

PROCEDURE RepsTime () : CARDINAL ;
BEGIN
   RETURN( BinInstUnit(i486.repmovs, i486.PDWord, 1) )
END RepsTime ;


(*
   CodeUnbounded - codes the creation of an unbounded parameter
                   variable.
                   Pushes Op3 onto the stack and sets Op1 to the start
                   of this data.
*)

PROCEDURE CodeUnbounded (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   PrepareToPushUnboundedData ;
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConstString(Operand3)
   THEN
      PushStringToStack(Operand3, 0)
   ELSIF IsArray(GetType(Operand3))
   THEN
      Size := PushOperandMode(Operand3)
   ELSIF IsUnbounded(GetType(Operand3))
   THEN
      PushUnbounded(Start, End, Operand3)
   ELSE
      WriteError('String, Array or Unbounded expected for Unbounded Parameter')
   END ;
   (* Set Op1 to the start of the copy of Op3 *)
   PushReg(sp) ;
   Size := PopOperandMode(Operand1) ;
   PushSize(Operand1) ;
   Assert(PopCard()=AddressSize)
END CodeUnbounded ;


(*
   PushUnbounded - pushes an Unbounded Array Sym data onto the stack.
*)

PROCEDURE PushUnbounded (Start, End: CARDINAL; Sym: CARDINAL) ;
VAR
   Size: CARDINAL ;
BEGIN
   Size := PushOperandMode(Sym) ;
   IF APUUseMemCopy
   THEN
      (* use memcopy to effect push *)
      PushUnboundedWithMemCopy
   ELSE
      PushUnboundedWithLoad(Start, End)
   END
END PushUnbounded ;


(*
   PushUnboundedWithMemCopy - pushes an unbounded array with the memcopy instruction
*)

PROCEDURE PushUnboundedWithMemCopy ;
BEGIN
   PopReg(rb) ;                          (* rb := ArrayAddress *)
   PushImmSize(1+1, DWORDSIZE) ;
   EmitZero(add) ;                       (* 1+1 because HIGH+1 = size & 1 for nul *)
   PopReg(rd) ;                          (* rd := High         *)

   PushReg(sp) ;
   PushReg(rd) ;
   EmitZero(sub) ;
   PopReg(sp) ;
   PushReg(rb) ;                         (* source             *)
   PushReg(sp) ;
   PushImmSize(DWORDSIZE, DWORDSIZE) ;   (* need to add DWORDSIZE as memcopy *)
                                         (* operands will disappear          *)
   EmitZero(add) ;                       (* destination        *)
   PushReg(rd) ;                         (* number of bytes    *)
   IF LocalProfiling
   THEN
      PushReg(rd) ;
      PushImmSize(RepsTime(), DWORDSIZE) ;
      EmitZero(mult) ;
      EmitZero(pr)
   END ;
   EmitZero(memcopy)
END PushUnboundedWithMemCopy ;


(*
   PushUnboundedWithLoad - pushes an unbounded array with multiple loads and tests.
*)

PROCEDURE PushUnboundedWithLoad (Start, End: CARDINAL) ;
BEGIN
   PopReg(rb) ;                                 (* rb := ArrayAddress *)
   PushImmSize(1+1, DWORDSIZE) ;
   EmitZero(add) ;                              (* 1+1 because HIGH+1 = size & 1 for nul *)
   PopReg(rd) ;                                 (* rd := High         *)
   IF LocalProfiling
   THEN
      PushReg(rd) ;
      PushImmSize(RepsTime(), DWORDSIZE) ;
      EmitZero(mult) ;
      EmitZero(pr)
   END ;
   EmitLabel(Start, End, CreateLabelName('Ll', CurrentQuad)) ;
   PushReg(rd) ;                                (* while rd#0 do      *)
   PushReg(rb) ;
   EmitZero(add) ;
   Emit(load, indexed, st, byte, byte, Zero, NulSym) ;  (*    Push Array[i]   *)
   PushImmSize(0, DWORDSIZE) ;
   PushReg(rd) ;
   EmitZero(equ) ;
   Brt(Start, End, CreateLabelName('Le', CurrentQuad)) ;
   DecReg(rd) ;                                 (*    dec(rd)         *)
   Jump(Start, End, CreateLabelName('Ll', CurrentQuad)) ;   (* end                *)
   EmitLabel(Start, End, CreateLabelName('Le', CurrentQuad))
END PushUnboundedWithLoad ;


(*
   IncReg - increments a register.
*)

PROCEDURE IncReg (r: Registers) ;
BEGIN
   PushReg(r) ;
   PushImmSize(1, DWORDSIZE) ;
   EmitZero(add) ;
   PopReg(r)
END IncReg ;


(*
   DecReg - increments a register.
*)

PROCEDURE DecReg (r: Registers) ;
BEGIN
   PushReg(r) ;
   PushImmSize(1, DWORDSIZE) ;
   EmitZero(sub) ;
   PopReg(r)
END DecReg ;

   
(*
   RemoveAnyUnboundedData - attempts to remove any undounded data.
                            Unbounded data will be on the stack if the
                            boolean, UnboundedOnStack, is true.
                            In which case the SP := RC.
*)

PROCEDURE RemoveAnyUnboundedData ;
BEGIN
   IF UnboundedOnStack
   THEN
      PushReg(rc) ;
      PopReg(sp) ;
      UnboundedOnStack := FALSE
   END
END RemoveAnyUnboundedData ;


(*
   PrepareToPushUnboundedData - prepares to push some unbounded data.
                                If this has not been called before
                                then Register RC := SP.
*)

PROCEDURE PrepareToPushUnboundedData ;
BEGIN
   IF NOT UnboundedOnStack
   THEN
      (*
         Not done yet -
         so do this now and never again before the procedure call.
      *)
      PushReg(sp) ;
      PopReg(rc) ;
      UnboundedOnStack := TRUE
   END
END PrepareToPushUnboundedData ;


(*
   CodeHigh - codes the procedure HIGH for a static array.
              Op1 is a cardinal or variable which can contain
                  a numerial ie INTEGER, etc.
              Op2 is NulSym
              Op3 contains the Array Symbol which should have its
                  highest bounds calculated.
*)

PROCEDURE CodeHigh ;
BEGIN
   InternalError('compile time assignment - add to M2Optimize', __FILE__, __LINE__)
END CodeHigh ;


PROCEDURE CodeAddr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   IF IsConst(Operand3)
   THEN
      WriteError('Not expecting a constant')
   ELSE
      IF GetMode(Operand3)=RightValue
      THEN
         (*
            Mem[Op1] := Op3
         *)
         PushADR(Operand3) ;
         Assert(PopAPU(Operand1)=DWORDSIZE)
      ELSE
         (*
            Mem[Op1] := Mem[Op3]
         *)
         Assert(GetMode(Operand3)=LeftValue) ;
         ADR(Operand3) ;   (* Was PushAPU(Operand3) ; *)
         Assert(PopAPU(Operand1)=DWORDSIZE) ;
         (*
            Consistancy check to see whether the SIZE(Operand1)=SIZE(ADDRESS)
            If this should ever fail then we know the compiler would produce
            incorrect code.
            The reason I changed PushAPU -> ADR was that PushAPU fails when
            a record structure is passed as a VAR parameter and is altered.
            ADR will only push the address of the record thus implementing
            the VAR parameter correctly, however, we must make sure that
            Operand1 is the same size as ADDRESS otherwise on a stack machine
            all hell breaks out!
         *)
         PushSize(Operand1) ;
         Assert(PopCard()=AddressSize)
      END
   END
END CodeAddr ;


PROCEDURE CodeBecomes ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size3   : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand3)
   THEN
      PushValue(Operand3) ;
      PopValue(Operand1)
   ELSIF IsConstString(Operand3)
   THEN
      PushSize(GetType(Operand1)) ;  (* Make sure the string is the same size as Operand1 *)
      PushStringToStack(Operand3, PopCard()) ;
      Size1 := PopAPU(Operand1)
   ELSIF IsConst(Operand3)
   THEN
      Size3 := PushAPU(Operand3) ;
      PopAPUFrom(Operand1, Size3)
   ELSE
      Size3 := PushAPU(Operand3) ;
      Size1 := PopAPU(Operand1) ;
      IF Size1#Size3
      THEN
         WriteError('Different sizes in :=')
      END
   END
END CodeBecomes ;


(*
   PushStringToStack - pushes a string, StringSym, onto the runtime stack.
                       Max characters are pushed onto the stack, if the string
                       is too small then it is padded out with <nul>s.
                       If it is too large then an error is reported.
*)

PROCEDURE PushStringToStack (StringSym: CARDINAL; Max: CARDINAL) ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
   b: ARRAY [0..3] OF CHAR ;
   i     : INTEGER ;
   Pack,
   String,
   Length: CARDINAL ;
BEGIN
   String := GetString(StringSym) ;
   Length := GetStringLength(StringSym) ;
   IF Max=0
   THEN
      Max := Length+1       (* no fixed limit - therefore place nul on top *)
   END ;
   GetKey(String, a) ;

   Assert(MaxString>=Max) ;

   (* now place nuls from a[Length..Max] *)
   i := Length ;
   WHILE i<Max DO
      a[i] := nul ;
      INC(i)
   END ;

   (* Fill in 0..Length-1 *)
   i := Max ;
   CASE (i MOD DWORDSIZE) OF

   1:  DEC(i) ;
       Pack := PackString(a, Max, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=1) |
   2:  DEC(i, 2) ;
       Pack := PackString(a, Max, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=2) |
   3:  DEC(i) ;
       Pack := PackString(a, Max, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=1) ;
       DEC(i, 2) ;
       Pack := PackString(a, Max, CARDINAL(i), b) ;
       PushString(Pack, b) ;
       Assert(Pack=2)

   ELSE
   END ;
   IF Max>=DWORDSIZE
   THEN
      DEC(i, DWORDSIZE) ;
      WHILE i>=0 DO
         Pack := PackString(a, Max, CARDINAL(i), b) ;
         Assert(Pack=DWORDSIZE) ;
         PushString(Pack, b) ;
         DEC(i, DWORDSIZE)
      END
   ELSE
      Assert(i=0)
   END
END PushStringToStack ;


(*
   PackString - packs a string 'a' into another string 'b',
                starting at position i in string 'a'.
                The number of characters packed into b are returned.
                String b is assummed to contain 4 elements.
*)

PROCEDURE PackString (a: ARRAY OF CHAR; DestSize, i: CARDINAL;
                      VAR b: ARRAY OF CHAR) : CARDINAL ;
VAR
   SourceSize,
   Pack, j   : CARDINAL ;
BEGIN
   SourceSize := StrLen(a) ;
   IF (DestSize>SourceSize) AND ((SourceSize-i=3) OR (SourceSize-i=1))
   THEN
      (* Padd out remaining boundary byte with the nul character *)
      Pack := SourceSize-i+1
   ELSE
      Pack := GetUnitOfTransfer(DestSize-i)
   END ;
   FOR j := 0 TO Pack-1 DO
      IF i+j>=SourceSize
      THEN
         b[j] := nul
      ELSE
         b[j] := a[i+j]
      END
   END ;
   IF Pack<HIGH(b)
   THEN
      b[Pack] := nul     (* May need to terminate the string, b *)
   END ;
   RETURN( Pack )
END PackString ;


(*
   GetUnitOfTransfer - returns the number of bytes that can be transferred
                       at a time, if Size bytes are to be transferred.
*)

PROCEDURE GetUnitOfTransfer (Size: CARDINAL) : CARDINAL ;
BEGIN
   IF Size=1
   THEN
      RETURN( BYTESIZE )
   ELSIF (Size=2) OR (Size=3)
   THEN
      RETURN( WORDSIZE )
   ELSIF Size>=DWORDSIZE
   THEN
      RETURN( DWORDSIZE )
   ELSE
      InternalError('Cannot transfer 0 bytes of memory!', __FILE__, __LINE__)
   END
END GetUnitOfTransfer ;


(*
   PushString - pushes a 1..4 byte string to the stack.
*)

PROCEDURE PushString (Pack: CARDINAL; a: ARRAY OF CHAR) ;
VAR
   i,
   Word: CARDINAL ;
   b   : ARRAY [0..MaxString] OF CHAR ;
BEGIN
   Word := 0 ;
   FOR i := Pack-1 TO 0 BY -1 DO
      Word := Word*0100H ;
      IF i<=HIGH(a)
      THEN
         INC(Word, ORD(a[i]))
      END
   END ;
   GenerateCommentString(a, Pack-1, b) ;
   Comment(b) ;
   PushImmSize(INTEGER(Word), Pack)
END PushString ;


(*
   GenerateCommentString - generates a comment string, b, from a string, a,
                           which contains, n, characters.
*)

PROCEDURE GenerateCommentString (a: ARRAY OF CHAR; n: CARDINAL;
                                 VAR b: ARRAY OF CHAR) ;
VAR
   i: CARDINAL ;
   c: ARRAY [0..1] OF CHAR ;
BEGIN
   StrCopy('"', b) ;
   FOR i := 0 TO n DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, '"   ( ', b) ;
   FOR i := n TO 0 BY -1 DO
      IF a[i]=nul
      THEN
         StrConCat(b, '<nul>', b)
      ELSE
         c[0] := a[i] ;
         c[1] := nul ;
         StrConCat(b, c, b)
      END
   END ;
   StrConCat(b, ' )', b)
END GenerateCommentString ;


(*
   PushProcName - Sym is a procedure, PushProcName pushes
                  the address of the procedure onto the runtime
                  stack.
*)

PROCEDURE PushProcName (Sym: CARDINAL) : CARDINAL ;
VAR
   Str : ARRAY [0..30] OF CHAR ;
BEGIN
   PushImmLabel(CreateLabelReference(Sym, 0)) ;
   PushSize(Address) ;
   RETURN( PopCard() )
END PushProcName ;


(*
   PushOperandModeDword - pushes an Operand according to its mode it ensures that the
                          operand has the size dword on the APU stack.
                          This function is used to place operands on the APU stack
                          for future comparison.
*)

PROCEDURE PushOperandModeDword (Operand: CARDINAL) : CARDINAL ;
VAR
   Offset: INTEGER ;
   Type,
   Father,
   Size,
   i     : CARDINAL ;
   Mode  : ModeOfAddr ;
   Str   : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsProcedure(Operand)
   THEN
      Size := PushProcName(Operand)
(*
   can we catch this case in the IsConst condition below?
   ELSIF IsConstString(Operand)
   THEN
      Size := GetStringLength(Operand) ;
      PushStringToStackDword(Operand, Size) ;
      Size := DWORDSIZE
*)
   ELSIF IsConst(Operand)
   THEN
      PushValue(Operand) ;
      PushImmSize(PopInt(), DWORDSIZE) ;
      Size := DWORDSIZE
   ELSE
      Mode := GetMode(Operand) ;
      IF Mode=RightValue
      THEN
         (*
             Mem[Operand] :=
         *)
         Type := GetType(Operand) ;
         PushSize(Type) ;
         Size := PopCard() ;

         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            PushOffset(Operand) ;
            Offset := PopInt() ;
            PushUsingDword(Operand, Offset, Size, indexed, ra)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            PushGlobalDword(Operand, Size)
         END ;
         Size := DWORDSIZE
      ELSE
         Assert(Mode=LeftValue) ;
         (*
            Mem[Mem[Operand]] :=
         *)
         Type := GetType(Operand) ;
         PushSize(Type) ;
         Size := PopCard() ;

         PushOffset(Operand) ;
         Offset := PopInt() ;
         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            Emit(load, indexed, ra, dword, dword, CreateNumber(Offset), Operand)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            Emit(load, direct, no, dword, dword, CreateLabelReference(Operand, 0), Operand)
         END ;
         PopReg(rb) ;
         PushUsingDword(NulSym, 0, Size, indexed, rb) ;
         Size := DWORDSIZE
      END
   END ;
   RETURN( Size )
END PushOperandModeDword ;


(*
   PushOperandMode - pushes an Operand according to its mode and returns
                     the pushed Operand size.
*)

PROCEDURE PushOperandMode (Operand: CARDINAL) : CARDINAL ;
VAR
   Value,
   Offset: INTEGER ;
   Type,
   Father,
   Size,
   i     : CARDINAL ;
   Mode  : ModeOfAddr ;
   Str   : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsProcedure(Operand)
   THEN
      Size := PushProcName(Operand)
   ELSIF IsConstString(Operand)
   THEN
      Size := GetStringLength(Operand) ;
      PushStringToStack(Operand, Size)
   ELSIF IsConst(Operand)
   THEN
      PushValue(Operand) ;
      Value := PopInt() ;
      Type := GetType(Operand) ;
      IF Type#NulSym
      THEN
         PushSize(Type) ;
         Size := PopCard() ;
         PushImmSize(Value, Size)
      ELSE
         PushSize(Cardinal) ;
         Size := PopCard() ;
         PushImmSize(Value, Size)
      END
   ELSE
      Mode := GetMode(Operand) ;
      IF Mode=RightValue
      THEN
         (*
             Mem[Operand] :=
         *)
         Type := GetType(Operand) ;
         PushSize(Type) ;
         Size := PopCard() ;

         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            PushOffset(Operand) ;
            Offset := PopInt() ;
            PushUsing(Operand, Offset, Size, indexed, ra)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            PushGlobal(Operand, Size)
         END
      ELSE
         Assert(Mode=LeftValue) ;
         (*
            Mem[Mem[Operand]] :=
         *)
         Type := GetType(Operand) ;
         PushSize(Type) ;
         Size := PopCard() ;

         PushOffset(Operand) ;
         Offset := PopInt() ;
         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            Emit(load, indexed, ra, dword, dword, CreateNumber(Offset), Operand)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            Emit(load, direct, no, dword, dword, CreateLabelReference(Operand, 0), Operand)
         END ;
         PopReg(rb) ;
         PushUsing(NulSym, 0, Size, indexed, rb)
      END
   END ;
   RETURN( Size )
END PushOperandMode ;


(*
   PopOperandMode - pops an Operand according to its mode and returns
                    the poped Operand size.
*)

PROCEDURE PopOperandMode (Operand: CARDINAL) : CARDINAL ;
VAR
   Offset: INTEGER ;
   Type,
   Father,
   Size,
   i     : CARDINAL ;
   Mode  : ModeOfAddr ;
BEGIN
   IF IsConst(Operand)
   THEN
      WriteError('Trying to assign to a constant')
   ELSE
      Mode := GetMode(Operand) ;
      IF Mode=RightValue
      THEN
         (*
                 := Mem[Operand]
         *)
         Type := GetType(Operand) ;
         Assert(Type#NulSym) ;
         PushSize(Type) ;
         Size := PopCard() ;

         PushOffset(Operand) ;
         Offset := PopInt() ;
         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            PopUsing(Operand, Offset, Size, indexed, ra)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            UseStoreForPopGlobal(Operand, Size) ;
            (* was PopUsing(Offset, Size, direct, no, GetSymName(Operand)) *)
         END
      ELSE
         Assert(Mode=LeftValue) ;
         (*
            := Mem[Mem[Operand]]
         *)
         Type := GetType(Operand) ;
         PushSize(Type) ;
         Size := PopCard() ;

         Father := GetVarFather(Operand) ;
         IF IsProcedure(Father)
         THEN
            PushOffset(Operand) ;
            Offset := PopInt() ;
            Emit(load, indexed, ra, dword, dword, CreateNumber(Offset), Operand)
         ELSE
            PushSize(Father) ;
            TotalDataSize := PopCard() ;

            Emit(load, direct, ra, dword, dword, CreateLabelReference(Operand, 0), Operand)
         END ;
         PopReg(rb) ;
         PopUsing(NulSym, 0, Size, indexed, rb)
      END
   END ;
   RETURN( Size )
END PopOperandMode ;


(*
   CodeXIndr - 

   Xindr Operator  ( *a = b)
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<V>   Copy   Sym2<I>     Meaning     Mem[Sym1<I>] := constant
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
*)

PROCEDURE CodeXIndr ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   PushSize(Operand2) ;    (* Operand2 is the TYPE of the variable being copied *)
   Size := PopCard() ;
   IF IsConstString(Operand3)
   THEN
      PushStringToStack(Operand3, Size) ;
      PushReg(sp) ;
(*
      PushImmSize(INTEGER(Size), DWORDSIZE) ;
      EmitZero(add) ;                                  (* source      *)
*)
      PushAPUTo(Operand1, DWORDSIZE) ;                 (* destination *)
      PushImmSize(INTEGER(Size), DWORDSIZE) ;          (* size        *)
      EmitZero(memcopy) ;
      IncreaseStack(Size)                    (* remove stacked string *)
   ELSIF IsConst(Operand3)
   THEN
      PushAPUTo(Operand1, DWORDSIZE) ;
      PushAPUTo(Operand3, Size) ;
      ManipulateUsingUnary(Zero, store, Size, Size, indexed, st, NulSym)
   ELSIF IsConst(Operand1)
   THEN
      (* FIXME - this is legal *)
      InternalError('this needs more thought gaius FIXME indirection via a constant', __FILE__, __LINE__) ;
   ELSE
      (*
         Mem[Mem[Op1]] := Mem[Op3]
      *)
      IF WorthUsingMemCopy(Size)
      THEN
         PushADR(Operand3) ;                               (* source      *)
         PushAPUTo(Operand1, DWORDSIZE) ;                  (* destination *)
         PushImmSize(INTEGER(Size), DWORDSIZE) ;           (* size        *)
         EmitZero(memcopy)
      ELSE
         IF IsProcedure(GetVarFather(Operand1))
         THEN
            IF IsUnarySize(Size)
            THEN
               PushAPUTo(Operand1, DWORDSIZE) ;
               PushAPUTo(Operand3, Size) ;
               ManipulateUsingUnary(Zero, store, Size, Size, indexed, st, NulSym)
            ELSE
               PushUsing(Operand3, 0, Size, indexed, ra) ;       (* PushAPUTo(Operand3, Size) ; *)
               PushAPUTo(Operand1, DWORDSIZE) ;
               PopReg(rb) ;
               PopUsing(NulSym, 0, Size, indexed, rb)
            END
         ELSE
            IF IsUnarySize(Size)
            THEN
               (* must also be global not local *)
               PushAPUTo(Operand3, Size) ;
               ManipulateUsingUnary(CreateLabelReference(Operand1, 0), store,
                                    Size, Size, indirect, no, Operand1)
            ELSE
               PushUsing(Operand3, 0, Size, direct, no) ;        (* PushAPUTo(Operand3, Size) ; *)
               PushAPUTo(Operand1, DWORDSIZE) ;
               PopReg(rb) ;
               PopUsing(NulSym, 0, Size, indexed, rb)
            END
         END
      END
   END
END CodeXIndr ;


(*
   CodeIndrX - 
------------------------------------------------------------------------------
   IndrX Operator  (a = *b)   where <X> means any value
------------------------------------------------------------------------------
*)

PROCEDURE CodeIndrX ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size    : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   (*
      Follow the Quadruple rules:
   *)

   IF IsConstString(Operand3)
   THEN
      InternalError('not expecting to index through a constant string', __FILE__, __LINE__)
   ELSIF IsConst(Operand3)
   THEN
      InternalError('this needs more thought gaius FIXME indirection via a constant', __FILE__, __LINE__)
   ELSE
      PushSize(Operand2) ;    (* Operand2 is the TYPE of the variable being copied *)
      Size := PopCard() ;
      (*
         Mem[Op1] := Mem[Mem[Op3]]
      *)


      IF WorthUsingMemCopy(Size)
      THEN
         PushAPUTo(Operand3, DWORDSIZE) ;              (* source      *)
         PushADR(Operand1) ;                           (* destination *)
         PushImmSize(INTEGER(Size), DWORDSIZE) ;       (* size        *)
         EmitZero(memcopy)
      ELSE
         PushAPUTo(Operand3, DWORDSIZE) ;     (* source address     *)

         (* convert source into value *)
         IF IsUnarySize(Size)
         THEN
            ManipulateUsingUnary(Zero, load, Size, Size, indexed, st, NulSym)
         ELSE
            PopReg(rb) ;
            PushUsing(NulSym, 0, Size, indexed, rb)
         END ;

         IF IsProcedure(GetVarFather(Operand1))
         THEN
            IF IsUnarySize(Size)
            THEN
               PushOffset(Operand1) ;
               ManipulateUsingUnary(CreateNumber(PopInt()), store, Size, Size, indexed, ra, Operand1)
            ELSE
               PushOffset(Operand1) ;
               PopUsing(Operand1, PopInt(), Size, indexed, ra)
            END
         ELSE
            IF IsUnarySize(Size)
            THEN
               ManipulateUsingUnary(CreateLabelReference(Operand1, 0), store,
                                    Size, Size, direct, no, Operand1)
            ELSE
               PopGlobal(Operand1, Size)
            END
         END
      END
   END
END CodeIndrX ;


PROCEDURE CodeGoto (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   FlushTime ;
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   Jump(Start, End, CreateLabelName('L', Operand3))
END CodeGoto ;


PROCEDURE CodeOffset ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1)
   THEN
      PushOffset(Operand3) ;
      PopValue(Operand1)
   ELSE
      PushOffset(Operand3) ;
      PushImmSize(PopInt(), DWORDSIZE) ;
      Assert(PopAPU(Operand1)=DWORDSIZE)
   END
END CodeOffset ;


(*
   CodeBase - base operation is now handled inside M2Optimize
*)

PROCEDURE CodeBase ;
BEGIN
   InternalError('expecting BaseOp to have been optimized away during M2Optimize', __FILE__, __LINE__) ;
END CodeBase ;


(*
   PushADR - pushes the location in memory of the variable Operand onto
             the runtime stack.
*)

PROCEDURE PushADR (Operand: CARDINAL) ;
VAR
   Str: ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsProcedureVariable(Operand)
   THEN
      PushReg(ra) ;
      PushOffset(Operand) ;
      PushImmSize(PopInt(), DWORDSIZE) ;
      EmitZero(add)
   ELSE
      PushImmLabel(CreateLabelReference(Operand, 0))
   END
END PushADR ;


(*
   ADR - if Operand is a LeftValue then the Mem[Sym<I>] is pushed
         onto the runtime stack ie the Address of the item to
         which Operand is pointing.
         elsif Operand is an Offset address then Sym<O> is pushed
         onto the runtime stack.
*)

PROCEDURE ADR (Operand: CARDINAL) ;
VAR
   Str: ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsProcedureVariable(Operand)
   THEN
      IF GetMode(Operand)=LeftValue
      THEN
         PushOffset(Operand) ;
         Emit(load, indexed, ra, dword, dword, CreateNumber(PopInt()), Operand)
      ELSE
         PushReg(ra) ;
         PushOffset(Operand) ;
         PushImmSize(PopInt(), DWORDSIZE) ;
         EmitZero(add)
      END
   ELSE
      IF GetMode(Operand)=LeftValue
      THEN
         Emit(load, direct, no, dword, dword, CreateLabelReference(Operand, 0), Operand)
      ELSE
         PushImmLabel(CreateLabelReference(Operand, 0))
      END
   END
END ADR ;


PROCEDURE CodeElementSize ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Subscript: CARDINAL ;
   Type     : CARDINAL ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   Type := GetType(Operand2) ;
   IF NoOfElements(Type)>=Operand3
   THEN
      Subscript := GetNth(Type, Operand3) ;
      PushSize(Subscript) ;
      IF IsConst(Operand1)
      THEN
         PopValue(Operand1)
      ELSE
         PushImmSize(PopInt(), DWORDSIZE) ;
         Assert(PopAPU(Operand1)=DWORDSIZE)
      END
   ELSE
      WriteError('Too many elements') ;
   END
END CodeElementSize ;


(*
   CodeIfIn - codes a bit test determining whether operand1 is in operand2.
*)

PROCEDURE CodeIfIn (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF SetIn()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(and) ;
      PushImmSize(0, DWORDSIZE) ;
      EmitZero(notequ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfIn ;


(*
   CodeIfNotIn - codes a bit test determining whether operand1 is not
                 in operand2.
*)

PROCEDURE CodeIfNotIn (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF SetIn()
      THEN
         (* Do nothing since never arrises! *)
      ELSE
         CodeGoto(Start, End)
      END
   ELSE
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(and) ;
      PushImmSize(0, DWORDSIZE) ;
      EmitZero(equ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfNotIn ;


PROCEDURE CodeIfEqu (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Equ()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(equ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfEqu ;


PROCEDURE CodeIfNotEqu (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF NotEqu()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(notequ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfNotEqu ;


PROCEDURE CodeIfLess (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Less()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(less) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfLess ;


PROCEDURE CodeIfGre (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF Gre()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(gre) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfGre ;


PROCEDURE CodeIfLessEqu (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF LessEqu()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(lessequ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfLessEqu ;


PROCEDURE CodeIfGreEqu (Start, End: CARDINAL) ;
VAR
   Operator: QuadOperator ;
   Operand1,
   Operand2,
   Operand3: CARDINAL ;
   Size1,
   Size2   : CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   GetQuad(CurrentQuad, Operator, Operand1, Operand2, Operand3) ;
   IF IsConst(Operand1) AND IsConst(Operand2)
   THEN
      PushValue(Operand1) ;
      PushValue(Operand2) ;
      IF GreEqu()
      THEN
         CodeGoto(Start, End)
      ELSE
         (* Do nothing since never arrises! *)
      END
   ELSE
      FlushTime ;
      (* Only allow Base types to be compared - at the moment *)
      Size1 := PushOperandModeDword(Operand1) ;
      Size2 := PushOperandModeDword(Operand2) ;
      EmitZero(greequ) ;
      Brt(Start, End, CreateLabelName('L', Operand3))
   END
END CodeIfGreEqu ;



(*
   PushAPU - pushes symbol onto the APU stack. It treats Value, Offset
             symbols the same. It returns the number of bytes pushed.
*)

PROCEDURE PushAPU (Sym: CARDINAL) : CARDINAL ;
VAR
   Offset: INTEGER ;
   Father,
   i,
   Size  : CARDINAL ;
   Str   : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsConstString(Sym)
   THEN
      Size := GetStringLength(Sym) ;
      IF Size=BYTESIZE
      THEN
         GetKey(GetString(Sym), Str) ;
         PushImmSize(ORD(Str[0]), Size)
      ELSE
         WriteError('Cannot PushAPU a string with more than 1 character in it')
      END
   ELSIF IsConst(Sym)
   THEN
      PushValue(Sym) ;
      PushImmSize(PopInt(), DWORDSIZE) ;
      Size := DWORDSIZE
   ELSE
      Father := GetVarFather(Sym) ;
      IF GetMode(Sym)=LeftValue
      THEN
         Size := DWORDSIZE
      ELSE
         PushSize(Sym) ;
         Size := PopCard()
      END ;

      IF IsProcedure(Father)
      THEN
         PushOffset(Sym) ;
         Offset := PopInt() ;
         PushUsing(Sym, Offset, Size, indexed, ra)
      ELSE
         PushGlobal(Sym, Size)
      END
   END ;
   RETURN( Size )
END PushAPU ;


PROCEDURE PopAPU (Sym: CARDINAL) : CARDINAL ;
VAR
   Size,
   Father: CARDINAL ;
   Offset: INTEGER ;
   Str   : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsConst(Sym)
   THEN
      WriteError('Cannot pop into a constant')
   ELSE
      Father := GetVarFather(Sym) ;
      IF GetMode(Sym)=LeftValue
      THEN
         Size := DWORDSIZE
      ELSE
         PushSize(Sym) ;
         Size := PopCard()
      END ;

      IF IsProcedure(Father)
      THEN
         PushOffset(Sym) ;
         Offset := PopInt() ;
         PopUsing(Sym, Offset, Size, indexed, ra)
      ELSE
         PopGlobal(Sym, Size)
      END
   END ;
   RETURN( Size )
END PopAPU ;


(*
   PushAPUTo - pushes symbol onto the APU stack. It treats Value, Offset
               symbols the same. It pushes the symbol using ToSize bytes.
*)

PROCEDURE PushAPUTo (Sym: CARDINAL; ToSize: CARDINAL) ;
VAR
   Offset  : INTEGER ;
   Father,
   i,
   FromSize: CARDINAL ;
   Str     : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsConstString(Sym)
   THEN
      FromSize := GetStringLength(Sym) ;
      IF FromSize=1
      THEN
         GetKey(GetString(Sym), Str) ;
         PushImmSize(ORD(Str[0]), ToSize)
      ELSE
         WriteError('Cannot push a string with more than 1 character in it')
      END
   ELSIF IsConst(Sym)
   THEN
      PushValue(Sym) ;
      PushImmSize(PopInt(), ToSize)
   ELSIF IsProcedure(Sym)
   THEN
      Assert(ToSize=PushProcName(Sym))
   ELSE
      Father := GetVarFather(Sym) ;
      IF IsProcedure(Father)
      THEN
         PushVarSize(Sym) ;
         FromSize := PopCard() ;
         PushOffset(Sym) ;
         Offset := PopInt() ;

         ManipulateUsingUnary(CreateNumber(Offset), load,
                              FromSize, ToSize, indexed, ra, Sym)
      ELSE
         PushSize(Father) ;
         TotalDataSize := PopCard() ;

         PushVarSize(Sym) ;
         FromSize := PopCard() ;
         ManipulateUsingUnary(CreateLabelReference(Sym, 0), load,
                              FromSize, ToSize, direct, no, Sym)
      END
   END
END PushAPUTo ;


PROCEDURE PopAPUFrom (Sym: CARDINAL; FromSize: CARDINAL) ;
VAR
   ToSize,
   Father: CARDINAL ;
   Offset: INTEGER ;
   Str   : ARRAY [0..20] OF CHAR ;
BEGIN
   IF IsConst(Sym)
   THEN
      WriteError('Cannot pop into a constant')
   ELSE
      Father := GetVarFather(Sym) ;
      IF IsProcedure(Father)
      THEN
         PushVarSize(Sym) ;
         ToSize := PopCard() ;
         PushOffset(Sym) ;
         Offset := PopInt() ;
         ManipulateUsingUnary(CreateNumber(Offset), store,
                              FromSize, ToSize, indexed, ra, Sym)
      ELSE
         PushSize(Father) ;
         TotalDataSize := PopCard() ;

         PushVarSize(Sym) ;
         ToSize := PopCard() ;
         ManipulateUsingUnary(CreateLabelReference(Sym, 0), store,
                              FromSize, ToSize, direct, no, Sym)
      END
   END
END PopAPUFrom ;


(*
   InitializationCode - displays the initialization code.
*)

PROCEDURE InitializationCode ;
BEGIN
   OpenAssemblyFileAndInitialize   
END InitializationCode ;


(*
   Comment - writes a comment to the assembly file
*)

PROCEDURE Comment (a: ARRAY OF CHAR) ;
BEGIN
   WriteString('# ') ; WriteString(a) ; WriteLn
END Comment ;


(*
   OpenAssemblyFileAndInitialize - displays the initialization code.
*)

PROCEDURE OpenAssemblyFileAndInitialize ;
VAR
   ModuleName,
   FileName  : ARRAY [0..80] OF CHAR ;
   a         : ARRAY [0..10] OF CHAR ;
BEGIN
   IF GetEnvironment('DEBUGCODE', a)
   THEN
      PushOutput(IO.Write) ;
      Debugging := TRUE
   ELSE
      OpenAssemblyFile(GetSymName(GetMainModule()))
   END ;
   WriteLn ;
   WriteLn ;
   Comment('       M-2 -> assembly language') ;
   Comment('');
   WriteLn ;
   WriteString('        .file   "') ;
   GetKey(GetSymName(GetMainModule()), ModuleName) ;
   CalculateFileName(ModuleName, 'mod', FileName) ;
   WriteString(FileName) ;
   Write('"') ; WriteLn ;
   Flush(0, 0) ;
   FileInfo(GetMainModule()) ;
   WriteLn ;
   TypeInfo ;   (* Display the GDB info for the types used in this MODULE *)
   DeclareGlobalVariables(GetMainModule()) ;
   ForeachExportedDo(GetMainModule(), DisplayExportedProcedure) ;
   WriteString('   .global ') ; WriteString('__M2_') ;
   WriteKey(GetSymName(GetMainModule())) ; WriteString('_init') ; WriteLn ;
   ForeachImportedDo(GetMainModule(), DisplayImportedVar) ;
   ForeachImportedDo(GetMainModule(), DisplayImportedProcedure) ;
   IF GetSymName(GetMainModule())#MakeKey('M2RTS')
   THEN
      WriteString('   .extern    _M2RTS_HALT') ; WriteLn ;
      WriteString('   .extern    _M2RTS_Terminate') ; WriteLn
   END ;
   WriteLn
END OpenAssemblyFileAndInitialize ;


(*
   CloseAssemblyFileAndTerminate - displays termination code
*)

PROCEDURE CloseAssemblyFileAndTerminate ;
BEGIN
   IF Debugging
   THEN
      PopOutput
   ELSE
      CloseAssemblyFile
   END
END CloseAssemblyFileAndTerminate ;


(*
   DisplayImportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM imported form.
*)

PROCEDURE DisplayImportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('   .extern   ') ; WriteAsmName(Sym) ; WriteLn
   END
END DisplayImportedProcedure ;


(*
   DisplayExportedProcedure - displays a symbol, Sym, (if a procedure)
                              in a MASM exported form.
*)

PROCEDURE DisplayExportedProcedure (Sym: CARDINAL) ;
BEGIN
   IF IsProcedure(Sym) AND (NOT IsPseudoSystemFunction(Sym))
   THEN
      WriteString('   .global  ') ; WriteAsmName(Sym) ;
      WriteLn
   END
END DisplayExportedProcedure ;


(*
   DisplayExportedVar - displays a symbol, Sym, (if a variable)
                        in a GDB exported form.
*)

PROCEDURE DisplayExportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('   .global   ') ; WriteAsmName(Sym) ; WriteLn
   END
END DisplayExportedVar ;


(*
   DisplayImportedVar - displays a symbol, Sym, (if a variable)
                        in a MASM exported form.
*)

PROCEDURE DisplayImportedVar (Sym: CARDINAL) ;
BEGIN
   IF IsVar(Sym)
   THEN
      WriteString('   .extern   ') ; WriteAsmName(Sym) ; WriteLn
   END
END DisplayImportedVar ;


(*
   WritePrologue - writes the prologue code before the
                   initialization code of a module.
*)

PROCEDURE WritePrologue (ModSym: CARDINAL;
                         CurrentQuad: CARDINAL; Line: CARDINAL) ;
BEGIN
   Flush(0, 0) ;
   MainInfo(ModSym, Line) ;
   WriteString('  .global  ') ;
   WriteKey(GetModuleInitName(ModSym)) ; WriteLn ;
   WriteKey(GetModuleInitName(ModSym)) ; Write(':') ; WriteLn ;

   PushReg(ra) ;                (* save old frame pointer *)
   PushReg(sp) ;                (* make new frame pointer *)
   PopReg(ra) ;

   (* Push Register *)
   PushReg(rc) ;
   Flush(0, 0) ;

   BeginInfo
END WritePrologue ;


(*
   WriteEpilogue - writes the termination code after the initialization
                   of a module.
*)

PROCEDURE WriteEpilogue (ModSym: CARDINAL) ;
BEGIN
   Flush(0, 0) ;
   EndInfo ;

   IF IsModule(GetMainModule()) AND (GetMainModule()=ModSym)
   THEN
      Comment(' clean up using Modula-2 M2RTS termination') ;
      WriteString('  load.d.d  $_M2RTS_Terminate') ; WriteLn ;
      WriteString('  call') ; WriteLn
   END ;
   (* Pop Registers *)
   PopReg(rc) ;                 (* restore register          *)

   PopReg(ra) ;                 (* restore old frame pointer *)
   Ret(0, 0) ;
   EmitZero(nop) ;  (* so that our program counter never exceeds its limits *)
   Flush(0, 0)
END WriteEpilogue ;


(*
   GetSize - returns the size of a symbol, Sym.
             returns the amount of space variable, Sym, requires.
             NOTE this is not the same as GetVarTypeSize since
                  if Sym is LeftValue then the size will be
                  TSIZE(ADDRESS) since Sym is in fact a "pointer".
                  ie VAR parameter...
*)

PROCEDURE GetSize (Sym: CARDINAL) : CARDINAL ;
BEGIN
   IF IsVar(Sym)
   THEN
      PushVarSize(Sym)
   ELSE
      PushSize(Sym)
   END ;
   RETURN( PopCard() )
END GetSize ;


(*
   GetOffset - returns the offset for a symbol, Sym.
*)

PROCEDURE GetOffset (Sym: CARDINAL) : INTEGER ;
BEGIN
   IF IsConst(Sym) OR IsProcedure(Sym)
   THEN
      RETURN( 0 )
   ELSE
      PushOffset(Sym) ;
      RETURN( PopInt() )
   END
END GetOffset ;


(*
   DeclareGlobalVariables - lists the Global variables for Module ModSym
                            together with their offset.
*)

PROCEDURE DeclareGlobalVariables (ModSym: CARDINAL) ;
VAR
   s,
   n,
   Son: CARDINAL ;
BEGIN
   n := 1 ;
   Son := GetNth(ModSym, n) ;
   WHILE Son#NulSym DO
      Flush(0, 0) ;
      VariableInfo(ModSym, Son) ;   (* Generate GDB debugging info *)
      IF IsExported(ModSym, Son)
      THEN
         (* Actually exported variables are defined as COMM ... *)
         (* was - WriteString('EXTRN  ') *)
         WriteString('.comm  ')
      ELSE
         WriteString('.lcomm   ')
      END ;
      WriteAsmName(Son) ; WriteString('  ') ;
      s := GetSize(Son) ;
      Write(',') ; WriteSize(s) ;
      WriteString(' # offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(ModSym, n)
   END ;
   ForeachInnerModuleDo(ModSym, DeclareGlobalVariables)
END DeclareGlobalVariables ;


(*
   DeclareLocalVariables - lists the Local variables for Procedure Sym
                           together with their offset.
*)

PROCEDURE DeclareLocalVariables (Sym: CARDINAL) ;
VAR
   s,
   n,
   Son: CARDINAL ;
BEGIN
   n := 1 ;   (* Later collect function return param *)
   Son := GetNth(Sym, n) ;
   WHILE Son#NulSym DO
      WriteString('#   ') ;
      WriteAsmName(Son) ; Write(':') ;
      s := GetSize(Son) ;
      WriteSize(s) ;
      WriteString(' ; offset') ;
      WriteInt(GetOffset(Son), 6) ;
      IF GetMode(Son)=LeftValue
      THEN
         WriteString(' (Pointer To ') ; WriteKey(GetSymName(GetType(Son))) ;
         Write(')') ;
      END ;
      WriteLn ;
      INC(n) ;
      Son := GetNth(Sym, n)
   END
END DeclareLocalVariables ;


(*
   WriteSize - writes the Size in bytes, but in a human form.
*)

PROCEDURE WriteSize (Size: CARDINAL) ;
BEGIN
   WriteCard(Size, 0)
END WriteSize ;


(*
   DeclareVariable - declares a variable Sym together with its offset.
*)

PROCEDURE DeclareVariable (Sym: CARDINAL) ;
BEGIN
   WriteAsmName(Sym) ;
   WriteString(' .skip ') ;
   PushSize(Sym) ;
   WriteCard(PopCard(), 0) ;
   WriteLn
END DeclareVariable ;


(*
   TerminationCode - displays termination code
*)

PROCEDURE TerminationCode ;
BEGIN
   Flush(0, 0) ;    (* Flush all instructions *)
   WriteString('  .end') ; WriteLn ;
   CloseAssemblyFileAndTerminate
END TerminationCode ;


(*
   Time - creates a Pr instruction which will request the processor
          to perform processing for a time, Duration.
*)

PROCEDURE Time (Duration: CARDINAL) ;
BEGIN
   IF LocalProfiling
   THEN
      INC(OutstandingDuration, Duration)
   END
END Time ;


(*
   FlushTime - flushes the outstanding time instruction if it exists.
*)

PROCEDURE FlushTime ;
BEGIN
   IF OutstandingDuration>0
   THEN
      PushImmSize(INTEGER(OutstandingDuration), DWORDSIZE) ;
      EmitZero(pr) ;
      OutstandingDuration := 0
   END
END FlushTime ;


PROCEDURE Init ;
BEGIN
   Debugging           := FALSE ;
   Zero                := MakeKey('0') ;
   One                 := MakeKey('1') ;
   TotalDataSize       := 0 ;
   UnboundedOnStack    := FALSE ;
   OutstandingDuration := 0
END Init ;


PROCEDURE GenerateQuadLabel (Start, End: CARDINAL) ;
VAR
   Str: ARRAY [0..20] OF CHAR ;
BEGIN
   Flush(Start, End) ;
   Write('L') ; WriteCard(CurrentQuad, 0) ; Write(':') ; WriteLn
END GenerateQuadLabel ;


(* -------------------------------------------------------------------- *)

(*
   Emit routines to emit APU code
*)

CONST
   MaxCodeBuffer = 5000 ;

TYPE
   Instr = RECORD
              Inst  : Instructions ;
              Mode  : Modes ;
              Reg   : Registers ;
              Value : CARDINAL ;
              From,
              To    : Sizes ;
              Sym   : CARDINAL ;
              LineNo: CARDINAL ;
           END ;

VAR
   Instruction: ARRAY [1..MaxCodeBuffer] OF Instr ;
   CurrentInst: CARDINAL ;


(*
   Halt - creates a halt instruction.
*)

PROCEDURE Halt ;
BEGIN
   InternalError('why are we generating a halt instruction?', __FILE__, __LINE__)
END Halt ;


(*
   Ret - creates a return instruction.
*)

PROCEDURE Ret (Start, End: CARDINAL) ;
BEGIN
   PopReg(pc) ;
   Flush(Start, End)
END Ret ;


(*
   Brt - creates a branch if true instruction.
*)

PROCEDURE Brt (Start, End: CARDINAL; Name: CARDINAL) ;
BEGIN
   PushImmLabel(Name) ;
   Flush(Start, End) ;
   WriteString('  bift') ; WriteLn
END Brt ;


(*
   Jump - creates a jump instruction to a label, Name.
*)

PROCEDURE Jump (Start, End: CARDINAL; Name: CARDINAL) ;
BEGIN
   PushImmLabel(Name) ;
   PopReg(pc) ;
   Flush(Start, End)
END Jump ;


(*
   PushGlobalDword - pushes global data onto the stack. Destination is dword.
*)

PROCEDURE PushGlobalDword (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   IF Size=DWORDSIZE
   THEN
      Emit(load, direct, no, dword, dword, CreateLabelReference(Sym, 0), Sym)
   ELSIF Size=WORDSIZE
   THEN
      Emit(load, direct, no, word, dword, CreateLabelReference(Sym, 0), Sym)
   ELSIF Size=BYTESIZE
   THEN
      Emit(load, direct, no, byte, dword, CreateLabelReference(Sym, 0), Sym)
   ELSE
      InternalError('should only have size 1..4 bytes', __FILE__, __LINE__)
   END
END PushGlobalDword ;


(*
   PushGlobal - pushes global data onto the stack.
*)

PROCEDURE PushGlobal (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   IF WorthUsingMemCopy(Size)
   THEN
      UseMemCopyForPushGlobal(Sym, Size)
   ELSE
      UseLoadForPushGlobal(Sym, Size)
   END
END PushGlobal ;


(*
   UseLoadPushGlobal - pushes global data onto the stack.
*)

PROCEDURE UseLoadForPushGlobal (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   WHILE Size>0 DO
      IF Size>=DWORDSIZE
      THEN
         DEC(Size, DWORDSIZE) ;
         Emit(load, direct, no, dword, dword, CreateLabelReference(Sym, INTEGER(Size)), Sym)
      ELSIF Size>=WORDSIZE
      THEN
         DEC(Size, WORDSIZE) ;
         Emit(load, direct, no, word, word, CreateLabelReference(Sym, INTEGER(Size)), Sym)
      ELSE
         DEC(Size) ;
         Emit(load, direct, no, byte, byte, CreateLabelReference(Sym, INTEGER(Size)), Sym)
      END
   END
END UseLoadForPushGlobal ;


(*
   UseMemCopyForPushGlobal - encode a push using the memcopy instruction.
*)

PROCEDURE UseMemCopyForPushGlobal (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   (* firstly make space on the stack for Sym *)
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;
   PopReg(sp) ;
   PushImmLabel(CreateLabelReference(Sym, 0)) ;  (* source *)
   PushReg(sp) ;
   PushImmSize(DWORDSIZE, DWORDSIZE) ;
   EmitZero(add) ;                               (* destination *)
   PushImmSize(INTEGER(Size), DWORDSIZE) ;       (* number of bytes *)
   EmitZero(memcopy)
END UseMemCopyForPushGlobal ;


(*
   PopGlobal - pops global data from the stack.
*)

PROCEDURE PopGlobal (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   IF WorthUsingMemCopy(Size)
   THEN
      UseMemCopyForPopGlobal(Sym, Size)
   ELSE
      UseStoreForPopGlobal(Sym, Size)
   END
END PopGlobal ;


(*
   UseStoreForPopGlobal - pops global data from the stack.
*)

PROCEDURE UseStoreForPopGlobal (Sym: CARDINAL; Size: CARDINAL) ;
VAR
   i: INTEGER ;
BEGIN
   i := 0 ;
   WHILE i<Size DO
      IF Size-i>=DWORDSIZE
      THEN
         Emit(store, direct, no, dword, dword, CreateLabelReference(Sym, i), Sym) ;
         INC(i, DWORDSIZE)
      ELSIF Size-i>=WORDSIZE
      THEN
         Emit(store, direct, no, word, word, CreateLabelReference(Sym, i), Sym) ;
         INC(i, WORDSIZE)
      ELSE
         Emit(store, direct, no, byte, byte, CreateLabelReference(Sym, i), Sym) ;
         INC(i)
      END
   END
END UseStoreForPopGlobal ;


(*
   UseMemCopyForPopGlobal - encode a pop using the memcopy instruction.
*)

PROCEDURE UseMemCopyForPopGlobal (Sym: CARDINAL; Size: CARDINAL) ;
BEGIN
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;   (* source *)
   PushImmLabel(CreateLabelReference(Sym, 0)) ;  (* destination *)
   PushImmSize(INTEGER(Size), DWORDSIZE) ;   (* number of bytes *)
   EmitZero(memcopy) ; (* after memcopy has completed the previous three operands will have gone *)
   (* adjust stack to the position as if we had performed multiple pops *)
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;
   PopReg(sp)
END UseMemCopyForPopGlobal ;


(*
   PushUsing - pushes, Size, bytes to the stack starting at offset.
               Mode and register, m and r.
               It attempts to push DWORDSIZE items at a time.
*)

PROCEDURE PushUsing (Sym: CARDINAL;
                     Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   IF WorthUsingMemCopy(Size)
   THEN
      UseMemCopyForPush(Offset, Size, m, r)
   ELSE
      UseLoadForPush(Sym, Offset, Size, m, r)
   END
END PushUsing ;


(*
   WorthUsingMemCopy - is it worth using the memcopy instruction to push data?
                       We can only use this instruction if we are pushing
                       more than 3*DWORDSIZE bytes. See UseMemCopyForPush for
                       the reason why. (The memcopy requires 3 operands on the stack)
*)

PROCEDURE WorthUsingMemCopy (Size: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( APUUseMemCopy AND (Size>DWORDSIZE*3) )
END WorthUsingMemCopy ;


(*
   UseMemCopyForPush - encode a push using the memcopy instruction.
*)

PROCEDURE UseMemCopyForPush (Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   (* enlarge the stack so to emulate multiple pushes *)
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;
   PopReg(sp) ;

   PushReg(sp) ;
   PopReg(rd) ;     (* remember destination, memcopy works from low..high addresses *)

   IF (m=indexed) AND (r#no)
   THEN
      (* we must calculate the address of r(offset) ourselves *)
      Emit(load, immediate, no, dword, dword, CreateNumber(Offset), NulSym) ;
      PushReg(r) ;
      EmitZero(add)    (* source *)
   ELSE
      Emit(load, m, r, dword, dword, CreateNumber(Offset), NulSym)   (* source *)
   END ;
   PushReg(rd) ;   (* destination *)
   PushImmSize(INTEGER(Size), DWORDSIZE) ;   (* number of bytes *)
   EmitZero(memcopy)
END UseMemCopyForPush ;


(*
   UseLoadForPush - pushes the data to stack using multiple loads
*)

PROCEDURE UseLoadForPush (Sym: CARDINAL;
                          Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   WHILE Size>0 DO
      IF Size>=DWORDSIZE
      THEN
         DEC(Size, DWORDSIZE) ;
         Emit(load, m, r, dword, dword, CreateNumber(Offset+Size), Sym)
      ELSIF Size>=WORDSIZE
      THEN
         DEC(Size, WORDSIZE) ;
         Emit(load, m, r, word, word, CreateNumber(Offset+Size), Sym)
      ELSE
         DEC(Size) ;
         Emit(load, m, r, byte, byte, CreateNumber(Offset+Size), Sym)
      END
   END
END UseLoadForPush ;


(*
   PushUsingDword - pushes, Size, bytes to the stack starting at offset.
                    Mode and register, m and r. The destination is a dword.
*)

PROCEDURE PushUsingDword (Sym: CARDINAL;
                          Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   IF Size>=DWORDSIZE
   THEN
      Emit(load, m, r, dword, dword, CreateNumber(Offset), Sym)
   ELSIF Size>=WORDSIZE
   THEN
      Emit(load, m, r, word, dword, CreateNumber(Offset), Sym)
   ELSIF Size=BYTESIZE
   THEN
      Emit(load, m, r, byte, dword, CreateNumber(Offset), Sym)
   ELSE
      InternalError('should only have size 1..4 bytes', __FILE__, __LINE__)
   END
END PushUsingDword ;


(*
   PopUsing - pops, Size, bytes from the stack starting at offset.
              Mode and register, m and r.
              It attempts to pop DWORDSIZE items at a time.
*)

PROCEDURE PopUsing (Sym: CARDINAL;
                    Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   IF WorthUsingMemCopy(Size)
   THEN
      UseMemCopyForPop(Offset, Size, m, r)
   ELSE
      UseStoreForPop(Sym, Offset, Size, m, r)
   END
END PopUsing ;


(*
   UseMemCopyForPop - encode a pop using the memcopy instruction.
*)

PROCEDURE UseMemCopyForPop (Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
BEGIN
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;   (* source *)

   IF (m=indexed) AND (r#no)
   THEN
      (* we must calculate the address of r(offset) ourselves *)
      Emit(load, immediate, no, dword, dword, CreateNumber(Offset), NulSym) ;
      PushReg(r) ;
      EmitZero(add)    (* destination *)
   ELSE
      Emit(load, m, r, dword, dword, CreateNumber(Offset), NulSym)   (* destination *)
   END ;

   PushImmSize(INTEGER(Size), DWORDSIZE) ;     (* number of bytes *)
   EmitZero(memcopy) ; (* after memcopy has completed the previous three operands will have gone *)
   (* adjust stack to the position as if we had performed multiple pops *)
   PushReg(sp) ;
   PushImmSize(INTEGER(Size), DWORDSIZE) ;
   EmitZero(sub) ;
   PopReg(sp)
END UseMemCopyForPop ;


(*
   UseStoreForPop - pops the data to stack using multiple stores
*)

PROCEDURE UseStoreForPop (Sym: CARDINAL;
                          Offset: INTEGER; Size: CARDINAL; m: Modes; r: Registers) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<Size DO
      IF Size-i>=DWORDSIZE
      THEN
         Emit(store, m, r, dword, dword, CreateNumber(Offset+i), Sym) ;
         INC(i, DWORDSIZE)
      ELSIF Size-i>=WORDSIZE
      THEN
         Emit(store, m, r, word, word, CreateNumber(Offset+i), Sym) ;
         INC(i, WORDSIZE)
      ELSE
         Emit(store, m, r, byte, byte, CreateNumber(Offset+i), Sym) ;
         INC(i)
      END
   END
END UseStoreForPop ;


(*
   IsUnarySize - returns TRUE if size is byte, word or dword.
*)

PROCEDURE IsUnarySize (Size: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN( (Size=DWORDSIZE) OR (Size=WORDSIZE) OR (Size=BYTESIZE) )
END IsUnarySize ;


(*
   ManipulateUsingUnary - push/pops the data from stack using a single inst.
                          Can only be achieved if both operands are 1, 2, 4
                          bytes.
*)

PROCEDURE ManipulateUsingUnary (Name: CARDINAL;
                                Inst: Instructions;
                                FromSize, ToSize: CARDINAL;
                                m: Modes; r: Registers;
                                Sym: CARDINAL) ;
VAR
   Source,
   Dest  : Sizes ;
BEGIN
   Assert(IsUnarySize(FromSize)) ;
   Assert(IsUnarySize(ToSize)) ;
   Dest := CardToSize(ToSize) ;
   Source := CardToSize(FromSize) ;
   Emit(Inst, m, r, Source, Dest, Name, Sym)
END ManipulateUsingUnary ;


(*
   CardToSize - converts a CARDINAL into Sizes.
*)

PROCEDURE CardToSize (s: CARDINAL) : Sizes ;
BEGIN
   CASE s OF

   DWORDSIZE:  RETURN( dword ) |
   WORDSIZE :  RETURN( word ) |
   BYTESIZE :  RETURN( byte )

   ELSE
      WriteError('cannot convert symbol to a machine sized operand')
   END
END CardToSize ;


(*
   PushZero - push, n, bytes of zeros onto the APU stack.
*)

PROCEDURE PushZero (n: INTEGER) ;
BEGIN
   IF n>0
   THEN
      IF n=DWORDSIZE
      THEN
         Emit(load, immediate, no, dword, dword, Zero, NulSym) ;
         DEC(n, DWORDSIZE)
      END ;
      IF n>=WORDSIZE
      THEN
         Emit(load, immediate, no, word, word, Zero, NulSym) ;
         DEC(n, WORDSIZE)
      END ;
      IF n>0
      THEN
         Emit(load, immediate, no, byte, byte, Zero, NulSym) ;
         DEC(n)
      END
   ELSIF n<0
   THEN
      InternalError('should not be attempting to push < 0 bytes', __FILE__, __LINE__)
   END
END PushZero ;


(*
   PushImmLabel - 
*)

PROCEDURE PushImmLabel (Name: CARDINAL) ;
BEGIN
   Emit(load, immediate, no, dword, dword, Name, NulSym)
END PushImmLabel ;


(*
   PushImmSize - pushes a number, n, onto the SMP stack.
                 Pushed as a, s.
*)

PROCEDURE PushImmSize (n: INTEGER; s: CARDINAL) ;
BEGIN
   CASE s OF

   BYTESIZE : Emit(load, immediate, no, byte, byte, CreateNumber(n), NulSym) |
   WORDSIZE : Emit(load, immediate, no, word, word, CreateNumber(n), NulSym) |
   DWORDSIZE: Emit(load, immediate, no, dword, dword, CreateNumber(n), NulSym)

   ELSE
      InternalError('incorrect size s', __FILE__, __LINE__)
   END
END PushImmSize ;


(*
   PushReg - pushes register, r, to the stack.
*)

PROCEDURE PushReg (r: Registers) ;
BEGIN
   Emit(load, internal, r, dword, dword, Zero, NulSym)
END PushReg ;


(*
   PopReg - pops register, r, from the stack.
*)

PROCEDURE PopReg (r: Registers) ;
BEGIN
   Emit(store, internal, r, dword, dword, Zero, NulSym)
END PopReg ;


(*
   EmitLabel - places a label into the instruction stream
*)

PROCEDURE EmitLabel (Start, End: CARDINAL; Name: CARDINAL) ;
BEGIN
   Flush(Start, End) ;
   WriteKey(Name) ; Write(':') ; WriteLn
END EmitLabel ;


(*
   CreateLabelName - creates a label name and returns the result.
*)

PROCEDURE CreateLabelName (Prefix: ARRAY OF CHAR; Postfix: CARDINAL) : CARDINAL ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   CardToStr(Postfix, 0, a) ;
   StrConCat(Prefix, a, a) ;
   RETURN( MakeKey(a) )
END CreateLabelName ;


(*
   CreateLabelReference - creates a label reference and adds an offset, o, to the label.
*)

PROCEDURE CreateLabelReference (Sym: CARDINAL; o: INTEGER) : CARDINAL ;
VAR
   Num,
   Name: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IF o=0
   THEN
      RETURN( GetAsmName(Sym) )
   ELSE
      GetKey(GetAsmName(Sym), Name) ;
      IntToStr(o, 0, Num) ;
      IF o>0
      THEN
         StrConCat(Name, '+', Name)
      END ;
      StrConCat(Name, Num, Name) ;
      RETURN( MakeKey(Name) )
   END
END CreateLabelReference ;


(*
   CreateNumber - creates a number and returns the name.
*)

PROCEDURE CreateNumber (n: INTEGER) : CARDINAL ;
VAR
   a: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   IntToStr(n, 0, a) ;
   RETURN( MakeKey(a) )
END CreateNumber ;


(*
   CreateString - creates a number and returns the name.
                  The number has a string comment attached for debugging purposes only.
*)

PROCEDURE CreateString (n: CARDINAL) : CARDINAL ;
VAR
   a, b: ARRAY [0..MaxString] OF CHAR ;
BEGIN
   CardToStr(n, 0, a) ;
   StrConCat(a, '#  "', a) ;  (* comment *)
   IF n=0
   THEN
      StrCopy('<nul>', b)
   ELSE
      b[0] := VAL(CHAR, n MOD 0100H) ;
      b[1] := VAL(CHAR, (n DIV 0100H) MOD 0100H) ;
      b[2] := VAL(CHAR, (n DIV 010000H) MOD 0100H) ;
      b[3] := VAL(CHAR, (n DIV 01000000H) MOD 0100H) ;
      b[4] := nul
   END ;
   StrConCat(a, b, a) ;
   StrConCat(a, '"', a) ;
   RETURN( MakeKey(a) )
END CreateString ;


(*
   EmitZero - places Zero operand instruction, inst, into the instruction buffer.
*)

PROCEDURE EmitZero (inst: Instructions) ;
BEGIN
   Emit(inst, direct, no, dword, dword, Zero, NulSym)
END EmitZero ;


(*
   Emit - all generated APU code should pass through this routine.
          Emit buffers all Basic Blocks and calls flush at the start of a new
          Basic Block.
          All Emitted code that has a label attached to it will be placed
          into the code buffer, regardless of LocalCoding.
*)

PROCEDURE Emit (inst: Instructions; mode: Modes; reg: Registers;
                from, to: Sizes; value: CARDINAL; sym: CARDINAL) ;
BEGIN
   IF LocalCoding
   THEN
      IF CurrentInst=MaxCodeBuffer
      THEN
         InternalError('Increase MaxCodeBuffer', __FILE__, __LINE__)
      ELSE
         INC(CurrentInst) ;
         WITH Instruction[CurrentInst] DO
            Inst   := inst ;
            Mode   := mode ;
            Reg    := reg ;
            Value  := value ;
            From   := from ;
            To     := to ;
            Sym    := sym ;
            LineNo := CurrentLineNo
         END
      END
   END
END Emit ;


(*
   Flush - flushes all emitted code buffered so far.
*)

PROCEDURE Flush (Start, End: CARDINAL) ;
BEGIN
   IF Optimizing
   THEN
      OptAPU(Start, End)
   END ;
   ChuckAPU ;
   CurrentLineNo := 0
END Flush ;


(*
   OptAPU - perform very simple peephole optimizations on the APU code.
            This is only called if the -O flag is set.
*)

PROCEDURE OptAPU (Start, End: CARDINAL) ;
VAR
   i  : CARDINAL ;
   Opt: BOOLEAN ;
BEGIN
   REPEAT
      Opt := FALSE ;
      i := 1 ;
      WHILE (i<=CurrentInst) AND (NOT Opt) DO
         IF ReduceLoadStore(Start, End, i)
         THEN
            Opt := TRUE
         ELSIF ReduceAdd(i)
         THEN
            Opt := TRUE
         ELSIF ReduceSub(i)
         THEN
            Opt := TRUE
         ELSIF ReduceMult(i)
         THEN
            Opt := TRUE
         END ;
         INC(i)
      END
   UNTIL NOT Opt ;
END OptAPU ;


(*
   IsSymNeeded - returns TRUE if the symbol, Sym, is required in future quadruples.
*)

PROCEDURE IsSymNeeded (Sym: CARDINAL; Start, End: CARDINAL) : BOOLEAN ;
VAR
   ReadStart, ReadEnd: CARDINAL ;
BEGIN
   IF IsTemporary(Sym)
   THEN
      GetVarReadQuads(Sym, ReadStart, ReadEnd) ;
      IF (ReadEnd>End) OR (ReadStart<Start)
      THEN
         (* read outside of our basic block, we must flush symbol *)
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   END ;
   RETURN( IsVar(Sym) )
END IsSymNeeded ;


(*
   Dummy - 
*)

PROCEDURE Dummy ;
BEGIN
END Dummy ;


(*
   IsSymUsed - returns TRUE if symbol, Sym, is used in the instruction sequence, i..j
*)

PROCEDURE IsSymUsed (Sym: CARDINAL; i, j: CARDINAL) : BOOLEAN ;
BEGIN
   WHILE i<=j DO
      IF Instruction[i].Sym=Sym
      THEN
         RETURN( TRUE )
      ELSE
         INC(i)
      END
   END ;
   RETURN( FALSE )
END IsSymUsed ;


(*
   ReduceLoadStore - tests for redundant Load and Store, or, Store and Load
                     operations at instruction, i.
                     return TRUE if the optimization has taken
                     place otherwise FALSE.
*)

PROCEDURE ReduceLoadStore (Start, End: CARDINAL; i: CARDINAL) : BOOLEAN ;
BEGIN
   IF i+1<=CurrentInst
   THEN
      IF (Instruction[i].Sym#NulSym) AND (GetSymName(Instruction[i].Sym)=MakeKey('_T20'))
      THEN
         Dummy
      END ;
      IF Debugging
      THEN
         IF (Instruction[i].Inst=load) AND (Instruction[i+1].Inst=store) AND EquModeRegValue(i, i+1)
         THEN
            IF Instruction[i].Sym=NulSym
            THEN
               WriteString('# found load store with <NulSym>') ; WriteLn
            ELSE
               WriteString('# found load store with ') ; WriteAsmName(Instruction[i].Sym) ; WriteLn
            END
         ELSIF (Instruction[i].Inst=store) AND (Instruction[i+1].Inst=load) AND EquModeRegValue(i, i+1)
         THEN
            IF Instruction[i].Sym=NulSym
            THEN
               WriteString('# found store load with <NulSym>') ; WriteLn
            ELSE
               WriteString('# found store load with ') ; WriteAsmName(Instruction[i].Sym) ; WriteLn
            END
         END
      END ;
      IF (Instruction[i].Inst=load) AND (Instruction[i+1].Inst=store) AND
         EquModeRegValue(i, i+1)
      THEN
         IF Debugging
         THEN
            WriteString('# removing load/store on symbol, ') ; WriteAsmName(Instruction[i].Sym) ; WriteLn
         END ;

         (*
             load   foo            =>  nothing
             store  foo            =>  nothing
         *)
         Delete(i+1) ;
         Delete(i) ;
         RETURN( TRUE )
      ELSIF (Instruction[i].Inst=store) AND (Instruction[i+1].Inst=load) AND
            EquModeRegValue(i, i+1) AND (Instruction[i].Sym#NulSym) AND
            (NOT IsSymNeeded(Instruction[i].Sym, Start, End)) AND
            (NOT IsSymUsed(Instruction[i].Sym, i+2, CurrentInst))
      THEN
         IF Debugging
         THEN
            WriteString('# removing store/load on unused symbol, ') ; WriteAsmName(Instruction[i].Sym) ; WriteLn
         END ;

         (*
             we can remove these instructions because Sym is not used outside basic block and
             foo is a temporary variable.

             store  foo            =>  nothing
             load   foo            =>  nothing
         *)
         Delete(i+1) ;
         Delete(i) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END ReduceLoadStore ;


(*
   ReduceAdd - tests for redundant Add operations ie Adding 0 to a variable.
               returns TRUE if the optimization has taken
               place otherwise FALSE.
*)

PROCEDURE ReduceAdd (i: CARDINAL) : BOOLEAN ;
VAR
   t: Instr ;
BEGIN
   IF i+2<=CurrentInst
   THEN
      (*
         i+2 instruction exists
      *)
      IF InstLoadImm(i+1, Zero) AND InstAdd(i+2)
      THEN
         (* Chuck Load imm 0 and Add *)
         Delete(i+2) ;
         Delete(i+1) ;
         RETURN( TRUE )
      ELSIF InstLoadImm(i, Zero) AND InstAdd(i+2)
      THEN
         (* Chuck Load imm 0 and Add *)
         Delete(i+2) ;
         Delete(i) ;
         RETURN( TRUE )
      ELSIF IsInstLoadImm(i) AND InstAdd(i+2) AND (NOT IsInstLoadImm(i+1))
      THEN
         t := Instruction[i] ;
         Instruction[i] := Instruction[i+1] ;
         Instruction[i+1] := t ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END ReduceAdd ;


(*
   ReduceSub - tests for redundant Sub operations ie Subtracting 0
               from  a variable.
               returns TRUE if the optimization has taken
               place otherwise FALSE.
*)

PROCEDURE ReduceSub (i: CARDINAL) : BOOLEAN ;
BEGIN
   IF i+1<=CurrentInst
   THEN
      (*
         i+1 instruction exists
      *)
      IF InstLoadImm(i+1, Zero) AND InstSub(i)
      THEN
         Delete(i+1) ;
         Delete(i) ;
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END ReduceSub ;


(*
   ReduceMult - tests for redundant Mult operations ie
                Multiplying by 0 or 1
                returns TRUE if the optimization has taken
                place otherwise FALSE.
*)

PROCEDURE ReduceMult (i: CARDINAL) : BOOLEAN ;
BEGIN
   IF i+2<=CurrentInst
   THEN
      (*
         i+2 instruction exists
      *)
      IF InstLoadImm(i+1, One) AND InstMult(i+2)
      THEN
         Delete(i+2) ;
         Delete(i+1) ;
         RETURN( TRUE )
      ELSIF InstLoadImm(i, One) AND InstMult(i+2)
      THEN
         Delete(i+2) ;
         Delete(i) ;
         RETURN( TRUE )
      ELSIF InstLoadImm(i, Zero) AND InstMult(i+2)
      THEN
         Delete(i+2) ;
         Delete(i+1) ;
         (* Only Leave the Load imm 0 *)
         RETURN( TRUE )
      ELSIF InstLoadImm(i+1, Zero) AND InstMult(i+2)
      THEN
         Delete(i+2) ;
         Delete(i) ;
         (* Only Leave the Load imm 0 *)
         RETURN( TRUE )
      ELSE
         RETURN( FALSE )
      END
   ELSE
      RETURN( FALSE )
   END
END ReduceMult ;


(*
   ChuckAPU - displays the APU code that remains in the buffer.
*)

PROCEDURE ChuckAPU ;
VAR
   LastLineNo: CARDINAL ;
   i         : CARDINAL ;
BEGIN
   i := 1 ;
   LastLineNo := 0 ;
   WHILE i<=CurrentInst DO
      WITH Instruction[i] DO
         IF (LineNo#0) AND (LastLineNo#LineNo)
         THEN
            LastLineNo := LineNo ;
            LineInfo(LineNo)
         END ;
         WriteString('  ') ;
         WriteInstruction(Inst) ;
         IF (Inst=load) OR (Inst=store)
         THEN
            WriteOperand(i) ;
            IF Debugging
            THEN
               WriteString('  #  <') ;
               IF Sym=NulSym
               THEN
                  WriteString('NulSym>')
               ELSE
                  WriteAsmName(Sym) ; Write('>')
               END
            END
         END ;
         WriteLn
      END ;
      INC(i)
   END ;
   CurrentInst := 0
END ChuckAPU ;


(*
   WriteInstruction - writes an instruction, i.
*)

PROCEDURE WriteInstruction (i: Instructions) ;
BEGIN
   CASE i OF

   sal    :  WriteString('sal') |
   sar    :  WriteString('sar') |
   or     :  WriteString('or') |
   and    :  WriteString('and') |
   xor    :  WriteString('xor') |
   not    :  WriteString('not') |
   add    :  WriteString('add') |
   sub    :  WriteString('sub') |
   mult   :  WriteString('mult') |
   div    :  WriteString('div') |
   mod    :  WriteString('mod') |
   load   :  WriteString('load') |
   store  :  WriteString('store') |
   pr     :  WriteString('pr') |
   equ    :  WriteString('equ') |
   notequ :  WriteString('notequ') |
   less   :  WriteString('less') |
   gre    :  WriteString('gre') |
   lessequ:  WriteString('lessequ') |
   greequ :  WriteString('greequ') |
   pr     :  WriteString('pr') |
   memcopy:  WriteString('memcopy') |
   nop    :  WriteString('nop')

   ELSE
      InternalError('instruction not known', __FILE__, __LINE__)
   END
END WriteInstruction ;


(*
   WriteOperand - writes a load or store operand.
*)

PROCEDURE WriteOperand (i: CARDINAL) ;
BEGIN
   WITH Instruction[i] DO
      WriteMode(From) ;
      WriteMode(To) ;
      IF Inst=load
      THEN
         Write(' ')
      END ;
      CASE Mode OF

      immediate:  WriteString('  $') ; WriteKey(Value) |
      direct   :  WriteString('  ') ; WriteKey(Value) |
      indirect :  IF Reg=no
                  THEN
                     WriteString('  (') ; WriteKey(Value) ; Write(')')
                  ELSE
                     InternalError('not expecting to use this mode for instruction - use indexed (reg) instead',
                                   __FILE__, __LINE__)
                  END |
      indexed  :  WriteString('  ') ; WriteKey(Value) ; Write('(') ; WriteRegister(Reg) ; Write(')') |
      internal :  WriteString('  ') ; WriteRegister(Reg)

      ELSE
         InternalError('unknown Mode', __FILE__, __LINE__)
      END
   END
END WriteOperand ;


(*
   WriteMode - writes the mode of the operand.
*)

PROCEDURE WriteMode (m: Modes) ;
BEGIN
   CASE m OF

   byte :  WriteString('.b') |
   word :  WriteString('.w') |
   dword:  WriteString('.d')

   ELSE
      InternalError('unknown mode', __FILE__, __LINE__)
   END
END WriteMode ;


(*
   WriteRegister - writes the register, r.
*)

PROCEDURE WriteRegister (r: Registers) ;
BEGIN
   CASE r OF

   ra: WriteString('%ra') |
   rb: WriteString('%rb') |
   rc: WriteString('%rc') |
   rd: WriteString('%rd') |
   pc: WriteString('%pc') |
   sp: WriteString('%sp') |
   em: WriteString('%em') |
   im: WriteString('%im') |
   st: WriteString('%st') |
   no:

   ELSE
      InternalError('unknown register', __FILE__, __LINE__)
   END
END WriteRegister ;


(*
   EquModeRegValue - returns true if all Mode, Reg, Value of Instruction
                     i and j are the same.
*)

PROCEDURE EquModeRegValue(i, j: CARDINAL) : BOOLEAN ;
BEGIN
   RETURN(
          (Instruction[i].Mode =Instruction[j].Mode)  AND
          (Instruction[i].Reg  =Instruction[j].Reg)   AND
          (Instruction[i].Value=Instruction[j].Value) AND
          (Instruction[i].From =Instruction[j].From)  AND
          (Instruction[i].To   =Instruction[j].To)
         )
END EquModeRegValue ;


(*
   InstLoadImm - tests whether instruction, i, loads an immediate
                 value, Val. The boolean test result is returned.
*)

PROCEDURE InstLoadImm (i: CARDINAL; Val: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( IsInstLoadImm(i) AND (Value=Val) )
   END
END InstLoadImm ;


(*
   IsInstLoadImm - tests whether instruction, i, loads an immediate
                   value.
*)

PROCEDURE IsInstLoadImm (i: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( (Inst=load) AND (Mode=immediate) )
   END
END IsInstLoadImm ;


(*
   InstAdd - tests whether instruction, i, is an Add instruction.
             The boolean test result is returned.
*)

PROCEDURE InstAdd (i: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( Inst=add )
   END
END InstAdd ;


(*
   InstSub - tests whether instruction, i, is a Sub instruction.
             The boolean test result is returned.
*)

PROCEDURE InstSub (i: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( Inst=sub )
   END
END InstSub ;


(*
   InstMult - tests whether instruction, i, is a Mult instruction.
              The boolean test result is returned.
*)

PROCEDURE InstMult (i: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( Inst=mult )
   END
END InstMult ;


(*
   InstDiv - tests whether instruction, i, is a Div instruction.
             The boolean test result is returned.
*)

PROCEDURE InstDiv (i: CARDINAL) : BOOLEAN ;
BEGIN
   WITH Instruction[i] DO
      RETURN( Inst=div )
   END
END InstDiv ;


(*
   Delete - removes an Instruction i from the code buffer.
*)

PROCEDURE Delete (i: CARDINAL) ;
VAR
   j: CARDINAL ;
BEGIN
   IF i<=CurrentInst
   THEN
      IF i<CurrentInst
      THEN
         j := i+1 ;
         WHILE j<=CurrentInst DO
            Instruction[i] := Instruction[j] ;
            INC(i) ;
            INC(j)
         END
      END ;
      DEC(CurrentInst)
   END
END Delete ;


BEGIN
   (* No Instructions in the buffer *)
   CurrentInst   := 0 ;
   Init
END M2GenAPU.
