IMPLEMENTATION MODULE M2Time ;


FROM M2GenAPU IMPORT Time ;
FROM M2Quads IMPORT QuadOperator, GetQuad ;

FROM SymbolTable IMPORT IsProcedureVariable, IsProcType, IsTemporary,
                        GetType,
                        ModeOfAddr, GetMode, IsProcedure,
                        PushOffset, PushSize ;

FROM M2ALU IMPORT PopCard, PopInt ;
FROM M2Base IMPORT Cardinal ;
FROM M2Lexical IMPORT InternalError ;
FROM M2Debug IMPORT Assert ;
FROM StrIO IMPORT WriteString, WriteLn ;


(*
   Throughout the MC68000 coding is assumed to be without any optimization.
   :-(
*)

(*
   Rules for Quadruples and Modes
   ==============================

   Symbols have four modes

   Mode = (None, Value, Offset, Immediate)

   A symbol can only change its mode from:
   None -> [Value, Offset, Immediate]
     any other alteration should be regarded as an error.


   Description of Mode
   ===================

   Sym<O> => Offset    - the value of symbol is contained at location Offset.
   Sym<V> => Value     - the symbol value is referenced indirectly via the
                         memory location offset.
   Sym<I> => Immediate - symbol is a constant and the value is in Symbol Table.


   Rules
   =====

   All program declared variables are given the mode, Offset.
   All constants have mode, Immediate.

   Operators
   =========

------------------------------------------------------------------------------
   Array Operators
------------------------------------------------------------------------------
   Sym<I>   Base   a            Delivers a constant result if a is a
                                Global variable. If a is a local variable
                                then the Frame pointer needs to be added.
                                Base yields the effective location in memory
                                of, a, array [0,0, .. ,0] address.
   Sym<I>   ElementSize 1       Always delivers a constant. The number
                                indicates which specified element is chosen.
                                ElementSize is the TypeSize for that element.
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   Addr Operator  - contains the address of a variable - may need to add
------------------------------------------------------------------------------
   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   + - / * Operators
------------------------------------------------------------------------------
   Sym1<I>   +      Sym2<I>  Sym3<I>  meaning Sym1<I> := Sym2<I> + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<I>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<O>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<O>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<V>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
------------------------------------------------------------------------------
   Base Operator
------------------------------------------------------------------------------
   Sym1<O>   Base  Sym2   Sym3<O>     meaning     Mem[Sym1<I>] := Sym3<I>
   Sym1<V>   Base  Sym2   Sym3<O>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Sym3<I>
   Sym1<O>   Base  Sym2   Sym3<V>     meaning     Mem[Sym1<I>] := Mem[Sym3<I>]
   Sym1<V>   Base  Sym2   Sym3<V>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Mem[Sym3<I>]
                   Sym2 is the array type
------------------------------------------------------------------------------

Notes

1)   All references to memory Mem[Sym1<I>] may need to add the base
     pointer of the activation record if the symbol is a procedure variable.

2)   Addr provides the address of variables and therefore can be used to
     perform the pseudo system call function ADR( variable ), provided
     one obeys rule (1).
*)


(*
   All Times are in terms of T States.
   All Instructions are using long format ( .L )
*)


CONST
   ADDLImmDnTIME  =  16 ;
   ADDLDnDnTIME   =   8 ;
   ADDLImmAnTIME  =  16 ;
   ADDQLImmAnTIME =   8 ;

   BEQTIME        =  11 ;

   CMPLDnDnTIME   =   8 ;

   DIVLDnDnTIME   = 140 ;

   JMPLMemTIME    =  10 ;

   JSRMemTIME     =  18 ;
   JSRAnTIME      =  16 ;

   LINKRnImmTIME  =  16 ;

   MOVELMemDnTIME =  20 ;
   MOVELDnMemTIME =  20 ;
   MOVELDnIdxTIME =  16 ;
   MOVELIdxDnTIME =  16 ;
   MOVELDnDnTIME  =   4 ;
   MOVELImmDnTIME =  12 ;
   MOVELDnSPTIME  =   4 ;
   MOVELIdrDnTIME =  12 ;
   MOVELDnIdrTIME =  12 ;

   MULTLDnDnTIME  =  70 ;

   RTSTIME        =  16 ;

   SUBLDnDnTIME   =   8 ;

   UNLKTIME       =  12 ;

   MaxRegs        = 1000 ;

VAR
   Regs  : ARRAY [0..MaxRegs] OF RECORD
                                    Var     : CARDINAL ;
                                    Modified: BOOLEAN ;
                                 END ;
   RegPtr: CARDINAL ;

   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;

   CurTime: CARDINAL ;    (* Accumulated time for the current quadruple *)


(* %%%FORWARD%%%
PROCEDURE ADDLDnDn ; FORWARD ;
PROCEDURE ADDLImmAn ; FORWARD ;
PROCEDURE ADDLImmDn ; FORWARD ;
PROCEDURE ADDQLImmAn ; FORWARD ;
PROCEDURE AtEndOfProcedure () : BOOLEAN ; FORWARD ;
PROCEDURE BEQ ; FORWARD ;
PROCEDURE CMPLDnDn ; FORWARD ;
PROCEDURE DIVLDnDn ; FORWARD ;
PROCEDURE DoAdd ; FORWARD ;
PROCEDURE DoAddr ; FORWARD ;
PROCEDURE DoBase ; FORWARD ;
PROCEDURE DoBecomes ; FORWARD ;
PROCEDURE DoCall ; FORWARD ;
PROCEDURE DoCopy ; FORWARD ;
PROCEDURE DoDiv ; FORWARD ;
PROCEDURE DoElementSize ; FORWARD ;
PROCEDURE DoFunctValue ; FORWARD ;
PROCEDURE DoGoto ; FORWARD ;
PROCEDURE DoIfEqu ; FORWARD ;
PROCEDURE DoIfGre ; FORWARD ;
PROCEDURE DoIfGreEqu ; FORWARD ;
PROCEDURE DoIfLess ; FORWARD ;
PROCEDURE DoIfLessEqu ; FORWARD ;
PROCEDURE DoIfNotEqu ; FORWARD ;
PROCEDURE DoKillLocalVar ; FORWARD ;
PROCEDURE DoMult ; FORWARD ;
PROCEDURE DoNewLocalVar ; FORWARD ;
PROCEDURE DoOffset ; FORWARD ;
PROCEDURE DoParam ; FORWARD ;
PROCEDURE DoReturn ; FORWARD ;
PROCEDURE DoReturnValue ; FORWARD ;
PROCEDURE DoSub ; FORWARD ;
PROCEDURE Exists (Op: CARDINAL) : BOOLEAN ; FORWARD ;
PROCEDURE FetchReg (Op: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE FlushRegs ; FORWARD ;
PROCEDURE JMPLMem ; FORWARD ;
PROCEDURE JSRLAn ; FORWARD ;
PROCEDURE JSRLMem ; FORWARD ;
PROCEDURE LINKRnImm ; FORWARD ;
PROCEDURE LargeCopy ; FORWARD ;
PROCEDURE LoadReg (Op: CARDINAL) : CARDINAL ; FORWARD ;
PROCEDURE MOVELDnDn ; FORWARD ;
PROCEDURE MOVELDnIdr ; FORWARD ;
PROCEDURE MOVELDnIdx ; FORWARD ;
PROCEDURE MOVELDnMem ; FORWARD ;
PROCEDURE MOVELDnSP ; FORWARD ;
PROCEDURE MOVELIdrDn ; FORWARD ;
PROCEDURE MOVELIdxDn ; FORWARD ;
PROCEDURE MOVELImmDn ; FORWARD ;
PROCEDURE MOVELMemDn ; FORWARD ;
PROCEDURE MULTLDnDn ; FORWARD ;
PROCEDURE RTS ; FORWARD ;
PROCEDURE SUBLDnDn ; FORWARD ;
PROCEDURE SetReg (Op: CARDINAL; Altered: BOOLEAN) ; FORWARD ;
PROCEDURE SmallCopy ; FORWARD ;
PROCEDURE StoreReg (r, Op: CARDINAL) ; FORWARD ;
PROCEDURE UNLK ; FORWARD ;
   %%%FORWARD%%% *)


(*
   TimeBeforeQuad - sums the time the processor would spend executing a
                    quadruple, QuadNo.
                    StartBB and EndBB determine the range of the Basic Block.
                    Multiway branch depending upon the QuadNo.
                    All branches are timed before actually encoded as
                    APU code.
*)

PROCEDURE TimeBeforeQuad (QuadNo, StartBB, EndBB: CARDINAL) ;
BEGIN
   CurTime := 0 ;
   IF QuadNo=StartBB
   THEN
      RegPtr := 0
   END ;
   GetQuad(QuadNo, Op, Op1, Op2, Op3) ;
   CASE Op OF

   NegateOp      : DoNegate |
   AddOp         : DoAdd |
   SubOp         : DoSub |
   MultOp        : DoMult |
   ModOp,
   DivOp         : DoDiv |
   NewLocalVarOp : |
   KillLocalVarOp: DoKillLocalVar |
   CallOp        : DoCall |
   ParamOp       : DoParam |
   AddrOp        : DoAddr |
   ConvertOp,
   CoerceOp,
   BecomesOp     : DoBecomes |
   GotoOp        : DoGoto |
   IfEquOp       : DoIfEqu |
   IfNotEquOp    : DoIfNotEqu |
   IfGreEquOp    : DoIfGreEqu |
   IfLessEquOp   : DoIfLessEqu |
   IfLessOp      : DoIfLess |
   IfGreOp       : DoIfGre |
   BaseOp        : DoBase |
   ElementSizeOp : DoElementSize |
   OffsetOp      : DoOffset |
   FunctValueOp  : DoFunctValue |
   ReturnOp      : DoReturn |
   ReturnValueOp : DoReturnValue |
   UnboundedOp   : DoUnbounded |
   DummyOp       : |

   (* And the Compiler Flag Pseudo Quadruples *)
   StartDefFileOp,
   StartModFileOp,
   StartOp       ,
   EndOp         ,
   EndFileOp     ,
   LineNumberOp  ,
   InlineOp      ,
   CodeOnOp      ,
   CodeOffOp     ,
   ProfileOnOp   ,
   ProfileOffOp  ,
   OptimizeOnOp  ,
   OptimizeOffOp : ;

   ELSE
      InternalError('not implemented yet', __FILE__, __LINE__)
   END ;
   Time(CurTime)
END TimeBeforeQuad ;


(*
   TimeAfterQuad - sums the time the processor would spend executing a
                   quadruple, QuadNo.
                   StartBB and EndBB determine the range of the Basic Block.
                   Multiway branch depending upon the QuadNo.
                   All branches are timed after being encoded as
                   APU code.
*)

PROCEDURE TimeAfterQuad (QuadNo, StartBB, EndBB: CARDINAL) ;
BEGIN
   CurTime := 0 ;
   GetQuad(QuadNo, Op, Op1, Op2, Op3) ;
   CASE Op OF

   NegateOp      : |
   AddOp         : |
   SubOp         : |
   MultOp        : |
   ModOp         : |
   DivOp         : |
   NewLocalVarOp : DoNewLocalVar |
   KillLocalVarOp: |
   CallOp        : |
   ParamOp       : |
   AddrOp        : |
   ConvertOp     : |
   CoerceOp      : |
   BecomesOp     : |
   GotoOp        : |
   IfEquOp       : |
   IfNotEquOp    : |
   IfGreEquOp    : |
   IfLessEquOp   : |
   IfLessOp      : |
   IfGreOp       : |
   BaseOp        : |
   ElementSizeOp : |
   OffsetOp      : |
   FunctValueOp  : |
   ReturnOp      : |
   ReturnValueOp : |
   UnboundedOp   : |
   DummyOp       : |

   (* And the Compiler Flag Pseudo Quadruples *)
   StartDefFileOp,
   StartModFileOp,
   StartOp       ,
   EndOp         ,
   EndFileOp     ,
   LineNumberOp  ,
   InlineOp      ,
   CodeOnOp      ,
   CodeOffOp     ,
   ProfileOnOp   ,
   ProfileOffOp  ,
   OptimizeOnOp  ,
   OptimizeOffOp : ;

   ELSE
      InternalError('not implemented yet', __FILE__, __LINE__)
   END ;
   IF QuadNo=EndBB
   THEN
      FlushRegs
   END ;
   Time(CurTime)
END TimeAfterQuad ;


(*
   AddOp - briefly:
           AddOp   Op1  Op2  Op3
           where Op1 := Op2 + Op3
*)

PROCEDURE DoAdd ;
VAR
   R0, R1: CARDINAL ;
BEGIN
   (*
      Code:

      R0 := Op2 ;
      R1 := Op3 ;
      ADD.W  R0,R1
      Op1 := R0
   *)
   R0 := LoadReg(Op2) ;
   R1 := LoadReg(Op3) ;
   ADDLDnDn ;
   StoreReg(R0, Op1)
END DoAdd ;


(*
   NegateOp - briefly:
              NegateOp   Op1  Op2  Op3
              where Op1 :=  - Op3
*)

PROCEDURE DoNegate ;
VAR
   R0: CARDINAL ;
BEGIN
   (*
      Code:

      R0 := Op3 ;
      NEG.W  R0
      Op1 := R0
   *)
   R0 := LoadReg(Op3) ;
   ADDLDnDn ;
   StoreReg(R0, Op1)
END DoNegate ;


(*
   SubOp - briefly:
           SubOp   Op1  Op2  Op3
           where Op1 := Op2 - Op3
*)

PROCEDURE DoSub ;
VAR
   R0, R1: CARDINAL ;
BEGIN
   (*
      Code:

      R0 := Op2 ;
      R1 := Op3 ;
      SUB.L  R0,R1
      Op1 := R0
   *)
   R0 := LoadReg(Op2) ;
   R1 := LoadReg(Op3) ;
   SUBLDnDn ;
   StoreReg(R0, Op1)
END DoSub ;


(*
   DivOp - briefly:
           DivOp   Op1  Op2  Op3
           where Op1 := Op2 DIV Op3
*)

PROCEDURE DoDiv ;
VAR
   R0, R1: CARDINAL ;
BEGIN
   (*
      Code:

      R0 := Op2 ;
      R1 := Op3 ;
      DIV.L  R0,R1
      Op1 := R0
   *)
   R0 := LoadReg(Op2) ;
   R1 := LoadReg(Op3) ;
   DIVLDnDn ;
   StoreReg(R0, Op1)
END DoDiv ;


(*
   MultOp - briefly:
            MultOp   Op1  Op2  Op3
            where Op1 := Op2 * Op3
*)

PROCEDURE DoMult ;
VAR
   R0, R1: CARDINAL ;
BEGIN
   (*
      Code:

      R0 := Op2 ;
      R1 := Op3 ;
      MULT.L  R0,R1
      Op1 := R0
   *)
   R0 := LoadReg(Op2) ;
   R1 := LoadReg(Op3) ;
   MULTLDnDn ;
   StoreReg(R0, Op1)
END DoMult ;


(*
   DoCall - briefly:
            CallOp    Op3
*)

PROCEDURE DoCall ;
VAR
   R0: CARDINAL ;
BEGIN
   IF IsProcedure(Op3)
   THEN
      (*
         Code:

         JSR.L  proc
      *)
      JSRLMem
   ELSE
      Assert(IsProcType(GetType(Op3))) ;
      (*
         Code:

         MOVE.L  op,A0
         JSR.L   A0
      *)
      R0 := LoadReg(Op3) ;
      JSRLAn
   END
END DoCall ;


(*
   DoGoto - briefly:
            GotoOp   Op3
*)

PROCEDURE DoGoto ;
BEGIN
   (*
      JMP.L  Op3
   *)
   JMPLMem
END DoGoto ;


(*
   DoReturn - simply return from subroutine.
*)

PROCEDURE DoReturn ;
BEGIN
   RTS
END DoReturn ;


PROCEDURE AtEndOfProcedure () : BOOLEAN ;
BEGIN
   RETURN( (Op=KillLocalVarOp) OR (Op=ReturnOp) )
END AtEndOfProcedure ;


PROCEDURE FlushRegs ;
VAR
   i, r: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<RegPtr DO
      (*
         Dont bother saving tempories or variables that have not been
         modified or variables if we are at the end of a procedure.
      *)
      IF (NOT IsTemporary(Regs[i].Var)) AND (Regs[i].Modified) AND
         (NOT AtEndOfProcedure())
      THEN
         r := FetchReg(Regs[i].Var)
      END ;
      INC(i)
   END ;
   RegPtr := 0
END FlushRegs ;


PROCEDURE SetReg (Op: CARDINAL; Altered: BOOLEAN) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE (i<RegPtr) AND (Regs[i].Var#Op) DO
      INC(i)
   END ;
   IF (i<RegPtr) AND Altered
   THEN
      Regs[i].Modified := Altered
   ELSE
      Regs[RegPtr].Var := Op ;
      Regs[RegPtr].Modified := Altered ;
      INC(RegPtr) ;
      IF RegPtr>14
      THEN
         WriteString('// Register Overflow') ; WriteLn
      END
   END
END SetReg ;


PROCEDURE Exists (Op: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE i<RegPtr DO
      IF Regs[i].Var=Op
      THEN
         RETURN( TRUE )
      ELSE
         INC(i)
      END
   END ;
   RETURN( FALSE )
END Exists ;


(*
   Loadreg - loads a register with the operand, Op.
*)

PROCEDURE LoadReg (Op: CARDINAL) : CARDINAL ;
BEGIN
   (*
      Quick hack is to assume that all tempories are always in Registers
      This should then create code similar to the SUN M-2 compiler.
   *)
   IF Exists(Op)
   THEN
      (*
         Register Magically Found...
      *)
      RETURN( 0 )
   ELSE
      SetReg(Op, FALSE) ;
      RETURN( FetchReg(Op) )
   END
END LoadReg ;


PROCEDURE FetchReg (Op: CARDINAL) : CARDINAL ;
BEGIN
   IF GetMode(Op)=ImmediateValue
   THEN
      (*
         MOVE.W  #Op[i],R0
      *)
      MOVELImmDn
   ELSE
      (*
         Four combinations

         Local or Non Local variables

         Referenced by Value or Offset
      *)
      IF IsProcedureVariable(Op)
      THEN
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i](A6),R0
            *)
            MOVELIdxDn
         ELSE
            (*
               MOVE.W  Op[i](A6),A0
               MOVE.W  (A0),D0
            *)
            MOVELIdxDn ;
            MOVELIdrDn
         END
      ELSE
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i],R0
            *)
            MOVELMemDn
         ELSE
            (*
               MOVE.W  Op[i],A0
               MOVE.W  (A0),D0
            *)
            MOVELMemDn ;
            MOVELIdrDn
         END
      END
   END ;
   RETURN( 0 )
END FetchReg ;


(*
   StoreReg - stores a register, r, in the operand, Op.
*)

PROCEDURE StoreReg (r, Op: CARDINAL) ;
BEGIN
   IF Exists(Op)
   THEN
      (*
         Register Magically Stored...
      *)
   ELSE
      SetReg(Op, TRUE)
   END
END StoreReg ;


(*
   DoAddr - briefly
            Addr   Op1   Op3

            Op1 := ADR(Op3)
*)

PROCEDURE DoAddr ;
BEGIN
   SetReg(Op3, FALSE) ;  (* Op3 in a register *)
   (*
      Four combinations

      Local or Non Local variables

      Referenced by Value or Offset
   *)
   IF IsTemporary(Op3)
   THEN
   ELSE
      IF IsProcedureVariable(Op3)
      THEN
         IF GetMode(Op3)=RightValue
         THEN
            (*
               MOVE.W  A6,D0
               ADDQI   #Op[i],D0
            *)
            MOVELDnDn ;
            PushOffset(Op3) ;
            IF PopInt()#0
            THEN
               ADDQLImmAn
            END
         ELSE
            (*
               MOVE.W  Op[i](A6),D0
            *)
            MOVELIdxDn
         END
      ELSE
         IF GetMode(Op3)=RightValue
         THEN
            (*
               MOVE.W  #Op[i],D0
            *)
            MOVELImmDn
         ELSE
            (*
               MOVE.W  Op[i],D0
            *)
            MOVELMemDn
         END
      END
   END ;
   StoreReg(0, Op1)
END DoAddr ;


(*
   DoBase - calculates the base of an array.
            BaseOp   Op1   Op3

            where Op1 now contains the address of array Op3
            ie effectively ADR(Op3[0,0,0....])
*)

PROCEDURE DoBase ;
BEGIN
   IF IsProcedureVariable(Op3)
   THEN
      (*
         Code:

         MOVE.L  A6,D0
         ADDI  #Base,D0
      *)
      MOVELDnDn ;
      ADDLImmDn
   ELSE
      (*
         Code:

         MOVE.L  #Base,D0
      *)
      MOVELImmDn
   END ;
   StoreReg(0, Op1)
END DoBase ;


(*
   DoOffset - sets Op1 to the Value of Op3,
              used to calculate the array subscript address.
              OffsetOp   Op1   Op3
*)

PROCEDURE DoOffset ;
BEGIN
   StoreReg(LoadReg(Op3), Op1)
END DoOffset ;


(*
   DoBecomes - performs the operation Op1 := Op3
*)

PROCEDURE DoBecomes ;
VAR
   by, i, Size: CARDINAL ;
BEGIN
   PushSize(Cardinal) ;
   PushSize(Op1) ;
   IF PopCard()=PopCard()
   THEN
      StoreReg(LoadReg(Op3), Op1)
   ELSE
      PushSize(Cardinal) ;
      by := PopCard() ;
      PushSize(Op1) ;
      Size := PopCard() ;
      i := 0 ;
      WHILE i<Size DO
         StoreReg(LoadReg(Op3), Op1) ;
         INC(i, by)
      END
   END
END DoBecomes ;


PROCEDURE DoCopy ;
BEGIN
   PushSize(Cardinal) ;
   PushSize(Op1) ;
   IF PopCard()=PopCard()
   THEN
      SmallCopy
   ELSE
      LargeCopy
   END
END DoCopy ;


PROCEDURE SmallCopy ;
BEGIN
   (*
      Eight combinations

      Op3 Local or Non Local variables

      Op3 Referenced by Value or Offset

      Op1 Local or Procedural
   *)
   Assert((GetMode(Op3)=RightValue) OR (GetMode(Op3)=LeftValue) OR (GetMode(Op3)=ImmediateValue)) ;
   IF IsProcedureVariable(Op3)
   THEN
      IF GetMode(Op3)=RightValue
      THEN
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i](A6),D0
               MOVE.L  D0,Op1[i](A6)
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELIdxDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnIdx
            END
         ELSE
            (*
               MOVE.L  Op3[i](A6),D0
               MOVE.L  D0,Op1[i]
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELIdxDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnMem
            END
         END
      ELSE
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i](A6),A0
               MOVE.L  (A0),D0
               MOVE.L  D0,Op1[i](A6)
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELIdxDn
            END ;
            MOVELIdrDn ;
            IF Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnIdx
            END
         ELSE
            (*
               MOVE.L  Op3[i](A6),A0
               MOVE.L  (A0),D0
               MOVE.L  D0,Op1[i]
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELIdxDn
            END ;
            MOVELIdrDn ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnMem
            END
         END
      END
   ELSE
      IF GetMode(Op3)=RightValue
      THEN
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i],D0
               MOVE.L  D0,(A6)
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELMemDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnIdr
            END
         ELSE
            (*
               MOVE.L  Op3[i],D0
               MOVE.L  D0,Op1[i]
            *)
            IF NOT Exists(Op3)
            THEN
               SetReg(Op3, FALSE) ;
               MOVELMemDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnMem
            END
         END
      ELSE
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i],A0
               MOVE.L  (A0),D0        or MOVE.L #op3, D0


               MOVE.L  D0,Op1(A6)
            *)
            IF GetMode(Op3)=ImmediateValue
            THEN
               IF NOT Exists(Op3)
               THEN
                  SetReg(Op3, FALSE) ;
                  MOVELImmDn
               END
            ELSE
               IF NOT Exists(Op3)
               THEN
                  SetReg(Op3, FALSE) ;
                  MOVELMemDn
               END ;
               MOVELIdrDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnIdx
            END
         ELSE
            (*
               MOVE.L  Op3[i],A0
               MOVE.L  (A0),D0        or MOVE.L #op3, D0


               MOVE.L  D0,Op1[i]
            *)
            IF GetMode(Op3)=ImmediateValue
            THEN
               IF NOT Exists(Op3)
               THEN
                  SetReg(Op3, FALSE) ;
                  MOVELImmDn
               END
            ELSE
               IF NOT Exists(Op3)
               THEN
                  SetReg(Op3, FALSE) ;
                  MOVELMemDn
               END ;
               MOVELIdrDn
            END ;
            IF NOT Exists(Op1)
            THEN
               SetReg(Op1, TRUE) ;
               MOVELDnMem
            END
         END
      END
   END
END SmallCopy ;


PROCEDURE LargeCopy ;
VAR
   Size, i, by: CARDINAL ;
BEGIN
   PushSize(Cardinal) ;
   by := PopCard() ;
   PushSize(Op1) ;
   Size := PopCard() ;
   i := 0 ;
   WHILE i<Size DO
      SmallCopy ;
      INC(i, by)
   END
END LargeCopy ;


PROCEDURE DoIfEqu ;
VAR
   R0, R1: CARDINAL ;
BEGIN
   R0 := LoadReg(Op1) ;
   R1 := LoadReg(Op2) ;
   (*
      CMP.L  D0,D1
      BEQ    Label
   *)
   CMPLDnDn ;
   BEQ
END DoIfEqu ;


PROCEDURE DoIfNotEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfNotEqu ;


PROCEDURE DoIfLess ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfLess ;


PROCEDURE DoIfLessEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfLessEqu ;


PROCEDURE DoIfGre ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfGre ;


PROCEDURE DoIfGreEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfGreEqu ;


PROCEDURE DoParam ;
VAR
   R0: CARDINAL ;
BEGIN
   R0 := LoadReg(Op3) ;
   (*
      MOVE.L  R0,(sp)-
   *)
   MOVELDnSP
END DoParam ;


PROCEDURE DoNewLocalVar ;
BEGIN
   LINKRnImm ;
   ADDLImmDn
END DoNewLocalVar ;


PROCEDURE DoKillLocalVar ;
BEGIN
   (*
      ADDQL  #-,sp
      UNLK
   *)
   ADDQLImmAn ;
   UNLK
END DoKillLocalVar ;


PROCEDURE DoFunctValue ;
BEGIN
   (*
      Op3 := R0
   *)
   StoreReg(0, Op3)
END DoFunctValue ;


PROCEDURE DoReturnValue ;
VAR
   R0: CARDINAL ;
BEGIN
   R0 := LoadReg(Op3)
END DoReturnValue ;


PROCEDURE DoUnbounded ;
BEGIN
   (* this requires some thought *)
END DoUnbounded ;


PROCEDURE DoElementSize ;
BEGIN
   InternalError('should never get here', __FILE__, __LINE__)
END DoElementSize ;


PROCEDURE ADDLImmDn ;
BEGIN
   INC(CurTime, ADDLImmDnTIME)
END ADDLImmDn ;


PROCEDURE ADDLDnDn ;
BEGIN
   INC(CurTime, ADDLDnDnTIME)
END ADDLDnDn ;


PROCEDURE ADDLImmAn ;
BEGIN
   INC(CurTime, ADDLImmAnTIME)
END ADDLImmAn ;

PROCEDURE ADDQLImmAn ;
BEGIN
   INC(CurTime, ADDQLImmAnTIME)
END ADDQLImmAn ;


PROCEDURE CMPLDnDn ;
BEGIN
   INC(CurTime, CMPLDnDnTIME)
END CMPLDnDn ;


PROCEDURE SUBLDnDn ;
BEGIN
   INC(CurTime, SUBLDnDnTIME)
END SUBLDnDn ;


PROCEDURE MULTLDnDn ;
BEGIN
   INC(CurTime, MULTLDnDnTIME)
END MULTLDnDn ;


PROCEDURE DIVLDnDn ;
BEGIN
   INC(CurTime, DIVLDnDnTIME)
END DIVLDnDn ;


PROCEDURE UNLK ;
BEGIN
   INC(CurTime, UNLKTIME)
END UNLK ;


PROCEDURE LINKRnImm ;
BEGIN
   INC(CurTime, LINKRnImmTIME)
END LINKRnImm ;


PROCEDURE MOVELMemDn ;
BEGIN
   INC(CurTime, MOVELMemDnTIME)
END MOVELMemDn ;


PROCEDURE MOVELDnMem ;
BEGIN
   INC(CurTime, MOVELDnMemTIME)
END MOVELDnMem ;


PROCEDURE MOVELIdxDn ;
BEGIN
   INC(CurTime, MOVELIdxDnTIME)
END MOVELIdxDn ;


PROCEDURE MOVELDnIdx ;
BEGIN
   INC(CurTime, MOVELDnIdxTIME)
END MOVELDnIdx ;


PROCEDURE MOVELImmDn ;
BEGIN
   INC(CurTime, MOVELImmDnTIME)
END MOVELImmDn ;


PROCEDURE MOVELDnSP ;
BEGIN
   INC(CurTime, MOVELDnSPTIME)
END MOVELDnSP ;


PROCEDURE MOVELDnDn ;
BEGIN
   INC(CurTime, MOVELDnDnTIME)
END MOVELDnDn ;


PROCEDURE MOVELIdrDn ;
BEGIN
   INC(CurTime, MOVELIdrDnTIME)
END MOVELIdrDn ;


PROCEDURE MOVELDnIdr ;
BEGIN
   INC(CurTime, MOVELDnIdrTIME)
END MOVELDnIdr ;


PROCEDURE JSRLMem ;
BEGIN
   INC(CurTime, JSRMemTIME)
END JSRLMem ;


PROCEDURE JSRLAn ;
BEGIN
   INC(CurTime, JSRAnTIME)
END JSRLAn ;


PROCEDURE RTS ;
BEGIN
   INC(CurTime, RTSTIME)
END RTS ;


PROCEDURE BEQ ;
BEGIN
   INC(CurTime, RTSTIME)
END BEQ ;


PROCEDURE JMPLMem ;
BEGIN
   INC(CurTime, JMPLMemTIME)
END JMPLMem ;



END M2Time.
