MODULE m2f ;

(*
   Author     : Gaius Mulley
   Title      : m2f
   Date       : Thu May 23 09:02:09 gmt 1991
              : [$Date: 1998/12/18 15:42:16 $]
   SYSTEM     : UNIX SUN and Logitech M2
   Description: Provides the front end to the gm2 compiler.
   Version    : $Revision: 1.6 $
*)

(*
   Log        : $Log: m2f.mod,v $
   Log        : Revision 1.6  1998/12/18 15:42:16  gaius
   Log        : major alterations include:
   Log        :    + better 586/686 pipelining
   Log        :    + faster parameter passing (using words not bytes). Similar to gcc.
   Log        :    + alignment of parameters, data and stack. Significant speed improvements
   Log        :      for both 586 and 686 targets.
   Log        :    + 2 bugs found in M2GenDyn486
   Log        :    + pipelining now includes more instructions and understands the cache.
   Log        :      Pipelining can be used with the is32 code generator to predict performance
   Log        :      (accurate to within 8%)
   Log        :    + flag -pedantic will issue warnings if excessively slow code is being issued.
   Log        :      (Array parameters etc)
   Log        :    + is32 (apu) code generator has been modified slightly to bring it in line with
   Log        :      pentium alignment. Also a few code generator bugs fixed.
   Log        :    + new option -Oworst included, which will perform the worst pipelining possible.
   Log        :      Useful only as a comparison.
   Log        :    - downside is that the new paramater mechanism and data alignment has not
   Log        :      been stressed enough.
   Log        :
   Log        : Revision 1.5  1998/02/11 16:13:49  gaius
   Log        :     (i)  more pipeline bugs (test and cmp being at the end of a pipeline)
   Log        :     (ii) also fixed some error messages (string copied into too small a dest).
   Log        :          Fixed m2f so that it will link a module expressed as: m2f -l test.o
   Log        :
   Log        : Revision 1.4  1998/01/21 13:26:08  gaius
   Log        : fixed bug in static code generator. Improved smp code generator and
   Log        : making sure pentium execution times are accurately represented in smp
   Log        : instructions.
   Log        :
   Log        : Revision 1.3  1997/04/28 12:35:39  gaius
   Log        : fixed many of the errors detected by Stuart and his students.
   Log        : These include: untermined comments at the beginning of the module,
   Log        : identifiers must be exported and parameter types are checked.
   Log        :
   Log        : Revision 1.2  1996/11/28 15:47:12  gaius
   Log        : major modifications to M2Quads the indirection quadruples have been
   Log        : added and the mode of symbols do not now indicate indirection.
   Log        :
   Log        : many error messages have been tidied up and more errors are caught.
   Log        :
   Log        : the i586 code generator has been started but is in a very early stage
   Log        : (M2GenDyn486.mod)
   Log        :
   Log        : Revision 1.1.1.1  1996/05/28 10:13:02  gaius
   Log        : Modula-2 compiler sources imported
   Log        :
*)

IMPORT CmdArgs ;
FROM SYSTEM IMPORT ADR ;
FROM libc IMPORT system, exit, getpid ;
FROM StrLib IMPORT StrEqual, StrCopy, StrConCat, StrLen ;
FROM Args IMPORT Narg, GetArg ;
FROM FIO IMPORT Exists ;
FROM StrIO IMPORT WriteString, WriteLn ;
FROM M2FileName IMPORT CalculateFileName ;
FROM ASCII IMPORT nul ;
FROM M2Configure IMPORT SystemIsXenix ;
FROM M2Defaults IMPORT GetOptions ;
FROM NumberIO IMPORT CardToStr ;


(* %%%FORWARD%%%
PROCEDURE BuildCompiler (VAR a: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE DisplayHelp ; FORWARD ;
PROCEDURE StripModuleExtention (VAR a: ARRAY OF CHAR) ; FORWARD ;
PROCEDURE RemoveAssemblerFile ; FORWARD ;
   %%%FORWARD%%% *)

CONST
   Max          = 4096 ;      (* Maximum command line *)

VAR
   Module          : ARRAY [0..Max] OF CHAR ;
   Rest            : ARRAY [0..Max] OF CHAR ;
   SearchPath      : ARRAY [0..Max] OF CHAR ;
   SearchFound     : BOOLEAN ;
   Target          : ARRAY [0..Max] OF CHAR ;
   TargetFound     : BOOLEAN ;
   VerboseFound    : BOOLEAN ;   (* -v option found               *)
   DryRunFound     : BOOLEAN ;   (* -dryrun option found          *)
   Assemble        : BOOLEAN ;   (* FALSE if -S flag is found     *)
   DebugFound      : BOOLEAN ;   (* TRUE if -g flag is found      *)
   LinkFound       : BOOLEAN ;   (* TRUE if -l flag is found      *)
   ProfileFound    : BOOLEAN ;   (* TRUE if -p flag is found      *)
   OrderOnlyFound  : BOOLEAN ;   (* TRUE if -orderonly flag found *)
   OrderFileFound  : BOOLEAN ;   (* TRUE if -orderfile flag found *)
   ListModulesFound: BOOLEAN ;   (* TRUE if -m flag found         *)
   StartupFound    : BOOLEAN ;   (* TRUE if -startup found        *)
   VersionFound    : BOOLEAN ;   (* TRUE if -version found        *)
   GasFound        : BOOLEAN ;   (* TRUE if -gas found (default)  *)
   MasmFound       : BOOLEAN ;   (* TRUE if -masm found           *)
   BasFound        : BOOLEAN ;   (* TRUE if -bas found            *)
   ApuFound        : BOOLEAN ;   (* TRUE if -apu found            *)
   i386Found       : BOOLEAN ;   (* TRUE if -i386 found           *)
   UseSlashTmpFound: BOOLEAN ;   (* TRUE if -tmp is found         *)
   OmitM2RTSFound  : BOOLEAN ;   (* TRUE if -M2RTS is found       *)
   HelpFound       : BOOLEAN ;   (* TRUE if -h is found           *)
   OrderFile       : ARRAY [0..Max] OF CHAR ;
   StartupFile     : ARRAY [0..Max] OF CHAR ;


(*
   ProcessSearchPath - copies the search path into SearchPath and
                       sets the boolean SearchFound.
*)

PROCEDURE ProcessSearchPath (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(SearchPath, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   SearchFound := TRUE
END ProcessSearchPath ;


(*
   ProcessOrderFile - copies the specified order file name into OrderFile
                      and sets the boolean OrderFound.
*)

PROCEDURE ProcessOrderFile (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(OrderFile, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   OrderFileFound := TRUE
END ProcessOrderFile ;


(*
   ProcessStartupFile - copies the specified startup file name into StartupFile
                        and sets the boolean StartupFound.
*)

PROCEDURE ProcessStartupFile (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(StartupFile, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   StartupFound := TRUE
END ProcessStartupFile ;


(*
   ProcessTarget - copies the specified target file into Target
                   and sets the boolean TargetFound.
*)

PROCEDURE ProcessTarget (i: CARDINAL) ;
BEGIN
   IF NOT GetArg(Target, i+1)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   TargetFound := TRUE
END ProcessTarget ;


(*
   ProcessDefault - adds the current argument to the CommandLine.
*)

PROCEDURE ProcessDefault (i: CARDINAL) ;
VAR
   a: ARRAY [0..Max] OF CHAR ;
BEGIN
   StrConCat(Rest, ' ', Rest) ;
   IF NOT GetArg(a, i)
   THEN
      HALT  (* Should always be able to get argument *)
   END ;
   StrConCat(Rest, a, Rest)
END ProcessDefault ;


(*
   ProcessDefaultOptions - adds the current environment option to the CommandLine.
*)

PROCEDURE ProcessDefaultOptions (a: ARRAY OF CHAR; i: CARDINAL) ;
BEGIN
   StrConCat(Rest, ' ', Rest) ;
   StrConCat(Rest, a, Rest)
END ProcessDefaultOptions ;


(*
   IsAnOption - returns TRUE if an option was found and it was legal.
*)

PROCEDURE IsAnOption (a: ARRAY OF CHAR) : BOOLEAN ;
VAR
   IsLegal: BOOLEAN ;
BEGIN
   IF StrEqual(a, '-S')
   THEN
      (* Assemble found *)
      Assemble := FALSE ;      (* Do not assemble assembly file *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-v')
   THEN
      (* Verbose found *)
      VerboseFound := TRUE ;   (* Echo commands before executing *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-version')
   THEN
      Assemble := FALSE ;      (* dont assemble if -version found *)
      VersionFound := TRUE ;   (* version flag found *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-gas')
   THEN
      GasFound := TRUE ;       (* use gas assembler *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-bas')
   THEN
      BasFound := TRUE ;       (* use bas assembler *)
      DebugFound := FALSE ;
      IsLegal := TRUE
   ELSIF StrEqual(a, '-masm')
   THEN
      MasmFound := TRUE ;      (* use masm assembler *)
      DebugFound := FALSE ;
      IF NOT SystemIsXenix
      THEN
         Assemble := FALSE     (* do not assemble file *)
      END ;
      IsLegal := TRUE
   ELSIF StrEqual(a, '-apu')
   THEN
      i386Found := FALSE ;
      ApuFound := TRUE ;       (* generate code for APU processor *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-i386')
   THEN
      i386Found := TRUE ;      (* generate code for i386 processor *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-dryrun')
   THEN
      (* Dryrun found *)
      DryRunFound := TRUE ;    (* Echo commands without executing *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-ng')
   THEN
      DebugFound := FALSE ;    (* no debugging allows user to override env *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-g')
   THEN
      DebugFound := TRUE ;     (* Generate debugging information *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-p')
   THEN
      ProfileFound := TRUE ;   (* Generate profiling code *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-tmp')
   THEN
      UseSlashTmpFound := TRUE ; (* use /tmp for the assembly file *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-M2RTS')
   THEN
      OmitM2RTSFound := TRUE ; (* do NOT automatically include M2RTS for linkage (see gm2l) *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-m')
   THEN
      ListModulesFound := TRUE ; (* List modules when linking *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-l')
   THEN
      LinkFound := TRUE ;      (* Link module *)
      IsLegal := TRUE
   ELSIF StrEqual(a, '-h') OR StrEqual(a, '-help')
   THEN
      HelpFound := TRUE
   ELSE
      IsLegal := FALSE
   END ;
   RETURN( IsLegal )
END IsAnOption ;


(*
   ScanArguments - scans the command line arguments and builds up
                   CommandLine, SearchPath, Target and sets booleans.
*)

PROCEDURE ScanArguments ;
VAR
   i: CARDINAL ;
   a: ARRAY [0..Max] OF CHAR ;
BEGIN
   i := 1 ;
   WHILE GetArg(a, i) DO
      IF StrEqual(a, '-M') AND (Narg()>i)
      THEN
         ProcessSearchPath(i) ;
         INC(i) ; (* SearchPath found *)
      ELSIF StrEqual(a, '-o')
      THEN
         ProcessTarget(i) ;
         INC(i)                   (* Target found *)
      ELSIF StrEqual(a, '-orderfile')
      THEN
         ProcessOrderFile(i) ;
         INC(i)                   (* Order file found for initialization *)
      ELSIF StrEqual(a, '-orderonly')
      THEN
         OrderOnlyFound := TRUE   (* Only generate the order file   *)
      ELSIF StrEqual(a, '-startup')
      THEN
         ProcessStartupFile(i) ;  (* Use specified startup file not crt0.o *)
         INC(i)
      ELSIF IsAnOption(a)
      THEN
      ELSE
         (* append argument to CommandLine *)
         ProcessDefault(i)
      END ;
      INC(i)
   END ;
   (* Last argument MUST be module name *)
   IF NOT ((i>1) AND GetArg(Module, i-1))
   THEN
      WriteString('m2f: last argument must be a module') ; WriteLn ;
      exit(1)
   END ;
   StripModuleExtention(Module) ;
   StripModuleExtention(Rest)
END ScanArguments ;


(*
   ScanEnvironmentArgs - scans the args held in M2OPTIONS for any
      	       	     	 default options.

                         Note that it only scans M2OPTIONS if no flags
                         are issued on the command line.
*)

PROCEDURE ScanEnvironmentArgs ;
VAR
   Options,
   a      : ARRAY [0..Max] OF CHAR ;
   i      : CARDINAL ;
BEGIN
   IF Narg()=2
   THEN
      GetOptions(Options) ;
      i := 0 ;
      WHILE CmdArgs.GetArg(Options, i, a) DO
         IF IsAnOption(a)
         THEN
            (* do nothing as it has been registered *)
         ELSE
            (* append argument to CommandLine *)
            ProcessDefaultOptions(a, i)
         END ;
         INC(i)
      END
   END
END ScanEnvironmentArgs ;



(*
   ExtractExtension - given a filename, a, extract the extension, Ext,
                      with length ExtLength.
*)

PROCEDURE ExtractExtension (VAR a: ARRAY OF CHAR;
                            Ext: ARRAY OF CHAR; ExtLength: CARDINAL) ;
CONST
   MaxExtension    = 3 ;    (* must be the greatest of the constants declared below *)
                            (* (ModExtension and ObjectExtension)                   *)
VAR
   High,
   i   : CARDINAL ;
   ext : ARRAY [0..3] OF CHAR ;
BEGIN
   High := StrLen(a) ;
   IF High>ExtLength
   THEN
      i := 0 ;
      WHILE i<=ExtLength DO
         ext[ExtLength-i] := a[High-i-1] ;
         INC(i)
      END ;
      IF ExtLength<3
      THEN
         ext[ExtLength+1] := nul
      END ;
      IF StrEqual(ext, Ext)
      THEN
         a[High-ExtLength-1] := nul
      END
   END
END ExtractExtension ;


(*
   StripModuleExtention - strips a .mod from off the end of string, a.
*)

PROCEDURE StripModuleExtention (VAR a: ARRAY OF CHAR) ;
CONST
   ModExtension    = 3 ;
   ObjectExtension = 1 ;
BEGIN
   ExtractExtension(a, '.mod', ModExtension) ;
   ExtractExtension(a, '.o'  , ObjectExtension)
END StripModuleExtention ;


(*
   CheckVerbose - will echo string, a, if the -v flag was used,
                  otherwise the command will be executed and its
                  exit status returned.
*)

PROCEDURE CheckVerbose (a: ARRAY OF CHAR) : INTEGER ;
VAR
   Result: INTEGER ;
   Copy  : ARRAY [0..Max] OF CHAR ;
BEGIN
   IF DryRunFound
   THEN
      WriteString(a) ; WriteLn ;
      Result := 0
   ELSE
      StrCopy(a, Copy) ;  (* We copy from a dynamic array of char *)
      IF VerboseFound
      THEN
      	 WriteString(Copy) ; WriteLn
      END ;
      Result := system(ADR(Copy))
   END ;
   RETURN( Result )
END CheckVerbose ;


(*
   DisplayHelp - displays the options available for m2f.
*)

PROCEDURE DisplayHelp ;
VAR
   Result: INTEGER ;
   a     : ARRAY [0..Max] OF CHAR ;
BEGIN
   WriteString('Usage: m2f [-gas|-masm] [-i386|-apu] [-bounds] [-nomath]') ; WriteLn ;
   WriteString('           [-O] [-Ocse] [-Ouncalled] [-Omemcopy] [-Odynamic] [-Oi586]') ; WriteLn ;
   WriteString('           [-Oi686] [-Ocache] [-Oworst] [-16] [-32] [-p] [-verbose]') ; WriteLn ;
   WriteString('           [-statistics] [-pedantic] [-M SearchPath] [-m] [-o <file>]') ; WriteLn ;
   WriteString('           [-l] [-g] [-S] [-dryrun] [-v] [-orderonly] [-orderfile <file>]') ; WriteLn ;
   WriteString('           [-startup <file>] [-dryrun]') ; WriteLn ;
   WriteString('       -i386               generate code for the i386 processor') ; WriteLn ;
   WriteString('       -apu                generate code for the APU processor') ; WriteLn ;
   WriteString('       -gas                use the GAS assembler') ; WriteLn ;
   WriteString('       -masm               use the MASM assembler') ; WriteLn ;
   WriteString('       -bas                use the BAS (Bruce Evans) assembler') ; WriteLn ;
   WriteString('       -m                  generate listing of modules and object files') ; WriteLn ;
   WriteString('       -o <file>           destination executable is <file>') ; WriteLn ;
   WriteString('       -l                  link application') ; WriteLn ;
   WriteString('       -orderfile <file>   module initialization order specified by <file>') ; WriteLn ;
   WriteString('       -orderonly          only generate module initialization file') ; WriteLn ;
   WriteString('       -startup <file>     use <file> for startup sequence (instead of crt0.o)') ; WriteLn ;
   WriteString('       -dryrun             display all commands that the front end initiates') ; WriteLn ;
   WriteString('       -ng                 no debugging, overrides environment') ; WriteLn ;
   BuildCompiler(a) ;
   Result := CheckVerbose(a)
END DisplayHelp ;


(*
   AddSentance - adds a sentance, a, to sentance, Sentance, separated
                 by a space and enclosed by double quotes.
*)

PROCEDURE AddSentance (VAR Sentance: ARRAY OF CHAR; a: ARRAY OF CHAR) ;
BEGIN
   StrConCat(Sentance, ' "', Sentance) ;
   StrConCat(Sentance, a, Sentance) ;
   StrConCat(Sentance, '"', Sentance)
END AddSentance ;


(*
   AddWord - adds a word, a, to sentance, Sentance, separated
             by a space.
*)

PROCEDURE AddWord (VAR Sentance: ARRAY OF CHAR; a: ARRAY OF CHAR) ;
BEGIN
   StrConCat(Sentance, ' ', Sentance) ;
   StrConCat(Sentance, a, Sentance)
END AddWord ;


(*
   AddTempFile - adds a temporary file to string, a, with extension,
                 ext.
*)

PROCEDURE AddTempFile (VAR a: ARRAY OF CHAR; ext: ARRAY OF CHAR) ;
VAR
   pid: CARDINAL ;
   str: ARRAY [0..80] OF CHAR ;
BEGIN
   IF NOT StrEqual(a, '')
   THEN
      StrConCat(a, ' ', a)
   END ;
   StrConCat(a, '/tmp/', a) ;
   CardToStr(getpid(), 0, str) ;
   StrConCat(a, str, a) ;
   StrConCat(a, ext, a)
END AddTempFile ;


(*
   BuildCompiler - works out which compiler we should invoke.
*)

PROCEDURE BuildCompiler (VAR a: ARRAY OF CHAR) ;
BEGIN
   StrCopy('gm2', a) ;
   IF i386Found
   THEN
      StrConCat(a, '-i386', a) ;
      IF BasFound
      THEN
         StrConCat(a, '-bas', a)
      ELSIF MasmFound
      THEN
         StrConCat(a, '-masm', a)
      ELSIF GasFound
      THEN
         StrConCat(a, '-gas', a)
      END
   ELSIF ApuFound
   THEN
      StrConCat(a, '-apu', a) ;
      IF NOT HelpFound
      THEN
         AddWord(a, '-apu')
      END
   END
END BuildCompiler ;


(*
   BuildCompilerCommand - builds a command in the string, a,
                          to be sent to the shell for execution.
*)

PROCEDURE BuildCompilerCommand (VAR a: ARRAY OF CHAR) ;
BEGIN
   BuildCompiler(a) ;
   IF VersionFound
   THEN
      AddWord(a, '-version')
   ELSIF SearchFound
   THEN
      AddWord(a, '-M') ;
      AddSentance(a, SearchPath)
   END ;
   IF DebugFound
   THEN
      AddWord(a, '-g')
   END ;
   IF ProfileFound
   THEN
      AddWord(a, '-p')
   END ;
   IF UseSlashTmpFound
   THEN
      AddWord(a, '-o') ;
      AddTempFile(a, '.3')
   END ;
   AddWord(a, Rest)
END BuildCompilerCommand ;


(*
   InvokeCompiler - invokes the gm2 compiler with arguments specified.
*)

PROCEDURE InvokeCompiler ;
VAR
   CompilerCommand: ARRAY [0..Max] OF CHAR ;
   Result         : INTEGER ;
BEGIN
   BuildCompilerCommand(CompilerCommand) ;
   Result := CheckVerbose(CompilerCommand) ;
   IF Result#0
   THEN
      RemoveAssemblerFile ;
      exit(1)
   END
END InvokeCompiler ;


(*
   BuildAssemblerCommand - builds a command in the string, a,
                           to be sent to the shell for execution.
*)

PROCEDURE BuildAssemblerCommand (VAR a: ARRAY OF CHAR) ;
VAR
   FileName: ARRAY [0..Max] OF CHAR ;
BEGIN
   IF BasFound
   THEN
      StrCopy('as86 -j -w -a -0', a)
      (*
          -j = jumps long
          -w = warnings off
          -a = asld compatible
          -0 = 8086
      *)
   ELSIF ApuFound
   THEN
      StrCopy('as-is32', a)
   ELSE
      StrCopy('as', a)
   END ;
   IF DebugFound AND (NOT GasFound)
   THEN
      AddWord(a, '-g')
   END ;
   AddWord(a, '-o') ;
   IF NOT TargetFound
   THEN
      IF ApuFound
      THEN
         CalculateFileName(Module, 'obj', Target)
      ELSE
         CalculateFileName(Module, 'o', Target)
      END
   END ;
   AddWord(a, Target) ;
   IF UseSlashTmpFound
   THEN
      StrCopy('', FileName) ;
      AddTempFile(FileName, '.3')
   ELSE
      CalculateFileName(Module, 's', FileName)
   END ;
   IF (NOT Exists(FileName)) AND (NOT DryRunFound)
   THEN
      exit(1)   (* have not created assembler file therefore exit *)
   END ;
   AddWord(a, FileName)
END BuildAssemblerCommand ;


(*
   RemoveAssemblerFile - removes the assembler file if -S was absent
*)

PROCEDURE RemoveAssemblerFile ;
VAR
   Command,
   TemporaryFile: ARRAY [0..Max] OF CHAR ;
   Result       : INTEGER ;
BEGIN
   IF Assemble
   THEN
      IF UseSlashTmpFound
      THEN
         StrCopy('/bin/rm -f ', Command) ;
         AddTempFile(Command, '.3') ;
         Result := CheckVerbose(Command) ;
         IF Result#0
         THEN
            exit(1)
         END
      ELSE
         CalculateFileName(Module, 's', TemporaryFile) ;
         StrCopy('/bin/rm -f ', Command) ;
         AddWord(Command, TemporaryFile) ;
         Result := CheckVerbose(Command) ;
         IF Result#0
         THEN
            exit(1)
         END
      END
   END
END RemoveAssemblerFile ;


(*
   InvokeAssembler - invokes the GNU assembler with arguments specified.
*)

PROCEDURE InvokeAssembler ;
VAR
   AssemblerCommand: ARRAY [0..Max] OF CHAR ;
   Result          : INTEGER ;
BEGIN
   BuildAssemblerCommand(AssemblerCommand) ;
   Result := CheckVerbose(AssemblerCommand) ;
   RemoveAssemblerFile ;
   IF Result#0
   THEN
      exit(1)
   END
END InvokeAssembler ;


(*
   BuildOrderCommand - builds the order command.
*)

PROCEDURE BuildOrderCommand (VAR a: ARRAY OF CHAR) ;
BEGIN
   StrCopy('gm2l', a) ;
   IF SearchFound
   THEN
      AddWord(a, '-M') ;
      AddSentance(a, SearchPath)
   END ;
   IF OmitM2RTSFound
   THEN
      AddWord(a, '-M2RTS')
   END ;
   AddWord(a, Module) ;
   AddWord(a, '| gm2lsub >') ;
   AddWord(a, OrderFile)
END BuildOrderCommand ;


(*
   InvokeLinker - executes the linking phase of the Modula-2 compiler.
*)

PROCEDURE InvokeLinker ;
VAR
   a     : ARRAY [0..Max] OF CHAR ;
   Result: INTEGER ;
BEGIN
   IF (NOT OrderFileFound) OR OrderOnlyFound
   THEN
      (* No order file found or we want to create one *)
      BuildOrderCommand(a) ;
      Result := CheckVerbose(a) ;
      IF (Result#0) OR OrderOnlyFound
      THEN
      	 exit(Result)
      END
   END ;
   (* Continue to link, create initialization file *)
   IF ApuFound
   THEN
      StrCopy('gm2lgen -apu <', a) ;
      AddWord(a, OrderFile) ;
      AddWord(a, '> mod_init.asm')
   ELSE
      StrCopy('gm2lgen <', a) ;
      AddWord(a, OrderFile) ;
      AddWord(a, '> mod_init.c')
   END ;
   Result := CheckVerbose(a) ;
   IF (Result#0) OR OrderOnlyFound
   THEN
      exit(Result)
   END ;
   (* Compile this initialization file *)
   IF ApuFound
   THEN
      StrCopy('as-is32 -o mod_init.obj mod_init.asm', a)
   ELSE
      StrCopy('gcc -c', a) ;
      IF ProfileFound
      THEN
         AddWord(a, '-p')
      END ;
      IF DebugFound
      THEN
         AddWord(a, '-g')
      END ;
      AddWord(a, 'mod_init.c')
   END ;
   Result := CheckVerbose(a) ;
   IF (Result#0) OR OrderOnlyFound
   THEN
      exit(Result)
   END ;
   (* Create the linking script *)
   StrCopy('gm2lcc', a) ;
   IF ApuFound
   THEN
      AddWord(a, '-apu')
   END ;
   IF ProfileFound
   THEN
      AddWord(a, '-p')
   END ;
   IF DebugFound
   THEN
      AddWord(a, '-g')
   END ;
   IF TargetFound
   THEN
      AddWord(a, '-o') ;
      AddWord(a, Target)
   END ;
   IF SearchFound
   THEN
      AddWord(a, '-M') ;
      AddSentance(a, SearchPath)
   END ;
   IF StartupFound
   THEN
      AddWord(a, '-startup') ;
      AddWord(a, StartupFile)
   END ;
   AddWord(a, '<') ;
   AddWord(a, OrderFile) ;
   AddWord(a, '>') ;
   AddTempFile(a, '.2') ;
   Result := CheckVerbose(a) ;
   IF (Result#0) OR OrderOnlyFound OR ListModulesFound
   THEN
      IF (Result#0) OR ListModulesFound
      THEN
         StrCopy('gm2lcc', a) ;
         IF ApuFound
         THEN
            AddWord(a, '-apu')
         END ;
         AddWord(a, '-c') ;
         IF SearchFound
         THEN
            AddWord(a, '-M') ;
            AddSentance(a, SearchPath)
         END ;
         AddWord(a, '<') ;
         AddWord(a, OrderFile) ;
      	 Result := CheckVerbose(a)
      END ;
      IF (Result#0) OR OrderOnlyFound
      THEN
         exit(Result)
      END
   END ;
   (* Perform linkage *)
   StrCopy('/bin/sh', a) ;
   AddTempFile(a, '.2') ;
   Result := CheckVerbose(a) ;
   IF (Result#0) OR OrderOnlyFound
   THEN
      exit(Result)
   END ;
   (* Tidy up *)
   StrCopy('/bin/rm -f ', a) ;
   AddTempFile(a, '.2') ;
   IF NOT OrderFileFound
   THEN
      AddWord(a, OrderFile)   (* It was a temporary file *)
   END ;
   Result := CheckVerbose(a) ;
   IF (Result#0) OR OrderOnlyFound
   THEN
      exit(Result)
   END
END InvokeLinker ;


(*
   ExecuteSubComponents - executes the subcomponents of the modula-2
                          compiler.
*)

PROCEDURE ExecuteSubComponents ;
BEGIN
   IF HelpFound
   THEN
      DisplayHelp
   ELSIF LinkFound
   THEN
      InvokeLinker
   ELSE
      InvokeCompiler ;
      IF Assemble
      THEN
         InvokeAssembler
      END
   END
END ExecuteSubComponents ;


(*
   Init - initializes the variables used within M2Front.
*)

PROCEDURE Init ;
BEGIN
   StrCopy('', Rest) ;
   StrCopy('', SearchPath) ;
   SearchFound := FALSE ;
   Assemble := TRUE ;
   TargetFound := FALSE ;
   DebugFound := FALSE ;
   LinkFound := FALSE ;
   ProfileFound := FALSE ;
   StrCopy('', OrderFile) ;
   AddTempFile(OrderFile, '.1') ;
   VerboseFound := FALSE ;
   DryRunFound := FALSE ;
   OrderFileFound := FALSE ;
   OrderOnlyFound := FALSE ;
   ListModulesFound := FALSE ;
   StartupFound := FALSE ;
   VersionFound := FALSE ;
   GasFound := TRUE ;
   BasFound := FALSE ;
   MasmFound := FALSE ;
   ApuFound := FALSE ;
   i386Found := TRUE ;
   UseSlashTmpFound := FALSE ;
   OmitM2RTSFound := FALSE ;
   HelpFound := FALSE
END Init ;


BEGIN
   Init ;
   ScanEnvironmentArgs ;
   ScanArguments ;
   ExecuteSubComponents
END m2f.
