IMPLEMENTATION MODULE M2Time ;


FROM M2GenAPU IMPORT Time ;

FROM M2Quads IMPORT QuadOperator, GetQuad ;

FROM SymbolTable IMPORT IsProcedureVariable, IsProcType, GetType,
                        ModeOfAddr, GetMode, IsProcedure,
                        PushSize ;

FROM M2ALU IMPORT PopCard ;

FROM M2Base IMPORT Cardinal ;

FROM M2Lexical IMPORT WriteError ;

FROM M2Debug IMPORT Assert ;



(*
   Throughout the MC68000 coding is assumed to be without any optimization.
   :-(
*)

(*
   Rules for Quadruples and Modes
   ==============================

   Symbols have four modes

   Mode = (None, Value, Offset, Immediate)

   A symbol can only change its mode from:
   None -> [Value, Offset, Immediate]
     any other alteration should be regarded as an error.


   Description of Mode
   ===================

   Sym<O> => Offset    - the value of symbol is contained at location Offset.
   Sym<V> => Value     - the symbol value is referenced indirectly via the
                         memory location offset.
   Sym<I> => Immediate - symbol is a constant and the value is in Symbol Table.


   Rules
   =====

   All program declared variables are given the mode, Offset.
   All constants have mode, Immediate.

   Operators
   =========

------------------------------------------------------------------------------
   Array Operators
------------------------------------------------------------------------------
   Sym<I>   Base   a            Delivers a constant result if a is a
                                Global variable. If a is a local variable
                                then the Frame pointer needs to be added.
                                Base yields the effective location in memory
                                of, a, array [0,0, .. ,0] address.
   Sym<I>   ElementSize 1       Always delivers a constant. The number
                                indicates which specified element is chosen.
                                ElementSize is the TypeSize for that element.
------------------------------------------------------------------------------
   := Operator
------------------------------------------------------------------------------
   Sym1<I> := Sym2<I>           := produces a constant
   Sym1<O> := Sym2<O>           := has the effect Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O> := Sym2<V>           := has the effect
                                             Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V> := Sym2<O>           := has the effect
                                             Mem[Mem[Sym1<I>]] := Mem[Sym2<I>]
   Sym1<V> := Sym2<V>           := has the effect
                                        Mem[Mem[Sym1<I>]] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   Addr Operator  - contains the address of a variable - may need to add
------------------------------------------------------------------------------
   Yields the address of a variable - need to add the frame pointer if
   a variable is local to a procedure.

   Sym1<O>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<V>   Addr   Sym2<O>     meaning     Mem[Sym1<I>] := Sym2<I>
   Sym1<O>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Addr   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
------------------------------------------------------------------------------
   Copy Operator
------------------------------------------------------------------------------
   Sym1<O>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<V>   Copy   Sym2<O>     meaning     Mem[Sym1<I>] := Mem[Sym2<I>]
   Sym1<O>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
   Sym1<V>   Copy   Sym2<V>     meaning     Mem[Sym1<I>] := Mem[Mem[Sym2<I>]]
------------------------------------------------------------------------------
   + - / * Operators
------------------------------------------------------------------------------
   Sym1<I>   +      Sym2<I>  Sym3<I>  meaning Sym1<I> := Sym2<I> + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<I>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Sym3<I>
   Sym1<O>   +      Sym2<O>  Sym3<O>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<O>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<O>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
   Sym1<V>   +      Sym2<V>  Sym3<V>  meaning Mem[Sym1<I>] :=
                                                    Mem[Sym2<I>] + Mem[Sym3<I>]
------------------------------------------------------------------------------
   Base Operator
------------------------------------------------------------------------------
   Sym1<O>   Base  Sym2   Sym3<O>     meaning     Mem[Sym1<I>] := Sym3<I>
   Sym1<V>   Base  Sym2   Sym3<O>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Sym3<I>
   Sym1<O>   Base  Sym2   Sym3<V>     meaning     Mem[Sym1<I>] := Mem[Sym3<I>]
   Sym1<V>   Base  Sym2   Sym3<V>     meaning     Should Never Occur But If it did..
                                                  Mem[Mem[Sym1<I>]] := Mem[Sym3<I>]
                   Sym2 is the array type
------------------------------------------------------------------------------

Notes

1)   All references to memory Mem[Sym1<I>] may need to add the base
     pointer of the activation record if the symbol is a procedure variable.

2)   Addr provides the address of variables and therefore can be used to
     perform the pseudo system call function ADR( variable ), provided
     one obeys rule (1).
*)


(*
   All Times are in terms of T States.
   All Instructions are using long format ( .L )
*)


CONST
   ADDLImmDnTIME  =  16 ;
   ADDLDnDnTIME   =   8 ;
   ADDLImmAnTIME  =  16 ;
   ADDQLImmAnTIME =   8 ;

   BEQTIME        =  11 ;

   CMPLDnDnTIME   =   8 ;

   DIVLDnDnTIME   = 140 ;

   JMPLMemTIME    =  10 ;

   JSRMemTIME     =  18 ;
   JSRAnTIME      =  16 ;

   LINKRnImmTIME  =  16 ;

   MOVELMemDnTIME =  20 ;
   MOVELDnMemTIME =  20 ;
   MOVELDnIdxTIME =  16 ;
   MOVELIdxDnTIME =  32 ;
   MOVELDnDnTIME  =   4 ;
   MOVELImmDnTIME =  12 ;
   MOVELDnSPTIME  =   4 ;

   MULTLDnDnTIME  =  70 ;

   RTSTIME        =  16 ;

   SUBLDnDnTIME   =   8 ;

   UNLKTIME       =  12 ;



TYPE
   RegDesc = RECORD
                Var     : CARDINAL ;  (* Current Variable inside Register *)
                Modified: BOOLEAN ;   (* Determines whether variable has  *)
                                      (* been altered.                    *)
             END ;

VAR
   An : ARRAY [0..MaxAn] OF RegDesc ;
   Dn : ARRAY [0..MaxDn] OF RegDesc ;
   Op : QuadOperator ;
   Op1,
   Op2,
   Op3: CARDINAL ;

   CurTime: CARDINAL ;    (* Accumulated time for the current quadruple *)


(*
   TimeBeforeQuad - sums the time the processor would spend executing a
                    quadruple, QuadNo.
                    StartBB and EndBB determine the range of the Basic Block.
                    Multiway branch depending upon the QuadNo.
                    All branches are timed before actually encoded as
                    APU code.
*)

PROCEDURE TimeBeforeQuad (QuadNo, StartBB, EndBB: CARDINAL) ;
BEGIN
   CurTime := 0 ;
   GetQuad(QuadNo, Op, Op1, Op2, Op3) ;
   CASE Op OF

   AddOp         : DoAdd |
   SubOp         : DoSub |
   MultOp        : DoMult |
   DivOp         : DoDiv |
   NewLocalVarOp : |
   KillLocalVarOp: DoKillLocalVar |
   CallOp        : DoCall |
   ParamOp       : DoParam |
   AddrOp        : DoAddr |
   CopyOp        : DoCopy |
   BecomesOp     : DoBecomes |
   GotoOp        : DoGoto |
   EndOp         : |  (* No need to time the end quad *)
   IfEquOp       : DoIfEqu |
   IfNotEquOp    : DoIfNotEqu |
   IfGreEquOp    : DoIfGreEqu |
   IfLessEquOp   : DoIfLessEqu |
   IfLessOp      : DoIfLess |
   IfGreOp       : DoIfGre |
   BaseOp        : DoBase |
   ElementSizeOp : DoElementSize |
   OffsetOp      : DoOffset |
   FunctValueOp  : DoFunctValue |
   ReturnOp      : DoReturn |
   ReturnValueOp : DoReturnValue |

   (* And the Compiler Flag Pseudo Quadruples *)
   InlineOp      ,
   CodeOnOp      ,
   CodeOffOp     ,
   ProfileOnOp   ,
   ProfileOffOp  ,
   OptimizeOnOp  ,
   OptimizeOffOp : ;

   ELSE
      WriteError('Not implemented yet - in M2Time')
   END ;
   Time(CurTime)
END TimeBeforeQuad ;


(*
   TimeAfterQuad - sums the time the processor would spend executing a
                   quadruple, QuadNo.
                   StartBB and EndBB determine the range of the Basic Block.
                   Multiway branch depending upon the QuadNo.
                   All branches are timed after being encoded as
                   APU code.
*)

PROCEDURE TimeAfterQuad (QuadNo, StartBB, EndBB: CARDINAL) ;
BEGIN
   CurTime := 0 ;
   GetQuad(QuadNo, Op, Op1, Op2, Op3) ;
   CASE Op OF

   AddOp         : |
   SubOp         : |
   MultOp        : |
   DivOp         : |
   NewLocalVarOp : DoNewLocalVar |
   KillLocalVarOp: |
   CallOp        : |
   ParamOp       : |
   AddrOp        : |
   CopyOp        : |
   BecomesOp     : |
   GotoOp        : |
   EndOp         : |  (* No need to time the end quad *)
   IfEquOp       : |
   IfNotEquOp    : |
   IfGreEquOp    : |
   IfLessEquOp   : |
   IfLessOp      : |
   IfGreOp       : |
   BaseOp        : |
   ElementSizeOp : |
   OffsetOp      : |
   FunctValueOp  : |
   ReturnOp      : |
   ReturnValueOp : |

   (* And the Compiler Flag Pseudo Quadruples *)
   InlineOp      ,
   CodeOnOp      ,
   CodeOffOp     ,
   ProfileOnOp   ,
   ProfileOffOp  ,
   OptimizeOnOp  ,
   OptimizeOffOp : ;

   ELSE
      WriteError('Not implemented yet - in M2Time')
   END ;
   Time(CurTime)
END TimeAfterQuad ;


(*
   AddOp - briefly:
           AddOp   Op1  Op2  Op3
           where Op1 := Op2 + Op3
*)

PROCEDURE DoAdd ;
VAR
   Ri, Rj: CARDINAL ;
BEGIN
   (*
      Code:

      Ri := Op2 ;
      Rj := Op3 ;
      ADD.W  Ri,Rj
      Op1 := Ri
   *)
   Ri := LoadVarDn(Op2) ;
   Rj := LoadVarDn(Op3) ;
   ADDLDnDn ;
   StoreVarDn(Op1, Ri)
END DoAdd ;


(*
   SubOp - briefly:
           SubOp   Op1  Op2  Op3
           where Op1 := Op2 - Op3
*)

PROCEDURE DoSub ;
VAR
   Ri, Rj: CARDINAL ;
BEGIN
   (*
      Code:

      Ri := Op2 ;
      Rj := Op3 ;
      SUB.L  Ri,Rj
      Op1 := Ri
   *)
   Ri := LoadVarDn(Op2) ;
   Rj := LoadVarDn(Op3) ;
   SUBLDnDn ;
   StoreVarDn(Op1, Ri)
END DoSub ;


(*
   DivOp - briefly:
           DivOp   Op1  Op2  Op3
           where Op1 := Op2 DIV Op3
*)

PROCEDURE DoDiv ;
VAR
   Ri, Rj: CARDINAL ;
BEGIN
   (*
      Code:

      Ri := Op2 ;
      Rj := Op3 ;
      DIV.L  Ri,Rj
      Op1 := Ri
   *)
   Ri := LoadVarDn(Op2) ;
   Rj := LoadVarDn(Op3) ;
   ADDLDnDn ;
   StoreVarDn(Op1, Ri)
END DoDiv ;


(*
   MultOp - briefly:
            MultOp   Op1  Op2  Op3
            where Op1 := Op2 * Op3
*)

PROCEDURE DoMult ;
VAR
   Ri, Rj: CARDINAL ;
BEGIN
   (*
      Code:

      Ri := Op2 ;
      Rj := Op3 ;
      MULT.L  Ri,Rj
      Op1 := Ri
   *)
   Ri := LoadVarDn(Op2) ;
   Rj := LoadVarDn(Op3) ;
   MULTLDnDn ;
   StoreVarDn(Op1, Ri)
END DoMult ;


(*
   DoCall - briefly:
            CallOp    Op3
*)

PROCEDURE DoCall ;
VAR
   Ai: CARDINAL ;
BEGIN
   IF IsProcedure(Op3)
   THEN
      (*
         Code:

         JSR.L  proc
      *)
      JSRLMem
   ELSE
      Assert(IsProcType(GetType(Op3))) ;
      (*
         Code:

         MOVE.L  op,Ai
         JSR.L   Ai
      *)
      Ai := LoadVarAn(Op3) ;
      JSRLAn
   END
END DoCall ;


(*
   DoGoto - briefly:
            GotoOp   Op3
*)

PROCEDURE DoGoto ;
BEGIN
   (*
      JMP.L  Op3
   *)
   JMPLMem
END DoGoto ;


(*
   DoReturn - simply return from subroutine.
*)

PROCEDURE DoReturn ;
BEGIN
   RTS
END DoReturn ;




(*
   DoAddr - briefly
            Addr   Op1   Op2

            Op1 := ADR(Op2)
*)

PROCEDURE DoAddr ;
BEGIN
   (*
      Four combinations

      Local or Non Local variables

      Referenced by Value or Offset
   *)
   IF IsProcedureVariable(Op3)
   THEN
      IF GetMode(Op3)=RightValue
      THEN
         (*
            MOVE.W  A6,D0
            ADDI    #Op[i],D0
         *)
         MOVELDnDn ;
         ADDLImmDn
      ELSE
         (*
            MOVE.W  Op[i](A6),D0
         *)
         MOVELIdxDn
      END
   ELSE
      IF GetMode(Op3)=RightValue
      THEN
         (*
            MOVE.W  #Op[i],D0
         *)
         MOVELImmDn
      ELSE
         (*
            MOVE.W  Op[i],D0
         *)
         MOVELMemDn
      END
   END ;
   StoreReg(0, Op1)
END DoAddr ;


(*
   DoBase - calculates the base of an array.
            BaseOp   Op1   Op3

            where Op1 now contains the address of array Op3
            ie effectively ADR(Op3[0,0,0....])
*)

PROCEDURE DoBase ;
BEGIN
   IF IsProcedureVariable(Op3)
   THEN
      (*
         Code:

         MOVE.L  A6,D0
         ADDI  #Base,D0
      *)
      MOVELDnDn ;
      ADDLImmDn
   ELSE
      (*
         Code:

         MOVE.L  #Base,D0
      *)
      MOVELImmDn
   END ;
   StoreReg(0, Op1)
END DoBase ;


(*
   DoOffset - sets Op1 to the Value of Op3,
              used to calculate the array subscript address.
              OffsetOp   Op1   Op3
*)

PROCEDURE DoOffset ;
BEGIN
   StoreReg(LoadReg(Op3), Op1)
END DoOffset ;


(*
   DoBecomes - performs the operation Op1 := Op3
*)

PROCEDURE DoBecomes ;
VAR
   by, i, Size: CARDINAL ;
BEGIN
   PushSize(Cardinal) ;
   PushSize(Op1) ;
   IF PopCard()=PopCard()
   THEN
      StoreReg(LoadReg(Op3), Op1)
   ELSE
      PushSize(Cardinal) ;
      by := PopCard() ;
      PushSize(Op1) ;
      Size := PopCard() ;
      i := 0 ;
      WHILE i<Size DO
         StoreReg(LoadReg(Op3), Op1) ;
         INC(i, by)
      END
   END
END DoBecomes ;


PROCEDURE DoCopy ;
BEGIN
   PushSize(Cardinal) ;
   PushSize(Op1) ;
   IF PopCard()=PopCard()
   THEN
      SmallCopy
   ELSE
      LargeCopy
   END
END DoCopy ;


PROCEDURE SmallCopy ;
VAR
   i, j: CARDINAL ;
BEGIN
   (*
      Eight combinations

      Op3 Local or Non Local variables

      Op3 Referenced by Value or Offset

      Op1 Local or Procedural
   *)
   Assert((GetMode(Op3)=RightValue) OR (GetMode(Op3)=LeftValue)) ;
   IF IsProcedureVariable(Op3)
   THEN
      IF GetMode(Op3)=RightValue
      THEN
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i](A6),D0
               MOVE.L  D0,Op1[i](A6)
            *)
            i := GetFreeDn() ;
            MOVELIdxDn ;
            MOVELDnIdx
         ELSE
            (*
               MOVE.L  Op3[i](A6),D0
               MOVE.L  D0,Op1[i]
            *)
            i := GetFreeDn() ;
            MOVELIdxDn ;
            MOVELDnMem
         END
      ELSE
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i](A6),A0
               MOVE.L  0(A0),D0
               MOVE.L  D0,Op1[i](A6)
            *)
            MOVELIdxDn ;
            MOVELIdxDn ;
            MOVELDnIdx
         ELSE
            (*
               MOVE.L  Op3[i](A6),A0
               MOVE.L  0(A0),D0
               MOVE.L  D0,Op1[i]
            *)
            MOVELIdxDn ;
            MOVELIdxDn ;
            MOVELDnMem
         END
      END
   ELSE
      IF GetMode(Op3)=RightValue
      THEN
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i],D0
               MOVE.L  D0,0(A6)
            *)
            MOVELMemDn ;
            MOVELDnIdx
         ELSE
            (*
               MOVE.L  Op3[i],D0
               MOVE.L  D0,Op1[i]
            *)
            MOVELMemDn ;
            MOVELDnMem
         END
      ELSE
         IF IsProcedureVariable(Op1)
         THEN
            (*
               MOVE.L  Op3[i],A0
               MOVE.L  0(A0),D0
               MOVE.L  D0,0(A6)
            *)
            MOVELMemDn ;
            MOVELIdxDn ;
            MOVELDnIdx
         ELSE
            (*
               MOVE.L  Op3[i],A0
               MOVE.L  0(A0),D0
               MOVE.L  D0,Op1[i]
            *)
            MOVELMemDn ;
            MOVELIdxDn ;
            MOVELDnMem
         END
      END
   END
END SmallCopy ;


PROCEDURE LargeCopy ;
VAR
   Size, i, by: CARDINAL ;
BEGIN
   PushSize(Cardinal) ;
   by := PopCard() ;
   PushSize(Op1) ;
   Size := PopCard() ;
   i := 0 ;
   WHILE i<Size DO
      SmallCopy ;
      INC(Size, by)
   END
END LargeCopy ;


PROCEDURE DoIfEqu ;
VAR
   Ri, Rj: CARDINAL ;
BEGIN
   Ri := LoadReg(Op1) ;
   Rj := LoadReg(Op2) ;
   (*
      CMP.L  Di,Dj
      BEQ    Label
   *)
   CMPLDnDn ;
   BEQ
END DoIfEqu ;


PROCEDURE DoIfNotEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfNotEqu ;


PROCEDURE DoIfLess ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfLess ;


PROCEDURE DoIfLessEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfLessEqu ;


PROCEDURE DoIfGre ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfGre ;


PROCEDURE DoIfGreEqu ;
BEGIN
   (* Same time as DoIfEqu *)
   DoIfEqu
END DoIfGreEqu ;


PROCEDURE DoParam ;
VAR
   Ri: CARDINAL ;
BEGIN
   Ri := LoadRegDn(Op3) ;
   (*
      MOVE.L  R0,(sp)-
   *)
   MOVELDnSP
END DoParam ;


PROCEDURE DoNewLocalVar ;
BEGIN
   LINKRnImm ;
   ADDLImmDn
END DoNewLocalVar ;


PROCEDURE DoKillLocalVar ;
BEGIN
   (*
      ADDQL  #-,sp
      UNLK
   *)
   ADDQLImmAn ;
   UNLK
END DoKillLocalVar ;


PROCEDURE DoFunctValue ;
BEGIN
   (*
      Op3 := R0
   *)
   StoreVarDn(Op3, 0)
END DoFunctValue ;


PROCEDURE DoReturnValue ;
VAR
   R0: CARDINAL ;
BEGIN
   R0 := LoadRegDn(Op3)
END DoReturnValue ;


PROCEDURE DoElementSize ;
BEGIN
   WriteError('Should never get here - DoElementSize in M2Time')
END DoElementSize ;


(*
   VarExistsAn - returns true if an Address register contains the operand Op.
*)

PROCEDURE VarExistsAn (Op: CARDINAL) : BOOLEAN ;
VAR
   i: CARDINAL ;
BEGIN
   i := 0 ;
   WHILE (i<=MaxAn) AND (An[i].Var#Op) DO
      INC(i)
   END ;
   RETURN( (i<=MaxAn) )
END VarExistsAn ;


(*
   LoadVarAn - returns an Address register which contains operand Op.
*)

PROCEDURE LoadVarAn (Op: CARDINAL) : CARDINAL ;
BEGIN
   IF VarExistsAn(Op)
   THEN
      RETURN( GetRegAn(Op) )
   ELSE
      RETURN( LoadRegAn(Op) )
   END
END LoadVarAn ;


(*
   LoadRegAn - loads an Address register with the operand, Op.
*)

PROCEDURE LoadRegAn (Op: CARDINAL) : CARDINAL ;
VAR
   Ai: CARDINAL ;
   Aj: CARDINAL ;
BEGIN
   Ai := GetFreeReg(Op) ;
   IF GetMode(Op)=ImmediateValue
   THEN
      (*
         MOVE.W  #Op[i],A0
      *)
      MOVELImmAn
   ELSE
      (*
         Four combinations

         Local or Non Local variables

         Referenced by Value or Offset
      *)
      IF IsProcedureVariable(Op)
      THEN
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i](A6),Ai
            *)
            MOVELIdxAn
         ELSE
            (*
               MOVE.W  Op[i](A6),Aj
               MOVE.W  0(Aj),Ai
            *)
            MOVELIdxDn ;
            MOVELIdxDn
         END
      ELSE
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i],Ai
            *)
            MOVELMemDn
         ELSE
            (*
               MOVE.W  Op[i],Aj
               MOVE.W  0(Aj),Ai
            *)
            MOVELMemDn ;
            MOVELIdxDn
         END
      END
   END ;
   RETURN( Ai )
END LoadRegAn ;


(*
   Loadreg - loads a register with the operand, Op.
*)

PROCEDURE LoadReg (Op: CARDINAL) : CARDINAL ;
BEGIN
   IF GetMode(Op)=ImmediateValue
   THEN
      (*
         MOVE.W  #Op[i],R0
      *)
      MOVELImmDn
   ELSE
      (*
         Four combinations

         Local or Non Local variables

         Referenced by Value or Offset
      *)
      IF IsProcedureVariable(Op)
      THEN
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i](A6),R0
            *)
            MOVELIdxDn
         ELSE
            (*
               MOVE.W  Op[i](A6),A0
               MOVE.W  0(A0),D0
            *)
            MOVELIdxDn ;
            MOVELIdxDn
         END
      ELSE
         IF GetMode(Op)=RightValue
         THEN
            (*
               MOVE.W  Op[i],R0
            *)
            MOVELMemDn
         ELSE
            (*
               MOVE.W  Op[i],A0
               MOVE.W  0(A0),D0
            *)
            MOVELMemDn ;
            MOVELIdxDn
         END
      END
   END ;
   RETURN( 0 )
END LoadReg ;


(*
   StoreReg - stores a register, r, in the operand, Op.
*)

PROCEDURE StoreReg (r, Op: CARDINAL) ;
BEGIN
   (*
      Approximately the Mirror immage of the LoadReg
      as regards timings!
   *)
   r := LoadReg(Op)
END StoreReg ;
PROCEDURE ADDLImmDn ;
BEGIN
   INC(CurTime, ADDLImmDnTIME)
END ADDLImmDn ;


PROCEDURE ADDLDnDn ;
BEGIN
   INC(CurTime, ADDLDnDnTIME)
END ADDLDnDn ;


PROCEDURE ADDLImmAn ;
BEGIN
   INC(CurTime, ADDLImmAnTIME)
END ADDLImmAn ;

PROCEDURE ADDQLImmAn ;
BEGIN
   INC(CurTime, ADDQLImmAnTIME)
END ADDQLImmAn ;


PROCEDURE CMPLDnDn ;
BEGIN
   INC(CurTime, CMPLDnDnTIME)
END CMPLDnDn ;


PROCEDURE SUBLDnDn ;
BEGIN
   INC(CurTime, SUBLDnDnTIME)
END SUBLDnDn ;


PROCEDURE MULTLDnDn ;
BEGIN
   INC(CurTime, MULTLDnDnTIME)
END MULTLDnDn ;


PROCEDURE DIVLDnDn ;
BEGIN
   INC(CurTime, DIVLDnDnTIME)
END DIVLDnDn ;


PROCEDURE UNLK ;
BEGIN
   INC(CurTime, UNLKTIME)
END UNLK ;


PROCEDURE LINKRnImm ;
BEGIN
   INC(CurTime, LINKRnImmTIME)
END LINKRnImm ;


PROCEDURE MOVELMemDn ;
BEGIN
   INC(CurTime, MOVELMemDnTIME)
END MOVELMemDn ;


PROCEDURE MOVELDnMem ;
BEGIN
   INC(CurTime, MOVELDnMemTIME)
END MOVELDnMem ;


PROCEDURE MOVELIdxDn ;
BEGIN
   INC(CurTime, MOVELIdxDnTIME)
END MOVELIdxDn ;


PROCEDURE MOVELDnIdx ;
BEGIN
   INC(CurTime, MOVELDnIdxTIME)
END MOVELDnIdx ;


PROCEDURE MOVELImmDn ;
BEGIN
   INC(CurTime, MOVELImmDnTIME)
END MOVELImmDn ;


PROCEDURE MOVELDnSP ;
BEGIN
   INC(CurTime, MOVELDnSPTIME)
END MOVELDnSP ;


PROCEDURE MOVELDnDn ;
BEGIN
   INC(CurTime, MOVELDnDnTIME)
END MOVELDnDn ;


PROCEDURE JSRLMem ;
BEGIN
   INC(CurTime, JSRMemTIME)
END JSRLMem ;


PROCEDURE JSRLAn ;
BEGIN
   INC(CurTime, JSRAnTIME)
END JSRLAn ;


PROCEDURE RTS ;
BEGIN
   INC(CurTime, RTSTIME)
END RTS ;


PROCEDURE BEQ ;
BEGIN
   INC(CurTime, RTSTIME)
END BEQ ;


PROCEDURE JMPLMem ;
BEGIN
   INC(CurTime, JMPLMemTIME)
END JMPLMem ;




END M2Time.
