MODULE yac ;

(*
   Author     : Gaius Mulley
   Title      : Yac
   Date       : 10/1/85
   LastEdit   : 12/11/86
   Description: Takes as input a BNF description and outputs recursive
                descent Modula-2.

                Usage: Yac [-e] [-o] [-code] [-comment] BnfFile

                -e       Produces error statemants in the recursive decent code.

                -o       Optimizes the inmost

                         IF something            to     Success := somthing
                         THEN
                            Success := TRUE
                         ELSE
                            Success := FALSE
                         END

                         Useful if no more M2 code is to be added after Yac usage.

                -code    omit code generation.

                -comment omit comments.


   Known bugs : if you specify a rule as:

                a := b | ( [ c d ] ) =:     probably not LALR(1) anyway
*)

FROM StrLib IMPORT StrEqual, StrCopy, StrLen ;
FROM StrIO IMPORT WriteString, WriteLn, ReadString ;
FROM StdIO IMPORT Write ;
FROM NumberIO IMPORT WriteCard ;
FROM Scan IMPORT OpenSource, CloseSource,
                 GetNextSymbol, WriteError ;
FROM Args IMPORT GetArg ;
FROM NameKey IMPORT WriteKey, MakeKey, LengthKey, GetKey, NulName ;
FROM YacTok IMPORT InitTok, PutTok ;
FROM PrettyEBNF IMPORT PrettyPrintRule ;


CONST
   MaxStack        =  10 ;
   MaxPath         =   5 ;
   MaxTable        = 100 ;
   HighWaterIndent =  65 ;
   MaxSymbolChar   = 256 ;


TYPE
   State    = (Unknown, True, False, Dummy) ;

   Operator = (CbraBeginSingular, CbraEndSingular, CbraBeginPlural, CbraEndPlural,
               SbraBeginSingular, SbraEndSingular, SbraBeginPlural, SbraEndPlural,
               IdBeginSingular,   IdEndSingular,   IdBeginPlural,   IdEndPlural,
               LitBeginSingular,  LitEndSingular,  LitBeginPlural,  LitEndPlural,
               BarBeginSingular,  BarEndSingular,  BarBeginPlural,  BarEndPlural,
               DefBeginSingular,  DefEndSingular,  DefBeginPlural,  DefEndPlural,
               NoOp) ;

   Atom     = (Id, Lit) ;

   Entry    = RECORD
                 operator : Operator ;
                 atom     : Atom ;
                 operand  : CARDINAL ;
                 success  : State ;
                 level    : CARDINAL ;
                 quote    : CHAR ;
                 error    : BOOLEAN ;
             END ;

(* %%%FORWARD%%%
PROCEDURE CheckForBar (start, end: CARDINAL) ; FORWARD ;
PROCEDURE Init ; FORWARD ;
PROCEDURE InitTable ; FORWARD ;
PROCEDURE Bnf () : BOOLEAN ; FORWARD ;
PROCEDURE Production () : BOOLEAN ; FORWARD ;
PROCEDURE FillAtom (a: Atom) : BOOLEAN ; FORWARD ;
PROCEDURE AssignAtom (n: CARDINAL) ; FORWARD ;
PROCEDURE Expression () : BOOLEAN ; FORWARD ;
PROCEDURE Term () : BOOLEAN ; FORWARD ;
PROCEDURE Factor () : BOOLEAN ; FORWARD ;
PROCEDURE Literal () : BOOLEAN ; FORWARD ;
PROCEDURE Ident () : BOOLEAN ; FORWARD ;
PROCEDURE AtomEndErrorCondition (begin, end: CARDINAL) ; FORWARD ;
PROCEDURE PotentialError (begin, end: CARDINAL ; er: BOOLEAN) ; FORWARD ;
PROCEDURE LegalID () : BOOLEAN ; FORWARD ;
PROCEDURE Assert (a: Atom) ; FORWARD ;
PROCEDURE Push (s: State) ; FORWARD ;
PROCEDURE Pop () : State ; FORWARD ;
PROCEDURE Put (o: Operator ; k: CARDINAL; a: Atom ; st: State) : CARDINAL ; FORWARD ;
PROCEDURE SymIs (a: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE IsSym (a: ARRAY OF CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE SymCharIs (ch: CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE IsCharSym (ch: CHAR) : BOOLEAN ; FORWARD ;
PROCEDURE GetSym () : BOOLEAN ; FORWARD ;
PROCEDURE Merge (s1, s2: State) : State ; FORWARD ;
PROCEDURE TruePath () : State ; FORWARD ;
PROCEDURE Quote (k: CARDINAL) : CHAR ; FORWARD ;
PROCEDURE OptSbra (b, e: CARDINAL) ; FORWARD ;
PROCEDURE OptCbra (b, e: CARDINAL) ; FORWARD ;
PROCEDURE OptProduction (b, e: CARDINAL) ; FORWARD ;
PROCEDURE OptLiteral (b, e: CARDINAL) ; FORWARD ;
PROCEDURE OptIdent (b, e: CARDINAL) ; FORWARD ;
PROCEDURE ChuckTable (b, e: CARDINAL) ; FORWARD ;
PROCEDURE DisplayEntry (e: CARDINAL) ; FORWARD ;
PROCEDURE DisplayOperator (o: Operator) ; FORWARD ;
PROCEDURE DisplayAtom (a: Atom) ; FORWARD ;
PROCEDURE DisplayState (s: State) ; FORWARD ;
PROCEDURE GenerateCode (b, e: CARDINAL) ; FORWARD ;
PROCEDURE CodeLine (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeBarBeginPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeBarBeginSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeCbraBeginPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeCbraEndPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeCbraBeginSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeCbraEndSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeSbraBeginPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeSbraEndPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeSbraBeginSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeSbraEndSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeIdBeginPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeIdEndPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeIdBeginSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeIdEndSingular ; FORWARD ;
PROCEDURE CodeSuccess (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeError (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeDefBeginPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeDefEndPlural (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeDefBeginSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeDefEndSingular (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeOperand (i: CARDINAL) ; FORWARD ;
PROCEDURE CodeKey (k: CARDINAL ; SOS: BOOLEAN) ; FORWARD ;
PROCEDURE CodeString (a: ARRAY OF CHAR ; SOS: BOOLEAN) ; FORWARD ;
PROCEDURE CodeLn (EOS: BOOLEAN) ; FORWARD ;
PROCEDURE IncIndent ; FORWARD ;
PROCEDURE DecIndent ; FORWARD ;
PROCEDURE BuildComment ; FORWARD ;
PROCEDURE EndLastLine (SOS: BOOLEAN) ; FORWARD ;
   %%%FORWARD%%% *)

VAR
   Stack                : ARRAY [0..MaxStack] OF State ;
   StackPtr             : CARDINAL ;
   Table                : ARRAY [0..MaxTable] OF Entry ;
   TablePtr             : CARDINAL ;
   Path                 : ARRAY [0..MaxPath] OF CARDINAL ;
   PathPtr              : CARDINAL ;
   isneeded             : BOOLEAN ;
   Global               : State ;
   NoOfPaths            : CARDINAL ;
   Symbol               : ARRAY [0..MaxSymbolChar] OF CHAR ;
   Sym                  : CARDINAL ;
   NextAtom             : CARDINAL ;
   NoOfTimesSuccessUsed : CARDINAL ;
   Level                : CARDINAL ;
   NewLine              : BOOLEAN ;
   EndOfStatement       : BOOLEAN ;
   Optimiz              : BOOLEAN ;
   Error                : BOOLEAN ;
   FailureAllowed       : BOOLEAN ;
   CommentGeneration    : BOOLEAN ;
   CodeGeneration       : BOOLEAN ;
   Debugging            : BOOLEAN ;


(*
   Table of code structures:
   =========================


   Rules:

   1)  Always set Success := TRUE as late on in the procedure as possible.
       The expection to this rule is with CbraPlural Begin .. End.
       Where the value of Success must be known at the point of
       REPEAT ... UNTIL Success ; Therefore must know value before
       UNTIL is executed. Must never destroy the value of Success
       within this construct later on in the procedure. Therefore
       at the close of ) } we destroy the Path[Top] which points
       to the last occurence of Success := TRUE.

   2)  Only need one occurance of Success := ??? for each Path
       of control in the procedure. Unless conflicting with (1).

   3)  Keep a Global value of Success at compile time to hopefully
       allow program to attempt to optimize obvious sequences.

   4)  Success is often attempted to be set to true - the last
       occurence of Success := TRUE is scratched and a new
       Success := TRUE is set. Unless conflicting with (1).

   5)  Error checking is performed within both types of brackets [] & {},
       the parser generated attempts a one lookahead scan. Therefore
       if a symbol is declared illegal inside the brackets and this
       symbol is NOT the first symbol within the last bracket
       then we declare an error since we cannot recover the previous
       state before the bracket option.



               Begin                           End
             +-------------------------------+---------------------------
CbraSingular | WHILE 'Atom' DO               | END ;
             |    ...                        | [ Success := TRUE ]
             |                               |
             |                               |
             |                               |
CbraPlural   |  REPEAT                       | UNTIL NOT Success ;
             |     ...                       | Success := TRUE
             |                               |
             |                               | must force the setting of
             |                               | Success within the loop
             |                               | or before UNTIL. This
             |                               | must never later be
             |                               | destroyed.
             +-------------------------------+---------------------------
SbraSingular | IF 'Atom'                     | END ;
             | THEN                          | [ Success := TRUE ]
             |                               |
             | may be optimized to:          |
             | Success := 'Atom' ;           |
             | [ Success := TRUE ; ]         |
             |                               |
             |                               |
SbraPlural   | IF 'Atom'                     | END ;
             | THEN                          | [ Success := TRUE ]
             |                               |
             |                               |
             +-------------------------------+---------------------------
BarSingular  | ELSIF 'Atom'                  | never gets here!
             | THEN                          |
             |    [ Success := TRUE ]        |
             |                               |
             |                               |
BarPlural    | ELSE                          |
             |    IF 'Atom'                  |
             |    THEN                       |    END ;
             |       [ Success := TRUE ]     |
             |                               |
             +-------------------------------+---------------------------
LitSingular  | IF 'Atom'                     | ELSE
             | THEN                          |    Success := FALSE
             |                               |    [ potential error statement ]
             |                               |    if satisfy rule 5
             |    [ Success := TRUE ]        | END
             |                               |
             | may be optimised to:          |
             |                               |
             | Success := 'Atom'             |
             |                               |
LitPlural    | IF 'Atom'                     | ELSE
             | THEN                          |    Success := FALSE
             |                               |    [ potential error statement ]
             |                               |    if satisfy rule 5
             |    [ Success := TRUE ]        | END
             |                               |
             +-------------------------------+---------------------------
IdSingular   | same idea as Lit except with  |
             | ident                         |
             |                               |
             |                               |
             |                               |
IdPlural     |                               |
             |                               |
             |                               |
             |                               |
             +-------------------------------+---------------------------

*)


(*
   Init - initialize the options from the command line.
*)

PROCEDURE Init ;
VAR
   i : CARDINAL ;
   ok: BOOLEAN ;
BEGIN
   i := 1 ;
   ok := FALSE ;
   Error := FALSE ;
   Optimiz := FALSE ;
   CodeGeneration := TRUE ;
   CommentGeneration := TRUE ;
   Debugging := FALSE ;
   REPEAT
      IF GetArg(Symbol, i)
      THEN
         IF StrEqual(Symbol, '-o')
         THEN
            Optimiz := TRUE ;
            INC(i)
         ELSIF StrEqual(Symbol, '-e')
         THEN
            Error := TRUE ;
            INC(i)
         ELSIF StrEqual(Symbol, '-code')
         THEN
            CodeGeneration := FALSE ;
            INC(i)
         ELSIF StrEqual(Symbol, '-comment')
         THEN
            CommentGeneration := FALSE ;
            INC(i)
         ELSIF StrEqual(Symbol, '-d')
         THEN
            Debugging := TRUE ;
            INC(i)
         ELSE
            ok := TRUE
         END
      ELSE
         ok := TRUE
      END
   UNTIL ok ;
   IF GetArg(Symbol, i)
   THEN
      IF NOT OpenSource(Symbol)
      THEN
         WriteString('File not found') ; WriteLn ;
         WriteString(Symbol) ; WriteString(' does not exist') ; WriteLn ;
         HALT
      END
   END ;
   IF Error AND Optimiz
   THEN
      WriteString('Illegal options: Cannot Optimize and Catch Errors') ; WriteLn ;
      HALT
   END
END Init ;


(*
   InitTable - initialize variables ready for the next BNF rule.
*)

PROCEDURE InitTable ;
BEGIN
   StackPtr := 0 ;
   TablePtr := 0 ;
   isneeded := FALSE ;
   NextAtom := 0 ;
   PathPtr := 0 ;
   Path[PathPtr] := 0 ;
   NoOfPaths := 0 ;
   NoOfTimesSuccessUsed := 0 ;
   Level := 0 ;
   NewLine := TRUE ;
   EndOfStatement := FALSE ;
   FailureAllowed := TRUE
END InitTable ;


(*
   BuildComment - build a comment for the BNF rule.
*)

PROCEDURE BuildComment ;
BEGIN
   PrettyPrintRule
END BuildComment ;
   
   
(*
   Bnf := " BNF " { Production } " FNB " =:
          while we have a Production generate code.
*)

PROCEDURE Bnf () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
BEGIN
   InitTable ;
   IF SymIs('BNF')
   THEN
      WHILE Production() DO
         InitTable
      END ;
      Success := StrEqual('FNB', Symbol)
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Bnf ;


(*
   Production := Symbol " := " Expression " := "
*)

PROCEDURE Production () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
   begin,
   end     : CARDINAL ;
   copy    : CARDINAL ;
BEGIN
   copy := MakeKey(Symbol) ;
   IF LegalID()
   THEN
      Success := GetSym() ;
      Global := True ;
      INC( Level ) ;
      begin := Put( DefBeginPlural, copy, Id, Dummy ) ;
      Push( Global ) ;
      IF SymIs(':=')
      THEN
         IF Expression()
         THEN
            Global := Merge( Global, Pop() ) ;
            end := Put( DefEndPlural, NulName, Id, Dummy ) ;
            DEC( Level ) ;
            OptProduction( begin, end ) ;
            GenerateCode( begin, end ) ;
            IF SymIs('=:')
            THEN
               InitTok      (* reset token stream for comments *)
            ELSE
               WriteError('expecting =: terminator') ;
               Success := FALSE
            END
         ELSE
            WriteError('expecting expression inside definition') ;
            Success := FALSE
         END
      ELSE
         WriteError('expecting :=') ;
         Success := FALSE
      END
   ELSE
      Success := FALSE
   END ;
   IF Success AND Debugging
   THEN
      ChuckTable(1, TablePtr)
   END ;
   RETURN( Success )
END Production ;


PROCEDURE FillAtom (a: Atom) : BOOLEAN ;
BEGIN
   IF NextAtom=0
   THEN
      RETURN( FALSE )
   ELSE
      WITH Table[NextAtom] DO
         operand := MakeKey(Symbol) ;
         atom := a ;
         quote := Quote(MakeKey(Symbol))
      END ;
      NextAtom := 0 ;
      RETURN( TRUE )
   END
END FillAtom ;


PROCEDURE AssignAtom (n: CARDINAL) ;
BEGIN
   IF NextAtom=0
   THEN
      NextAtom := n
   ELSE
      WriteError('Grammer expecting Atom NOT conditional')
   END
END AssignAtom ;


(* ******************* start of yac generated rules ***************)


(*
   Expression := Term { " | " Term } 
*)

PROCEDURE Expression() : BOOLEAN ;
VAR
   Success: BOOLEAN ;
   start,
   end  : CARDINAL ;
BEGIN
   IF Term()
   THEN
      start := TablePtr ;
      WHILE SymCharIs('|') DO
         (* Pseudo Pop Push except retain Global *)
         Path[PathPtr] := 0 ;
         Global := Merge( Global, True ) ;
         DEC( Level ) ;
         end := Put( BarBeginSingular, NulName, Id, TruePath() ) ;
         INC( Level ) ;
         AssignAtom( end ) ;
         (* 
            Must check to see whether the [ or { has ocurred after a |
            in which case we must alter the Bar state to Plural.
            Also the Atom Fill must be stopped as these | plural
            only depend upon Success. Decided by multiple Atoms.

            We just peep at the next symbol
         *)

         IF IsCharSym('[') OR IsCharSym('{')
         THEN
            WITH Table[end] DO
               operator := BarBeginPlural ;
               INC(level) ;
               NextAtom := 0
            END
         END ;

         IF Term()
         THEN
         ELSE
            WriteError('Term - expected')
         END ;
         CheckForBar(start, end) ;
         start := TablePtr
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Expression ;


(*
   SwapPath - swaps any entries in the array Path pointing to, a, and, b.
*)

PROCEDURE SwapPath (a, b: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   i := 1 ;
   WHILE i<=PathPtr DO
      IF Path[i]=a
      THEN
         Path[i] := b
      ELSIF Path[i]=b
      THEN
         Path[i] := a
      END ;
      INC(i)
   END
END SwapPath ;


(*
   CheckForBar - checks to see whether:

                 (i)  start+1=end=TablePtr    and IdEnd followed by a Bar
                                              in which case we should swap
                                              to make ELSIF work

                 (ii) end#TablePtr            we take the IdEnd away and
                                              add it to after end
*)

PROCEDURE CheckForBar (start, end: CARDINAL) ;
VAR
   new  : CARDINAL ;
   spare: Entry ;
BEGIN
   IF (Table[start].operator=IdEndSingular) OR
      (Table[start].operator=IdEndPlural) OR
      (Table[start].operator=LitEndSingular) OR
      (Table[start].operator=LitEndPlural)
   THEN
      IF (start+1=end) AND (end=TablePtr)
      THEN
         IF Debugging
         THEN
            ChuckTable(1, TablePtr) ;
            WriteString('-------------------------') ; WriteLn
         END ;
         spare := Table[end] ;
         Table[end] := Table[start] ;
         Table[start] := spare ;
         IF Debugging
         THEN
            WriteString('=========================') ; WriteLn ;
            ChuckTable(1, TablePtr)
         END ;
         SwapPath(start, end)
      ELSE
         WITH Table[start] DO
            new := Put(operator, operand, atom, success)
         END ;
         SwapPath(start, new) ;
         Table[start].operator := NoOp    (* remove origional *)
      END
   END
END CheckForBar ;


(*
   Term := Factor { Factor } 
*)

PROCEDURE Term () : BOOLEAN ;
VAR
   Success: BOOLEAN ;
BEGIN
   IF Factor()
   THEN
      WHILE Factor() DO
      END ;
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Term ;


(* ******************* end of yac generated rules ***************)

(*
   Factor := Symbol | " { " ( Expression " } " ) | 
             " [ " ( Expression " ] " ) | 
             " ( " ( Expression " ) " ) 
*)

PROCEDURE Factor () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
   begin,
   end     : CARDINAL ;
BEGIN
   IF SymCharIs('(')
   THEN
      IF Expression()
      THEN
         Success := SymCharIs(')')
      ELSE
         Success := FALSE
      END
   ELSIF SymCharIs('[')
   THEN
      FailureAllowed := TRUE ;  (* Only For the next atom *)
      INC( Level ) ;
      begin := Put( SbraBeginPlural, NulName, Id, Dummy ) ;
      Push( Global ) ;
      Global := True ;
      AssignAtom( begin ) ;
      IF Expression()
      THEN
         IF SymCharIs(']')
         THEN
            Success := TRUE ;
            Global := Pop() ;
            Global := True ;
            end := Put( SbraEndPlural, NulName, Id, TruePath() ) ;
            DEC( Level ) ;
            OptSbra( begin, end )
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSIF SymCharIs('{')
   THEN
      FailureAllowed := TRUE ;  (* Only For the next atom *)
      INC( Level ) ;
      begin := Put( CbraBeginSingular, NulName, Id, TruePath() ) ;
      Push( Global ) ;
      Global := True ;
      AssignAtom( begin ) ;
      (* Must check to see whether the ( has ocurred after a {      *)
      (* in which case we must alter the Cbra state to Plural.      *)
      (* Also the Atom Fill must be stopped as these { plural       *)
      (* only depend upon Success. Decided by multiple Atoms.       *)
      IF SymCharIs('(')
      THEN
         WITH Table[begin] DO
            operator := CbraBeginPlural ;
            NextAtom := 0
         END ;
         IF Expression()
         THEN
            Success := SymCharIs(')')
         ELSE
            Success := FALSE
         END
      ELSE
         Success := Expression()
      END ;
      IF Success
      THEN
         IF SymCharIs('}')
         THEN
            Global := Pop() ;
            IF Table[begin].operator=CbraBeginPlural
            THEN
               (* Must now immortialise the Success := TRUE   *)
               (* before and also force a new Success := TRUE *)
               (* now because at the end of } we know that    *)
               (* Success must have been FALSE to exit the    *)
               (* loop.                                       *)
               Path[PathPtr] := 0 ;
               (* Chop off old Path. Therefore Success := TRUE *)
               end := Put( CbraEndPlural, NulName, Id, TruePath() ) ;
               Global := True
            ELSE
               end := Put( CbraEndSingular, NulName, Id, TruePath() ) ;
               Global := True
            END ;
            DEC( Level ) ;
            OptCbra( begin, end )
         ELSE
            Success := FALSE
         END
      ELSE
         Success := FALSE
      END
   ELSIF Literal()
   THEN
      Success := TRUE
   ELSIF Ident()
   THEN
      Success := TRUE
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Factor ;


PROCEDURE Literal () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
   begin,
   end     : CARDINAL ;
   er      : BOOLEAN ;
BEGIN
   IF SymCharIs('"')
   THEN
      er := NOT FailureAllowed ; (* Must only set FailureAllowed if we *)
      FailureAllowed := FALSE ;  (* know literal is going to succeed.  *)
      IF FillAtom( Lit )
      THEN
         Success := GetSym() ;
         IF NOT SymCharIs('"')
         THEN
            WriteError('" - expected')
         END
      ELSE
         INC( Level ) ;
         begin := Put( LitBeginPlural, MakeKey(Symbol), Lit, TruePath() ) ;
         Success := GetSym() ;  (* Chuck Literal *)
         IF SymCharIs('"')
         THEN
            Success := Expression() ;
            Success := TRUE ;
            end := Put( LitEndPlural, NulName, Lit, False ) ;
            PotentialError(begin, end, er) ;
            DEC( Level ) ;
            (* Now place the end statement for Literal *)
            AtomEndErrorCondition(begin, end) ;
            OptLiteral( begin, end )
         ELSE
            WriteError('" - expected')
         END
      END
   ELSIF SymCharIs("'")
   THEN
      er := NOT FailureAllowed ; (* Must only set FailureAllowed if we *)
      FailureAllowed := FALSE ;  (* know literal is going to succeed.  *)
      IF FillAtom( Lit )
      THEN
         Success := GetSym() ;
         IF NOT SymCharIs("'")
         THEN
            WriteError("' - expected")
         END
      ELSE
         INC( Level ) ;
         begin := Put( LitBeginPlural, MakeKey(Symbol), Lit, TruePath() ) ;
         Success := GetSym() ;  (* Chuck Literal *)
         IF SymCharIs("'")
         THEN
            Success := Expression() ;
            Success := TRUE ;
            end := Put( LitEndPlural, NulName, Lit, False ) ;
            PotentialError(begin, end, er) ;
            DEC( Level ) ;
            (* Now place the end statement for Literal *)
            AtomEndErrorCondition(begin, end) ;
            OptLiteral( begin, end )
         ELSE
            WriteError("' - expected")
         END
      END
   ELSE
      Success := FALSE
   END ;
   RETURN( Success )
END Literal ;


PROCEDURE Ident () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
   begin,
   end     : CARDINAL ;
   er      : BOOLEAN ;
BEGIN
   IF LegalID()
   THEN
      er := NOT FailureAllowed ; (* Must only set FailureAllowed if we *)
      FailureAllowed := FALSE ;  (* know ident is going to succeed.    *)
      IF FillAtom( Id )
      THEN
         Success := GetSym()
      ELSE
         INC( Level ) ;
         begin := Put( IdBeginPlural, MakeKey(Symbol), Id, TruePath() ) ;
         Success := GetSym() ;
         Success := Expression() ;
         end := Put( IdEndPlural, NulName, Id, False ) ;
         PotentialError(begin, end, er) ;
         DEC( Level ) ;
         (* Now place the end statement for Ident *)
         AtomEndErrorCondition(begin, end) ;
         OptIdent( begin, end )
      END ;
      Success := TRUE
   ELSE
      Success := FALSE  
   END ;
   RETURN( Success )
END Ident ;


(* Normally at the end of the ATOM we set an else clause to catch *)
(* errors. However if the BarBeginPlural has been used we cannot  *)
(* use the else since it has already been used.                   *)

PROCEDURE AtomEndErrorCondition (begin, end: CARDINAL) ;
VAR
   i  : CARDINAL ;
   bar: BOOLEAN ;
BEGIN
   bar := FALSE ;
   FOR i := begin TO end DO
      WITH Table[i] DO
         IF (operator=BarBeginPlural) AND (level=Level)
         THEN
            IF bar
            THEN
               WriteError('Expression too complex - rearrange | clause') ;
               WriteString('Too many ELSEs will occur') ; WriteLn
            ELSE
               bar := TRUE
            END
         END
      END
   END ;
   IF bar
   THEN
      (* Cannot use:
                     ELSE
                        Success := FALSE
                     END
         since the BarPlural has already used an ELSE clause
      *)
      WITH Table[end] DO
         success := Dummy ;
         Error := FALSE ;   (* No error allowed either *)
      END
   ELSE
      (* set Success := FALSE *)
      Global := Merge( Global, False ) ;
      Path[PathPtr] := end
   END
END AtomEndErrorCondition ;


PROCEDURE PotentialError (begin, end: CARDINAL ; er: BOOLEAN) ;
BEGIN
   WITH Table[end] DO
      error := Error AND er ;
      IF error
      THEN
         operand := Table[begin].operand ;
         quote := Table[begin].quote
      END
   END
END PotentialError ;


PROCEDURE LegalID () : BOOLEAN ;
VAR
   Success : BOOLEAN ;
BEGIN
   IF IsCharSym('{') OR IsCharSym('}') OR
      IsCharSym('[') OR IsCharSym(']') OR
      IsCharSym('(') OR IsCharSym(')') OR
      IsCharSym('"')
   THEN
      Success := FALSE
   ELSIF
      IsSym('BNF') OR IsSym('FNB') OR
      IsSym(':=') OR IsSym('=:') OR
      IsCharSym('|')
   THEN
      Success := FALSE
   ELSE
      Success := TRUE
   END ;
   RETURN( Success )
END LegalID ;


PROCEDURE Assert (a: Atom) ;
BEGIN
   WITH Table[TablePtr] DO
      operand := MakeKey(Symbol) ;
      atom := a
   END
END Assert ;


PROCEDURE Push (s: State) ;
BEGIN
   Stack[StackPtr] := s ;
   IF StackPtr=MaxStack
   THEN
      WriteError('Max Stack Exceeded')
   ELSE
      INC( StackPtr )
   END ;
   IF PathPtr=MaxPath
   THEN
      WriteError('Max Path Exceeded')
   ELSE
      INC( PathPtr ) ;
      Path[PathPtr] := Path[PathPtr-1] ;
      IF NoOfPaths<PathPtr
      THEN
         NoOfPaths := PathPtr
      END
   END
END Push ;


PROCEDURE Pop () : State ;
BEGIN
   DEC( PathPtr ) ;
   DEC( StackPtr ) ;
   RETURN( Stack[StackPtr] )
END Pop ;


(*
   Put - place an operator, operand, atom and state into the table
         at the end of the table list.
*)

PROCEDURE Put (o: Operator ; k: CARDINAL; a: Atom ; st: State) : CARDINAL ;
BEGIN
   IF TablePtr=MaxTable
   THEN
      WriteError('Max Table Exceeded') ;
      HALT
   ELSE
      INC( TablePtr ) ;
      WITH Table[TablePtr] DO
         operator := o ;
         operand := k ;
         atom := a ;
         quote := Quote(k) ;
         success := st ;
         level := Level ;
         error := FALSE
      END ;
      IF st#Dummy
      THEN
         Path[PathPtr] := TablePtr
      END
   END ;
   RETURN( TablePtr )
END Put ;


PROCEDURE SymIs (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   IF StrEqual( a, Symbol )
   THEN
      RETURN( GetSym() )
   ELSE
      RETURN( FALSE )
   END
END SymIs ;


PROCEDURE IsSym (a: ARRAY OF CHAR) : BOOLEAN ;
BEGIN
   RETURN( StrEqual( a, Symbol ) )
END IsSym ;


PROCEDURE SymCharIs (ch: CHAR) : BOOLEAN ;
BEGIN
   IF (StrLen(Symbol)=1) AND (ch=Symbol[0])
   THEN
      RETURN( GetSym() )
   ELSE
      RETURN( FALSE )
   END
END SymCharIs ;


PROCEDURE IsCharSym (ch: CHAR) : BOOLEAN ;
BEGIN
   RETURN ((StrLen(Symbol)=1) AND (ch=Symbol[0]))
END IsCharSym ;


(*
   SkipComments - skips any comments.
*)

PROCEDURE SkipComments ;
VAR
   CommentLevel: CARDINAL ;
BEGIN
   IF StrEqual(Symbol, '(*')
   THEN
      CommentLevel := 1 ;
      WriteLn ;
      WriteString('(* ') ;
      REPEAT
         GetNextSymbol(Symbol) ;
         WriteString(Symbol) ;
         Write(' ') ;
         IF StrEqual(Symbol, '(*')
         THEN
            INC(CommentLevel)
         ELSIF StrEqual(Symbol, '*)')
         THEN
            DEC(CommentLevel) ;
            WriteLn
         END
      UNTIL StrEqual(Symbol, '') OR
            ((CommentLevel=0) AND StrEqual(Symbol, '*)')) ;
      IF StrEqual(Symbol, '') AND (CommentLevel>0)
      THEN
         WriteError('end of file reached inside a comment')
      ELSIF StrEqual(Symbol, '*)') AND (CommentLevel=0)
      THEN
         GetNextSymbol(Symbol)
      END
   END
END SkipComments ;


(*
   GetSym - consumes a symbol and places it into the symbol stack.
*)

PROCEDURE GetSym () : BOOLEAN ;
BEGIN
   GetNextSymbol( Symbol ) ;
   SkipComments() ;
   Sym := MakeKey(Symbol) ;
   PutTok(Sym) ;
   RETURN( TRUE )
END GetSym ;


(*
   Merge - merges two states, s1 and s2, provided they
           are the same. Otherwise Unknown is returned.
*)

PROCEDURE Merge (s1, s2: State) : State ;
BEGIN
   IF s1=s2
   THEN
      RETURN( s1 )
   ELSE
      RETURN( Unknown )
   END
END Merge ;


PROCEDURE TruePath () : State ;
VAR
   e: CARDINAL ;
BEGIN
   e := Path[PathPtr] ;
   IF e=0
   THEN
      (* No Path Therefore set this to True *)
      RETURN( True )
   ELSIF Table[e].success=True
   THEN
      (* Already a True Path therefore return True  *)
      (* and replace this True path with Dummy.     *)
      Path[PathPtr] := 0 ;
      Table[e].success := Dummy ;
      RETURN( True )
   ELSE
      RETURN( Dummy )
   END
END TruePath ;


(*
   Quote - returns the character which will be used for quoting the string
           denoted by, k.
*)

PROCEDURE Quote (k: CARDINAL) : CHAR ;
VAR
   i, high: CARDINAL ;
   double,
   single : BOOLEAN ;
   a      : ARRAY [0..MaxSymbolChar] OF CHAR ;
BEGIN
   GetKey(k, a) ;
   double := FALSE ;
   single := FALSE ;
   high := StrLen(a) ;
   i := 0 ;
   WHILE i<high DO
      IF a[i]='"' THEN double := TRUE END ;
      IF a[i]="'" THEN single := TRUE END ;
      INC(i)
   END ;
   IF single AND double
   THEN
      WriteString(a) ; WriteLn ;
      WriteError('Literal has both kinds of quotes') ;
      RETURN(' ')
   ELSIF single
   THEN
      RETURN('"')
   ELSE
      RETURN("'")
   END
END Quote ;


PROCEDURE OptSbra (b, e: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
(*   ChuckTable(b, e) ; *)
   IF (Table[b].operator=SbraBeginPlural) AND (b+1=e) AND Optimiz
   THEN
      (* Deal with the Square Brackets *)
      Table[b].operator := SbraBeginSingular ;
      Table[e].operator := SbraEndSingular ;
      INC( NoOfTimesSuccessUsed )
   END ;
   i := Path[PathPtr] ;
   IF i#0
   THEN
      Table[i].success := Dummy
   END ;
   Table[e].success := True ;
   Global := True ;
   Path[PathPtr] := e ;
   (* Go back as far as possible getting rid of reduntant Success := ?? *)
   i := e-1 ;
   WHILE (i>0) AND (Table[i].level>=Level) AND
         (Table[i+1].operator#CbraEndPlural)
   DO
      Table[i].success := Dummy ;
      DEC(i)
   END
END OptSbra ;


PROCEDURE OptCbra (b, e: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
(*   ChuckTable(b, e) ; *)
   IF Table[b].operator=CbraBeginPlural
   THEN
      Table[e].operator := CbraEndPlural ;
      INC( NoOfTimesSuccessUsed ) ;
      IF Table[e].success=True
      THEN
         INC( NoOfTimesSuccessUsed )
      END
   ELSE
      i := Path[PathPtr] ;
      IF i#0
      THEN
         Table[i].success := Dummy
      END ;
      Table[e].success := True ;
      Global := True ;
      Path[PathPtr] := e ;
      (* Go back as far as possible getting rid of reduntant Success := ?? *)
      i := e-1 ;
      WHILE (i>0) AND (Table[i].level>=Level) AND
            (Table[i+1].operator#CbraEndPlural)
      DO
         Table[i].success := Dummy ;
         DEC(i)
      END
   END ;
(*   ChuckTable(b, e) ; *)
END OptCbra ;


PROCEDURE OptProduction (b, e: CARDINAL) ;
VAR
   found: BOOLEAN ;
   i    : CARDINAL ;
   op   : Operator ;
BEGIN
   found := FALSE ;
   i := e-1 ;
   WHILE (NOT found) AND (i>b) DO
      op := Table[i].operator ;
      IF (op=IdBeginSingular) OR
         (op=LitBeginSingular)
      THEN
         found := TRUE
      ELSIF op=NoOp
      THEN
         DEC(i)
      ELSE
         (* terminate if End... found *)
         (* only deal with level 0    *)
         i := b
      END
   END ;
   IF found
   THEN
      Table[e].atom := Table[i].atom ;
      Table[e].operand := Table[i].operand ;
      Table[i].operator := NoOp ;
      DEC( NoOfTimesSuccessUsed )
   END ;
   IF NoOfTimesSuccessUsed=0
   THEN
      IF Global=True
      THEN
         Table[e].success := True ;
         Table[b].operator := DefBeginSingular ;
         Table[e].operator := DefEndSingular ;
         FOR i := b TO e-1 DO
            Table[i].success := Dummy
         END
      ELSIF found
      THEN
         Table[e].success := Unknown ;
         Table[b].operator := DefBeginSingular ;
         Table[e].operator := DefEndSingular
      END
   ELSIF Global=True
   THEN
      i := e-1 ;
      REPEAT
         Table[i].success := Dummy ;
         DEC(i)
      UNTIL (i<b) OR (Table[i+1].operator=CbraEndPlural) ;
      Table[e].success := True
   END ;
(*
   ChuckTable(b, e) ;
   WriteString('NoOfTimesSuccessUsed') ; WriteCard( NoOfTimesSuccessUsed, 3 ) ; WriteLn
 *)
END OptProduction ;


PROCEDURE OptLiteral (b, e: CARDINAL) ;
BEGIN
(* ChuckTable(b, e) ; *)
   IF (b+1=e) AND Optimiz
   THEN
      Table[b].operator := LitBeginSingular ;
      WITH Table[e] DO
         operator := NoOp
      END ;
      INC( NoOfTimesSuccessUsed )
   END
END OptLiteral ;


PROCEDURE OptIdent (b, e: CARDINAL) ;
BEGIN
(* ChuckTable(b, e) ; *)
   IF (b+1=e) AND Optimiz
   THEN
      Table[b].operator := IdBeginSingular ;
      WITH Table[e] DO
         operator := NoOp
      END ;
      INC( NoOfTimesSuccessUsed )
   END
END OptIdent ;


PROCEDURE ChuckTable (b, e: CARDINAL) ;
VAR
   t: CARDINAL ;
BEGIN
  FOR t := b TO e DO
     DisplayEntry( t )
  END   
END ChuckTable ;


PROCEDURE DisplayEntry (e: CARDINAL) ;
BEGIN
   WriteCard( e, 4 ) ;
   WriteString('  ') ;
   WITH Table[e] DO
      DisplayOperator( operator ) ;
      WriteString('  ') ;
      DisplayAtom( atom ) ;
      WriteString('  ') ;
      DisplayState( success ) ;
      WriteString('  ') ;
      WriteKey(operand) ;
      WriteCard(level, 3) ;
      WriteLn
   END
END DisplayEntry ;


PROCEDURE DisplayOperator (o: Operator) ;
BEGIN
   CASE o OF

   CbraBeginSingular : WriteString('CbraBeginSingular') |
   CbraEndSingular   : WriteString('CbraEndSingular  ') |
   CbraBeginPlural   : WriteString('CbraBeginPlural  ') |
   CbraEndPlural     : WriteString('CbraEndPlural    ') |

   SbraBeginSingular : WriteString('SbraBeginSingular') |
   SbraEndSingular   : WriteString('SbraEndSingular  ') |
   SbraBeginPlural   : WriteString('SbraBeginPlural  ') |
   SbraEndPlural     : WriteString('SbraEndPlural    ') |

   IdBeginSingular   : WriteString('IdBeginSingular  ') |
   IdEndSingular     : WriteString('IdEndSingular    ') |
   IdBeginPlural     : WriteString('IdBeginPlural    ') |
   IdEndPlural       : WriteString('IdEndPlural      ') |

   BarBeginSingular  : WriteString('BarBeginSingular ') |
   BarEndSingular    : WriteString('BarEndSingular   ') |
   BarBeginPlural    : WriteString('BarBeginPlural   ') |
   BarEndPlural      : WriteString('BarEndPlural     ') |

   LitBeginSingular  : WriteString('LitBeginSingular ') |
   LitEndSingular    : WriteString('LitEndSingular   ') |
   LitBeginPlural    : WriteString('LitBeginPlural   ') |
   LitEndPlural      : WriteString('LitEndPlural     ') |

   DefBeginSingular  : WriteString('DefBeginSingular ') |
   DefEndSingular    : WriteString('DefEndSingular   ') |
   DefBeginPlural    : WriteString('DefBeginPlural   ') |
   DefEndPlural      : WriteString('DefEndPlural     ') |

   NoOp              : WriteString('NoOp             ')

   ELSE
      WriteString('Undefined Operator') ; WriteLn ;
      HALT
   END
END DisplayOperator ;


PROCEDURE DisplayAtom (a: Atom) ;
BEGIN
   CASE a OF

   Lit : WriteString('Lit') |
   Id  : WriteString('Id ')

   END
END DisplayAtom ;


PROCEDURE DisplayState (s: State) ;
BEGIN
   CASE s OF

   Unknown : WriteString('Unknown') |
   True    : WriteString('True   ') |
   False   : WriteString('False  ') |
   Dummy   : WriteString('Dummy  ') 

   END
END  DisplayState ;


(*
   GenerateCode - generate Modula-2 code for productions, b, to e.
*)

PROCEDURE GenerateCode (b, e: CARDINAL) ;
VAR
   i: CARDINAL ;
BEGIN
   IF CommentGeneration
   THEN
      BuildComment
   END ;
   IF CodeGeneration
   THEN
      FOR i := b TO e DO
         CodeLine( i )
      END
   END
END GenerateCode ;


PROCEDURE CodeLine (i: CARDINAL) ;
BEGIN
   CASE Table[i].operator OF
   
   DefBeginPlural   : CodeDefBeginPlural( i )    |
   DefEndPlural     : CodeDefEndPlural( i )      |
   DefBeginSingular : CodeDefBeginSingular( i )  |
   DefEndSingular   : CodeDefEndSingular( i )    |
   IdBeginPlural    : CodeIdBeginPlural( i )     |
   IdEndPlural      : CodeIdEndPlural( i )       |
   IdBeginSingular  : CodeIdBeginSingular( i )   |
   IdEndSingular    : CodeIdEndSingular |
   LitBeginPlural   : CodeIdBeginPlural( i )     |
   LitEndPlural     : CodeIdEndPlural( i )       |
   LitBeginSingular : CodeIdBeginSingular( i )   |
   LitEndSingular   : CodeIdEndSingular |
   SbraBeginPlural  : CodeSbraBeginPlural( i )   |
   SbraEndPlural    : CodeSbraEndPlural( i )     |
   SbraBeginSingular: CodeSbraBeginSingular( i ) |
   SbraEndSingular  : CodeSbraEndSingular( i )   |
   CbraBeginPlural  : CodeCbraBeginPlural( i )   |
   CbraEndPlural    : CodeCbraEndPlural( i )     |
   CbraBeginSingular: CodeCbraBeginSingular( i ) |
   CbraEndSingular  : CodeCbraEndSingular( i )   |
   BarBeginPlural   : CodeBarBeginPlural( i )    |
   BarBeginSingular : CodeBarBeginSingular( i )

   ELSE
   END
END CodeLine ;

PROCEDURE CodeBarBeginPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      DecIndent ;
      CodeString('ELSE', FALSE) ; CodeLn(FALSE) ;
      IncIndent ;
      CodeSuccess( i )
   END
END CodeBarBeginPlural ;


PROCEDURE CodeBarBeginSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      DecIndent ;
      CodeString('ELSIF ', FALSE) ; CodeOperand( i ) ; CodeLn(FALSE) ;
      CodeString('THEN', FALSE) ; CodeLn(FALSE) ;
      IncIndent ;
      CodeSuccess( i )
   END
END CodeBarBeginSingular ;


PROCEDURE CodeCbraBeginPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeSuccess( i ) ;
      CodeString('REPEAT', TRUE) ; CodeLn(FALSE) ;
      IncIndent ;
   END
END CodeCbraBeginPlural ;


PROCEDURE CodeCbraEndPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      DecIndent ;
      CodeString('UNTIL NOT Success', FALSE) ; CodeLn(TRUE) ;
      CodeSuccess( i )
   END
END CodeCbraEndPlural ;


PROCEDURE CodeCbraBeginSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('WHILE ', TRUE) ; CodeOperand( i ) ; CodeString(' DO', FALSE) ; CodeLn(FALSE) ;
      IncIndent ;
      CodeSuccess( i )
   END
END CodeCbraBeginSingular ;


PROCEDURE CodeCbraEndSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      DecIndent ;
      CodeString('END', FALSE) ; CodeLn(TRUE) ;
      CodeSuccess( i )
   END
END CodeCbraEndSingular ;


PROCEDURE CodeSbraBeginPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('IF ', TRUE) ; CodeOperand( i ) ; CodeLn(FALSE) ;
      CodeString('THEN', FALSE) ; CodeLn(FALSE) ;
      IncIndent ;
      CodeSuccess( i )
   END
END CodeSbraBeginPlural ;


PROCEDURE CodeSbraEndPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      DecIndent ;
      CodeString('END', FALSE) ; CodeLn(TRUE) ;
      CodeSuccess( i )
   END
END CodeSbraEndPlural ;


PROCEDURE CodeSbraBeginSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('Success := ', TRUE) ; CodeOperand( i ) ; CodeLn(TRUE) ;
      CodeSuccess( i )
   END
END CodeSbraBeginSingular ;


PROCEDURE CodeSbraEndSingular (i: CARDINAL) ;
BEGIN
   CodeSuccess( i )
END CodeSbraEndSingular ;


PROCEDURE CodeIdBeginPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('IF ', TRUE) ; CodeOperand( i ) ; CodeLn(FALSE) ;
      CodeString('THEN', FALSE) ; CodeLn(FALSE) ;
      IncIndent ;
      CodeSuccess( i )
   END
END CodeIdBeginPlural ;


PROCEDURE CodeIdEndPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF (success=Dummy) AND (NOT error)
      THEN
         DecIndent ;
         CodeString('END', FALSE) ; CodeLn(TRUE)
      ELSE
         DecIndent ;
         CodeString('ELSE', FALSE) ; CodeLn(FALSE) ;
         IncIndent ;
         CodeSuccess( i ) ;
         CodeError( i ) ;
         DecIndent ;
         CodeString('END', FALSE) ; CodeLn(TRUE)
      END
   END
END CodeIdEndPlural ;


PROCEDURE CodeIdBeginSingular (i: CARDINAL) ;
BEGIN
   CodeString('Success := ', TRUE) ; CodeOperand( i ) ; CodeLn(TRUE)
END CodeIdBeginSingular ;


PROCEDURE CodeIdEndSingular ;
BEGIN
   WriteString('Error should never get here') ; WriteLn ;
   HALT
END CodeIdEndSingular ;


PROCEDURE CodeSuccess (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF success=True
      THEN
         CodeString('Success := ', TRUE) ; CodeString('TRUE', FALSE) ; CodeLn( TRUE )
      ELSIF success=False
      THEN
         CodeString('Success := ', TRUE) ; CodeString('FALSE', FALSE) ; CodeLn( TRUE )
      END
   END
END CodeSuccess ;


PROCEDURE CodeError (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF error=TRUE
      THEN
         IF quote="'"
         THEN
            CodeString("WriteError('", TRUE) ; CodeKey(operand, FALSE) ; CodeString(" - expected')", FALSE) ; CodeLn(TRUE)
         ELSE
            CodeString('WriteError("', TRUE) ; CodeKey(operand, FALSE) ; CodeString(' - expected")', FALSE) ; CodeLn(TRUE)
         END
      END
   END
END CodeError ;


PROCEDURE CodeDefBeginPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('PROCEDURE ', FALSE) ;
      CodeKey(operand, FALSE) ; CodeString(' () : BOOLEAN ;', FALSE) ; CodeLn(FALSE) ;
      CodeString('VAR', FALSE) ; CodeLn(FALSE) ;
      CodeString('   Success: BOOLEAN ;', FALSE) ; CodeLn(FALSE) ;
      CodeString('BEGIN', FALSE) ; CodeLn(FALSE) ;
      IncIndent
   END
END CodeDefBeginPlural ;


PROCEDURE CodeDefEndPlural (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF success=True
      THEN
         CodeString('RETURN( TRUE )', TRUE) ; CodeLn(FALSE)
      ELSIF (operand=NulName) AND (atom=Id)
      THEN
         CodeString('RETURN( Success )', TRUE) ; CodeLn(FALSE)
      ELSE
         CodeString('RETURN( ', FALSE) ; CodeOperand(i) ; CodeString(' )', FALSE) ; CodeLn(FALSE)
      END ;
      DecIndent ;
      CodeString('END ', FALSE) ; CodeKey(Table[1].operand, FALSE) ; CodeString(' ;', FALSE) ; CodeLn(FALSE) ;
      CodeString('', FALSE)
   END
END CodeDefEndPlural ;


PROCEDURE CodeDefBeginSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      CodeString('PROCEDURE ', FALSE) ; CodeKey(operand, FALSE) ; CodeString(' () : BOOLEAN ;', FALSE) ; CodeLn(FALSE) ;
      CodeString('BEGIN', FALSE) ; CodeLn(FALSE) ;
      IncIndent
   END
END CodeDefBeginSingular ;


PROCEDURE CodeDefEndSingular (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF Global=True
      THEN
         CodeString('RETURN( TRUE )', TRUE) ; CodeLn(FALSE)
      ELSE
         CodeString('RETURN( ', TRUE) ; CodeOperand( i ) ; CodeString(' )', FALSE) ; CodeLn(FALSE)
      END ;
      DecIndent ;
      CodeString('END ', FALSE) ; CodeKey(Table[1].operand, FALSE) ; CodeString(' ;', FALSE) ; CodeLn(FALSE) ;
      CodeString('', FALSE)
   END
END CodeDefEndSingular ;


PROCEDURE CodeOperand (i: CARDINAL) ;
BEGIN
   WITH Table[i] DO
      IF atom=Lit
      THEN
         IF quote="'"
         THEN
            CodeString("SymIs('", FALSE) ; CodeKey(operand, FALSE) ; CodeString("')", FALSE)
         ELSE
            CodeString('SymIs("', FALSE) ; CodeKey(operand, FALSE) ; CodeString('")', FALSE)
         END
      ELSE
         CodeKey(operand, FALSE) ; CodeString('()', FALSE)
      END
   END
END CodeOperand ;


(*
   CodeKey - writes out key and correctly terminates the previous line of code.
*)

PROCEDURE CodeKey (k: CARDINAL; SOS: BOOLEAN) ;
BEGIN
   EndLastLine(SOS) ;
   WriteKey(k)
END CodeKey ;


(*
   CodeString - writes out string, a, and correctly terminates the previous
                line of code.
*)

PROCEDURE CodeString (a: ARRAY OF CHAR ; SOS: BOOLEAN) ;
BEGIN
   EndLastLine(SOS) ;
   WriteString(a)
END CodeString ;


(*
   EndLastLine - terminates the previous line of code. It adds a ;
                 if we have just ended a statement and we are about
                 to start a new statement.
*)

PROCEDURE EndLastLine (SOS: BOOLEAN) ;
VAR
   i: CARDINAL ;
BEGIN
   IF NewLine
   THEN
      IF SOS AND EndOfStatement
      THEN
         WriteString(' ;') ;
         EndOfStatement := FALSE
      END ;
      WriteLn ;
      NewLine := FALSE ;
      i := Level ;
      WHILE i>0 DO
         Write(' ') ;
         DEC(i)
      END
   END
END EndLastLine ;
   

PROCEDURE CodeLn (EOS: BOOLEAN) ;
BEGIN
   EndOfStatement := EOS ;
   NewLine := TRUE
END CodeLn ;


PROCEDURE IncIndent ;
BEGIN
   INC( Level, 3 )
END IncIndent ;


PROCEDURE DecIndent ;
BEGIN
   DEC( Level, 3 )
END DecIndent ;


BEGIN
   Init ;
   GetNextSymbol( Symbol ) ;
   IF NOT Bnf()
   THEN
      WriteError('Parser Failed')
   END
END yac.
