
/*
 *
 * Author     : Gaius Mulley
 * Title      : epson bitmap manipulation program
 * Date       : 31/10/93
 * Description: Twofold purpose:
 *
 *              (i)  takes as input a bitmap generated by say:
 *                   #!/bin/sh
 *                   cat << EOFEOF | gs -q -sDEVICE=bit -r$1 -dNOPAUSE -sOutputFile=op $2
 *                   quit
 *                   EOFEOF
 *              
 *                   op is a bitmap file which can be translated onto the epson
 *                   bitmap format.
 *
 *              (ii) will take the bitmap input and display it onto
 *                   the vga screen using vgalib.
 *                   NOTE you have to be root to run this option.
 *
 */


#include <stdio.h>
#include <vga.h>
#include <sys/types.h>
#include <unistd.h>

#define O_RDONLY  0
#define ByteSize  8
#define TRUE      (1==1)
#define FALSE     (1==0)

#define WIDTH     8    /* inches wide         */
#define HEIGHT   11    /* inches in height    */

/*
 * devinfo - this structure is used to define resolutions and
 *           maximum horizontal and vertical pixels of the bitmap
 *           page.
 */

struct devinfo {
  int Xres;               /* dots per inch horizontally */
  int Yres;               /* dots per inch vertically */
  int MaxX;               /* 5 inches wide */
  int MaxY;               /* 11 inches long */
  int BlockSize;          /* no of bytes required to represent a horizontal line */
  char *Resolution;       /* device ascii resolution */
  char *DevName;          /* user device name */
};

#define MaxDevices   14

static struct devinfo Devices[MaxDevices] =
{
  {  32,  32,  32*WIDTH,  32*HEIGHT,  32*8/8+2,  "32x32",   "tiny"        },
  {  64,  64,  64*WIDTH,  64*HEIGHT,  64*8/8+4,  "64x64",   "tinyx2"      },
  { 100, 100, 100*WIDTH, 100*HEIGHT, 100*8/8+9,  "100x100", "tinyx3"      },
  { 120, 120, 120*WIDTH, 120*HEIGHT, 120*8/8+8,  "120x120", "epsondouble" },
  { 160, 160, 160*WIDTH, 160*HEIGHT, 160*8/8+10, "160x160", "double"      },
  { 160,  96, 160*WIDTH,  96*HEIGHT, 160*8/8+10, "160x96",  "doubleprint" },
  { 320, 320, 320*WIDTH, 320*HEIGHT, 320*8/8+20, "320x320", "quad"        },
  { 320,  96, 320*WIDTH,  96*HEIGHT, 320*8/8+20, "320x96",  "quadprint"   },
  {  44,  44,  44*WIDTH,  44*HEIGHT,  44*8/8+3,  "44x44",   "A4"          },
  { 108,  48,  48*WIDTH, 108*HEIGHT, 108*8/8+7,  "108x48",  "fat"         },
  {  80,  48,  48*WIDTH,  80*HEIGHT,  80*8/8+5,  "80x48",   "allfat"      },
  {  80,  80,  80*WIDTH,  80*HEIGHT,  80*8/8+5,  "80x80",   "single"      },
  {  70,  70,  70*WIDTH,  70*HEIGHT,  70*8/8+5,  "70x70",   "test"        },
  { 108, 108, 108*WIDTH, 108*HEIGHT, 108*8/8+7,  "108x108", "fullwidth"   },
};

#define MaxBlockSize 0x10000


static char block[MaxBlockSize] ;  /* current bitmap X line.               */
static int  no=0;                  /* number of Y lines we have read.      */
static int  VGA_MAX_X;             /* the maximum X point on the vga card  */
static int  VGA_MAX_Y;             /* the maximum Y point on the vga card  */
static int  InFile=0;              /* file number of the input file.       */
static int  CurrentDevice  = 0;    /* the device the user requested        */
static struct devinfo *Dev = (struct devinfo *)&Devices[0];
                                   /* pointer to the device the user wants */
static int  IsEpson = 0;           /* are we generating an epson bitmap?   */


/*
   ReadBlock - read a block - a block is the number of bytes required to
               store a complete X line of bitmap. (If IsEpson is set then
               we have ByteSize * BlockSize as we need to transpose 8 bits
               at a time).
*/

static int ReadBlock (char *b)
{
  int block=Dev->BlockSize;

  if (IsEpson) {
    block = Dev->BlockSize*ByteSize;
    if (block > MaxBlockSize) {
      fprintf(stderr, "increase MaxBlockSize to at least 0x%x\n", block);
    }
  }
  if (read(InFile, b, block) == block) {
    return( TRUE );
  } else {
    return( FALSE );
  }
}


/*
   incl - include a bit, i, in byte, b. Set operation b := b + {i}.
*/

static void incl (char *b, int i)
{
  (*b) |= (1<<i);
}


/*
   IsSet - returns true if bit, j, is in byte, i.
*/

static int IsSet (char *b, int i, int j)
{
  return( b[i] & (1<<j));
}


/*
   here - is a debugging stub.
*/

static void here () {}


/*
   IsAnythingToPrint - checks to see whether the output buffer
                       consists of an empty bitmap line, if so
                       we return FALSE.
*/

static int IsAnythingToPrint(char *OutputBuffer, int LineLength)
{
  int i;

  for (i=0; i<LineLength; i++) {
    if (OutputBuffer[i] != '\000') {
      return( TRUE );
    }
  }
  return( FALSE );
}

/*
   ConvertAndWrite - if we are an epson then take each 8 vertical bits
                     and write this byte to stdout. If we are displaying
                     the bitmap to screen then simply dump each bit to
                     a pixel.
*/

static void ConvertAndWrite (char *b)
{
  int i, j;
  char OutputBuffer[MaxBlockSize];
  char CommandBuffer[20];
  char byte;

  no = no+1;

  if (! IsEpson) {
    /* check for end of page on the vga screen */
    if (no == VGA_MAX_Y+1) {
      vga_screenon();
      vga_getch();
      vga_clear();
      vga_screenoff(); 
      no = 0;
    }
    for (i=0; i<Dev->MaxX; i++) {
      if ((i<VGA_MAX_X) && (no < VGA_MAX_Y)) {
	if (IsSet(b, i/ByteSize, ByteSize-(i % ByteSize))) {
	  vga_drawpixel(i, no);
	}
      }
    }
  } else {
    /* write binary bitmap to printer */
    for (i=0; i<Dev->MaxX; i++) {
      byte = 0;
      for (j=0; j<ByteSize; j++) {
	if (IsSet(b, i/ByteSize+(j*Dev->BlockSize), ByteSize-(i % ByteSize))) {
	  incl(&byte, (ByteSize-1)-j);
	}
      }
      if (i > MaxBlockSize) {
	fprintf(stderr, "increase MaxBlockSize to at least 0x%x\n", i);
	exit(1);
      }
      OutputBuffer[i] = byte;
    }
    if (IsAnythingToPrint(OutputBuffer, Dev->MaxX*ByteSize)) {
      /* condensed print \033\017  - puts the printer into 320x320 bits/inch */
      CommandBuffer[0] = '\033';
      if (strcmp(Dev->DevName, "doubleprint") == 0) {
	CommandBuffer[1] = 'L';   /* L for 160x108 */
      } else if (strcmp(Dev->DevName, "quadprint") == 0) {
	CommandBuffer[1] = 'Z';   /* Z for 320x216 */
      } else {
	fprintf(stderr, "must use 160x108 or 320x216 for -epson option\n");
	exit(1);
      }
      CommandBuffer[2] = Dev->MaxX*ByteSize % 256;
      CommandBuffer[3] = Dev->MaxX*ByteSize / 256;
      /* bitmap graphics mode 3 (320 dots/inch)  \033*3 n%256 n/256 */
      write(1, CommandBuffer, 4);
      write(1, &OutputBuffer, Dev->MaxX*ByteSize);
    }
    write(1, "\012", 1);
    if (no == Dev->MaxY/8) {
      /* send form feed at end of page */
      write(1, "\14\015", 2);
      no = 0;
    }
  }
}


/*
   SetUpPrinter - sends control codes to put the printer into graphic
                  mode.
*/

static void SetUpPrinter (void)
{
  char CommandBuffer[20];

  if (IsEpson) {
    /* compressed mode \033 \017 */
    CommandBuffer[0] = '\033';
    CommandBuffer[1] = '\017';
    CommandBuffer[2] = '\015';
    /* line spacing 24  \033 3 \030 */
    CommandBuffer[3] = '\033';
    CommandBuffer[4] =    '3';
    CommandBuffer[5] = '\030';  /* was 10 */
    CommandBuffer[6] = '\015';
    write(1, &CommandBuffer, 7);
  }
}

/*
   ResetPrinter - sends control codes to reset the printer and clear
                  the print buffer (not the input buffer!).
*/

static void ResetPrinter (void)
{
  char CommandBuffer[20];

  if (IsEpson) {
    /* reset printer \033 @ */
    CommandBuffer[0] = '\033';
    CommandBuffer[1] = '@';
    write(1, &CommandBuffer, 2);
  }
}


/*
   Translate - for each line of bitmap do convert.
*/

static void Translate (void)
{
  ResetPrinter();
  SetUpPrinter();
  while (ReadBlock(block)) {
    ConvertAndWrite(block);
  }
  ResetPrinter();
}


/*
   SearchForDevice - given a name, n, search for the device or resolution
                     that the user requested.
                     It returns true if a valid device was found.
*/

static int SearchForDevice (char *n)
{
  CurrentDevice = 0;
  while ((CurrentDevice < MaxDevices) &&
	 (strcmp(n, Devices[CurrentDevice].DevName) != 0) &&
	 (strcmp(n, Devices[CurrentDevice].Resolution) != 0)) {
    CurrentDevice++;
  }
  if (CurrentDevice == MaxDevices) {
    CurrentDevice = 0;
    printf("choice of resolutions and device names are:\n");
    for (CurrentDevice = 0; CurrentDevice < MaxDevices; CurrentDevice++) {
      printf("   %s   %s\n",
	     Devices[CurrentDevice].Resolution,
	     Devices[CurrentDevice].DevName);
    }
    CurrentDevice = 0;
    return( FALSE );
  } else {
    Dev = &Devices[CurrentDevice];
    return( TRUE );
  }
}


/*
   ScanArguments - checks to see whether the correct number of
                   parameters have been passed. In particular
                   it checks for a valid device and filename.
                   It also checks to see whether the user has requested
                   that the epson format bitmap is to be created.
*/

static void ScanArguments (int argc, char *argv[])
{
  IsEpson = FALSE;
  if (argc > 1) {
    InFile = open(argv[2], O_RDONLY);
    if ((InFile < 0) || (argc < 3) || (! SearchForDevice(argv[1]))) {
      fprintf(stderr, "Usage: %s <resolution> <inputfile> [-epson]\n",
	      argv[0]);
      exit(0);
    }
    if (argc == 4) {
      IsEpson = (strcmp(argv[3], "-epson") == 0);
    }
  } else {
    if (SearchForDevice("not found")) {}
    fprintf(stderr, "Usage: %s <resolution> <inputfile> [-epson]\n",
	    argv[0]);
    exit(0);
  }
}


/*
   main - check arguments, set up VGA_MAX variables and then call translate.
*/

main (int argc, char *argv[])
{
  ScanArguments(argc, argv);

  if (! IsEpson) {
    vga_setmode(G640x480x16);
    vga_screenoff();

    VGA_MAX_X = vga_getxdim()-1;
    VGA_MAX_Y = vga_getydim()-1;

    vga_drawline(0, 0, VGA_MAX_X, 0);
    vga_drawline(VGA_MAX_X, 0, VGA_MAX_X, VGA_MAX_Y);

    Translate();

    vga_screenon(); 

    vga_getch();
    vga_setmode(TEXT);
  } else {
    VGA_MAX_X = Dev->MaxX;
    VGA_MAX_Y = Dev->MaxY;

    Translate();
  }
}
